/***************************************************************************
                          vigrademo.cpp  -  description
                             -------------------
    begin                : 14.11.2002
    copyright            : (C) 2002 by Fabian Wenzel, Tobias Frommfeld
    email                : tuvisionlib@tuvision.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
/***************************************************************************/
/*
 *   This example shows how to use the functions of the generic VIGRA
 *   computer vision library with the TUVision Image Processing Library.
 *
 *   The main part which makes up the connection between the two lib's is
 *   an iterator which can be used with VIGRA generated with a
 *   TUVisionlib - image and an accessor which defines how pixels in the
 *   image are accessed.
 *
 *   This example is split in two separate demos: One for
 *   connecting Grayscale images to VIGRA, the second one
 *   for connecting RGB images to VIGRA.
 *
 *   WARNING: MSVC 6.0 cannot produce a release-version, it crashes with
 *   an "internal compiler error". So currently only debug-versions can
 *   be made. Solution: Don't use MSVC ;-)
 *
 *                                 Fabian Wenzel
 */
/***************************************************************************/

#include <cstdlib>

#include <tuv/CFileReadUtils.h>
#include <tuv/CFileWritePng.h>
#include <tuv/CImageByteGray.h>
#include <tuv/CImageByteRGB.h>

// This is needed in any case
#include <vigra/imageiterator.hxx>

// This is needed for the first (grayscale) example
#include <vigra/edgedetection.hxx>

// This is needed for the second (RGB) example
#include <vigra/rgbvalue.hxx>
#include <vigra/cornerdetection.hxx>
#include <vigra/localminmax.hxx>

using namespace std;
using namespace TUVision;
using namespace vigra;


/********************************* GRAY IMAGE DEMO ****************************/
/*
 *  This function demonstrates how to connect grayscale images with VIGRA
 *  algorithms. Basically, there are only two steps:
 *
 *  1. Define an iterator which can navigate on the TUVisionlib-image-content
 *
 *  2. Define an accessor which tells VIGRA how to access each pixel. Usually
 *     this is easy when dealing with grayscale images (see below)
 *     
 * As an algorithm, the VIGRA canny detector is used.
 *
 ******************************************************************************/
void grayscaledemo()
{
    // First, read input image
    CImageByteGray *inputimage  = readImageByteGray("test.png");

    int width =inputimage->getW();
    int height=inputimage->getH();

    // The results are saved in the output image
    CImageByteGray *outputimage = new CImageByteGray(width,height);
    outputimage->clearImage();

    // Canny algorithm:
    // The first argument specifies the beginning of the input area
    // The second argument specifies the end of the input area
    //
    // In this case, the whole image gets processed (by using Diff2D(width,height))
    //
    // The third argument specifies the destination start (here outputimage->getRow(0))
    //
    // Look in VIGRA - documentation for the last three parameters
    // (they are scale, gradientthreshold and value of the edge in dst-image)

    cannyEdgeImage(srcImageRange(*inputimage),
        destImage(*outputimage),
        0.8,4.0,255);


    // Done. Now save image
    CFileWritePng pngwriter;
    pngwriter.open("bird_edges.png");
    pngwriter.write(*outputimage);
    pngwriter.close();

    delete outputimage;
    delete inputimage;
}

/********************************* RGB IMAGE DEMO ****************************/
/*
 *  This function demonstrates how to connect RGB images with VIGRA
 *  algorithms. Again, there are only two steps:
 *
 *  1. Define an iterator which can navigate on the TUVisionlib-image-content
 *
 *  2. Define an accessor which tells VIGRA how to access each pixel. In the
 *     following example, a special accessor is used (for demonstration
 *     purposes) that accesses the "virtual" gray value of an RGB pixel. That
 *     means colorspace conversion is done on-the-fly.
 *     
 *  As an algorithm, the VIGRA corner response function is used which then
 *  gets searched for local maxima (the "corners"). After that, only highly
 *  responding corners (i.e. not "almost-corners") are kept and the maxima
 *  which have been found before are discarded by image multiplication.
 *  In the end, the result gets saved in an image file.
 *
 ******************************************************************************/
void rgbdemo()
{
    CImageByteRGB *inputimage = readImageByteRGB("lena_std.png");
    int width =inputimage->getW();
    int height=inputimage->getH();

    CImageByteGray *outputimage = new CImageByteGray(width,height);
    outputimage->clearImage();

    // Intermediate VIGRA image needed for float values of corner response
    FImage corners(width,height);

    // Set up Vigra Accessors

    typedef RGBValue< unsigned char > RGBByte;

    // Set up accessors for the image. The corner response functions
    // does not operate on RGB-triples, so by setting up an accessor
    // which accesses the "gray" value of an RGB pixel, the intensity
    // is used for corner detection. Other accessors (RedChannelAccessor,...)
    // are theoretically possible, too.

    RGBToGrayAccessor< RGBByte       > RGBtoByteGray;

    // Now four lines to perform the algorithm (mostly taken from the docs)

    // find corner response at scale 1.0
    cornerResponseFunction(srcImageRange(*inputimage,RGBtoByteGray),
        destImage(corners), 1.0);

    // find local maxima of corner response, mark corresponding pixels in
    // the destination image with 1
    localMaxima(srcImageRange(corners),destImage(*outputimage));

    // threshold corner response to keep only strong corners (above 400.0)
    // and mark them with 255.0
    // To avoid three iterator and parameters (see last vigra function calls)
    // a template can be used when using "native" VIGRA images (srcImageRange)
    transformImage(srcImageRange(corners),destImage(corners),
        Threshold< double , double >( 400.0, 100000000.0,
        0.0  , 255.0)); 

    // multiply byte image with gray image (i.e. only keep local maxima)
    combineTwoImages(srcImageRange(*outputimage),
        srcImage(corners),
        destImage(*outputimage), std::multiplies<float>());

    CFileWritePng pngwriter;
    pngwriter.open("lenacorners.png");
    pngwriter.write(*outputimage);
    pngwriter.close();

    delete outputimage;
    delete inputimage;
}

int main(int argc, char *argv[])
{
    grayscaledemo();
    rgbdemo();
    return EXIT_SUCCESS;
}
