#include <tuv/CFileReadAll.h>   // We want to read a png - file
#include <tuv/CFileWritePng.h>  // We want to write the output images
#include <tuv/CImageByteGray.h> // We want to use a grayscale image

#include <iostream>



using namespace std;


/* The following part defines a class CAverageDifferenceOperator. Given a grayscale
 * image it computes the average (or difference) of two adjacent pixels, either
 * horizontally or vertically and stores them in a new output image. This means
 * that the image size of the output image will be different from the input file:
 * The width or the height will be halved - depending on the direction.
 *
 * All classes of the TUVision library should be in the namespace TUVision.
 * This is why the class code is inside the "namespace" - statement
 */

namespace TUVision{

    /* This is the class definition. Usually it can be found in a corresponding ".h"
     * - file, but it can also be written inside a .cpp - function as we don't
     * want to change the library here. */
    class CAverageDifferenceOperator{
    public:
        enum Tdirection {
            horizontal,
            vertical
        };
        void setType(Tdirection direction);
        CAverageDifferenceOperator (Tdirection direction) { setType(direction); }
        CImageByteGray * average   (CImageByteGray &inputimage);
        CImageByteGray * difference(CImageByteGray &inputimage);
    private:
        Tdirection d_direction;
    };

    /* The following code implements the class' functionality */
    void CAverageDifferenceOperator::setType(Tdirection direction){
        d_direction = direction;
    }

    /* The average - function */
    CImageByteGray * CAverageDifferenceOperator::average(CImageByteGray &inputimage){
        CImageByteGray *outputimage = 0;
        switch(d_direction){
    case horizontal: {
        /* The new image is generated with a difference width (we are in 'horizontal')
         * Afterwards, the image is being looped through row- and columnwise */
        outputimage = new CImageByteGray(inputimage.getW()/2,inputimage.getH());
        /* Outside loop for rows */
        for(int y=0;y<inputimage.getH();y++){
            /* Usually it is possible to get the pixel value by calling getPixel(x,y,0),
             * but it slows down the program as the transformation (x,y)->memory has
             * to be done at each access. This is why getRow(y) and row[x] is used:
             * getRow(y) gives the pointer to the first pixel in row y, and row[x] adds
             * the x-value to it. This way (0,y)->memory has to be done only once
             * per row */
            unsigned char *inputrow =inputimage  .getRow(y);
            unsigned char *outputrow=outputimage->getRow(y);
            /* Inside loop for columns */
            for(int x=0; x<inputimage.getW()/2;x++){
                outputrow[x]=(inputrow[2*x]+inputrow[2*x+1])/2;
            }
        }
        break; }
    case vertical: {
        outputimage = new CImageByteGray(inputimage.getW(),inputimage.getH()/2);
        for(int y=0;y<inputimage.getH()/2;y++){
            unsigned char *firstinputrow =inputimage  .getRow(2*y);
            unsigned char *secondinputrow=inputimage  .getRow(2*y+1);
            unsigned char *outputrow     =outputimage->getRow(y);
            for(int x=0; x<inputimage.getW();x++){
                outputrow[x]=(firstinputrow[x]+secondinputrow[x])/2;
            }
        }
        break; }
    default: break;
        }
        return outputimage;
    }

    /* The difference function - identical to the one above except that the output
     * pixels are calculated as difference */
    CImageByteGray * CAverageDifferenceOperator::difference(CImageByteGray &inputimage){
        CImageByteGray *outputimage = 0;
        switch(d_direction){
    case horizontal: {
        outputimage = new CImageByteGray(inputimage.getW()/2,inputimage.getH());
        for(int y=0;y<inputimage.getH();y++){
            unsigned char *inputrow =inputimage  .getRow(y);
            unsigned char *outputrow=outputimage->getRow(y);
            for(int x=0; x<inputimage.getW()/2;x++){
                outputrow[x]=abs(inputrow[2*x]-inputrow[2*x+1])/2;
            }
        }
        break; }
    case vertical: {
        outputimage = new CImageByteGray(inputimage.getW(),inputimage.getH()/2);
        for(int y=0;y<inputimage.getH()/2;y++){
            unsigned char *firstinputrow =inputimage  .getRow(2*y);
            unsigned char *secondinputrow=inputimage  .getRow(2*y+1);
            unsigned char *outputrow     =outputimage->getRow(y);
            for(int x=0; x<inputimage.getW();x++){
                outputrow[x]=abs(firstinputrow[x]-secondinputrow[x])/2;
            }
        }
        break; }
    default: break;
        }
        return outputimage;
    }
}


/* This is where the main program starts */

/* We want to use the TUVision namespace in which all library functions are stored
 * This can be done by the "using namespace" - statement. Otherwise each class of
 * the TUVision - library has to be prefixed with "TUVision::" and the code
 * would be much less readable */

using namespace TUVision;

/* Main function */
int main(int argc, char *argv[]){

    /* Read the input image using CFileReadPng */
    try{
        CFileReadPng mypngreader;
        CFileReadAll myFileReader;
        myFileReader.open("input.png");
        CImageByteGray *inputimage = (CImageByteGray *) myFileReader.read();

        /* Decompose image and write results */
        CFileWritePng myFileWriter;
        CImageByteGray *outputimage;
        CAverageDifferenceOperator myOperator(CAverageDifferenceOperator::horizontal);

        outputimage   = myOperator.average   (*inputimage);
        myFileWriter.open("horizontal_average.png");
        myFileWriter.write(*outputimage);
        myFileWriter.close();

        /* The operator (see above) returns a NEW image, so we have to delete it
         * as soon as we don't need it anymore, otherwise it will stay in memory */
        delete outputimage;

        myOperator.setType(CAverageDifferenceOperator::vertical);
        outputimage = myOperator.difference(*inputimage);
        myFileWriter.open("vertical_difference.png");
        myFileWriter.write(*outputimage);
        myFileWriter.close();

        myFileReader.close();

        delete outputimage;
        delete inputimage;

    } catch ( CException &e ){
        cout << e.what() << endl;
    }
    return 0;
}
