/***************************************************************************
                          CImage.h  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde, Eric Sommerlade, Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde, Eric Sommerlade, PhD Shdaifat

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_CIMAGE_H__
#define __TUVISION_CIMAGE_H__ 1

#include "TUVisionDef.h"

namespace TUVision {

class CRect;
class CImageAttributes;

/**********************************************************/
//!  Base class defining an image in the memory.
/*! 
 *   This base class defines an image in the memory.
 *   From this class you have to derive classes for the
 *   different number formats like unsigned char, int,
 *   float etc. 
 *
 *   \ingroup image
 *
 *   \author Marco Budde
 *   \author Eric Sommerlade
 *   \author Islam Shdaifat
 */
/**********************************************************/

class _CCTUVISION CImage {

public:

    /******************************************************/
    //! Destructor.
    /******************************************************/
    virtual ~CImage();

    /******************************************************/
    //! Get width of the image.
    /******************************************************/
    inline int getW()  const;

    /******************************************************/
    //! Full image width (no windowing)
    /******************************************************/
    inline int width() const;

    /******************************************************/
    //! Get height of the image.
    /******************************************************/
    inline int getH()   const;

    /******************************************************/
    //! Full image height (no windowing)
    /******************************************************/
    inline int height() const;

    /******************************************************/
    //! Get a reference to the image attributes.
    /******************************************************/
    virtual CImageAttributes& getAttributes();

    /******************************************************/
    //! Get a reference to the image attributes 
    //! (const version).
    /******************************************************/
    virtual const CImageAttributes& getAttributes() const;

    /******************************************************/
    //! Returns the number of channels.
    /******************************************************/
    inline int getNumOfChannels() const { return d_nchannels; }

    /******************************************************/
    //! Sets the window of an image.
    /******************************************************/
    virtual void setWindow (int x_0, int y_0, int width, int height);

    /******************************************************/
    //! Sets the window of an image.
    /******************************************************/
    virtual void setWindow (const CRect& rect);

    /******************************************************/
    //! Gets the window of an image.
    /******************************************************/
    virtual CRect getWindow ();

    /******************************************************/
    //! Deletes the window of an image.
    /******************************************************/
    virtual void resetWindow ();

    /******************************************************/
    //! Makes a virtual copy of the image
    /*!
     *  This function which has to be implemented in all
     *  instantiable child classes returns a copy of itself
     *
     *  \param copycontent Defines if the image data should
     *         be copied also (see copy constructor)
     *
     */
    /******************************************************/
    virtual CImage *clone(bool copycontent=true) const =0;

    /******************************************************/
    //! Clears the active region with the default value.
    /*!
     *  This function sets all data in the active region 
     *  to the default value specified in the specific 
     *  image type.
     *  
     *  The function must be implemented in the 
     *  CImage<Type> classes.
     */
    /******************************************************/
    virtual void clearImage () = 0;      

protected:
    /******************************************************/
    //! New image constructor.
    /******************************************************/
    CImage (int w, int h, int nchannels);

    /******************************************************/
    //! Copy constructor.
    /******************************************************/
    CImage (const CImage &img, bool copy);

    /******************************************************/
    //! Adjusts the pointer to the first row.
    /*!
     *  This function sets a pointer (d_img_pnt) to the 
     *  data of the first pixel in the first row. This
     *  pointer is equal to the first byte of the image
     *  buffer (d_img_data), when the user has not 
     *  created a window using the setWindow() function.
     *  Otherwise it points to the first pixel in the
     *  window.
     *  
     *  The function must be implemented in the 
     *  CImage<Type> classes.
     */
    /******************************************************/
    virtual void adjustRowPtr() = 0;

protected:
    const int d_width;      //!< image width (in pixels) 
    const int d_height;     //!< image height (in pixels) 
    const int d_nchannels;  //!< number of color channels 
    int d_win_x;      //!< window start point (x coordinate) 
    int d_win_y;      //!< window start point (y coordinate) 
    int d_win_width;  //!< width of the window 
    int d_win_height; //!< height of the window 
    CImageAttributes* d_img_attributes; //!< string and double attributes 

};


// implementation of the inline code

/**********************************************************/
/*! 
 *   Get the windowed width of the image (in pixels). 
 *   This width needs not to be the physical width of the 
 *   image.
 *
 *   \return window-width of the image
 */
/**********************************************************/

inline int 
CImage::getW (void) const
{
    return d_win_width; 
}


/**********************************************************/
/*! 
 *   Get the image width of the image (in pixels). 
 *   This width does not respect windows
 *
 *   \return window-width of the image
 */
/**********************************************************/
inline int
CImage::width (void) const
{
    return d_width;
}

/**********************************************************/
/*! 
 *   Get the windowed height of the image (in pixels). 
 *   This height needs not to be the physical height of 
 *   the image.
 * 
 *   \return window-height of the image
 */
/**********************************************************/

inline int 
CImage::getH (void) const
{ 
    return d_win_height; 
}


/**********************************************************/
/*! 
 *   Get the height of the image (in pixels). 
 *   This height does not consider windows.
 * 
 *   \return window-height of the image
 */
/**********************************************************/
inline int 
CImage::height (void) const
{ 
    return d_height;
}

} // namespace TUVision

#endif // __TUVISION_CIMAGE_H__
