/***************************************************************************
                          CImageTemplate.h  -  description
                             -------------------
    changed              : 13.11.2002
    copyright            : (C) 2002 Tobias Frommfeld
                         : based on Diplomarbeit Marco Budde,
                           Eric Sommerlade, PhD Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : Studienarbeit Tobias Frommfeld

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_TEMPLATE_H__
#define __TUVISION_TEMPLATE_H__ 1

#include "TUVisionDef.h"
#include "vigra/accessor.hxx"
#include "vigra/imageiterator.hxx"
#include "vigra/tinyvector.hxx"
#include "vigra/utilities.hxx"
#include "vigra/rgbvalue.hxx"

#include <cstring>

#include "tuv/CImageType.h"
#include "tuv/CImageData.h"
#include "tuv/CExceptionRange.h"

namespace TUVision {
/**********************************************************/
//!  Marks a Colorspace for CImageTemplate and defines
/*!  number of channels
 *
 *	Colorspaces are classes used for being able to seperate image types
 *	Example:
 *	- typedef CImageTemplate<unsigned char, RGB> CImageByteRGB
 *	- typedef CImageTemplate<unsigned char, BRG> CImageByteBGR
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
struct _CCTUVISION UnspecifiedColorspace { enum { CHANNELS = 1 }; };

/**********************************************************/
//!  Marks a Colorspace for CImageTemplate and defines
/*!  number of channels
 *
 *	Colorspaces are classes used for being able to seperate image types
 *	Example:
 *	- typedef CImageTemplate<unsigned char, RGB> CImageByteRGB
 *	- typedef CImageTemplate<unsigned char, BRG> CImageByteBGR
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
struct _CCTUVISION Gray    { enum { CHANNELS = 1}; };

/**********************************************************/
//!  Marks a Colorspace for CImageTemplate and defines
/*!  number of channels
 *
 *	Colorspaces are classes used for being able to seperate image types
 *	Example:
 *	- typedef CImageTemplate<unsigned char, RGB> CImageByteRGB
 *	- typedef CImageTemplate<unsigned char, BRG> CImageByteBGR
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
struct _CCTUVISION Bilevel { enum { CHANNELS = 1}; };
/**********************************************************/
//!  Marks a Colorspace for CImageTemplate and defines
/*!  number of channels
 *
 *	Colorspaces are classes used for being able to seperate image types
 *	Example:
 *	- typedef CImageTemplate<unsigned char, RGB> CImageByteRGB
 *	- typedef CImageTemplate<unsigned char, BRG> CImageByteBGR
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
struct _CCTUVISION RGB     { enum { CHANNELS = 3}; };

/**********************************************************/
//!  Marks a Colorspace for CImageTemplate and defines
/*!  number of channels
 *
 *	Colorspaces are classes used for being able to seperate image types
 *	Example:
 *	- typedef CImageTemplate<unsigned char, RGB> CImageByteRGB
 *	- typedef CImageTemplate<unsigned char, BRG> CImageByteBGR
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
struct _CCTUVISION YUV     { enum { CHANNELS = 3}; };

/**********************************************************/
//!  Defines traversers and Accessors for CImageTemplate
/*!
 *	Uses macros -> MS VC++ does not support
 *	partial template specialization
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
template<class TYPE, class COLORSPACE, int CHANNELS>
struct VigraAccessTraits {
    typedef typename
        vigra::TinyVector< TYPE, CHANNELS > 	result_type;
    typedef typename
        vigra::ImageIterator<
        vigra::TinyVector< TYPE, CHANNELS > >	traverser;
    typedef typename
        vigra::ConstImageIterator<
        vigra::TinyVector< TYPE, CHANNELS > >	const_traverser;
    typedef typename
        vigra::StandardAccessor<
        vigra::TinyVector< TYPE, CHANNELS > >	Accessor;
    typedef typename
        vigra::StandardConstAccessor<
        vigra::TinyVector< TYPE, CHANNELS > >	ConstAccessor;
};

template<class TYPE> struct VigraAccessTraits< TYPE, RGB, 3 > {
    typedef vigra::RGBValue< TYPE> 	result_type;
    typedef vigra::ImageIterator<
        vigra::RGBValue< TYPE> >	traverser;
    typedef vigra::ConstImageIterator<
        vigra::RGBValue< TYPE > >	const_traverser;
    typedef vigra::StandardAccessor<
        vigra::RGBValue< TYPE > >	Accessor;
    typedef vigra::StandardConstAccessor<
        vigra::RGBValue< TYPE > >	ConstAccessor;
};


template<class TYPE, class COLORSPACE>
struct VigraAccessTraits< TYPE, COLORSPACE,1 > {

    typedef TYPE result_type;
    typedef vigra::ImageIterator<TYPE>		traverser;
    typedef vigra::ConstImageIterator<TYPE>		const_traverser;
    typedef vigra::StandardAccessor<TYPE>		Accessor;
    typedef vigra::StandardConstAccessor<TYPE>	ConstAccessor;
};

/**********************************************************/
//!  Generic Image based on old TUVisionlib.
/*!
 *	This class defines a template for generating arbitrary
 *	images. Easy access to VIGRA is granted by typedefs for
 *	accessors and iterators.
 *
 *   \ingroup image
 *
 *   \param TYPE Base-type of class
 *   \param COLORSPACE Type of colorspace
 *   \param CHANNELS Nr. of channels (ex. 3 channels: 123123123..)
 *
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
template< class TYPE,
class COLORSPACE = UnspecifiedColorspace,
    int CHANNELS = COLORSPACE::CHANNELS,
class VTR = VigraAccessTraits<TYPE, COLORSPACE, CHANNELS> >
class CImageTemplate : public CImageType<TYPE>
{
public:
    using CImageType<TYPE>::d_img_pnt;
    using CImageType<TYPE>::d_img_data;
    using CImageType<TYPE>::d_nchannels;
    using CImageType<TYPE>::d_width;
    using CImageType<TYPE>::d_height;
    using CImageType<TYPE>::d_win_width;
    using CImageType<TYPE>::d_win_height;
    /*! provides a matching vigra iterator */
    typedef typename VTR::traverser		traverser;
    /*! provides a matching constant vigra iterator */
    typedef typename VTR::const_traverser	const_traverser;
    /*! provides a matching vigra accessor */
    typedef typename VTR::Accessor		Accessor;
    /*! provides a matching constant vigra accessor */
    typedef typename VTR::ConstAccessor	ConstAccessor;
    /*! provides a matching vigra pixel-type for the colorspace of the image (e.g. int, TinyVektor<int, 3>, ...) */
    typedef typename VTR::result_type	PixelType;
    /*! provides basic datatype of the image memory */
    typedef TYPE			value_type;

    /*! number of channels used in this class */
    enum { NCHANNELS = CHANNELS };


    /**********************************************************/
    /*!
     *   Returns a vigra iterator for upper left pixel.
     *	The iterator is created for 'PixelType' which means:
     *	- for one channel the template parameter TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  vigra::RGBValue<TYPE>
     *	- otherwise it is virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline traverser upperLeft () {
        return traverser((PixelType *) d_img_pnt
            ,d_width);
    }

    /**********************************************************/
    /*!
     *   Returns a vigra iterator for lower right pixel.
     *	The iterator is created for 'PixelType' which means:
     *	- for one channel the template parameter TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  vigra::RGBValue<TYPE>
     *	- otherwise it is virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline traverser lowerRight () {
        return traverser((PixelType *) d_img_pnt
            ,d_width)
            + vigra::Diff2D(d_win_width,d_win_height);
    }

    /**********************************************************/
    /*!
     *   Returns a constant vigra iterator for upper left pixel.
     *	The iterator is created for 'PixelType' which means:
     *	- for one channel the template parameter TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  vigra::RGBValue<TYPE>
     *	- otherwise it is virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline const_traverser upperLeft () const {
        return const_traverser((PixelType *) d_img_pnt, d_width);
    }

    /**********************************************************/
    /*!
     *   Returns a constant vigra iterator for lower right pixel.
     *	The iterator is created for 'PixelType' which means:
     *	- for one channel the template parameter TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  vigra::RGBValue<TYPE>
     *	- otherwise it is virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline const_traverser lowerRight () const {
        return const_traverser((PixelType *) d_img_pnt, d_width)
            + vigra::Diff2D(d_win_width,d_win_height);
    }

    /**********************************************************/
    /*!
     *   Returns a STL- vector compatible iterator (pointer)
     *   for the image pointing to the first pixel in memory
     */
    /**********************************************************/
    inline PixelType * begin () {
        return (PixelType *) d_img_pnt;
    }

    /**********************************************************/
    /*!
     *   Returns a STL- vector compatible iterator (pointer)
     *   for the image pointing to the first pixel in memory
     *   (const version)
     */
    /**********************************************************/
    inline const PixelType * begin () const {
        return (const PixelType *) d_img_pnt;
    }

    /**********************************************************/
    /*!
     *   Returns a STL- vector compatible iterator (pointer)
     *   for the image pointing past the last pixel in memory
     */
    /**********************************************************/
    inline PixelType * end () {
        return ((PixelType *) d_img_pnt) + d_height*d_width;
    }


    /**********************************************************/
    /*!
     *   Returns a STL- vector compatible iterator (pointer)
     *   for the image pointing past the last pixel in memory
     */
    /**********************************************************/
    inline const PixelType * end () const {
        return ((const PixelType *) d_img_pnt) + d_height*d_width;
    }


    /**********************************************************/
    /*!
     *   Returns an vigra accessor depending on the template parameter
     *	TYPE. This means:
     *	- for one channel the accessor for TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  an accessor for vigra::RGBValue<TYPE>
     *	- otherwise an accessor for virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline Accessor accessor() {
        return Accessor();
    }

    /**********************************************************/
    /*!
     *   Returns an constant vigra accessor depending on the
     *	template parameter TYPE. This means:
     *	- for one channel the accessor for TYPE
     *	- for 3 channels and the labe 'LabelIsRGB'  it is
     *	  an accessor for vigra::RGBValue<TYPE>
     *	- otherwise an accessor for virga::TinyVektor<TYPE>
     */
    /**********************************************************/
    inline ConstAccessor accessor() const {
        return ConstAccessor();
    }

    /**********************************************************/
    /*!
     *   Constructor: create new image object with the
     *   specified dimensions.
     *
     *   \param w width (in pixels)
     *   \param h length (in pixels)
     */
    /**********************************************************/
    CImageTemplate (int w, int h)
        : CImageType<TYPE> (w, h, NCHANNELS) { }


        /**********************************************************/
        /*!
         *   Constructor: create new image object with the
         *   specified dimensions, copy the passed in data.
         *
         *   Use this constructor only as an interface to existing
         *   code. New programs should use the other available
         *   constructors.
         *
         *   \param w width (in pixels)
         *   \param h length (in pixels)
         *   \param data the data that is used to set up this
         *               image
         */
        /**********************************************************/
        CImageTemplate (int w, int h, const TYPE* data)
            : CImageType<TYPE> (w, h, NCHANNELS)
        {
            memcpy ((TYPE*)(*d_img_data), data,
                w * h * d_nchannels * sizeof (TYPE));
        }


        /**********************************************************/
        /*!
         *   Constructor: copy an image to the local object.
         *   It is possible to share the image buffer memory
         *   between objects.
         *
         *   If the buffer is shared, the whole buffer of the
         *   original image is shared, i.e. resetting the window
         *   size results in access to the full image buffer for
         *   the specified object.
         *
         *   If the buffer is copied, only the active image region
         *   is copied. I.e. resetting the window size might have
         *   no effect.
         *
         *   \bug Check with requirements if the behaviour
         *        described above is correct.
         *
         *   \param x img image to be copied
         *   \param copy If true, then copy the image buffer
         *               between the image objects.
         *               If false, the buffer is shared.
         */
        /**********************************************************/
        CImageTemplate (const CImageTemplate<TYPE, COLORSPACE, CHANNELS> &img, bool copy = true)
            : CImageType<TYPE> (img, copy) {}


            /**********************************************************/
            /*!  \internal
             *
             *   Deletes the image data array, if it is not
             *   referenced by any other CImageByteRGB.
             */
            /**********************************************************/
            ~CImageTemplate ()
            {
                delete d_img_data;
            }

            /******************************************************/
            //!  Cloned an existing image (C++ style).
            /******************************************************/
            inline CImage *clone(bool copycontent=true) const {
                return new CImageTemplate(*this,copycontent);
            }

            /**********************************************************/
            /*!
             *   Inserts an existing image in this image object.
             *   The specified image is copied into this image.
             *
             *   \param img source image
             *   \param x x coordinate of start point
             *   \param y y coordinate of start point
             */
            /**********************************************************/
            inline void insertSubImage (const CImageTemplate &img, int x, int y) {
                CImageType<TYPE>::insertSubImage (img, x, y);
            }

            /**********************************************************/
            /*!
             *   Check if this image is equal to another image.
             *   The dimensions and all values of the image are
             *   compared.
             *
             *   \param  img the image to be compared with
             *   \return "true" if the images are equal, "false" if
             *           the images are not equal
             */
            /**********************************************************/
            inline bool operator== (const CImageTemplate &img) const
            {
                return (CImageType<TYPE>::isEqual (img));
            }


            /**********************************************************/
            /*!
             *   Check if this image is not equal to another image.
             *   The dimensions and all values of the image are
             *   compared.
             *
             *   \param img the image to be compared with
             *
             *   \return "false" if the images are equal, "true" if
             *                       the image are not equal
             */
            /**********************************************************/
            inline bool operator!= (const CImageTemplate &img) const
            {
                return (CImageType<TYPE>::isUnequal (img));
            }

private:

    /******************************************************/
    //! Prohibits assignment to this class.
    /******************************************************/
    CImageTemplate& operator= (const CImageTemplate &);
};



} // namespace TUVision

// Parameter factories have to be in namespace vigra for compatibility
namespace vigra{
    using namespace TUVision;
    
template < class TYPE, class COLORSPACE, int CHANNELS >
    inline vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::ConstAccessor >
    srcImageRange(CImageTemplate< TYPE, COLORSPACE, CHANNELS > const & img)
{
    return vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::ConstAccessor >
        (img.upperLeft(), img.lowerRight(), img.accessor());
}

template < class TYPE, class COLORSPACE, int CHANNELS, class Accessor >
    inline vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
    Accessor >
    srcImageRange(CImageTemplate< TYPE, COLORSPACE, CHANNELS > const & img, Accessor &a)
{
    return vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS>::const_traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS>::const_traverser,
        Accessor >
        (img.upperLeft(), img.lowerRight(), a);
}

template < class TYPE, class COLORSPACE, int CHANNELS >
    inline std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::ConstAccessor >
    srcImage(CImageTemplate< TYPE, COLORSPACE, CHANNELS > const & img)
{
    return std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::const_traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::ConstAccessor >
        (img.upperLeft(), img.accessor());
}

template < class TYPE, class COLORSPACE, int CHANNELS, class Accessor >
    inline std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS>::const_traverser,
    Accessor >
    srcImage(CImageTemplate< TYPE, COLORSPACE, CHANNELS > const & img, Accessor & a)
{
    return std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS>::const_traverser,
        Accessor>
        (img.upperLeft(), a);
}

template < class TYPE, class COLORSPACE, int CHANNELS >
    inline vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::Accessor>
    destImageRange(CImageTemplate< TYPE, COLORSPACE, CHANNELS > & img)
{
    return vigra::triple<typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::Accessor>
        (img.upperLeft(), img.lowerRight(), img.accessor());
}

template < class TYPE, class COLORSPACE, int CHANNELS, class Accessor >
    inline vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    Accessor >
    destImageRange(CImageTemplate< TYPE, COLORSPACE, CHANNELS > & img, Accessor &a )
{
    return vigra::triple< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        Accessor >
        (img.upperLeft(), img.lowerRight(), a);
}


template < class TYPE, class COLORSPACE, int CHANNELS >
    inline std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::Accessor >
    destImage(CImageTemplate< TYPE, COLORSPACE, CHANNELS > & img)
{
    return std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::Accessor>
        (img.upperLeft(), img.accessor());
}

template < class TYPE, class COLORSPACE, int CHANNELS, class Accessor >
    inline std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
    Accessor >
    destImage(CImageTemplate< TYPE, COLORSPACE, CHANNELS > & img, Accessor & a)
{
    return std::pair< typename CImageTemplate< TYPE, COLORSPACE, CHANNELS >::traverser,
        Accessor >
        (img.upperLeft(), a);
}

}

#endif // __TUVISION_CIMAGETEMPLATE_H__
