/***************************************************************************
                          CVertex.h  -  description
                             -------------------
    begin                : 18.05.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_CVERTEX_H__
#define __TUVISION_CVERTEX_H__ 1

#include "TUVisionDef.h"
#include <cmath>
#include "newmat/newmat.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

namespace TUVision {

////////////////////////////////////////////////////////////
// NOTE:
//
// This template file does not use the DLL-Export via
// _CCTUVISION. Otherwise each template had to be 
// instantiated (e.g. for a CPoint3d of type double: 
// template class __declspec(dllexport) CPoint3D<double>;
//)

/**********************************************************/ 
//!  Defines a twodimensional vector
/*! 
 *   This class defined a 2D vector which can
 *   be used for mathematical computations. While there has
 *   been a discussion among some developers a year ago
 *   whether the coordinates of a CPoint should be able to be
 *   added, CPoint2D is for general purpose.
 *
 *   \author Fabian Wenzel
 */

template< class T > class CPoint2D
{
public:
    CPoint2D() { x=0; y=0; }
    CPoint2D (T newx, T newy) { x=newx; y=newy; }
    CPoint2D (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); }
    void set(T newx, T newy) { x=newx; y=newy; }
    CPoint2D &operator= (const CPoint2D<T>  &rhs) { x=rhs.x; y=rhs.y; return *this; }
    CPoint2D &operator= (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); return *this; }
    const bool operator== (const CPoint2D  &pt) const { return (x==pt.x && y==pt.y); }
    const bool operator!= (const CPoint2D  &pt) const { return (x!=pt.x || y!=pt.y); }
    const bool operator < (const CPoint2D  &pt) const { return abs() < pt.abs(); }
    CPoint2D& operator*= (const double factor) { x*=(T)factor; y*=(T)factor; return *this; }
    CPoint2D& operator/= (const double factor) { x=(T) ((double)x/factor); y=(T) ((double)y/factor); return *this; }
    CPoint2D& operator-=(const CPoint2D<T> & p) { x-=p.x; y-=p.y; return *this; }
    CPoint2D& operator+=(const CPoint2D<T> & p) { x+=p.x; y+=p.y; return *this; }
    CPoint2D operator+(const CPoint2D<T> & r) const { return CPoint2D<T>(x+r.x,y+r.y); }
    CPoint2D operator-(const CPoint2D<T> & r) const { return CPoint2D<T>(x-r.x,y-r.y); }
    CPoint2D operator*(const double factor) const { return CPoint2D<T>((T) (x*factor),(T)(y*factor)); }
    CPoint2D operator/(const double factor) const { return CPoint2D<T>((T) ((double)x/factor), (T) ((double)y/factor )); }
    NEWMAT::ColumnVector toColumnVector() const { NEWMAT::ColumnVector result(2); result << x << y; return result; }
    double abs() const { return (T) sqrt( (double) x*x + y*y ); }
    double dot(const CPoint2D<T> & p) const { return x*p.x + y*p.y; }
    double arg() const { return (T) atan((double)y/ (double) x)+((x<0)?M_PI:0); }
    CPoint2D& normalize() { *this/=abs(); return *this; }
    T x;
    T y;
};

typedef CPoint2D< int    > CVertex2i;
typedef CPoint2D< float  > CVertex2f;
typedef CPoint2D< double > CVertex2d;

/**********************************************************/ 
//!  Defines a threedimensional vector
/*! 
 *   This class defined a 3D vector which can
 *   be used for mathematical computations.
 *
 *
 *   \author Fabian Wenzel
 */

template< class T > class CPoint3D : public CPoint2D<T>
{
public:
    using CPoint2D<T>::x;
    using CPoint2D<T>::y;
    CPoint3D() : CPoint2D<T>(), z(0) {}
    CPoint3D (T newx, T newy, T newz) : CPoint2D<T>(newx,newy), z(newz) {}
    CPoint3D (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); z=rhs(3);}
    void set(T newx, T newy, T newz) { CPoint2D<T>::set(newx,newy); z=newz; }
    CPoint3D &operator= (const CPoint3D &rhs) { x=rhs.x; y=rhs.y; z=rhs.z; return *this; }
    CPoint3D &operator= (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); z=rhs(3); return *this; }
    const bool operator== (const CPoint3D &pt) const { return (x==pt.x && y==pt.y && z==pt.z); }
    const bool operator!= (const CPoint3D &pt) const { return (x!=pt.x || y!=pt.y || z!=pt.z); }
    const bool operator < (const CPoint3D &pt) const { return abs() < pt.abs(); }
    CPoint3D& operator*= (const double factor) { x=(T) ((double)x*factor); y=(T) ((double)y*factor); z=(T) ((double)z*factor); return *this; }
    CPoint3D& operator/= (const double factor) { x=(T) ((double)x/factor); y=(T) ((double)y/factor); z=(T) ((double)z/factor); return *this; }
    CPoint3D& operator-=(const CPoint3D& p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    CPoint3D& operator+=(const CPoint3D& p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    CPoint3D  operator+ (const CPoint3D& r) const { return CPoint3D(x+r.x,y+r.y,z+r.z); }
    CPoint3D  operator- (const CPoint3D& r) const { return CPoint3D(x-r.x,y-r.y,z-r.z); }
    CPoint3D cross(const CPoint3D& other) const { return CPoint3D((y*other.z)-(z*other.y),(z*other.x)-(x*other.z),(x*other.y)-(y*other.x)); }
    CPoint3D operator*(const double factor) const { return CPoint3D((T) (x*factor),(T) (y*factor),(T) (z*factor)); }
    CPoint3D operator/(const double factor) const { return CPoint3D((T) ((double)x/factor),(T) ((double)y/factor),(T) ((double)z/factor)); }
    NEWMAT::ColumnVector toColumnVector() const { NEWMAT::ColumnVector result(3); result << x << y << z; return result; }
    double abs() const { return (T) sqrt( (double) x*x + y*y +z*z); }
    double dot(const CPoint3D& p) const { return x*p.x + y*p.y +z*p.z; }
    CPoint3D &normalize() { *this/=abs(); return *this; }
    T z;
};

typedef CPoint3D< int    > CVertex3i;
typedef CPoint3D< float  > CVertex3f;
typedef CPoint3D< double > CVertex3d;


/**********************************************************/ 
//!  Defines a fourdimensional vector
/*! 
 *   This class defined a 4D vector which can
 *   be used for mathematical computations.
 *
 *
 *   \author Thomas Beilicke
 */

template< class T > class CPoint4D : public CPoint3D<T>
{
public:
    using CPoint3D<T>::x;
    using CPoint3D<T>::y;
    using CPoint3D<T>::z;
    CPoint4D() : CPoint3D<T>(), t(0) {}
    CPoint4D (T newx, T newy, T newz, T newt) : CPoint3D<T>(newx,newy,newz), t(newt) {}
    CPoint4D (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); z=rhs(3); t=rhs(4); }
    void set(T newx, T newy, T newz, T newt) { CPoint3D<T>::set(newx,newy,newz); t=newt; }
    CPoint4D &operator= (const CPoint4D &rhs) { x=rhs.x; y=rhs.y; z=rhs.z; t=rhs.t; return *this; }
    CPoint4D &operator= (const NEWMAT::ColumnVector &rhs) { x=rhs(1); y=rhs(2); z=rhs(3); t=rhs(4); return *this; }
    const bool operator== (const CPoint4D &pt) const { return (x==pt.x && y==pt.y && z==pt.z && t==pt.t); }
    const bool operator!= (const CPoint4D &pt) const { return (x!=pt.x || y!=pt.y || z!=pt.z || t!=pt.t); }
    const bool operator < (const CPoint4D &pt) const { return abs() < pt.abs(); }
    CPoint4D& operator*= (const double factor) { x=(T) ((double)x*factor); y=(T) ((double)y*factor); z=(T) ((double)z*factor); t=(T) ((double)t*factor); return *this; }
    CPoint4D& operator/= (const double factor) { x=(T) ((double)x/factor); y=(T) ((double)y/factor); z=(T) ((double)z/factor); t=(T) ((double)t/factor); return *this; }
    CPoint4D& operator-=(const CPoint4D& p) { x-=p.x; y-=p.y; z-=p.z; t-=p.t; return *this; }
    CPoint4D& operator+=(const CPoint4D& p) { x+=p.x; y+=p.y; z+=p.z; t+=p.t; return *this; }
    CPoint4D  operator+ (const CPoint4D& r) const { return CPoint4D(x+r.x,y+r.y,z+r.z,t+r.t); }
    CPoint4D  operator- (const CPoint4D& r) const { return CPoint4D(x-r.x,y-r.y,z-r.z,t-r.t); }
    CPoint4D operator*(const double factor) const { return CPoint4D((T) (x*factor),(T) (y*factor),(T) (z*factor),(T) (t*factor)); }
    CPoint4D operator/(const double factor) const { return CPoint4D((T) ((double)x/factor),(T) ((double)y/factor),(T) ((double)z/factor),(T) ((double)t/factor)); }
    double abs() const { return (T) sqrt( (double) x*x + y*y + z*z + t*t); }
    double dot(const CPoint4D& p) const { return x*p.x + y*p.y + z*p.z + t*p.t; }
    NEWMAT::ColumnVector toColumnVector() const {NEWMAT::ColumnVector result(4); result << x << y << z << t; return result;}
    CPoint4D &normalize() { *this/=abs(); return *this; }
    T t;
};

typedef CPoint4D< int    > CVertex4i;
typedef CPoint4D< float  > CVertex4f;
typedef CPoint4D< double > CVertex4d;

}

#endif
