/***************************************************************************
                          CCurveSearch.h  -  description
                             -------------------
    begin                : 1.1.2000
    copyright            : (C) 2000 by Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
//CCurveSearch.h
#ifndef __TUVISION_CCURVESEARCH_H_
#define __TUVISION_CCURVESEARCH_H_ 1

#include "TUVisionDef.h"
#include "tuv/CImageDatatypes.h"


#include <vector>
#include "tuv2d/CPointXY.h"
#include "tuv2d/CCurve.h"
#include "tuv2d/CShape.h"

namespace TUVision {

/**********************************************************/
/*! 
 * This class should be the output of the curve search
 * method implemented in CCurveSearch class.
 *
 *   \ingroup extractors
 * \author Islam Shdaifat
 */
/**********************************************************/
class _CCTUVISION COut {
public:
    int x;       //! the x coordinate of the curve start
    int y;       //! the y coordinate of the curve start
    int xf;       //! the x coordinate of the curve start
    int yf;       //! the y coordinate of the curve start
    int Hits;    //! the hit points of the curve
    int NonHits; //! the non hit points of the curve
    CCurve *crv; //! apointer to the curve itself

    int condition; //! this variable is used for post processing
    double vote; //! the persentage that the curve simmilar to the edge image
};

/**********************************************************/
/*! 
 *    This class implements the search method for a curve
 *    in a binary image i.e edge image. The edge image is
 *    definded as 0 no edge and 255 edge.
 *    d_image is a pointer to the image object to be search.
 *    d_curve is a pointer to the Curve object to be Search for
 *    see definition of CCurve
 *    The curve is definded by a vector of x,y points.
 *    For each point in the image, the curve anchor is set
 *    to the first point in the curve. We run across 
 *    the curve points one by one and in the image if sufficient
 *    overlap between the curve and the two was found, we
 *    regester the point and the anchor point. If withen the
 *    search the non hit points were greater than a certain 
 *    threshold we break it and continue to with the next
 *    anchor of the curve.
 *
 *    \author Islam Shdaifat
 */
/**********************************************************/
class _CCTUVISION CCurveSearch: public CShape {

public:

    //! default constructor
    CCurveSearch(){delCurveFromImge = false;};

    virtual ~CCurveSearch();

    //! construct a CurveSearch object
    CCurveSearch(std::vector< CCurve > vecSrCrs,
        double dThr0 = 0.2 ,
        double dThr1 = 0.8);

    //! construct a CurveSearch object
    CCurveSearch(CShape& shape,
        double dThr0 = 0.2,
        double dThr1 = 0.8);

    //! insert new shape in the shape pattern i.e in vecCurvesLUT
    void insertShape(CShape& shape);

    //! set the current search image
    void setSearchInImage( CImageByteGray &img);

    //! set the current search curve
    void virtual setSearchCurveVector( CCurve *crv);

    //! set the thresholds of serch
    void setThreshold(double Thr0, double Thr1 );

    //! set the curves start iterator
    void setItCurveStart(std::vector< CCurve >::iterator it_s) {
        itCurveStart = it_s;
    };

    //! set the curves end iterator
    void setItCurveEnd(std::vector< CCurve >::iterator it_e) {
        itCurveEnd = it_e;
    };

    //! draw a curve in the image at x,y
    void drawCurveInImage(CImageByteGray& img, CCurve& crv,
        int x, int y, int color = 128);

    //! search the curves in the Image
    virtual void search(CImageByteGray* img, int dx = 0, int dy = 0);

    //! specialized search for each kind of curves
    /*!
     * specialized search for each kind of curves, should be
     * implemented in the inheritor class, cause e.g. search for
     * line curves is different than for the circle
     */
    virtual void searchSpecial() = 0;

    //! iterates through curves if more than one curve
    virtual void checkCurves_itStart_To_itEnd();

    //! gives the count of hit and non hit point
    virtual void getHitAndNonHitCount();

    //! the vector of the output see defintion. of COut
    std::vector<COut> vecSearchOut;

    bool delCurveFromImge;

protected:

    //! iterator ponits to a curve being search for
    std::vector< CCurve >::iterator d_it_crv;

    //! counter for non hit points
    int d_Hit;

    //! counter for hit points
    int d_NonHit; 

    //! current position of the curve origin in the image
    int d_x , d_y;

    /**********************************************************/
    /*! 
     * check Special Condition, this may be definded by the user latter
     * this helps to break the search if a condition is not full filled
     * also the user can set the group of curves to search for
     * if a condition is known in advance the user can restrect
     * the search for only few curves
     */
    /**********************************************************/
    virtual bool checkSpecialCondition () {
        return true;
    };

    //! check the Hit and Non Hit counts whether they are beyond the thresholds
    bool isSuffecientHitPoints();

    //! write the result of the search for a curve to the output stream
    virtual void writeToFeatureStream();

    //! a pointer to the image to search in
    CImageByteGray* d_image;

    //! the begin of the curves
    std::vector< CCurve >::iterator itCurveStart;

    //! the the end of the curves
    std::vector< CCurve >::iterator itCurveEnd;

    //! a pointer to the current curve to be serched for in the image
    CCurve *d_curve;

    //! max allowed NON HIT points, count 0's
    int d_Max_NonHitCount;

    //! min allowed HIT points, count 1's
    int d_Min_HitCount;

};

} // TUVision

#endif  // __TUVISION_CCURVESEARCH_H_
