/***************************************************************************
                           CHoughLineSearch.h  -  description
                              -------------------
     begin                : 30.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Diploma thesis Marc Hensel
 
***************************************************************************/

#ifndef __TUVISION_CHOUGHLINESEARCH_H__
#define __TUVISION_CHOUGHLINESEARCH_H__	1

#include <tuv/CImageByteGray.h>

#include <TUVisionDef.h>

namespace TUVision 
{

/**********************************************************/
//!  Hough transform line search.
/*! 
 *   This class implements the hough transform line search.
 *
 *   Lines are represented by the parameters s (distance to
 *   origin) and \f$ \theta \f$ ('theta'; angle between distance
 *   and x-axes):
 *
 *   \f[
 *     s = x \cdot \cos\theta + y \cdot \sin\theta.
 *   \f]
 *
 *   The internal variables are suited to store the resulting
 *   lines in the arrays s and theta. However, at this point
 *   of development only the line with max. value in hough
 *   space is stored, i.e. only results s[0] and theta[0]
 *   (result_nr = 1) are available.
 *
 *   NOTE: In most applications an edge detection filter as 
 *   e.g. sobel or must be used to create an edge image 
 *   BEFORE calling the method createHoughImage.
 *
 *   Example code (extract):
 *   
 *   \code
 *   #include <tuv2d/CHoughLineSearch.h>
 *   ...
 *   bool method_name(CImageByteGray &img)
 *   {
 *		CHoughLineSearch	hough;
 *		double s,theta;
 *   	if (!hough.createHoughImage(img,128))	return false;
 *   	if (!hough.findMaxLine())	return false;
 *		hough.getLineAngle(1,theta);
 *		hough.getLineDistance(1,s);
 *   	...
 *   	return true;
 *   }
 *   \endcode
 *
 *   \ingroup extractors
 *
 *   \author Marc Hensel
 *
 *   \class CHoughLineSearch CHoughLineSearch.h tuv2d/CHoughLineSearch.h
 */
/**********************************************************/

class _CCTUVISION CHoughLineSearch
{
private:
    // Hough space data
    int				*hough_space;	//!< Hough space data: Pointer matrix (row sorted array) holding hough space data
    int				hwidth;			//!< Hough space data: Width of hough space data matrix
    int				hheight;		//!< Hough space data: Height of hough space data matrix
    // Coordinate transform to image center (reduces hought data dimension hheight)
    int				s_max;			//!< Image coordinate transform: Maximum distance to centered origin in original image (Needed for hought matrix height)
    int				center_x;		//!< Image coordinate transform: x-coordinate of center of image data matrix
    int				center_y;		//!< Image coordinate transform: y-coordinate of center of image data matrix
    // Result: Line parameters ("Hessesche Normalform")
    double			*s;				//!< Result: Distance(s) of line to origin
    double			*theta;			//!< Result: Angle(s) of normal vector of line to x-axes in rad (x-axes == 0; y-axes == pi/2; ...)
    unsigned char	valid_results;	//!< Number of valid results in *s and *theta

public:
    //! Constructor
    CHoughLineSearch	();

    //! Destructor
    ~CHoughLineSearch	();


    /**********************************************************
     *
     * Readout variables and results
     *
     **********************************************************/

    int				*getHoughSpaceData				()	{ return hough_space; }		//!< Returns pointer to hough soace data array
    int				getHoughSpaceWidth				()	{ return hwidth; }			//!< Returns width of hough space data
    int				getHoughSpaceHeight				()	{ return hheight; }			//!< Returns height of hough space data
    int				getCenteredImageMaxDistance		()	{ return s_max; }			//!< Returns maximum distance to centered origin in original image
    int				getImageCenterX					()	{ return center_x; }		//!< Returns x-coordinate of center of image data matrix
    int				getImageCenterY					()	{ return center_y; }		//!< Returns y-coordinate of center of image data matrix
    int				getMaxDistance					()	{ return s_max; }			//!< Returns maximum distance to centered origin in original image
    unsigned char	getNumberOfValidResults			()	{ return valid_results; }	//!< Returns number of valid results in *s and *theta

    bool			getLineDistance	(unsigned char result_nr,double &distance);		//!< Returns distance s of line number result_nr to origin
    bool			getLineAngle	(unsigned char result_nr,double &angle);		//!< Returns angle theta of normal vector of line number result_nr to x-axes


    /**********************************************************
     *
     * Create results (line parameters)
     *
     **********************************************************/

    //! Create the hough space data array
    /*!
     *  Pixels with value >= threshold are regarded as belonging to an edge.
     *
     *  The return value is 'false' if hough data memory could not be allocated.
     *
     *  NOTE: In most applications an edge detection filter as e.g. sobel or 
     *  must be used to create an edge image BEFORE calling createHoughImage.
     *
     *  \param	image	Input image (to be searched for line)
     *	\param	threshold	Pixel threshold
     */
    bool	createHoughImage	(CImageByteGray *image,unsigned char threshold);

    //! Search hough space data for line with maximum value
    /*!
     *  If no valid hough space data exists 'false' is returned. Otherwise 
     *  'true' is returned after storing the result (line number = 1) in s[0]
     *  and theta[0].
     */
    bool	findMaxLine			();

    //! Improve line parameters by subpixel maximum location in hough space.
    /*!
     *  Line distance and angle are replaced by center of gravity in 5x5 
     *  element window in hough space data matrix. Window size is reduced 
     *  at matrix boarders.
     *
     *  \param	result_nr	Line number; elements in *s and *theta (Count starting with 1)
     */
    bool	improveLineSubpixel	(unsigned char result_nr);


    /**********************************************************
     *
     * Evaluate results
     *
     **********************************************************/

    //! Evaluate line function.
    /*!
     *  If no valid hough space data exists 'false' is returned. Otherwise 
     *  'true' is returned after storing \f$ y_0 = f(x_0) \f$, where f is
     *  the line described by parameter/result set result_nr.
     *
     *  \param	result_nr	Line number; elements in *s and *theta (Count starting with 1)
     *  \param	x0	x-coordinate \f$ x_0 \f$ to be evaluated
     *  \param	y0	Resulting y-coordinate \f$ y_0 \f$
     */
    bool			getFunctionOfX	(unsigned char result_nr,double x0,double &y0);

    //! Evaluate line function.
    /*!
     *  If no valid hough space data exists 'false' is returned. Otherwise 
     *  'true' is returned after storing \f$ x_0 = f^{-1}(y_0) \f$, where f
     *  is the line described by parameter/result set result_nr.
     *
     *  \param	result_nr	Line number; elements in *s and *theta (Count starting with 1)
     *  \param	y0	y-coordinate \f$ y_0 \f$ to be evaluated
     *  \param	x0	Resulting x-coordinate \f$ x_0 \f$
     */
    bool			getFunctionOfY	(unsigned char result_nr,double y0,double &x0);

    //! Draw line.
    /*!
     *  Draw a line found in the hough space in an image.
     *
     *  \param	image	Image in which the line is to be drawn
     *  \param	result_nr	Line number; elements in *s and *theta (Count starting with 1)
     *  \param	gray_intensity	Pixel value of line luminance (in [0,255], where 0 is black and 255 is white)
     */
    bool	drawLine			(CImageByteGray *image,unsigned char result_nr,unsigned char gray_intensity);
};

}	// namespace TUVision


#endif	// __TUVISION_CHOUGHLINESEARCH_H__

