/***************************************************************************
                          COperatorFFT.h  -  description
                             -------------------
    begin                : 27.11.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de
    origin               : Bildverarbeitung 1, WS 2001/2002

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_COPERATORFFT_H__
#define __TUVISION_COPERATORFFT_H__ 1

#include "TUVisionDef.h"
#include "tuv/CImageDatatypes.h"
#include <vector>
#include <complex>
#include <cmath>

#ifndef M_PI
#define M_PI           3.14159265358979323846  /* pi */
#endif

namespace TUVision {

/**********************************************************/
//!  Computes FFT of an image
/*!
 *   This operator computes the FFT and IFFT of grayscale
 *   images. The underlying algorithm is the Radix-2-FFT
+   with decimation in time, this means that input images
*   will be resized to a width and height of a power of 2
*   if they are smaller.
*
*   \ingroup operator
*
*   \author Fabian Wenzel
*
*   \class COperatorFFT COperatorFFT.h tuv2d/COperatorFFT.h
*/
/**********************************************************/

class _CCTUVISION COperatorFFT{
public:

    //! Constructor does nothing special
    COperatorFFT();

    //! Destructor does nothing special
    ~COperatorFFT();

    //! Computes the FFT of a double grayscale image
    CImageComplexGray fft(CImageDoubleGray &image,
        int output_width=0,
        int output_height=0);

    //! Computes the FFT of a byte grayscale image
    CImageComplexGray fft(CImageByteGray &image,
        int output_width=0,
        int output_height=0);

    //! Computes the FFT of a complex grayscale image
    CImageComplexGray fft(CImageComplexGray &image,
        int output_width=0,
        int output_height=0);

    //! Computes the IFFT of a complex grayscale image
    CImageComplexGray ifft(CImageComplexGray &image,
        int output_width=0,
        int output_height=0);

    //! Shifts an image such that the zero frequency will be centered
    static bool fftshift(CImageComplexGray &image);

private:

    //! Common code for FFT and IFFT: Does a columnwise 1D FFT of an image
    CImageComplexGray *fft_1d_main(CImageComplexGray &srcimage);

    CImageComplexGray fft_main(CImageComplexGray &srcimage);

    //! Initializes bitreversal table for bitreversed input values
    void init_bitreversal(int width, std::vector< int > &bitrevsal_table);

    //! Initializes twiddle table with complex coefficients
    void init_twiddletable(int width,
        std::vector< std::complex<double> > &twiddle_table);

    //! Does the core buttlerfly operation
    inline void butterfly_in_place(std::complex<double> *firstwing,
        std::complex<double> *secondwing,
        const std::complex<double> &twiddle);

    //! Helper function: Computes floor(ld(value))
    static inline int ld_floor(int value);

    //! Helper function: Computes ceil(ld(value))
    static inline int ld_ceil(int value);

    //! Helper function: Checks if value is power of 2
    static inline bool is_ld(int value);

    //! Does in-place FFT operation
    inline void fft_in_place_core(int width,std::complex<double> *row,
        std::vector< std::complex<double> > &twiddles);
};

}
#endif //__TUVISION_COPERATORFFT_H__
