/***************************************************************************
                          COperatorMultiplex.h  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_COPERATORMULTIPLEX_H__
#define __TUVISION_COPERATORMULTIPLEX_H__ 1

#include "TUVisionDef.h"
#include "tuv/CImageDatatypes.h"
#include <vector>

namespace TUVision {

/**********************************************************/
//!  Multiplex and demultiplex a CImage.
/*! 
 *   Sometimes it is necessary to extract one color 
 *   channel from an image with several color channels
 *   or to combine several images with only one color
 *   channel to one image with several color channels.
 *   
 *   This class contains a multiplexer and a demultiplexer
 *   for this purpose.
 *   
 *   \ingroup operator
 *
 *   \author Marco Budde
 *   
 */
/**********************************************************/

class  _CCTUVISION COperatorMultiplex {

public:

    /******************************************************/
    //! Demultiplex a byte RGB image.
    /******************************************************/
    void demultiplex (const CImageByteRGB &src,
        std::vector <CImageByteGray*> &dst) const;

    /******************************************************/
    //! Demultiplex an integer RGB image.
    /******************************************************/
    void demultiplex (const CImageIntRGB &src,
        std::vector <CImageIntGray*> &dst) const;

    /******************************************************/
    //! Demultiplex a double RGB image.
    /******************************************************/
    void demultiplex (const CImageDoubleRGB &src,
        std::vector <CImageDoubleGray*> &dst) const;


    /******************************************************/
    //! Demultiplex a byte YUV image.
    /******************************************************/
    void demultiplex (const CImageByteYUV &src,
        std::vector <CImageByteGray*> &dst) const;

    /******************************************************/
    //! Demultiplex an integer YUV image.
    /******************************************************/
    void demultiplex (const CImageIntYUV &src,
        std::vector <CImageIntGray*> &dst) const;

    /******************************************************/
    //! Demultiplex a double YUV image.
    /******************************************************/
    void demultiplex (const CImageDoubleYUV &src,
        std::vector <CImageDoubleGray*> &dst) const;


    /******************************************************/
    //! Multiplex a byte RGB image.
    /******************************************************/
    void multiplex (const std::vector <CImageByteGray*> &src,
        CImageByteRGB **dst) const;

    /******************************************************/
    //! Multiplex an integer RGB image.
    /******************************************************/
    void multiplex (const std::vector <CImageIntGray*> &src,
        CImageIntRGB **dst) const;

    /******************************************************/
    //! Multiplex a double RGB image.
    /******************************************************/
    void multiplex (const std::vector <CImageDoubleGray*> &src,
        CImageDoubleRGB **dst) const;


    /******************************************************/
    //! Multiplex a byte YUV image.
    /******************************************************/
    void multiplex (const std::vector <CImageByteGray*> &src,
        CImageByteYUV **dst) const;

    /******************************************************/
    //! Multiplex an integer YUV image.
    /******************************************************/
    void multiplex (const std::vector <CImageIntGray*> &src,
        CImageIntYUV **dst) const;

    /******************************************************/
    //! Multiplex a double YUV image.
    /******************************************************/
    void multiplex (const std::vector <CImageDoubleGray*> &src,
        CImageDoubleYUV **dst) const;


    /******************************************************/
    //! Get one channel of a byte image.
    /******************************************************/
    CImageByteGray *getChannel (const CImageByte &src,
        int nchannels,
        int channel) const;

    /******************************************************/
    //! Get one channel of an integer image.
    /******************************************************/
    CImageIntGray *getChannel (const CImageInt &src,
        int nchannels,
        int channel) const;

    /******************************************************/
    //! Get one channel of a double image.
    /******************************************************/
    CImageDoubleGray *getChannel (const CImageDouble &src,
        int nchannels,
        int channel) const;


    /******************************************************/
    //! Insert image as color channel.
    /******************************************************/
    void setChannel (const CImageByteGray &src, CImageByte &dst,
        int nchannels, int channel) const;

    /******************************************************/
    //! Insert image as color channel.
    /******************************************************/
    void setChannel (const CImageIntGray &src, CImageInt &dst,
        int nchannels, int channel) const;

    /******************************************************/
    //! Insert image as color channel.
    /******************************************************/
    void setChannel (const CImageDoubleGray &src, CImageDouble &dst,
        int nchannels, int channel) const;


    /******************************************************/
    //! Demultiplex a byte image.
    /******************************************************/
    void demultiplexGeneric (const CImageByte &src,
        std::vector <CImageByteGray*> &dst,
        int nchannels) const;

    /******************************************************/
    //! Demultiplex an integer image.
    /******************************************************/
    void demultiplexGeneric (const CImageInt &src,
        std::vector <CImageIntGray*> &dst,
        int nchannels) const;

    /******************************************************/
    //! Demultiplex a double image.
    /******************************************************/
    void demultiplexGeneric (const CImageDouble &src, 
        std::vector <CImageDoubleGray*> &dst,
        int nchannels) const;


    /******************************************************/
    //! Multiplex a byte image.
    /******************************************************/
    void multiplexGeneric 
        (const std::vector <CImageByteGray*> &src,
        CImageByte &dst) const;

    /******************************************************/
    //! Multiplex an integer image.
    /******************************************************/
    void multiplexGeneric 
        (const std::vector <CImageIntGray*> &src,
        CImageInt &dst) const;

    /******************************************************/
    //! Multiplex a double image.
    /******************************************************/
    void multiplexGeneric 
        (const std::vector <CImageDoubleGray*> &src,
        CImageDouble &dst) const;


    /******************************************************/
    //! Test the dimension of the images and the number
    //! of images in the vector.
    /******************************************************/
    void testDimensions (const std::vector <CImage*> &src,
        int nchannels) const;
};
} 
#endif // __TUVISION_COPERATORMULTIPLEX_H__
