/***************************************************************************
                           CCalibrationGapDetection.h  -  description
                              -------------------
     last modified        : 30.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel
 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_CCALIBRATIONGAPDETECTION_H__
#define __TUVISION_CCALIBRATIONGAPDETECTION_H__	1

/*
 * +-----------------------+
 * | DEMONSTRATOR PROGRAM! |
 * +-----------------------+
 *
 * Note: Private methods can be defined as public so that single
 * processing steps can be visualized by the demonstrator program.
 */
#define	ALLOW_TO_ACCESS_PRIVATE_CCALIBRATIONGAPDETECTION_METHODS	0

#include <cmath>								// sin(), cos(), sqrt()
#include <tuv/CImageByteGray.h>
#include <tuv2d/CFilterSymmetricCovariance.h>	// Symmetric covariance filter
#include <tuv2d/CFilterSobel.h>					// Sobel filter (edge detection)
#include <tuv2d/COperatorThreshold.h>			// Threshold image
#include <tuv2d/CFilterSumInversion.h>			// Filter: Sums pixels, inverts and normalizes result

#include <TUVisionDef.h>

namespace TUVision 
{

/**********************************************************/
//!  Gap detection (pre calibration).
/*! 
 *   This class returns the image coordinates corresponding
 *   to world coordinates O(0,0,0), X(1,0,0), and Y(0,1,0)
 *   in 2D, and additionally Z(0,0,1) in 3D.
 *
 *   Thus single and multiple cameras find the same defined
 *   points as world coordinate system and can therefore be
 *   calibrated without manual user input.
 *
 *   The required 'gap calibration object' can be deriviated
 *   from a checker by simply swapping a black with a
 *   neighboring white square. The 3D object is folded
 *   rectangular two units left, i.e. in 'white direction',
 *   of the black/white edge in the calibration pattern.
 *
 *   Meaning of parameters/steps detecting the gap object:
 *   -# Symmetric covariance filter (symmetric corner image)
 *   -# Threshold points (amplify points in corner image)
 *   -# Sum inversion filter (high values in gaps)
 *   -# Threshold area (discard everything except areas in gaps)
 *   -# Pick biggest connected area as gap (discard points in center of checker squares)
 *
 *   Images after step (1) and (2) are also used to find the axes.
 *
 *	 Troubleshooting if gap detection fails:
 *   -# Make sure the image shows a gap detection calibration object
 *   -# Use the step by step demonstration of the demonstrator program pcX with your image
 *       to find out in which step(s) the detection parameter(s) need to be changed
 *	 -# Make use of the boolian return values (e.g. in a for-construct):
 *	     Choose hight sum inversion filter width and decrease (e.g. 17, 15, 13, ..., 5)
 *	 -# Symmetric covariance filter width:
 *	     Try 5 if calibration object is small and 7 if object is big in image
 *	 -# Arrange setting so that angle of view is not too steep in any direction
 *	 -# Try using an object with bigger square size"
 *
 *   Example code (Analyse 2D gap calibration object):
 *   
 *   \code
 *   #include "CCalibrationGapDetection.h"
 *   ...
 *   bool method_name(CImageByteGray &img)
 *   {
 *   	double Ox,Oy;  // Origin (0;0;0)
 *   	double Xx,Xy;  // X-axis (1;0;0)
 *   	double Yx,Yy;  // Y-axis (0;1;0)
 *   	CCalibrationGapDetection gap;
 *   	if (!gap.initGap(img)) return false;
 *   	if (!gap.findAxes2D(img)) return false;
 *   	gap.getOrigin(Ox,Oy);  // Read results
 *   	gap.getAxisX(Xx,Xy);
 *   	gap.getAxisY(Yx,Yy);
 *   	...
 *   	return true;
 *   }
 *   \endcode
 *
 *   \ingroup 3d
 *
 *   \author Marc Hensel
 *
 *   \class CCalibrationGapDetection CCalibrationGapDetection.h tuv3d/CCalibrationGapDetection.h
 */
/**********************************************************/

class _CCTUVISION CCalibrationGapDetection
{
private:
    int				fWidthSymCov;	//!< Width of symmetric covariance filter
    int				fWidthSumInv;	//!< Width of sum inversion filter
    unsigned char	thresPoints;	//!< Threshhold for amplifying corners (symmetric covariance filtered image)
    unsigned char	thresArea;		//!< Threshhold for discarding everything except max. gap areas (sum inversion filtered image)

    int		gap_left;		//!< Min. bounding box x-coordinate of detected gap area
    int		gap_right;		//!< Max. bounding box x-coordinate of detected gap area
    int		gap_top;		//!< Min. bounding box y-coordinate of detected gap area
    int		gap_bottom;		//!< Max. bounding box y-coordinate of detected gap area
    int		gap_x,gap_y;	//!< Image coordinates of starting position at black/white edge in gap

    double	dx,dy;			//!< Current direction (dx,dy)T in search for feature points (length:=1)

    double	Ox,Oy;			//!< Image coordinates of world system origin (0;0;0)
    double	Xx,Xy;			//!< Image coordinates of world system x-axis (1;0;0)
    double	Yx,Yy;			//!< Image coordinates of world system y-axis (0;1;0)
    double	Zx,Zy;			//!< Image coordinates of world system z-axis (0;0;1)

    /*
     * +-----------------------+
     * | DEMONSTRATOR PROGRAM! |
     * +-----------------------+
     *
     * Note: Private methods can be defined as public so that single
     * processing steps can be visualized by the demonstrator program.
     */
#ifdef ALLOW_TO_ACCESS_PRIVATE_CCALIBRATIONGAPDETECTION_METHODS
public:
#endif

    //! Rotate direction (dx,dy)T by angle (in rad)
    void			rotateDirection	(double angle);

    //! Detect gap area and return gap area image (gap area pixels=255, other pixels=0).
    CImageByteGray*	detectGap		(CImageByteGray &orig_img);
    //! Change pixel values of connected area containing (cx,cy).
    /*!
     *  Connected area means an area where each pixel has the same value.
     *  The method is used by detectGap() to find the biggest area.
     *
     *  \param	cx,cy	Point in area
     *	\param	new_value	Value each pixel in area will have after method returns
     *
     *	\return	Number of pixels in area
     */
    int				transposeArea	(CImageByteGray &img,int cx,int cy,unsigned char new_value);
    //! Find and set bounding box coordinates of detected gap using gap area image.
    bool			frameGap		(CImageByteGray &gap_img);
    //! Find and set starting position at black/white edge in gap using bounding box coordinates.
    /*!
     *  \param gap_image	Gap image (gap area pixels=255, other pixels=0)
     */
    void			initGapPosition	(CImageByteGray &orig_img);
    //! Find and set approx. direction (dx,dy)T of edge between black and white.
    /*!
     *  Method uses (gap_x,gap_y) and choses orientation so that the white area is on the
     *  left side of the motion. Vector (dx,dy)T is normalized to length 1.
     */
    void			initGapDirection(CImageByteGray &orig_img);

    //! Find next point in direction (dx, dy).
    /*!
     *  The point image can be created by symmetric covariance filtering followed by
     *  thresholding.
     *
     *  \param pnt_image	Point image (only pixels around corner points have value 255)
     */
    /*
     *  (1.a) Follow direction until point is detected (update direction vector)
     *  (1.b) Each step: Center position orthogonally to motion on edge image
     *  (2)   Delete origin and follow direction until y-axis is detected (update direction vector)
     */
    bool			findNextPoint	(CImageByteGray &edge_img,CImageByteGray &pnt_img,double x0,double y0,double &Px,double &Py);

public:
    //! Constructor
    /*!
     *  Sets all values to zero except for filter widths and thresholds:
     *
     *  fWidthSymCov = 7
     *  fWidthSumInv = 15
     *  thresPoints = 32
     *  thresArea = 245
     */
    CCalibrationGapDetection		();
    ~CCalibrationGapDetection		();							//!< Destructor

    void			setWidthSCF		(int width);				//!< Set width of symmetric covariance filter
    void			setWidthSIF		(int width);				//!< Set width of sum inversion filter
    void			setThreshPoints	(unsigned char threshold);	//!< Set threshold to amplify corners
    void			setThreshArea	(unsigned char threshold);	//!< Set threshold to discard except max. gap areas

    int				getWidthSCF		();							//!< Return width of symmetric covariance filter
    int				getWidthSIF		();							//!< Return width of sum inversion filter
    unsigned char	getThreshPoints	();							//!< Return threshold to amplify corners
    unsigned char	getThreshArea	();							//!< Return threshold to discard except max. gap areas

    void			getGapFrame		(int &left,int &right,int &top,int &bottom);	//!< Return bounding box coordinates of detected gap area
    void			getGapPosition	(int &x,int &y);								//!< Return starting position (gap_x,gap_y)T at black/white edge in gap
    void			getDirection	(double &x,double &y);							//!< Return direction (dx,dy)
    void			getOrigin		(double &x,double &y);		//!< Return image coordinates of world system origin (0;0;0)
    void			getAxisX		(double &x,double &y);		//!< Return image coordinates of world system x-axis (1;0;0)
    void			getAxisY		(double &x,double &y);		//!< Return image coordinates of world system y-axis (0;1;0)
    void			getAxisZ		(double &x,double &y);		//!< Return image coordinates of world system z-axis (0;0;1)

    //! Detect and analyses gap object.
    /*!
     *  Sets bounding box (gap_left, gap_right, gap_top, gap_bottom),
     *  starting position (gap_x,gap_y) and initial direction (dx,dy).
     */
    bool			initGap			(CImageByteGray &orig_img);

    //! Find origin (0;0;0), (1;0;0), and (0;1;0) in 2D gap calibration object.
    /*!
     *  Starting position (gap_x,gap_y)T and initial direction (dx,dy)T
     *  must have been set before by initGap().
     */
    bool			findAxes2D		(CImageByteGray &orig_img);
    //! Find origin (0;0;0), (1;0;0), (0;1;0), and (0;0;1) in 3D gap calibration object.
    /*!
     *  Starting position (gap_x,gap_y)T and initial direction (dx,dy)T
     *  must have been set before by initGap().
     */
    bool			findAxes3D		(CImageByteGray &orig_img);
};

}

#endif // __TUVISION_CCALIBRATIONGAPDETECTION_H__ 
