/***************************************************************************
                          CCameraMatrix.h  -  description
                             -------------------
    begin                : 12.03.2003
    copyright            : (C) 2003 by Thomas Beilicke
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Thomas Beilicke
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_CCAMERAMATRIX_H_
#define __TUVISION_CCAMERAMATRIX_H_ 1

#include "TUVisionDef.h"

#include "newmat/newmat.h"
#include "tuv/CVertex.h"
#include "tuv3d/CIntrinsicMatrix.h"
#include "tuv3d/CCorrespondence.h"
#include <vector>

namespace TUVision {

/***********************************************************/
//!  Class to contain a camera matrix.
/*! 
 *   This class is a NEWMAT matrix to contain a 3x4 camera
 *   matrix. It supports creation of a synthetic matrix as 
 *   well as assignment from an existing NEWMAT matrix.   
 *
 *   The intrinsic and extrinsic parameters can be calculated
 *   and points can be projected.   
 *
 *   \ingroup 3d
 * 
 *   \author Thomas Beilicke 
 */
/***********************************************************/

class _CCTUVISION CCameraMatrix : public NEWMAT::Matrix {
public:	
    // Exports for RANSAC : Its defining Datatype, codimension and minimum
    // count of building data
    // For camera calculation out of Worldpoint to imagepoint correspondences
    typedef std::vector< CCorrespondence > TDataVector;   
    enum { codimension=1 };
    enum { minimumdata=6 }; // 6 correspondences needed

    CCameraMatrix();	
    //!< The default constructor constructs a canonical camera matrix.
    CCameraMatrix(const CPoint3D<double> &origin, const CPoint3D<double> &lookat, 
        double aspect, double focal, double kx, double u0, double v0, double skew);
    //!< Constructs a synthetic matrix out of the given parameters.
    CCameraMatrix(const std::vector<CCorrespondence> &corresp, unsigned int index=0);
    //!< Calculates the camera matrix out of 6 world point - image point correspondences.
    CCameraMatrix(const NEWMAT::Matrix &P); 
    //!< Creates a new camera matrix out of an existing NEWMAT matrix.
    CCameraMatrix &operator=(const NEWMAT::Matrix &P);
    //!< Copies an existing NEWMAT matrix.

    void createCamera(const CPoint3D<double> &origin, const CPoint3D<double> &lookat, 
        double aspect, double focal, double kx, double u0, double v0, double skew);
    //!< Calculates a synthetic matrix out of the given parameters.
    void calculateCamera(const std::vector<CCorrespondence> &corresp, unsigned int index=0);
    //!< Calculates the camera matrix out of 6 world point - image point correspondences.
    static void setUpVector(const CPoint3D<double> &upvector);
    //!< Sets the up-vector for synthetical camera creation.

    NEWMAT::ColumnVector getCentre() const;
    //!< Calculates the camera centre position.
    NEWMAT::ColumnVector project(const NEWMAT::ColumnVector &point) const;
    //!< Projects a homogeneous point.
    NEWMAT::ColumnVector project(const CPoint4D<double> &point) const;
    //!< Projects a homogeneous point.
    void getIntrinsicRotation(CIntrinsicMatrix &K, NEWMAT::Matrix &R) const; 
    //!< Calculates the intrinsic matrix and the rotation matrix.
    void getIntrinsicExtrinsic(CIntrinsicMatrix &K, NEWMAT::Matrix &E) const;
    //!< Calculates the intrinsic and the extrinsic matrix.
    void getIntrinsic(CIntrinsicMatrix &K) const;
    //!< Calculates the intrinsic matrix.
    void getExtrinsic(NEWMAT::Matrix &E) const;
    //!< Calculates the extrinsic matrix.
    void getRotation(NEWMAT::Matrix &R) const;
    //!< Calculates the rotation matrix.
    NEWMAT::ColumnVector getPrincipalAxis() const;
    //!< Calculates the camera's principal axis.
    NEWMAT::ColumnVector getPrincipalPoint() const;
    //!< Calculates the camera's principal point.
    double depth(const NEWMAT::ColumnVector &worldpoint, const NEWMAT::ColumnVector &projectedpoint) const;
    //!< Calculates the depth of the worldpoint X from this camera. Use this directly after a projection x=P*X!
    double depth(const NEWMAT::ColumnVector &worldpoint) const;
    //!< Calculates the depth of the worldpoint X from this camera.

    double calculateDistance( const TDataVector::value_type &sample, unsigned int index=0);
    //!< Calculates the projection error for CRansac.

private:
    void rq(NEWMAT::Matrix &K, NEWMAT::Matrix &R) const;
    //!< Decomposes the camera's upper 3x3 submatrix into an upper right and an orthogonal matrix.

    static CVertex3d upVector;
};

}

#endif
