/***************************************************************************
                          CCorrespondence.h  -  description
                             -------------------
    begin                : 11.03.2003
    copyright            : (C) 2001 by Thomas Beilicke
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Thomas Beilicke

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_CCORRESPONDENCE_H__
#define __TUVISION_CCORRESPONDENCE_H__ 1

#include <vector>
#include "tuv/CVertex.h"
#include "tuv/CExceptionIllegalPoint.h"

#include "TUVisionDef.h"

namespace TUVision {
/**********************************************************/ 
//!  Defines a checking aspect for CCorrespondence
/*! 
 *   Usage: At the bottom of this h-file define your CCorrespondence
 *   class that you want to work with.
 *
 *   if you want do not want to have correpospondences that are checked
 *   with exceptions and so on and write
 *
 *   typedef CCorrespondenceBase CCorrespondence;
 *
 *   so that you can change easily between the two types.
 *   This aspect is written as a template such that more aspects
 *   can be added later, e.g. CCorrespondenceLogging< CCorrespondenceChecking< CCorrespondenceBase > >
 *
 *	
 *   \author Fabian Wenzel
 */

template <class Base> class CCorrespondenceChecking : public Base {
public:
    //! Default constructor
    CCorrespondenceChecking()
    {
    }

    //! Initializes the correspondence vector with a given size
    CCorrespondenceChecking(int size):
    Base(size)
    {
    }

    // --- image points --- //
    CPoint3D<double> getImagePoint(unsigned int index=0) const
    {
        if(index >= this->size())
            throw CExceptionIllegalPoint("Wrong index in correspondences");
        else
            return Base::getImagePoint(index);
    }

    CPoint2D<double> getEuclideanImagePoint(unsigned int index=0) const
    {
        if(index >= this->size())
            throw CExceptionIllegalPoint("Wrong index in correspondences");
        else{
            CPoint3D<double> tmp=(*this)[index];
            if (tmp.z == 0.0)
                throw CExceptionIllegalPoint("Homogeneous point represents direction!");
            else
                return Base::getEuclideanImagePoint(index);
        }
    }

    // --- world point --- //
    CPoint4D<double> getWorldPoint() const 
    {
        if (this->worldPoint==CPoint4D<double> (0.0, 0.0, 0.0, 0.0))
            throw CExceptionIllegalPoint("The world point has not been calculated yet!");
        else
            return Base::getWorldPoint();
    }

    CPoint3D<double> getEuclideanWorldPoint() const
    {
        if (this->worldPoint.t==0)
            throw CExceptionIllegalPoint("Homogeneous point represents direction!");
        else
            return Base::getEuclideanWorldPoint();
    }
};


/***********************************************************/
//!  Class to represent point correspondences.
/*! 
 *   This class is a std vector that can contain image points
 *   in homogeneous coordinates (3D). The corresponding points
 *   of each camera image should be stored in this vector.
 *   The corresponding world point (homogeneous, 4D) represented
 *   by the image points represent can also be stored.
 *   So the original world point and all its projections to
 *   different cameras can be stored in this class.
 *
 *   Use the aspect class CCorrespondenceChecking to enable  
 *   bounds and error checking.
 *
 *   \ingroup 3d
 * 
 *   \author Thomas Beilicke 
 */
/***********************************************************/
class _CCTUVISION CCorrespondenceBase : public std::vector < CPoint3D < double > > {
public:
    //! Default constructor
    CCorrespondenceBase()
    {
    }

    //! Initializes the correspondence vector with a given size
    CCorrespondenceBase(int size):
    std::vector < CPoint3D < double > >(size)
    {
    }

    // --- image points --- //
    CPoint3D < double > getImagePoint(unsigned int index=0) const; 
    //!< Returns the image point for camera #index.
    CPoint2D < double > getEuclideanImagePoint(unsigned int index=0) const;
    //!< Returns the image point for camera #index in euclidean coordinates.

    void addImagePoint(const CPoint3D < double > &point); 
    //!< Adds an image point to the correspondence.
    void addEuclideanImagePoint(const CPoint2D<double> &point);
    //!< Adds an image point in euclidean coordinates to the correspondence.

    // --- world point --- //
    CPoint4D < double > getWorldPoint() const;
    //!< Returns the world point.
    CPoint3D < double > getEuclideanWorldPoint() const;	
    //!< Returns the world point in euclidean coordinates.

    void setWorldPoint(const CPoint4D<double> &point);
    //!< Sets the world point.
    void setEuclideanWorldPoint(const CPoint3D<double> &point);
    //!< Sets the world point in euclidean coordinates.
protected:	
    CPoint4D < double > worldPoint;	//!< The 4D homogeneous world point.
};

//typedef CCorrespondenceChecking< CCorrespondenceBase >           CCorrespondence;

typedef CCorrespondenceBase CCorrespondence;

} // namespace TUVision


#endif
