/***************************************************************************
                          cfundamentalmatrix.h  -  description
                             -------------------
    begin                : Die Apr 9 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, wenzel@tu-harburg.de
    origin               : PhD Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef CFUNDAMENTALMATRIX_H
#define CFUNDAMENTALMATRIX_H

#include "newmat/newmat.h"
#include "tuv3d/CCorrespondence.h"
#include "tuv/CVertex.h"
#include "tuv3d/CCameraMatrix.h"
#include <vector>


namespace TUVision {

/***************************************************************************/
//! Fundamental Matrix - encapsulation for use with RANSAC
/*!
 *  It is based on a NEWMAT - Matrix and can be initialized given
 *  a set of point correspondences (seven, eight or more).
 *  It is not robust against outliers, i.e. they have to be
 *  filtered out first e.g. by using RANSAC. Point normalization is NOT
 *  part of the fundamental matrix determination, this has to be done
 *  elsewhere.
 *
 * \ingroup 3d
 *
 * @author  Fabian Wenzel
 *
 ***************************************************************************/
class _CCTUVISION CFundamentalMatrix : public NEWMAT::Matrix {
public:
    // Exports for RANSAC : Its defining Datatype, codimension and minimum
    // count of building data
    typedef std::vector< TUVision::CCorrespondence > TDataVector;
    enum { codimension=1 };
    enum { minimumdata=7 };

    CFundamentalMatrix();	//!< Creates a Fundamentalmatrix.
    CFundamentalMatrix ( const TDataVector &list, unsigned int firstindex=0, unsigned int secondindex=1);
    CFundamentalMatrix(const NEWMAT::Matrix &F); 
    //!< Creates a new fundamental matrix out of an existing NEWMAT matrix.
    CFundamentalMatrix(const CCameraMatrix &P1, const CCameraMatrix &P2);
    //!< Calculates the fundamental matrix (algebraically) out of two camera matrices.
    CFundamentalMatrix &operator=(const NEWMAT::Matrix &F);
    //!< Copies an existing NEWMAT matrix.

    //! Return left epipole of fundamental matrix
    const TUVision::CPoint3D< double > & getLeftEpipole() const
    {
        return m_leftepipole;
    }

    //! Return right epipole of fundamental matrix
    const TUVision::CPoint3D< double > & getRightEpipole() const
    {
        return m_rightepipole;
    }

    void create(const CCameraMatrix &P1, const CCameraMatrix &P2);
    //!< Calculates the fundamental matrix (algebraically) out of two camera matrices.
    ~CFundamentalMatrix();

    // Distance of a point pair to the internal model
    double calculateDistance( const TDataVector::value_type &sample, unsigned int firstindex=0, unsigned int secondindex=1);
private:

    // Used in the constructor based on the number of data items
    void sevenPointAlgorithm( const TDataVector &list, unsigned int firstindex=0, unsigned int secondindex=1 );
    void eightPointAlgorithm( const TDataVector &list, unsigned int firstindex=0, unsigned int secondindex=1 );

    // Internal helper function: Solves a polynomial of degree three
    // and returns real roots
    std::vector<double> *solve_cardano_real(const double &a3,
        const double &a2,
        const double &a1,
        const double &a0);

    TUVision::CPoint3D< double > m_leftepipole;
    TUVision::CPoint3D< double > m_rightepipole;
};

}
#endif
