/***************************************************************************
                          CHomography.h  -  description
                             -------------------
    begin                : July 17 2002
    copyright            : (C) 2002 by Marco Grimm
    email                : grimm@tu-harburg.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_CHOMOGRAPHY_H
#define __TUVISION_CHOMOGRAPHY_H

#include "newmat/newmat.h"
#include "tuv/CVertex.h"
#include "tuv3d/CIntrinsicMatrix.h"
#include "tuv3d/CCorrespondence.h"
#include <vector>

namespace TUVision
{

/***************************************************************************/
//! Homography - plane-to-image and planar homography
/*!
 *  It is based on a NEWMAT - Matrix and can be initialized given
 *  a set of point correspondences (more then 4) and the intrinsic camera
 *  parameters or given the rotation matrix R, translation t and the
 *  intrinsic camera parameter. The plane is located at Z=0 in
 *  worldcoordiantes.
 *
 *  \ingroup 3d
 *
 *  @author  Marco Grimm
 *
 ***************************************************************************/
class _CCTUVISION CHomography : public NEWMAT::Matrix
{
public:
    double getCondition();
    typedef std::vector< CCorrespondence > TDataVector;

    // minimum count of building data
    enum { minimumdata = 4 };

    // Codimension see Hartley p.103
    enum { codimension = 2 };

    //! Standard Constructor
    CHomography();

    //! Constructor to generate a plane to image homography from correspondences
    CHomography( const TDataVector &list, CIntrinsicMatrix A );

    //! Constructor to generate a plane to image homography from R and t
    CHomography( NEWMAT::Matrix R, NEWMAT::ColumnVector t,
        CIntrinsicMatrix A );

    //! Destructor does nothing special
    virtual ~CHomography();

    //! Calculate the rotation matrix from the homography
    NEWMAT::Matrix getR();

    //! Calculate the translation vector from the homography
    NEWMAT::Matrix getT();

    //! Set the homography to the input matrix
    void setHomography( NEWMAT::Matrix H, CIntrinsicMatrix A );

    //! The return value is the projection with the homography.
    CVertex2d project( const CVertex2d& point );

    //! The return value is the projection with the inverse homography.
    CVertex2d projectInverse( const CVertex2d& point );

    //! Set the intrinsic parameters of the first view
    void setIntrinsicParameters( CIntrinsicMatrix A );

    //! Get the intrinsic parameters of the first view
    CIntrinsicMatrix getIntrinsicParameters();

    //! Distance of a point pair to the internal model
    double calculateDistance( const TDataVector::value_type &sample,
        unsigned int firstindex = 0,
        unsigned int secondindex = 1 );

    static bool m_bUseNormalizer;

    //! Get the covariance matrix
    NEWMAT::Matrix getCovariance();

    //! Set the covariance matrix
    void setCovariance( const NEWMAT::Matrix& covariance );

private:

protected:
    //! Covariance (9x9) of the homoraphy as vector (9)
    NEWMAT::Matrix m_Covariance;

    // condition of the equation system
    double m_Condition;

    // helper function to calculate the rotation matrix from the plane-to-image homography
    NEWMAT::Matrix getRplaneToImageHomography();

    // helper function to calculate the translation vector from the plane-to-image homography
    NEWMAT::ColumnVector getTplaneToImageHomography();

    // helper function: calculates a orthogonal rotation matrix
    void calculateOrthogonalRotationMatrix( NEWMAT::Matrix& R );

    // helper function: calculates the cross product
    NEWMAT::ColumnVector crossProduct( NEWMAT::ColumnVector v1, NEWMAT::ColumnVector v2 );

    // calculation of the plane-to-image homography
    void planeToImageAlgorithm( const TDataVector &list );

    // helper function to calculate the homography from correspondences
    void calculateHomographyFromCorrespondences( const TDataVector &list,
        unsigned int firstindex=0,
        unsigned int secondindex=1);


    // intrinsic camera parameters according to the first camera
    CIntrinsicMatrix intrinsicParametersFirst;
};
}
#endif
