/***************************************************************************
                          CHomographyPlanar.h  -  description
                             -------------------
    begin                : March 13 2002
    copyright            : (C) 2002 by Marco Grimm
    email                : grimm@tu-harburg.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_CHOMOGRAPHY_PLANAR_H
#define __TUVISION_CHOMOGRAPHY_PLANAR_H

#include "newmat/newmat.h"
#include "tuv/CVertex.h"
#include "tuv3d/CIntrinsicMatrix.h"
#include "tuv3d/CHomography.h"
#include <vector>

namespace TUVision 
{

/***************************************************************************/
//! Homography - plane-to-image and planar homography
/*!
 *  It is based on a NEWMAT - Matrix and can be initialized given
 *  a set of point correspondences (more then 4) and the intrinsic camera 
 *  parameters or given the rotation matrix R, translation t and the 
 *  intrinsic camera parameter. The plane is located at Z=0 in 
 *  worldcoordiantes.
 *
 *  @author  Marco Grimm
 *
 ***************************************************************************/
class _CCTUVISION CHomographyPlanar : public CHomography   
{
public:
    //! Standard Constructor
    CHomographyPlanar();

    //! Constructor to generate a planar homography from correspondences
    CHomographyPlanar( const TDataVector &list );

    //! Constructor to generate a planar homography from correspondences
    CHomographyPlanar( const TDataVector &list, CIntrinsicMatrix Afirst, 
        CIntrinsicMatrix Asecond );

    //! Constructor to generate a planar homography from R, t
    CHomographyPlanar( NEWMAT::Matrix R, NEWMAT::ColumnVector t, 
        CIntrinsicMatrix Afirst, CIntrinsicMatrix Asecond, 
        NEWMAT::ColumnVector n, double distance );

    //! Constructor to generate a planar homography from R1, t1 and Rd, td
    CHomographyPlanar( NEWMAT::Matrix R1, NEWMAT::ColumnVector t1, 
        NEWMAT::Matrix Rd, NEWMAT::ColumnVector td,
        CIntrinsicMatrix Afirst, CIntrinsicMatrix Asecond );

    //! Destructor does nothing special
    virtual ~CHomographyPlanar();

    //! Calculate the rotation matrix from the homography
    NEWMAT::Matrix getR();

    //! Calculate the translation vector from the homography
    NEWMAT::Matrix getT();

    //! Set the homography to the input matrix
    void setHomography( NEWMAT::Matrix H, CIntrinsicMatrix A1, CIntrinsicMatrix A2 );

    //! Set the intrinsic parameters of the first and the second view
    void setIntrinsicParameters( CIntrinsicMatrix A1,
        CIntrinsicMatrix A2 );

    //! Get the intrinsic parameters of the first view (index=1) or the second (index=2)
    CIntrinsicMatrix getIntrinsicParameters( int index = 1 );

    //! Distance of a point pair to the internal model
    double calculateDistance( const TDataVector::value_type &sample,
        unsigned int firstindex  = 0 ,
        unsigned int secondindex = 1 );


private:

    // helper function to calculate the planar homography
    void planarAlgorithm( NEWMAT::Matrix R, NEWMAT::ColumnVector t, 
        CIntrinsicMatrix Afirst, CIntrinsicMatrix Asecond, 
        NEWMAT::ColumnVector n, double distance );

    // calculation of the planar Homography
    void planarHomography( const TDataVector &list );

    // intrinsic camera parameters according to the second camera
    CIntrinsicMatrix intrinsicParametersSecond;
};
}
#endif
