/***************************************************************************
                           CImageByteGrayTsaiDistortion.h  -  description
                              -------------------
     begin                : 30.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel

***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/


#ifndef __TUVISION_CIMAGEBYTEGRAYTSAIDISTORTION_H__
#define __TUVISION_CIMAGEBYTEGRAYTSAIDISTORTION_H__	1

#ifdef	TUV_USE_OPENGL

#include <cstdio>				// fprintf, stderr, define NULL as 0
#include <cmath>
#include <GL/glut.h>			// OpenGL Utility Tool
#include "tuv/CImageByteGray.h"

#include "TUVisionDef.h"

namespace TUVision 
{

	/**********************************************************/
	//!  Compensates radial distortion of the Tsai camera model.
	/*! 
	 *   This class divides an image into square blocks, calculates
	 *   the undistorted coordinates (xu,yu) of block corners using
	 *
	 *   \f[
	 *   (x_u,y_u)^T = (x_d,y_d)^T \cdot (1 + kappa1 * (x_d^2 + y_d^2))
	 *   \f]
	 *
	 *   and interpolates the interior area using OpenGL texture.
	 *
	 *   For further information about radial distortion in Tsai's
	 *   camera model refer to: "An Efficient and Accurate Camera
	 *   Calibration Technique for 3D Machine Vision", Roger Y. Tsai,
	 *   Proceedings of IEEE Conference on Computer Vision and Pattern
	 *   Recognition, 1986, pages 364-374.
	 *
	 *   Important note regarding OpenGL:
	 *   Class uses OpenGL and OpenGL Utility Toolkit (GLUT).
	 *   Please download library and install (http://www.opengl.org).
	 *   Include in own main file e.g. by #include<GL/glut.h>.
	 *
	 *   \ingroup 3d
	 *
	 *   \author Marc Hensel
	 *
	 *   \class CImageByteGrayTsaiDistortion CByteGrayTsaiDistortion.h tuv3d/CByteGrayTsaiDistortion.h
	 */
	/**********************************************************/

	class _CCTUVISION CImageByteGrayTsaiDistortion
	{
		private:
			/* Camera parameters */
			double kappa1;		//!< Distortion coefficient: Ru = Rd * ( 1 + kappa1 * Rd^2 ) [1/mm^2]
			double sx;			//!< Uncertainty calibration parameter (e.g. hardware timing mismatch) []
			double dpx;			//!< Center to center x-distances (sensor elements) [mm/pix]
			double dpy;			//!< Center to center y-distances (sensor elements) [mm/pix]
			double Cx;			//!< Center of distortion (Cx,Cy) in frame (origin = upper left corner) [pix]
			double Cy;			//!< Center of distortion (Cx,Cy) in frame (origin = upper left corner) [pix]
			/* Texture block */
			int	   size;		//!< Texture block edge length [pix]


			/*
			 * Geometric parameters used in texturing:
			 *
			 *  1 2 3 4 5=xsteps
			 *  +-+-+-+-+ 1
			 *  | | | | |
			 *  +-+-+-+-+ 2
			 *  | | | | |
			 *  +-+-+-+-+ 3=ysteps
			 * (| | | | |)<- part of image ommitted, if height no multiple of blocksize
			 *                (same in x-direction, if width no multiple of blocksize)
			 *
			 * tex_[width,height]:
			 * If image width or height are not 2^k the image is embeded in the next
			 * bigger tex_width x tex_height = 2^n x 2^m texture.
			 */
			int		xsteps;		//!< Number of used texture block corner points in x-direction
			int		ysteps;		//!< Number of used texture block corner points in y-direction
			int		tex_width;	//!< Width of texture embeding image (Texture must be 2^n x 2^m pixel)
			int		tex_height;	//!< Height of texture embeding image (Texture must be 2^n x 2^m pixel)


			/* Coordinate transforms */
			//! Calculate undistorted image coordinates u from observed coordinates d
			void	distorted_to_undistorted_image_coord	(double Xfd,double Yfd,double *Xfu,double *Yfu);
			//! Calculate undistorted sensor coordinates u from coordinates d
			void	distorted_to_undistorted_sensor_coord	(double Xd,double Yd,double *Xu,double *Yu);
			//! Calculate distorted image coordinates d from undistorted coordinates d
			void	undistorted_to_distorted_image_coord	(double Xfu,double Yfu,double *Xfd,double *Yfd);
			//! Calculate distorted sensor coordinates d from undistorted coordinates d
			void	undistorted_to_distorted_sensor_coord	(double Xu,double Yu,double *Xd,double *Yd);


			/* Calculate undistorted image */
			//! Init OpenGL and bind image as texture.
			void	initTexture	(CImageByteGray *img,GLuint& texName);


			/* Calculate maximum absolut interpolation error */
			/******************************************************/
			//! Calculate relevant corner (Common method for all error calculations)
			/*!
			 *  This method determines which of the texture blocks that
			 *  are at least partly visible in the undistorted image
			 *  (so called "relevant blocks") has the biggest distance
			 *  from the center of distortion (Cx,Cy).
			 *
			 *  For this block the maximum absolute distance (dx,dy) of
			 *  a corner point to (Cx,Cy) is returned.
			 *
			 *  \param	width	Image width [pix]
			 *  \param	height	Image height [pix]
			 *  \param	xp	Return value min{dx,dy}
			 *  \param	yp	Return value max{dx,dy}
			 */
			/******************************************************/
			void	maxCorner	(int width,int height,double &xp,double &yp);


		public:
			//! Constructor (Sets sx, dpx, dpy, and size to one and all other values to zero)
			CImageByteGrayTsaiDistortion  ();
			//! Destructor
			~CImageByteGrayTsaiDistortion ();


			//! Returns distortion coefficient kappa1.
			double	getKappa1			();
			//! Returns hardware uncertainty factor sx.
			double	getSx				();
			//! Returns center to center sensor distance dpx.
			double	getDpx				();
			//! Returns center to center sensor y-distance dpy.
			double	getDpy				();
			//! Returns x-coordinate Cx of center of distortion.
			double	getCenterX			();
			//! Returns y-coordinate Cy of center of distortion.
			double	getCenterY			();
			//! Returns edge length size of texture blocks.
			int		getBlockSize		();
			/* Get geometric parameters used for texture */
			//! Returns number of texture block corner points xsteps in x-direction.
			int		getNumberOfCornersX	();
			//! Returns number of texture block corner points ysteps in y-direction.
			int		getNumberOfCornersY	();


			//! Sets distortion coefficient kappa1.
			void	setKappa1			(double value);
			//! Sets hardware uncertainty factor sx.
			void	setSx				(double value);
			//! Sets center to center sensor distance dpx.
			void	setDpx				(double value);
			//! Sets center to center sensor distance dpy.
			void	setDpy				(double value);
			//! Sets x-coordinate Cx of center of distortion.
			void	setCenterX			(double value);
			//! Sets y-coordinate Cy of center of distortion.
			void	setCenterY			(double value);
			//! Sets edge length size of texture blocks.
			bool	setBlockSize		(int value);

			/*
			 * Method uses maxAlignedError and maxOrthogonalError.
			 */
			/******************************************************/
			//! Sets size of texture blocks so that maximum absolute error <= max_error.
			/*!
			 *  In calculation it is taken into account that width/blocksize and
			 *  height/blocksize must be integer if parts of the original images
			 *  must not be discarded in undistortion.
			 *
			 *  \param	width	Width of the image to be undistorted
			 *  \param	height	Height of the image to be undistorted
			 *  \param	max_error	Upper bound for maximum absolute interpolation error
			 */
			/******************************************************/
			bool setMaxBlockSize		(int width,int height,double max_error);


			/*
			 * CALCULATE UNDISTORTED IMAGE:
			 */

			/******************************************************/
			//! Undistortes an image.
			/*!
			 *  The following steps are taken:
			 *
			 *  -# Calculate and set texture parameters
			 *  -# Calculate undistorted coordinates of texture block corners
			 *  -# Perform undistortion by texturing
			 *  -# Store undistorted image to img
			 *
			 *  \param img	Pointer to original image (Gets overwritten with undistorted result!)
			 */
			/******************************************************/
			bool	undistortImage			(CImageByteGray *img);

			/******************************************************/
			//! Calculates and sets parameters for texturing.
			/*!
			 *  The following parameters are set:
			 *
			 *  - xsteps:	Number of used texture block corners in x-direction
			 *  - ysteps:	Number of used texture block corners in y-direction
			 *  - tex_width:	Width of texture embeding original image
			 *  - tex_height:	Height of texture embeding original image
			 *
			 *  OpenGL requires textures to be 2^n x 2^m pixels in size. The image
			 *  might get embeded into an tex_width x tex_height = 2^n x 2^m texture and
			 *  therefore tex_width and tex_height might differ from image dimensions.
			 *
			 *  \param	img	Pointer to image to be undistorted
			 */
			/******************************************************/
			void	setTextureParameters	(CImageByteGray *img);

			/******************************************************/
			//! Calculates undistorted coordinates of texture block corners.
			/*!
			 *  Note:
			 *  Texture parameters MUST HAVE BEEN SET before calling this method.
			 *
			 *  The results are written into the "new"-arrays:
			 *  -# Array dimension: getNumberOfCornersX() * getNumberOfCornersY()
			 *  -# Array organisation: Rowwise corner points, i.e. (0,0), (1,0),
			 *     ..., (getNumberOfCornersX(),0), (0,1), (1,1), ...,
			 *     (getNumberOfCornersX(),getNumberOfCornersY())
			 *
			 *  \param	img	Pointer to image to be undistorted
			 *  \param	new_x	Pointer to array of undistorted x-coordinates of texture block corners
			 *  \param	new_x	Pointer to array of undistorted y-coordinates of texture block corners
			 */
			/******************************************************/
			bool	undistortedCoordinates	(CImageByteGray *img,int *new_x,int *new_y);

			/*
			 * Note on isNext=true:
			 * Not new texture created, but texture content overwriten with new image instead (slightly faster).
			 *
			 * Note on runtime:
			 * Nearly ALL (!!!) of the time in textureImage is consumend for storing the textured image (glReadPixels).
			 */
			/******************************************************/
			//! Undistort image and saves result in img.
			/*!
			 *  Note:
			 *  Texture parameters MUST HAVE BEEN SET before calling this method.
			 *
			 *  There are two modes:
			 *  -# isNext=false: Undistort by performing the full calculations.
			 *  -# isNext=true: Undistort images with equal image dimensions and
			 *     parameters as in last undistortion, e.g. in video streams.
			 *  Note: Mode 2 is slighly faster but requires mode 1 to have been called before.
			 *
			 *  \param	img	Pointer to original image (Gets overwritten with undistorted result!)
			 *  \param	new_x	Pointer to array of undistorted x-coordinates of texture block corners
			 *  \param	new_x	Pointer to array of undistorted y-coordinates of texture block corners
			 *  \param	isNext	Mode (see above)
			 */
			/******************************************************/
			void	textureImage		(CImageByteGray *img,int *new_x, int *new_y,bool isNext=false);


			/*
			 * ERROR CALCULATIONS:
			 *
			 * The aim is to find an upper bound for max. error between image undistorted
			 * by interpolated blockwise OpenGL texturing and correcly calculated values.
			 *
			 * Error is calculated by max. distance between correctly calculated value and
			 * texturing + sqrt(0.5) for rounding the textured coordinate.
			 *
			 * Note: Comments located at the implementation of the methods include diagrams
			 * and further valuabe information to illustrate the theoretical background.
			 */

			/*
			 * maxAbsoluteError() calculates the maximum of all triangle error methods.
			 *
			 * Max. error in triangle: Both methods use maxCorner to find relevant corner
			 * P of max. distance to center of distortion.
			 *
			 * maxAlignedError divides block by a diagonal that includes corner P and finds
			 * max. absolut error in triangle including P by logarithmic search.
			 *
			 * maxOrthogonalError chooses other diagonal and triangle including P and corner Q
			 * second furthest from center of distortion.
			 *
			 * lambda (x-direction) and mu (y-direction) in [0,1] are parameters leading to the
			 * point of max. absolut error. CAUTION: The meaning of lambda and mu are different
			 * in maxAlignedError (starting from Q) and maxOrthogonalError (starting from P).
			 *
			 * The optional pointers *lambda_ret and *mu_ret make it possible to get the parameters
			 * e.g. for examinations where the maximum absolute error is located.
			 */

			/******************************************************/
			//! Returns upper bound for absolute maximum interpolation error.
			/*!
			 *  The error results due to texturing of blocks instead of
			 *  exact calculation for every pixel.
			 *
			 *  This method returns the maximum of maxAlignedError() and
			 *  maxOrthogonalError() added with 1.0 to consider rounding errors.
			 *
			 *  \param	width	Width of the image to be undistorted
			 *  \param	height	Height of the image to be undistorted
			 */
			/******************************************************/
			double maxAbsoluteError		(int width,int height);

			/******************************************************/
			//! Returns upper bound for abs. max. interpolation error in the aligned triangle.
			/*!
			 *  Please refer to Marc Hensel's study thesis.
			 *
			 *  \param	width	Width of the image to be undistorted
			 *  \param	height	Height of the image to be undistorted
			 *  \param	lambda_ret	Parameter in x-direction
			 *  \param	mu_ret	Parameter in y-direction
			 */
			/******************************************************/
			double maxAlignedError		(int width,int height,double *lambda_ret=NULL,double *mu_ret=NULL);

			/******************************************************/
			//! Returns upper bound for abs. max. interpolation error in the orthogonal triangle.
			/*!
			 *  Please refer to Marc Hensel's study thesis.
			 *
			 *  \param	width	Width of the image to be undistorted
			 *  \param	height	Height of the image to be undistorted
			 *  \param	lambda_ret	Parameter in x-direction
			 *  \param	mu_ret	Parameter in y-direction
			 */
			/******************************************************/
			double maxOrthogonalError	(int width,int height,double *lambda_ret=NULL,double *mu_ret=NULL);
	};

}

#endif // TUV_USE_OPENGL

#endif // __TUVISION_CIMAGEBYTEGRAYTSAIDISTORTION_H__