/***************************************************************************
                          cransac.h  -  description
                             -------------------
    begin                : Mon Apr 15 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, wenzel@tu-harburg.de
    origin               : PhD Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef CRANSAC_H
#define CRANSAC_H

#include <cmath>
#include <cstdlib>
#include <set>
#include <vector>

// Use MSAC instead of RANSAC
// (commend this line if the RANSAC algorithm should be used)
#define USE_MSAC

namespace TUVision {

/***************************************************************************/
//! RANSAC - Algorithm for outlier removal : Template - based method
/*!
 *  It is recommended to read the following passage if you want to work
 *  with this template - otherwise you won't get the meanings of the
 *  variable-types:
 *
 *  Hartley: Multiple View Geometry - page 101 - 107
 *
 *  This class is template-based and requires a small framework-like
 *  component setup: If you would like to run RANSAC for determining a new
 *  type of data model, you have to implement a class defining the data model
 *  which has to have the following five characteristics:
 *
 *  - It has to define (via a public enum called minimumdata) the minimum
 *    number of data which is necessary for the model to be built
 *  - It has to define the codimension (via public enum called codimension)
 *    for estimating the threshold based on the variance of the datas'
 *    measurements (see Hartley, page 103)
 *  - It has to define (via public typedef TDataVector) the datatype for
 *    its constructing data. There has to be an iterator for it such that
 *    this class can iterate through the sample-set
 *    (i.e. each STL-container is fine)
 *  - It has to be able to be constructed from a TDataVector having the
 *    minimum number of samples
 *  - It has to have a method called "calculateDistance" returning
 *    a double value representing the distance of a sample point to it
 *    and taking a TDataVector::value_type - kind of element (see STL-doc)
 *
 * FOR example: The first datamodel this algorithm was programmed for was
 * the fundamental matrix using CFundamentalMatrix as datamodel class. A
 * fundamental matrix can be built using either seven or eight point
 * correspondences, i.e. the enum "minimumdata" is set to 7 but can also be
 * passed to the instantiation of the template as second parameter. The
 * constructor gets a vector consisting of - seven or eight - point pairs
 * (which also gets exported as TDataVector - type). There is also a function
 * calculateDistance which is taking one point pair and that is returning
 * the sum of the distances of the point pairs to the corresponding epipolar
 * lines. This means that its codimension is one - this also gets exported as
 * enum. You might also want to look at the test - program linefit.
 *
 * \todo Only codimension of 1 is implemented, i.e. calculateDistance
 *       returns only one vector
 *
 * \ingroup 3d
 *
 * @author  Fabian Wenzel
 *
 ***************************************************************************/
template < class TDataModel,
    int buildingsamples = TDataModel::minimumdata > class CRansac {
    public:

        // Exports the DataVector - Type
        typedef typename TDataModel::TDataVector TDataVector;


        /***************************************************************************/
        //! Constructor takes the sample set of input data
        /*! It sets the datavector to the sample set and initializes the
         *  trustinterval which can determine the number of trials given a variance
         *  of the measurements to 0.99.
         ***************************************************************************/
        CRansac(const TDataVector &datavector) :
            m_trustinterval(0.99),
            m_samplecount(-1),
            m_initializedthreshold(false),
            m_initializedsamplecount(false),
            m_done(false),
            m_datavector(datavector),
            m_model(0)
        {
        }


        /***************************************************************************/
        //! Destructor deletes the model if there has been calculated one
        /*!
         *
         ***************************************************************************/
        ~CRansac(){
            if(m_model)
                delete m_model;
        };


        /***************************************************************************/
        //! The distance threshold for determining outliers get set directly
        /*! Another possibility is to set the variance of a normally distributed
         *  measured set by using the codimension of the model (see
         *  setVariance()
         ***************************************************************************/
        void setThreshold(const double &threshold)
        {
            m_threshold = threshold;
            m_initializedthreshold = true;
        }


        /***************************************************************************/
        //! The second possibility for setting the outlier threshold
        /*! It is assumed that the set of inliners is distributed normally about
         *  the true model. By knowing the codimension it is possible to determine
         *  the threshold based on a 0.95% - interval (see Hartley). This method
         *  usually is not used but an empiric threshold is applied
         *  (see setThreshold())
         ***************************************************************************/
        void setVariance (const double &variance )
        {
            switch( TDataModel::codimension ){
        case 1 : m_threshold = sqrt(3.84*variance); break;
        case 2 : m_threshold = sqrt(5.99*variance); break;
        case 3 : m_threshold = sqrt(7.81*variance); break;
        default: break;
            }
            m_initializedthreshold = true;
        }


        /***************************************************************************/
        //! The number of trials which has to be taken is determined here.
        /*! This is the upper bound for the following calculation: As soon as the
         *  "current" trial reveals an inlier-quote higher than the trust interval
         *  (which is set to 0.99 in the constructor but can be changed below), the
         *  algorithm stops. The second possibility of setting the sample count
         *  is setOutlierProb() - see below.
         ***************************************************************************/
        void setSampleCount(int samplecount)
        {
            m_samplecount = samplecount;
            m_initializedsamplecount = true;
        }

        /***************************************************************************/
        //! The number of trials gets calculated based on the "ratio" of inliers
        /*! This has nothing to do with the variance of the measured data
         *  but only to the "probability" that any selected point is an outlier
         *  (see Hartley, page 104). This is the second possibility of setting the
         *  sample count (see setSampleCount() above).
         *
         *  \warning It uses the trustinterval internally,
         *           i.e. the interval has to be changed first before calling
         *           setOutlierProb() !
         ***************************************************************************/
        void setOutlierProb(double outlierprob)
        {
            if (outlierprob != 0){
                m_samplecount = (int) (log(1-m_trustinterval)/
                    log(1-pow(1-outlierprob,buildingsamples)));
            } else {
                m_samplecount = 1;
            }

            m_initializedsamplecount = true;
        }


        /***************************************************************************/
        //! Sets the trust interval which is a halt-criterion when counting inliers
        /*! The default value for the trust interval is 0.99 (i.e. the algorithms
         *  stops as soon as more than 99% of inliers have been found). It is used
         *  during runtime as the halt-criterion but also in setOutlierProb() (see
         *  above) to estimate the number of trials which have to be taken for
         *  a desired success
         ***************************************************************************/
        void setTrustInterval( const double &interval )
        {
            m_trustinterval = interval;
        }


        /***************************************************************************/
        //! Returns a new TDataVector only containing the inliers based on the model
        /*!
         ***************************************************************************/
        TDataVector *getInliers ()
        {
            if ( !checkandrun() )
                return NULL;

            TDataVector *result = new TDataVector();

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample;

            for(currentsample =  m_datavector.begin();
                currentsample != m_datavector.end();
                ++currentsample)
                if( fabs(m_model->calculateDistance(*currentsample)) <= m_threshold )
                    result->push_back(*currentsample);

            return result;
        }


        /***************************************************************************/
        //! Returns a vector< int > containing inlier indices
        /*! This is helpful for applications in which the model is not built on
         *  the data vector directly but on transformed data which are not used
         *  afterwards
         ***************************************************************************/
        std::vector< int > getInlierIndices ()
        {
            if ( !checkandrun() )
                return std::vector< int >();

            std::vector< int > result;

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample = m_datavector.begin();
            int i=0;

            while(currentsample != m_datavector.end()){
                if( fabs(m_model->calculateDistance(*currentsample)) <= m_threshold )
                    result.push_back(i);
                ++currentsample,++i;
            }

            return result;
        }


        /***************************************************************************/
        //! Returns a new TDataVector only containing the outliers based on the model
        /*!
         ***************************************************************************/
        TDataVector *getOutliers()
        {
            if ( !checkandrun() )
                return NULL;

            TDataVector *result = new TDataVector();

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample;

            for(currentsample =  m_datavector.begin();
                currentsample != m_datavector.end();
                ++currentsample)
                if( fabs(m_model->calculateDistance(*currentsample)) > m_threshold )
                    result->push_back(*currentsample);

            return result;
        }


        /***************************************************************************/
        //! Returns a vector< int > containing outlier indices
        /*! This is helpful for applications in which the model is not built on
         *  the data vector directly but on transformed data which are not used
         *  afterwards
         ***************************************************************************/
        std::vector< int > getOutlierIndices ()
        {
            if ( !checkandrun() )
                return std::vector< int >();

            std::vector< int > result;

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample = m_datavector.begin();
            int i=0;

            while(currentsample != m_datavector.end()){
                if( fabs(m_model->calculateDistance(*currentsample)) > m_threshold )
                    result.push_back(i);
                ++currentsample,++i;
            }

            return result;
        }


        /***************************************************************************/
        //! Returns the model
        /*!
         ***************************************************************************/
        TDataModel *getModel()
        {
            if ( !checkandrun() )
                return NULL;
            return m_model;
        }

    protected:
        // The three parameters which determine the behavior of RANSAC
        double              m_threshold;
        double              m_trustinterval;
        int                 m_samplecount;

        // Two flags which determine the initialization state
        bool                m_initializedthreshold;
        bool                m_initializedsamplecount;
        bool                m_done;

        // The sample set
        const TDataVector  &m_datavector;

        // The model
        TDataModel         *m_model;



        /***************************************************************************/
        //! Runs the RANSAC-routine if initialized
        /*!
         ***************************************************************************/
        bool checkandrun()
        {
            if ( m_done )
                return true;

            if ( !m_initializedthreshold || !m_initializedsamplecount )
                return false;

            if ( m_model ){
                delete m_model;
                m_model = 0;
            }

            run();

            m_done = true;

            return true;
        }


        /***************************************************************************/
        //! The RANSAC-core
        /*! It successively builds a test model, counts the inlier-ratio
         *  and updates the best calculated model so far. It terminates
         *  if the inlier-ratio is sufficient or the sample/trial count has been
         *  reached
         ***************************************************************************/
        void run()
        {
            // only outliers first - score is negative such that
            // even an totally (score==zero) test model will be
            // stored. Otherwise there is no model at all and
            // the calling program may crash
            double score = -1.0;

            // solve for n iterations
            for(int n=0;n<m_samplecount;++n)
            {
                TDataModel *samplemodel = getTestModel();
                double newscore =         getScore(samplemodel);
                // only continue if the score is better
                if( newscore > score ){
                    if(m_model)
                        delete m_model;

                    // remember new model
                    m_model = samplemodel;
                    score   = newscore;

                    // runtime - halt - criterion
                    if ( newscore >= m_trustinterval )
                        return;
                }
                else{
                    if (samplemodel)
                        delete samplemodel;
                }

            }

        }


        /***************************************************************************/
        //! Builds a sample model based on random data of the data-vector
        /*!
         ***************************************************************************/
        TDataModel *getTestModel()
        {
            // First, build model based on the number of building samples
            std::set< int > indices;

            TDataVector samplevector;

            while(indices.size() < buildingsamples){
                int randomindex = (int) ((m_datavector.size())*( (double) rand()/(RAND_MAX+1.0)));
                if(indices.find(randomindex) == indices.end()){
                    indices.insert(randomindex);
                    samplevector.push_back(m_datavector[randomindex]);
                }
            }

            // build the new model based on the sample vector
            return new TDataModel(samplevector);
        }


        /***************************************************************************/
        //! Calculates a score, i.e. the ratio of inliers based on the sample model
        /*! With the define USE_MSAC the usage of the standard RANSAC algorithm can
         *  be switched to the usage of the MSAC algorithm. It is recommended to
         *  use the MSAC instead of the RANSAC algorithm, because MSAC needs less
         *  iterations to come to the same result.
         *  The MSAC algorithm is described in the paper:
         *  P.H.S. Torr and A. Zisserman: MLESAC: A New Robust Estimator with
         *  Application to Estimating Image Geometry, In Journal of Computer Vision
         *  and Image Understanding, pages 138-156, 78(1), (2000)
         ***************************************************************************/
#ifdef USE_MSAC
        double getScore(TDataModel *samplemodel){
            double score  = 0.0;

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample;

            for(currentsample =  m_datavector.begin();
                currentsample != m_datavector.end();
                ++currentsample)
            {
                double error = fabs(samplemodel->calculateDistance(*currentsample));
                if( error < m_threshold )
                {
                    score += error;
                }
                else
                {
                    score += m_threshold;
                }
            }

            double score_max = m_datavector.size() * m_threshold;

            return  ( score_max - score ) / score_max;
        }
#else
        double getScore(TDataModel *samplemodel){
            int inliers  = 0;

            typedef typename TDataVector::const_iterator TDataIterator;
            TDataIterator currentsample;

            for(currentsample =  m_datavector.begin();
                currentsample != m_datavector.end();
                ++currentsample)
                if( fabs(samplemodel->calculateDistance(*currentsample)) <= m_threshold ){
                    inliers++;
                }

                return (double) inliers / (double) m_datavector.size();
        }
#endif

    };

}
#endif

