/***************************************************************************
                           CTsaiDistortion.h  -  description
                              -------------------
     begin                : 05.09.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Marc Hensel

***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/


#ifndef __TUVISION_CTSAIDISTORTION_H__
#define __TUVISION_CTSAIDISTORTION_H__	1

#include <cstdio>					// fprintf, stderr, define NULL as 0
#include <cmath>
#include <tuv/CImageByteGray.h>
#include <tuv/CImageByteRGB.h>

#include "TUVisionDef.h"

namespace TUVision 
{

/*! \anchor ctsaidistortionsyntactic
 */  

/**********************************************************/
//!  Compensates radial distortion using the Tsai camera model.
/*! 
 *   Example code:
 *
 *   \code
 *   #include <tuv3d/CTsaiDistortion.h>
 *   ...
 *   void example_method(CImageByteRGB &image)
 *   {
 *     CTsaiDistortion tsai;
 *     double kappa1 = ...
 *
 *     // Set camera parameters
 *     tsai.setKappa1( kappa1 );
 *     ...
 *
 *     // Undistort image
 *     tsai.undistortImage( image );
 *   }
 *   \endcode
 *
 *   Semantic documentation and camera model parameter description:
 *   \ref ctsaidistortionsemantic "Semantic documentation"
 *
 *   \ingroup 3d
 *
 *   \author Marc Hensel
 *
 *   \class CTsaiDistortion CTsaiDistortion.h tuv3d/CTsaiDistortion.h
 */
/**********************************************************/

class _CCTUVISION CTsaiDistortion
{
private:
    /* Camera parameters */
    double kappa1;		//!< Distortion coefficient \f$ \kappa_1 \f$ [1/mm^2]
    double sx;			//!< Uncertainty calibration parameter (e.g. hardware timing mismatch) []
    double dpx;			//!< Center to center x-distances (sensor elements) [mm/pixel]
    double dpy;			//!< Center to center y-distances (sensor elements) [mm/pixel]
    double Cx;			//!< Center of distortion (Cx,Cy) in frame (origin = upper left corner) [pixel]
    double Cy;			//!< Center of distortion (Cx,Cy) in frame (origin = upper left corner) [pixel]

    /* Geometric mapping/undistortion parameters */
    bool	streamOpen;	//!< Stream is currently open ("true" = mapping data valid) or not ("false")
    int		width;		//!< Image width [pixel]
    int		height;		//!< Image height [pixel]
    int		*map_x;		//!< Mapping (rowwise): Nearest distorted image x-coordinate correspondending to undistorted coordinate
    int		*map_y;		//!< Mapping (rowwise): Nearest distorted image y-coordinate correspondending to undistorted coordinate

    /* Coordinate transforms */
    //! Calculate undistorted image coordinates u from observed image coordinates d
    void	distorted_to_undistorted_image_coord	(double Xfd,double Yfd,double *Xfu,double *Yfu);
    //! Calculate undistorted sensor coordinates u from sensor coordinates d
    void	distorted_to_undistorted_sensor_coord	(double Xd,double Yd,double *Xu,double *Yu);
    //! Calculate distorted image coordinates d from undistorted image coordinates d
    void	undistorted_to_distorted_image_coord	(double Xfu,double Yfu,double *Xfd,double *Yfd);
    //! Calculate distorted sensor coordinates d from undistorted sensor coordinates d
    void	undistorted_to_distorted_sensor_coord	(double Xu,double Yu,double *Xd,double *Yd);


public:
    //! Constructor (Sets sx, dpx, and dpy to one and all other values to zero)
    CTsaiDistortion  ();
    //! Destructor
    ~CTsaiDistortion ();


    //! Returns distortion coefficient kappa1.
    double	getKappa1			();
    //! Returns hardware uncertainty factor sx.
    double	getSx				();
    //! Returns center to center sensor distance dpx.
    double	getDpx				();
    //! Returns center to center sensor y-distance dpy.
    double	getDpy				();
    //! Returns x-coordinate Cx of center of distortion.
    double	getCenterX			();
    //! Returns y-coordinate Cy of center of distortion.
    double	getCenterY			();

    //! Sets distortion coefficient kappa1.
    void	setKappa1			(double value);
    //! Sets hardware uncertainty factor sx.
    void	setSx				(double value);
    //! Sets center to center sensor distance dpx.
    void	setDpx				(double value);
    //! Sets center to center sensor distance dpy.
    void	setDpy				(double value);
    //! Sets x-coordinate Cx of center of distortion.
    void	setCenterX			(double value);
    //! Sets y-coordinate Cy of center of distortion.
    void	setCenterY			(double value);


    /*
     * CALCULATE UNDISTORTED IMAGE:
     */
    /******************************************************/
    //! Undistort an 8-bit gray image and returns maximum interpolation error.
    /*!
     *  The following steps are taken:
     *
     *  -# Open stream (calculate mapping data)
     *  -# Undistort image
     *  -# Close stream (free memory storing mapping data)
     *
     *  \param img	Original image (gets overwritten with undistorted image)
     */
    /******************************************************/
    double	undistortImage	(CImageByteGray &img);
    //! Undistort an RGB image and returns maximum interpolation error.
    /*!
     *  The following steps are taken:
     *
     *  -# Open stream (calculate mapping data)
     *  -# Undistort image
     *  -# Close stream (free memory storing mapping data)
     *
     *  \param img	Original image (gets overwritten with undistorted image)
     */
    /******************************************************/
    double	undistortImage	(CImageByteRGB  &img);

    //! Calculate undistortion mapping.
    /*!
     *  Calculate mapping of undistorted to distorted pixel location in arrays 
     *  of type CImageByte. Returns maximum interpolation error.
     *
     *  \param iwidth	Image width in pixel
     *  \param iheight	Image height in pixel
     */
    /******************************************************/
    double	openStream		(int iwidth,int iheight);

    //! Undistort an image.
    /*!
     *  Undistort an image using the mapping of the current open stream.
     *
     *  If no stream is open method openStream() is called 
     *  before undistortion.
     *
     *  \param img	Original image (gets overwritten with undistorted image)
     */
    /******************************************************/
    bool	undistort		(CImageByteGray &img);
    bool	undistort		(CImageByteRGB &img);

    //! Frees memory of structures holding current mapping data.
    bool	closeStream		();
};

}


#endif // __TUVISION_CTSAIDISTORTION_H__
