/***************************************************************************
                          TsaiCalibrationBackend.h  -  description
                             -------------------
    begin                : 26.03.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_TSAICALIBRATIONBACKEND_H__
#define __TUVISION_TSAICALIBRATIONBACKEND_H__ 1

#include "TUVisionDef.h"
#include "TsaiCameraModel.h"

/* The following type definitions are taken from minpack/f2c.h, they are copied
 * as f2c.h should not be included here */

typedef long int integer;
typedef double doublereal;

/* Maximum number of data points allowed */
#define MAX_POINTS	500

/* An arbitrary tolerance factor */
#define EPSILON		1.0E-8

/* Commonly used macros */
#define ABS(a)          (((a) > 0) ? (a) : -(a))
#define SIGNBIT(a)      (((a) > 0) ? 0 : 1)
#define SQR(a)          ((a) * (a))
#define CUB(a)          ((a)*(a)*(a))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))

/* Some math libraries may not have the sincos() routine */
#ifdef _SINCOS_
void sincos();
#define SINCOS(x,s,c)   sincos(x,&s,&c)
#else
double sin(), cos();
#define SINCOS(x,s,c)   s=sin(x);c=cos(x)
#endif


/* Parameters controlling MINPACK's lmdif() optimization routine. */
/* See the file lmdif.f for definitions of each parameter.        */
#define REL_SENSOR_TOLERANCE_ftol    1.0E-5      /* [pix] */
#define REL_PARAM_TOLERANCE_xtol     1.0E-7
#define ORTHO_TOLERANCE_gtol         0.0
#define MAXFEV                       (1000*n)
#define EPSFCN                       1.0E-16     /* Do not set to zero! */
#define MODE                         1           /* variables are scalled internally */
#define FACTOR                       100.0 

/*******************************************************************************\
 *
 * Calibration data consists of the (x,y,z) world coordinates of a feature
 * point	(in mm) and the corresponding coordinates (Xf,Yf) (in pixels) of the
 * feature point in the image.  Point count is the number of points in the data
 * set.
 *
 *
 * Coplanar calibration:
 *
 * For coplanar calibration the z world coordinate of the data must be zero.
 * In addition, for numerical stability the coordinates of the feature points
 * should be placed at some arbitrary distance from the origin (0,0,0) of the
 * world coordinate system.  Finally, the plane containing the feature points
 * should not be parallel to the imaging plane.  A relative angle of 30 degrees
 * is recommended.
 *
 *
 * Noncoplanar calibration:
 *
 * For noncoplanar calibration the data must not lie in a single plane.
 *
\*******************************************************************************/
struct calibration_data {
int       point_count;	/* [points] 	 */
double    xw[MAX_POINTS];	/* [mm]          */
double    yw[MAX_POINTS];	/* [mm]          */
double    zw[MAX_POINTS];	/* [mm]          */
double    Xf[MAX_POINTS];	/* [pix]         */
double    Yf[MAX_POINTS];	/* [pix]         */
};

extern char   camera_type[];

namespace TUVision
{
class _CCTUVISION TsaiCalibrator{
public:
    void distorted_image_plane_error_stats (double *mean, double *stddev, double *max, double *sse);
    void undistorted_image_plane_error_stats (double *mean, double *stddev, double *max, double *sse);
    void object_space_error_stats (double *mean, double *stddev, double *max, double *sse);
    void normalized_calibration_error (double *mean, double *stddev);
    virtual bool calibration()=0;
    virtual bool calibration_with_full_optimization()=0;
    virtual bool extrinsic_parameter_estimation()=0;
    static SCameraParameters cp;
    static calibration_data cd;
    static SCalibrationConstants cc;
protected:
    virtual bool three_parm_optimization()=0;
    virtual bool nic_optimization()=0;
    virtual bool full_optimization()=0;
    static void Matrix2Euler();
    static void Euler2Matrix();
    static bool epe_optimize();
    bool epe_compute_Tx_Ty_Tz ();
    static int epe_optimize_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    static void undistorted_to_distorted_sensor_coord(double Xu, double Yu, double *Xd, double *Yd);
    static void distorted_to_undistorted_sensor_coord(double Xd, double Yd, double *Xu, double *Yu);
    static void undistorted_to_distorted_image_coord (double Xfu, double Yfu, double *Xfd, double *Yfd);
    static void distorted_to_undistorted_image_coord (double Xfd, double Yfd, double *Xfu, double *Yfu);
    static void world_coord_to_image_coord (double xw, double yw, double zw, double *Xf, double *Yf);
    static void image_coord_to_world_coord (double Xfd, double Yfd, double zw, double *xw, double *yw);
    static void world_coord_to_camera_coord (double xw, double yw, double zw, double *xc, double *yc, double *zc);
    static void camera_coord_to_world_coord (double xc, double yc, double zc, double *xw, double *yw, double *zw);
    static double CBRT(double x);
};

class _CCTUVISION TsaiCalibratorCoplanar: public TsaiCalibrator{
public:
    bool calibration();
    bool calibration_with_full_optimization();
protected:
    static void compute_Xd_Yd_and_r_squared();
    static bool compute_U();
    static void compute_Tx_and_Ty();
    static void compute_R();
    static bool compute_approximate_f_and_Tz();
    static int compute_exact_f_and_Tz_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    static bool compute_exact_f_and_Tz();
    bool three_parm_optimization();
    static void remove_sensor_plane_distortion_from_Xd_and_Yd();
    static int five_parm_optimization_with_late_distortion_removal_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool five_parm_optimization_with_late_distortion_removal();
    static int five_parm_optimization_with_early_distortion_removal_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool five_parm_optimization_with_early_distortion_removal();
    static int nic_optimization_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool nic_optimization();
    static int full_optimization_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool full_optimization();
    bool epe_compute_U(double U[]);
    void epe_compute_Tx_and_Ty(double U[]);
    void epe_compute_R(double U[]);
    bool epe_compute_approximate_f (double *f);
    bool extrinsic_parameter_estimation ();
};

class _CCTUVISION TsaiCalibratorNoncoplanar : public TsaiCalibrator{
public:
    bool calibration();
    bool calibration_with_full_optimization();
protected:
    void compute_sx();
    void compute_better_R();
    void compute_Xd_Yd_and_r_squared();
    static bool compute_U();
    static void compute_Tx_and_Ty();
    static void compute_R();
    static bool compute_approximate_f_and_Tz();
    static int compute_exact_f_and_Tz_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    static bool compute_exact_f_and_Tz();
    bool three_parm_optimization();
    static int nic_optimization_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool nic_optimization();
    static int full_optimization_error(integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused);
    bool full_optimization();
    bool epe_compute_U(double U[]);
    void epe_compute_Tx_and_Ty(double U[]);
    void epe_compute_R(double U[]);
    bool epe_compute_aproximate_f(double *f);
    bool extrinsic_parameter_estimation ();
};

}

#endif // TSAICALIBRATION_BACKEND_H__
