/***************************************************************************
                          CImage.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde, Eric Sommerlade, Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde, Eric Sommerlade, PhD Shdaifat

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv/CImage.h"
#include "tuv/CRect.h"
#include "tuv/CImageAttributes.h"
#include "tuv/CExceptionRange.h"

using namespace std;
using namespace TUVision;

static const CImageAttributes DEFAULT_ATTRIBUTES;


/**********************************************************/
/*!
 *  Constructs a new image.
 *
 *  \param w          image width
 *  \param h          image height
 *  \param nchannels  number of color channels
 */
/**********************************************************/

CImage::CImage (int w, int h, int nchannels)
: d_width (w), 
d_height (h), 
d_nchannels (nchannels)
{
    d_img_attributes = 0;
};


/**********************************************************/
/*!
 *  Constructs an object using an existing image.
 *
 *  \param img image to be copied
 *
 *  \warning The image attributes are always copied deep,  
 *           i.e. no reference counting here.
 *           So when an image is copied, you will always 
 *           have the overhead of copying this structure 
 *           (~48 bytes).
 */
/**********************************************************/

CImage::CImage (const CImage& img, bool copy)
: d_width (copy?img.d_win_width:img.d_width), 
d_height (copy?img.d_win_height:img.d_height), 
d_nchannels (img.d_nchannels)
{
    if(img.d_img_attributes==0)
        d_img_attributes = 0;
    else
        d_img_attributes = new CImageAttributes (*(img.d_img_attributes));
};


/**********************************************************/
/*!
 *  Deinitializes/frees memory occupied by the image 
 *  attributes.
 */
/**********************************************************/

CImage::~CImage()
{
    delete d_img_attributes;
};


/**********************************************************/
/*!
 *   Image attributes are used to store additional
 *   informations about an image. For example you can
 *   use this system to store informations like the
 *   resolution (DPI) of an image, the creation time of
 *   an image or its time code.
 *  
 *   \return reference to the image attributes
 */
/**********************************************************/

CImageAttributes&
CImage::getAttributes()
{
    if(d_img_attributes==0)
        d_img_attributes=new CImageAttributes();

    return *d_img_attributes;
};


/**********************************************************/
/*!
 *   Image attributes are used to store additional
 *   informations about an image. For example you can
 *   use this system to store informations like the
 *   resolution (DPI) of an image, the creation time of
 *   an image or its time code.
 *  
 *   \return reference to the image attributes
 */
/**********************************************************/

const CImageAttributes&
CImage::getAttributes() const
{
    if(d_img_attributes==0)
        return(DEFAULT_ATTRIBUTES);

    return *d_img_attributes;
};


/**********************************************************/
/*! 
 *   Creates a window in the original image. The user of 
 *   an object will see only the pixels in this window. 
 *   This is a fast method to get subimages of an image.
 *
 *   \param x_0 upper left corner of the window
 *              (x coordinate)
 *   \param y_0 upper left corner of the window
 *              (y coordinate)
 *   \param width width of the window
 *   \param height height of the window
 *
 */
/**********************************************************/

void 
CImage::setWindow (int x_0, int y_0, int width, int height)
{
    // test for correct parameters 
    if ((x_0 < 0) || (x_0 >= d_width))
        throw CExceptionRange ("You can not create a window with "
        "this start point (x dimension).", 
        x_0, 0, d_width - 1);
    if ((y_0 < 0) || (y_0 >= d_height))
        throw CExceptionRange ("You can not create a window with "
        "this start point (y dimension).", 
        y_0, 0, d_height - 1);
    if (x_0 + width > d_width)
        throw CExceptionRange ("You can not create a window with "
        "this width.", width, 0, 
        d_width - x_0 );
    if (y_0 + height > d_height)
        throw CExceptionRange ("You can not create a window with "
        "this height.", height, 0, 
        d_height - y_0 );


    d_win_x = x_0;
    d_win_y = y_0;
    d_win_width = width;
    d_win_height = height;

    adjustRowPtr();
}


/**********************************************************/
/*! 
 *   Creates a window in the original image. The user of 
 *   an object will see only the pixels in this window. 
 *   This is a fast method to get subimages of an image.
 *
 *   \param rect the rectangle that specifies the window
 */
/**********************************************************/

void 
CImage::setWindow (const CRect& rect)
{
    setWindow (rect.getX(), rect.getY(), rect.getW(), rect.getH());
}


/**********************************************************/
/*! 
 *   Gets the window rectangle which is set iside a large
 *   image.
 *
 *   \param rect the rectangle that specifies the window
 */
/**********************************************************/

CRect
CImage::getWindow ()
{
    return CRect(d_win_x, d_win_y, d_win_x + getW(), d_win_y + getH());
}


/**********************************************************/
/*!
 *   Deletes an existing window. After using this 
 *   function the user of an object will see the whole
 *   image in the image buffer.
 */
/**********************************************************/

void 
CImage::resetWindow ()
{
    d_win_x = 0;
    d_win_y = 0;
    d_win_width = d_width;
    d_win_height = d_height;

    adjustRowPtr();
}


