/***************************************************************************
                          CImageAttributes.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Eric Sommerlade
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Eric Sommerlade

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv/CImageAttributes.h"
#include <cstring>

#ifdef WIN32
#include <string>
#endif

#include <map>

namespace TUVision {

    /**********************************************************/
    /*! \internal 
     *  Maps strings to doubles, that is, given a string, you 
     *  get a double.
     *
     *  Wraps the standard template library maps.
     *  Useful to strip down error messages and for forward 
     *  declarations.
     *
     *  \Warning Design decision: what should be done if the 
     *           queried string is not found in the map?
     *           <br>
     *           First possibility: Throw an exception.
     *           <br>
     *           Second: Return -inf or NaN.
     *           Third: Return the iterator directly, and test 
     *                  if this is valid (i.e. by passing it 
     *                  back)
     *           <br>
     *           Fourth: Return a default value that is
     *                   newly inserted into the map, 
     *                   i.e. 0 or the empty string.
     *           <br>
     *           For now, the fourth method is implemented, since
     *           exception handling causes too much overhead imho.
     *           Also, the third method implies that the user 
     *           has to check back, and he has to deal with the stl type 
     *           iterator, which points to a pair.
     *           The second method has the disadvantage of checking against
     *           validity, and afterwards inserting anyway.
     *           Since the values in the map are non-const, they
     *           can be modified via references.
     *  
     *  \author Eric Sommerlade
     */
    /**********************************************************/

    class CImageAttributes::DoubleMap
        : public std::map<std::string, double> 
    {
    private:

        //! Defines the default value.
        /*! This value is going to be returned if
         *  the specified attribute name does not exist.
         */
        static const double DEFAULT;

    public:

        // GET-/SETTERS
        double& get(const std::string& name);
        void set(const std::string& name, double value);
    };


    /**********************************************************/
    /*! \internal 
     *  Maps strings to strings, that is, given a string, you 
     *  get a string.
     *
     *  Wraps the standard template library maps.
     *  Useful to strip down error messages and for forward 
     *  declarations.
     */
    /**********************************************************/

    class CImageAttributes::StringMap: 
        public std::map<std::string, std::string> 
    {
    private:

        //! defines the default value
        /*! this value is going to be returned if
         *  the specified attribute name does not exist.
         */
        static const std::string DEFAULT;


    public:

        // GET-/SETTERS
        std::string& get(const std::string& name);
        void set(const std::string& name, const std::string& value);

    };

    const double CImageAttributes::DoubleMap::DEFAULT=0.0;
    const std::string CImageAttributes::StringMap::DEFAULT="";
}


using namespace TUVision;

/**********************************************************/
/*! \internal 
 * 
 *  STL convenience wrapper: tries to find the 
 *  specified name as a key in the map.
 *  If not found, it sets up a new pair and returns the 
 *  DEFAULT value specified in this class.
 *
 *  \param name The string that identifies the value of the 
 *              attribute.
 *  \return If it existed : the value of the specified 
 *          attribute - else it returns the default value.
 */
/**********************************************************/

double& 
CImageAttributes::DoubleMap::get (const std::string& name) 
{
    CImageAttributes::DoubleMap::iterator i;
    i = find (name);

    if (i != end()) {
        return i->second;
    }
    insert(value_type (name,DEFAULT));

    return find(name)->second;          
}

/**********************************************************/
/*! \Internal 
 * 
 *  STL convenience wrapper: tries to insert the specified 
 *  key/name pair into the map. STL behaves as follows:
 *  If the key is not found, it sets up a new pair and 
 *  inserts it appropriately.
 *  If the key is found already, the map complains 
 *  second return value is false) and the already exisiting
 *  pair is changed.
 *
 *  \param name The string that identifies the attribute.
 *  \param Value the string that defines the new value of
 *         the specified attribute.
 */
/**********************************************************/

void 
CImageAttributes::DoubleMap::set (const std::string& name, double value) 
{
    std::pair<iterator,bool> pos=insert(value_type(name,value));
    if (pos.second == false) {
        pos.first->second=value;
    }
}


/**********************************************************/
/*! \internal 
 * 
 *  STL convenience wrapper: tries to find the specified 
 *  name as a key in the map. If not found, it sets up a 
 *  new pair and returns the DEFAULT value specified in 
 *  this class.
 *
 *  \param name The string that identifies the value of 
 *              the attribute.
 *  \return If it existed : the value of the specified 
 *          attribute - else it returns the default value.
 */
/**********************************************************/

std::string& 
CImageAttributes::StringMap::get(const std::string& name) 
{
    CImageAttributes::StringMap::iterator i;
    i=find(name);

    if (i != end()) {
        return i->second;
    }
    insert (CImageAttributes::StringMap::value_type (name,DEFAULT));

    return find(name)->second;          
}


/**********************************************************/
/*! \internal 
 * 
 *  STL convenience wrapper: tries to insert the 
 *  specified key/name pair into the map. STL behaves as 
 *  follows: If the key is not found, it sets up a new pair
 *  and inserts it appropriately.
 *  If the key is found already, the map complains 
 *  second return value is false) and the already exisiting
 *  pair is changed.
 *
 *  \param name The string that identifies the attribute.
 *  \param value The string that defines the new value of
 *               the specified attribute.
 */
/**********************************************************/

void 
CImageAttributes::StringMap::set (const std::string& name, 
                                  const std::string& value) 
{
    std::pair<iterator,bool> pos=insert(value_type(name,value));
    if (pos.second == false) {
        pos.first->second=value;
    }
}

/**********************************************************/
/*! 
 *  Constructs a CImageAttributes. 
 *  Allocates the memory for the attribute maps.
 */
/**********************************************************/

CImageAttributes::CImageAttributes()
{
    d_doubles = new DoubleMap();
    d_strings = new StringMap();
}

/**********************************************************/
/*! 
 *  Constructs a CImageAttributes. 
 *  Allocates the memory for the attribute maps and copies
 *  the maps from the input CImageAttributes.
 *
 *  \param in the input CImageAttributes
 */
/**********************************************************/
CImageAttributes::CImageAttributes (const CImageAttributes& in)
{
    d_doubles = new DoubleMap(*(in.d_doubles));
    d_strings = new StringMap(*(in.d_strings));
}


/**********************************************************/
/*! 
 *  Destructor.
 *  Destroys the CImageAttributes, cleans up the attribute
 *  maps.
 */
/**********************************************************/

CImageAttributes::~CImageAttributes()
{
    delete d_doubles;
    delete d_strings;
}

/**********************************************************/
/*! 
 *  Get a double valued attribute of the image:
 *
 *  \param name The name of the attribute.
 *              If the attribute does not exist, an new 
 *              attribute is inserted and a reference to it 
 *              is returned.
 *
 *  \return The reference to the double value of the 
 *          attribute which has the specified name.
 */
/**********************************************************/

double& 
CImageAttributes::getDoubleAttribute (const std::string& name) const
{ 
    return d_doubles->get (name);
}

/**********************************************************/
/*! 
 *   Set the specified double attribute of the image: 
 *
 *   \param name name of the attribute to be set
 *   \param value value of the attribute to be set
 *
 */
/**********************************************************/

void
CImageAttributes::setDoubleAttribute (const std::string& name, 
                                      double value)
{ 
    d_doubles->set (name,value);
}

/**********************************************************/
/*! 
 *   Get a string attribute of the image:
 *
 *   \param name The name of the attribute.
 *               If the attribute does not exist, a new 
 *               attribute is inserted and set up with an 
 *               empty string. The reference to this string 
 *               is returned.
 *
 *   \return The reference to the string value of the 
 *           attribute which has the specified name
 */
/**********************************************************/

std::string& 
CImageAttributes::getStringAttribute (const std::string& name) const
{ 
    return d_strings->get (name);
}


/**********************************************************/
/*! 
 *   Sets the specified string attribute of the image:
 *
 *   \param name name of the string attribute to be set
 *   \param value value of the string attribute to be set
 */
/**********************************************************/

void
CImageAttributes::setStringAttribute (const std::string& name, 
                                      const std::string& value)
{
    d_strings->set (name,value);
}
