/***************************************************************************
                          CRect.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Eric Sommerlade
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Eric Sommerlade

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv/CRect.h"

#include <cstdio>

//using namespace std;
using namespace TUVision;


/* *********************************************************/
/*!
 * construct a rectangle with four absolute coordinates
 * \param ulx the upper left x-position of the rectangle
 * \param uly the upper left y-position of the rectangle
 * \param lrx the lower right x-position of the rectangle
 * \param lry the lower right y-position of the rectangle
 */
/* *********************************************************/
CRect::CRect(int ulx, int uly, int lrx, int lry) : d_x(ulx), d_y(uly), d_w(lrx-ulx), d_h(lry-uly) {
}

/* *********************************************************/
/*!
 * construct a rectangle given the upper left and lower right corner point
 * \param ul the upper left corner point
 * \param lr the lower right corner point
 */
/* *********************************************************/
CRect::CRect(CPoint ul, CPoint lr) : d_x(ul.getX()), d_y(ul.getY()), d_w(lr.getX()-ul.getX()), d_h(lr.getY() - ul.getY()){
}

/* *********************************************************/
/*!
 * construct a rectangle with specified width and height
 * \param w width of the rectangle
 * \param h height of the rectangle
 */
/* *********************************************************/
CRect::CRect(int w, int h)
: d_x(0), d_y(0), d_w(w), d_h(h)
{
}

/* *********************************************************/
/*!
 * sets the upper left corner of the rectangle.
 * the width and height remain unchanged (a translation).
 * \param x x coordinate of the upper left rectangle corner
 * \param y y coordinate of the upper left rectangle corner
 * \return the changed rectangle.
 */
/* *********************************************************/
CRect&
CRect::setUpperLeft(int x, int y)
{
    d_x=x;
    d_y=y;

    return *this;
}

/* *********************************************************/
/*!
 * sets the lower right corner of the rectangle.
 * the width and height are changed therefore
 * \param x x coordinate of the lower right rectangle corner
 * \param y y coordinate of the lower right rectangle corner
 * \return the changed rectangle.
 */
/* *********************************************************/
CRect&
CRect::setLowerRight(int x, int y)
{
    d_w=x-d_x+1;
    d_h=y-d_y+1;

    return *this;
}

/* *********************************************************/
/*!
 * sets the center of the rectangle
 *
 * adjusts the upper left corner of the rectangle.
 * the width and height are rounded downwards
 * for adjustment.
 *
 * \param x x coordinate of the rectangle center
 * \param y y coordinate of the rectangle center
 * \return the changed rectangle.
 */
/* *********************************************************/
CRect& 
CRect::setCenter(int x, int y)
{
    d_x=x-(d_w/2);
    d_y=y-(d_h/2);
    return *this;
}


/* *********************************************************/
/*!
 * sets the four corner points to absolute values
 *
 * \param ulx the upper left x-position of the rectangle
 * \param uly the upper left y-position of the rectangle
 * \param lrx the lower right x-position of the rectangle
 * \param lry the lower right y-position of the rectangle
 */
/* *********************************************************/
void CRect::setRect(int ulx, int uly, int lrx, int lry){
    d_x=ulx;
    d_y=uly;
    d_w=lrx-ulx;
    d_h=lry-uly;
}

/* *********************************************************/
/*!
 * sets the two corners (i.e. the four corner points) to absolute values
 * \param ul the upper left corner point
 * \param lr the lower right corner point
 */
/* *********************************************************/
void CRect::setRect(CPoint ul, CPoint lr){
    d_x=ul.getX();
    d_y=ul.getY();
    d_w=lr.getX()-ul.getX();
    d_h=lr.getY()-ul.getY();
}

/* *********************************************************/
/*!
 * resizes the rectangle
 *
 * the left upper corner of the rectangle remains unchanged.
 * i.e. a translation and a scale.
 *
 * \param w the new width
 * \param h the new height
 *
 * \return the changed rectangle.
 */
/* *********************************************************/
CRect& 
CRect::setSize(int w, int h)
{
    d_w=w;
    d_h=h;

    return *this;
}

/* *********************************************************/
/*!
 * gets the lower right x coordinate of the rectangle
 *
 */
int CRect::getLRX() const
{
    return d_w+d_x;
}


/* *********************************************************/
/*!
 * gets the lower right y coordinate of the rectangle
 *
 */
int CRect::getLRY() const
{
    return d_y+d_h;
}

/* *********************************************************/
/*!
 * gets the width of the rectangle
 * \return the width.
 */
/* *********************************************************/
int 
CRect::getW() const
{
    return d_w;
}


/* *********************************************************/
/*!
 * gets the height of the rectangle
 * \return the height.
 */
/* *********************************************************/
int 
CRect::getH() const
{
    return d_h;
}

/* *********************************************************/
/*!
 * gets the x coordinate of the center of the rectangle
 * \return the x center.
 */
/* *********************************************************/
int 
CRect::getCenterX() const
{
    return d_x+(d_w>>1);
}

/* *********************************************************/
/*!
 * gets the y coordinate of the center of the rectangle
 * \return the y center.
 */
/* *********************************************************/
int 
CRect::getCenterY() const
{
    return d_y+(d_h>>1);
}

/* *********************************************************/
/*!
 * gets the upper left point of the rectangle
 */
/* *********************************************************/

CPoint CRect::getUpperLeft() const
{
    return CPoint(d_x,d_y);
}

/* *********************************************************/
/*!
 * gets the lower right point of the rectangle
 */
/* *********************************************************/
CPoint CRect::getLowerRight() const
{
    return CPoint(d_x+d_w-1,d_y+d_h-1);
}


/* *********************************************************/
/*!
 * gets the x coordinate of the upper left corner  of the rectangle
 * \return the x coordinate of the upper left corner.
 */
/* *********************************************************/
int 
CRect::getX() const
{
    return d_x;
}

/* *********************************************************/
/*!
 * gets the y coordinate of the upper left corner  of the rectangle
 * \return the y coordinate of the upper left corner.
 */
/* *********************************************************/
int 
CRect::getY() const
{
    return d_y;
}

/* *********************************************************/
/*!
 * checks wether the rectangles width or height is null.
 * \return true if the rectangles width or height is null.
 *         else it returns false.
 */
/* *********************************************************/
bool 
CRect::isNull() const
{
    return (d_w==0)&&(d_h==0);
}


/* *********************************************************/
/*!
 * intersects two rectangles and returns the intersection.
 *
 * \param r the rectangle to be checked against.
 *
 * \return the intersected rectangle.
 * \warning the rectangle may have no extension (i.e. 
 *          isNull() is true! 
 */
/* *********************************************************/
CRect 
CRect::intersect(const CRect& r) const
{
    if(!intersects(r))
        return CRect(0,0);

#define rmax(x,y) ((x)>(y)?(x):(y))
#define rmin(x,y) ((x)<(y)?(x):(y))

    int rr,ll,u,d;

    // left, right
    ll=rmax(d_x,r.d_x);
    rr=rmin(d_x+d_w,r.d_x+r.d_w);

    // up, down
    u=rmax(d_y,r.d_y);
    d=rmin(d_y+d_h,r.d_y+r.d_h);

    CRect intersection(rr-ll,d-u);

    return(intersection.setUpperLeft(ll,u));
}
/* *********************************************************/
/*!
 * checks wether the rectangles intersects the specified rectangle.
 *
 * \param r the rectangle to be checked against.
 *
 * \return true if the rectangles intersects the specified rectangle.
 *         else it returns false.
 */
/* *********************************************************/
bool 
CRect::intersects(const CRect& r) const
{
    if(isNull()||r.isNull()) return false;

    if(d_x+d_w<=r.d_x) return false;
    if(d_y+d_h<=r.d_y) return false;
    if(r.d_x+r.d_w<=d_x) return false;
    if(r.d_y+r.d_h<=d_y) return false;

    return true;
}

/* *********************************************************/
/*!
 * checks wether the rectangles contains the specified point.
 *
 * \param point the point to be checked against.
 *
 * \return true if the rectangles contains the specified point.
 *         else it returns false.
 */
/* *********************************************************/

bool CRect::contains(const CPoint point) const
{
    return contains(point.getX(),point.getY());
}

/* *********************************************************/
/*!
 * checks wether the rectangles contains the specified point.
 *
 * \param x the x coordinate of the specified point
 * \param y the y coordinate of the specified point
 *
 * \return true if the rectangles contains the specified point.
 *         else it returns false.
 */
/* *********************************************************/
bool 
CRect::contains(int x, int y) const
{
    if(isNull()) return false;

    if(x<d_x) return false;
    if(y<d_y) return false;
    if(d_x+d_w<=x) return false;
    if(d_y+d_h<=y) return false;

    return true;
}


/* *********************************************************/
/*!
 * checks wether the rectangles contains the specified rectangle.
 *
 * \param r the rectangle to be checked against.
 *
 * \return true if the rectangles contains the specified rectangle.
 *         else it returns false.
 */
/* *********************************************************/
bool 
CRect::contains(const CRect& r) const
{
    if(isNull()||r.isNull()) return false;

    //  printf("a\n");
    if(!contains(r.d_x,r.d_y)) return false;
    //  printf("b\n");
    if(!contains(r.d_x+r.d_w-1,r.d_y)) return false;
    //  printf("c\n");
    if(!contains(r.d_x,r.d_y+r.d_h-1)) return false;
    //  printf("d\n");
    if(!contains(r.d_x+r.d_w-1,r.d_y+r.d_h-1)) return false;

    return true;
}


/* *********************************************************/
/*!
 * fits the size of the rectangle 
 * so that it contains the specified point
 * \param x the x coordinate of the point
 * \param y the y coordinate of the point
 * \return the same rectangle, but fitted to the
 *         point.
 */
/* *********************************************************/
CRect& 
CRect::fitSize(int x, int y)
{
    if(x<d_x)
        d_x=x;
    if(x>=d_x+d_w)
        d_w=x-d_x+1;
    if(y<d_y)
        d_y=y;
    if(y>=d_y+d_h)
        d_h=y-d_y+1;

    return *this;
}

/* *********************************************************/
/*!
 * compares two rectangles.
 *
 * \param r1 rectangle 1
 * \param r2 rectangle 2
 * \return true, if both the rectangle have the same position and size.
 */
/* *********************************************************/
bool 
TUVision::operator==(const CRect& r1, const CRect& r2) 
{
    return( ( r1.getX()==r2.getX() ) && ( r1.getY()==r2.getY() ) 
        &&    ( r1.getW()==r2.getW() ) && ( r1.getH()==r2.getH() ) );

}

/* *********************************************************/
/*!
 * this function prints the rectangle parameters to stdout
 * 
 * \param r the rectangle to be printed.
 */
/* *********************************************************/
void 
TUVision::dump(const CRect& r) 
{
    printf("%i,%i - %i,%i\n",r.getX(),r.getY(),r.getW()+r.getX(),r.getH()+r.getY());
}
