/***************************************************************************
                          CFileReadJpeg.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include <string>
#include <cstdio>
#include <csetjmp> 
#include <exception>

extern "C" {       // fixme?
#include <jpeglib.h>
}

#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CFileReadJpeg.h"
#include "tuv/CException.h"
#include "tuv/CExceptionFile.h"


using namespace std;
using namespace TUVision;

const unsigned char CFileReadJpegRegistrator::d_id[]={(char) 0xff,(char) 0xd8,(char) 0xff,(char) 0xe0};
const int CFileReadJpegRegistrator::d_len=4;


/******************************************************/
/*!  \internal 
 *   
 *   Used to transfer data to the error handler.
 */
/******************************************************/

struct my_client_data {
    jmp_buf setjmp_buffer;     //!< for return to caller 
    string error_string;       //!< description of the error
};


#if JPEG_LIB_VERSION >= 62
/******************************************************/
/*!  \internal
 *
 *   This routine replaces the standard error_exit 
 *   method of libjpeg.
 */
/******************************************************/

static void my_error_exit (j_common_ptr cinfo)
{
    my_client_data  *data;
    char            buf[JMSG_LENGTH_MAX];

    data = (my_client_data*) cinfo->client_data;

    // Get description of the error as string.
    (*cinfo->err->format_message) (cinfo, buf);
    data->error_string = buf;
    data->error_string += ".";

    // Return control to the setjmp point.
    longjmp(data->setjmp_buffer, 1);
}
#endif


CFileReadJpeg::CFileReadJpeg (void)
{
    d_filename = "";
    d_hd = NULL;
}


CFileReadJpeg::~CFileReadJpeg()
{
    if (d_hd != NULL)
        fclose (d_hd);
}


void CFileReadJpeg::open (const string &filename)
throw (CException)
{
    d_filename = filename;

    d_hd = fopen (d_filename.c_str(), "rb");
    if (d_hd == NULL)
        throw CExceptionFileOpen (d_filename);
}


void CFileReadJpeg::close (void)
throw()
{
    if (d_hd != NULL)
    {
        fclose (d_hd);
        d_hd = NULL;
    }
}


CImage *CFileReadJpeg::read (int img_number)
throw (CException, exception)
{
    struct jpeg_decompress_struct  cinfo; 
    struct jpeg_error_mgr          jerr;
    my_client_data                 data;
    CImageByte                     *img = NULL;
    unsigned char                  *row;
    int                            y;


    // Valid file handle?
    if (d_hd == NULL)
        throw CExceptionFileRead (d_filename);

    // Install our own error handler. This handler is used
    // to get a description of the problem and throw an
    // exception.
    cinfo.err = jpeg_std_error (&jerr);
#if JPEG_LIB_VERSION >= 62
    jerr.error_exit = my_error_exit;
    cinfo.client_data = &data;

    if (setjmp (data.setjmp_buffer)) {
        // If we get here, the JPEG code has signaled an error.

        jpeg_destroy_decompress (&cinfo);

        if (img != NULL)
            delete img;

        throw CExceptionFileRead (d_filename, data.error_string);
    }
#endif

    // Init libjpeg object and read JPEG header.
    jpeg_create_decompress (&cinfo);
    jpeg_stdio_src (&cinfo, d_hd);
    jpeg_read_header (&cinfo, TRUE);
    jpeg_start_decompress (&cinfo);


    switch (cinfo.out_color_space) {
case JCS_GRAYSCALE:
    // Read a 8 bit gray image.
    { 
        try {
            img = new CImageByteGray (cinfo.output_width, 
                cinfo.output_height);
        }
        catch (...) {
            jpeg_destroy_decompress (&cinfo);
            throw;
        }

        for (y = 0; y < img->getH(); y++)
        {
            row = img->getRow (y);
            jpeg_read_scanlines (&cinfo, &row, 1);
        }
    }
    break;
case JCS_RGB:
    // Read a 8 bit RGB image.
    { 
        try {
            img = new CImageByteRGB (cinfo.output_width, 
                cinfo.output_height);
        }
        catch (...) {
            jpeg_destroy_decompress (&cinfo);
            throw;
        }

        for (y = 0; y < img->getH(); y++)
        {
            row = img->getRow (y);
            jpeg_read_scanlines (&cinfo, &row, 1);
        }
    }
    break;
default:
    // unsupported image format
    {
        jpeg_destroy_decompress(&cinfo);
        throw CExceptionFileRead (d_filename, 
            "The color model or the resolution of the image "
            "is not supported.");
    }
    }
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);

    return (img);
} 


int CFileReadJpeg::getNumberImages (void)
throw()
{
    return (1);
}


