/***************************************************************************
                          CFileReadMpeg.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Eric Sommerlade
    email                : tuvisionlib@tuvision.de
    origin               : Studienarbeit/Diplomarbeit Eric Sommerlade

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv/CFileReadMpeg.h"
#include "tuv/CImageByteYUV.h"
#include "tuv/CExceptionFile.h"

#include <libmpeg2.h>

using namespace TUVision;

const unsigned char CFileReadMpegRegistrator::d_id[]={0x00,0x00,0x01};
const int CFileReadMpegRegistrator::d_len=3;


/*! \internal
 *
 *  Not tested with interlaced images.
 *
 *  Conversion from 444 has not been tested - but it should work, it is only 
 *  interleaving.
 *
 *  Upconversion from 422 has been tested with
 *  ftp://ftp.sunet.se/pub/multimedia/Berkeley/mpeg2/conformance-bitstreams/video/bitstreams/422-profile/hhi/hhi_burst_422/hhi_burst_422_short.bits.gz
 *
 *  Upconversion from 420 (MP@ML) has been tested with
 *  ftp://ftp.sunet.se/pub/multimedia/Berkeley/mpeg2/conformance-bitstreams/video/bitstreams/main-profile/hhi/hhi_burst_short/hhi_burst_short.bits.gz
 *  
 */
static CImage* 
convertToYUV(TPictureInfo* pi, TStreamInfo* si) 
{
    CImageByteYUV* yuv=new CImageByteYUV(si->w,si->h);

    unsigned int i,j;
    for(j=0;j<si->h;++j) {
        unsigned char* dst=yuv->getRow(j);
        byte *y_src=pi->Y;
        byte *u_src=pi->Cr;
        byte *v_src=pi->Cb;

        y_src+=si->codedw*j;

        switch(si->chroma) {
      case chroma444:
          {
              u_src+=si->codedw*j;
              v_src+=si->codedw*j;
              for(i=0;i<si->w;++i) {
                  *dst++=*y_src++;
                  *dst++=*u_src++;
                  *dst++=*v_src++;
              }
          }
          break;
      case chroma422:
          {
              u_src+=si->codedw*(j>>1);
              v_src+=si->codedw*(j>>1);
              for(i=0;i<si->w;++i) {
                  *dst++=*y_src++;
                  *dst++=*(u_src+(i>>1));
                  *dst++=*(v_src+(i>>1));
              }
          }
          break;
          // falls through to default
      case chroma420:
      default:
          {
              u_src+=si->codedw*(j>>2);
              v_src+=si->codedw*(j>>2);
              for(i=0;i<si->w;++i) {
                  *dst++=*y_src++;
                  *dst++=*(u_src+(i>>1));
                  *dst++=*(v_src+(i>>1));
              }
          }
        } // end switch
    }
    return(yuv);
}

CFileReadMpeg::CFileReadMpeg ():
d_si(0), d_openedstream(false)
{
}

CFileReadMpeg::~CFileReadMpeg()
{
    if(d_openedstream){
        deinitializeSequence(d_si);
        deinitializeDecoder();
    }
}

void
CFileReadMpeg::open(const std::string &filename)
{
    if(!d_openedstream){
        char *args[3]={"a","-b",0};

        args[2]=const_cast<char*>(filename.c_str());
        if(initializeDecoder(3, args)) {
            throw CExceptionFileRead(filename,"could not initialize mpeg stream.");
        }
        d_si=initializeSequence();
        d_openedstream=true;
    } else {
        throw CExceptionFileRead(filename,"file has been opened already.");
    }
}

void 
CFileReadMpeg::close()
{
    if(d_openedstream){
        deinitializeSequence(d_si);
        deinitializeDecoder();
    }
    d_openedstream=false;
}



CImage* 
CFileReadMpeg::read(int img_number)
{
    TPictureInfo *pi;

    pi=nextPicture(d_si);

    if(pi==NULL)
        throw CExceptionFileRead("no more images in mpeg stream.");

    return(convertToYUV(pi,d_si));

}

int CFileReadMpeg::getNumberImages()
{
    return(-1);
}

/**********************************************************/
/*! \example io_mpeg.cpp 
 *   This gives an example on using the class 
 *   CFileReadMpeg.
 *
 *   This example shows how to read and write all images
 *   from a mpeg 1/2 or system stream file using the I/O classes.
 *
 *   The exceptions are caught to yield a good error
 *   resilience, even though this is only a example program.
 *
 *   The actual logic to splice up an mpeg stream into 
 *   its single pictures is about 12 lines only.
 *
 *   
 *
 *   \author Eric Sommerlade
 */
/**********************************************************/
/**********************************************************/
/*! \example getPicture.cpp 
 *   This gives an example on using the class 
 *   CFileReadMpeg.
 *
 *   This example shows how to read and write the first image
 *   from a mpeg 1/2 or system stream file using the I/O classes.
 *  
 *   The image is filtered and then saved as a tiff image.
 *
 *
 *   \author Eric Sommerlade
 */
/**********************************************************/
