/***************************************************************************
                          CFileReadTiff.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv/CImageByteBilevel.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CImageAttributes.h"
#include "tuv/CFileReadTiff.h"
#include "tuv/CException.h"
#include "tuv/CExceptionFile.h"

#include <string>
#include <exception>
#include <tiffio.h>

using namespace std;
using namespace TUVision;


const unsigned char CFileReadTiffBigEndianRegistrator::d_id[]={'M','M',0x00,0x2a};
const int CFileReadTiffBigEndianRegistrator::d_len=4;

const unsigned char CFileReadTiffLittleEndianRegistrator::d_id[]={'I','I',0x2a,0x00};
const int CFileReadTiffLittleEndianRegistrator::d_len=4;

CFileReadTiff::CFileReadTiff (void)
{
    d_filename = "";
    d_tif_hd = NULL;

    // Disable the error handler of the libtiff. We use
    // CException instead.
    TIFFSetErrorHandler (NULL);
}


CFileReadTiff::~CFileReadTiff()
{
    if (d_tif_hd != NULL)
        TIFFClose (d_tif_hd);
}


void CFileReadTiff::open (const string &filename) 
throw (CException)
{
    d_filename = filename;

    d_tif_hd = TIFFOpen (d_filename.c_str(), "r");
    if (d_tif_hd == NULL)
        throw CExceptionFileOpen (d_filename);
}


void CFileReadTiff::close (void) throw()
{
    if (d_tif_hd != NULL)
    {
        TIFFClose (d_tif_hd);
        d_tif_hd = NULL;
    }
}


/******************************************************/
/*! 
 *   Reads the following CImageAttributes from the
 *   TIFF file opened by this class: xdpi, and ydpi.
 *
 *   \param img image object
 */
/******************************************************/

void CFileReadTiff::readAttributes (CImage &img)
{
    float xdpi, ydpi;
    short unit;


    // Dots per inch or per centimeter?
    if (TIFFGetField (d_tif_hd, TIFFTAG_RESOLUTIONUNIT, &unit) != 1)
        unit = RESUNIT_INCH;

    // Get x/y resolution.
    if (TIFFGetField (d_tif_hd, TIFFTAG_XRESOLUTION, &xdpi) == 1)
    {
        if (unit == RESUNIT_CENTIMETER)
            xdpi *= 2,54;  //  cm -> inch
        img.getAttributes().setDoubleAttribute ("xdpi", xdpi); 
    }
    if (TIFFGetField (d_tif_hd, TIFFTAG_YRESOLUTION, &ydpi) == 1)
    {
        if (unit == RESUNIT_CENTIMETER)
            ydpi *= 2,54;  //  cm -> inch
        img.getAttributes().setDoubleAttribute ("ydpi", ydpi); 
    }
}


CImage *CFileReadTiff::read (int img_number) 
throw (exception, CException)
{
    unsigned int    size_x, size_y;
    unsigned short  bits_per_sample, samples_per_pixel;
    unsigned short  photometric;
    CImageByte      *img = NULL;
    unsigned char   *img_buffer;
    unsigned char   *my_img_buffer;
    unsigned int    i, x;


    // TIFF file handle valid?
    if (d_tif_hd == NULL)
        throw CExceptionFileRead (d_filename);

    // Set current image.
    if (TIFFSetDirectory (d_tif_hd, (unsigned int) img_number) != 1)
        throw CExceptionFileRead (d_filename);

    // Get some informations about image (width, height, 
    // color model, etc.)
    if (TIFFGetField (d_tif_hd, TIFFTAG_IMAGEWIDTH, &size_x) != 1)
        throw CExceptionFileRead (d_filename);
    if (TIFFGetField (d_tif_hd, TIFFTAG_IMAGELENGTH, &size_y) != 1)
        throw CExceptionFileRead (d_filename);
    if (TIFFGetField (d_tif_hd, TIFFTAG_BITSPERSAMPLE, 
        &bits_per_sample) != 1)
        throw CExceptionFileRead (d_filename);
    if (TIFFGetField (d_tif_hd, TIFFTAG_SAMPLESPERPIXEL, 
        &samples_per_pixel) != 1)
        throw CExceptionFileRead (d_filename);
    if (TIFFGetField (d_tif_hd, TIFFTAG_PHOTOMETRIC, 
        &photometric) != 1)
        throw CExceptionFileRead (d_filename);    


    // bilevel image?
    if (photometric == PHOTOMETRIC_MINISBLACK &&
        samples_per_pixel == 1 &&
        bits_per_sample == 1)
    {
        img = new CImageByteBilevel (size_x, size_y);

        try {
            my_img_buffer = new unsigned char [size_x/8 + 1];
        } 
        catch (...) {
            delete img;
            throw;
        }

        readAttributes (*img);

        for (i = 0; i < size_y; i++)
        {
            img_buffer = img->getRow (i);

            if (TIFFReadScanline (d_tif_hd, my_img_buffer, i) == -1)
            {
                delete img;
                delete [] my_img_buffer;
                throw CExceptionFileRead (d_filename);
            }  

            // Convert data format (file -> memory).
            for (x = 0; x < size_x; x++)
            {
                if ((my_img_buffer[x/8] & (1 << (7-(x % 8)))) ==
                    (1 << (7-(x % 8))))
                    img_buffer[x] = 1;
                else
                    img_buffer[x] = 0;
            }
        }

        delete [] my_img_buffer;
    } 

    // 8 bit gray level image? 
    else if (photometric == PHOTOMETRIC_MINISBLACK &&
        samples_per_pixel == 1 &&
        bits_per_sample == 8)
    {
        img = new CImageByteGray (size_x, size_y);

        readAttributes (*img);

        for (i = 0; i < size_y; i++)
        {
            img_buffer = img->getRow (i);

            if (TIFFReadScanline (d_tif_hd, img_buffer, i) == -1)
            {
                delete img;
                throw CExceptionFileRead (d_filename);
            }  
        }
    } 

    // 8 bit RGB image?
    else if (photometric == PHOTOMETRIC_RGB &&
        samples_per_pixel == 3 &&
        bits_per_sample == 8)
    {
        img = new CImageByteRGB (size_x, size_y);

        readAttributes (*img);

        for (i = 0; i < size_y; i++)
        {
            img_buffer = img->getRow (i);

            if (TIFFReadScanline (d_tif_hd, img_buffer, i) == -1)
            {
                delete img;
                throw CExceptionFileRead (d_filename);
            }  
        }
    } 

    // not supported image format
    else
        throw CExceptionFileRead (d_filename, 
        "The color model or the resolution of the image "
        "is not supported.");

    return (img);
} 


int CFileReadTiff::getNumberImages (void) 
throw (CException)
{
    tdir_t  current;
    int     i = 0;


    // TIFF file handle valid?
    if (d_tif_hd == NULL)
        throw CExceptionFileRead (d_filename);

    // Get current image number.
    current = TIFFCurrentDirectory (d_tif_hd);

    // Count number of images.
    while (TIFFSetDirectory (d_tif_hd, i) == 1)
        i++;

    // Restore old directory.
    if (TIFFSetDirectory (d_tif_hd, current) != 1)
        throw CExceptionFileRead (d_filename);

    return (i);
}


