/***************************************************************************
                          CFileWriteJpeg.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include <cstdio>
#include <string>
#include <csetjmp> 

extern "C" {             // fixme - MB?
#include <jpeglib.h>
}

#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CFileWriteJpeg.h"
#include "tuv/CException.h"
#include "tuv/CExceptionFile.h"


using namespace std;
using namespace TUVision;


/******************************************************/
/*!  \internal 
 *   
 *   Used to transfer data to the error handler.
 */
/******************************************************/

struct my_client_data {
    jmp_buf setjmp_buffer;     //!< for return to caller 
    string error_string;       //!< description of the error
};


/******************************************************/
/*!  \internal
 *
 *   This routine replaces the standard error_exit 
 *   method of libjpeg.
 *
 */
/******************************************************/

#if JPEG_LIB_VERSION >= 62
static void my_error_exit (j_common_ptr cinfo)
{
    my_client_data  *data;
    char            buf[JMSG_LENGTH_MAX];


    data = (my_client_data*) cinfo->client_data;

    // Get description of the error as string.
    (*cinfo->err->format_message) (cinfo, buf);
    data->error_string = buf;
    data->error_string += ".";

    // Return control to the setjmp point.
    longjmp(data->setjmp_buffer, 1);
}
#endif


CFileWriteJpeg::CFileWriteJpeg (void)
{
    d_filename = "";
    d_hd = NULL;
    d_quality = 75;
    d_progression = false;
}


CFileWriteJpeg::~CFileWriteJpeg()
{
    if (d_hd != NULL)
        fclose (d_hd);
}


void CFileWriteJpeg::open (const string &filename)
throw (CException)
{
    d_filename = filename;

    d_hd = fopen (d_filename.c_str(), "wb");
    if (d_hd == NULL)
        throw CExceptionFileOpen (d_filename);
}


void CFileWriteJpeg::close (void)
throw()
{
    if (d_hd != NULL)
    {
        fclose (d_hd);
        d_hd = NULL;
    }
}


/******************************************************/
/*! 
 *   This function writes the given image to a JPEG 
 *   graphic file. 
 *
 *   \param img         image to be saved
 *   \param color_space color space of the image:
 *                      grayscale (1) or RGB (3)
 */
/******************************************************/

void CFileWriteJpeg::write (const CImageByte &img, 
                            int color_space)
                            throw (CException)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr       jerr;
    JSAMPROW                    row_pointer[1]; 
    my_client_data              data;
    int                         y;     


    // Valid file handle?
    if (d_hd == NULL)
        throw CExceptionFileWrite (d_filename);

    // Install our own error handler. This handler is used
    // to get a description of the problem and throw an
    // exception.
    cinfo.err = jpeg_std_error (&jerr);
#if JPEG_LIB_VERSION >= 62
    jerr.error_exit = my_error_exit;
    cinfo.client_data = &data;

    if (setjmp (data.setjmp_buffer)) {
        // If we get here, the JPEG code has signaled an error.

        jpeg_destroy_compress (&cinfo);
        throw CExceptionFileWrite (d_filename, data.error_string);
    }
#endif
    // Init libjpeg object and create JPEG header.
    jpeg_create_compress (&cinfo);
    jpeg_stdio_dest (&cinfo, d_hd);
    cinfo.image_width = img.getW();
    cinfo.image_height = img.getH();
    cinfo.input_components = color_space;

    if (color_space == 1)
        cinfo.in_color_space = JCS_GRAYSCALE;
    else
        cinfo.in_color_space = JCS_RGB;

    jpeg_set_defaults (&cinfo);
    jpeg_set_quality (&cinfo, d_quality, TRUE);

    // Create a progressive JPEG?
    if (d_progression == true)
        jpeg_simple_progression (&cinfo);


    // Store image data.
    jpeg_start_compress(&cinfo, TRUE);

    for (y = 0; y < img.getH(); y++) 
    {
        row_pointer[0] = const_cast<unsigned char*> (img.getRow (y));
        jpeg_write_scanlines (&cinfo, row_pointer, 1);
    }

    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);
}


/******************************************************/
/*! 
 *   This function writes the given image (8 bit gray)
 *   to a JPEG graphic file. 
 *
 *   \param img image to be saved
 */
/******************************************************/

void CFileWriteJpeg::write (const CImageByteGray &img)
throw (CException)
{
    write (img, 1);
}


/******************************************************/
/*! 
 *   This function writes the given image (8 bit RGB)
 *   to a JPEG graphic file. 
 *
 *   \param img image to be saved
 */
/******************************************************/

void CFileWriteJpeg::write (const CImageByteRGB &img)
throw (CException)
{
    write (img, 3);
}


/******************************************************/
/*! 
 *   Sets the quality value of the JPEG file. The 
 *   value is expressed on a 0-100 scale. A larger
 *   value means a better quality. The default quality
 *   of this class is 75.
 *
 *   \param quality quality to be used
 */
/******************************************************/

void CFileWriteJpeg::setQuality (int quality)
throw ()
{
    if (d_quality > 100)
        d_quality = 100;
    else if (d_quality < 0)
        d_quality = 0;
    else
        d_quality = quality;
}


/******************************************************/
/*! 
 *   Sets whether whether a normal or a progressive
 *   JPEG file should be created.
 *
 *   Progressive JPEG rearranges the stored data into a 
 *   series of scans of increasing quality. This is
 *   interesting for the transmission of JPEG files 
 *   over slow communication links, because the decoder
 *   can generate a low-quality image very quickly from 
 *   the first scan and then gradually improve the 
 *   displayed quality as more scans are received.
 *
 *   \param progression Use progression?
 */
/******************************************************/

void CFileWriteJpeg::setProgression (bool progression)
throw ()
{
    d_progression = progression;
}
