/***************************************************************************
                          CExtractorShape.cpp  -  description
                             -------------------
    begin                : 10.04.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library's internal version. It is   *
 *   not permitted to publish this file at this point of time as the       *
 *   algorithms are meant to be shared between TUHH Vision Systems         *
 *   members only.                                                         *
 *                                                                         *
 ***************************************************************************/

#include "tuv2d/CExtractorShape.h"

using namespace TUVision;

/**********************************************************/
/*!
 *   Sets the current shape list empty
 */
/**********************************************************/
CExtractorShape::CExtractorShape()
{
    d_currentShape = 0;
}

CExtractorShape::~CExtractorShape()
{
}

/**********************************************************/
/*!
 *   Extracts shapes in an image. This function is the main
 *   routine of the extractor. It searches the whole image 
 *   for objects which either can be defined by non-zero pixels
 *   or by zero-pixels. Also it can be set if the shape of the
 *   object should be inside an object or outside. Note that when
 *   determining the outside shape of an object in the image,
 *   resulting shape - points can lie outside the image itself,
 *   so normally it is better to choose the "inside" - shape.
 *
 * \param image the image which is been searched for objects. Currently it has to be gray
 * \param ObjectContentIsZero flag which determines objects are defined by non-zero pixels or by zero pixels
 * \param ShapeInsideOfOjbect flag which determines if the contour of an object is supposed to be inside the object
 *
 */
/**********************************************************/
list <CPointShape *> *CExtractorShape::extractShapes(const CImageByteGray *image, bool ObjectContentIsZero, bool ShapeInsideOfObject)
{
    p_image=image;
    int x,y;
    const unsigned char *row;
    int *mask_row;

    d_objectContentIsZero=ObjectContentIsZero;
    d_shapeInsideOfObject=ShapeInsideOfObject;

    d_row.resize(image->getH());

    for(y=0;y<image->getH();y++)
        d_row[y]=image->getRow(y);

    d_shapes = new list<CPointShape *>;

    d_detectedShapesMask = new CImageIntGray(image->getW(),image->getH());
    d_detectedShapesMask->clearImage();
    d_maskHashVector     = new minmaxline[p_image->getH()];

    if(!d_shapeInsideOfObject){ // Even though the two following cases could easily be combined to one, extraction is faster when having as few operations as possible

        // Run around outside
        for(y=0;y<image->getH();y++){
            row=d_row[y];
            mask_row=d_detectedShapesMask->getRow(y);
            for(x=0;x<image->getW();x++){
                if( !mask_row[x] ){
                    if( (row[x]==0) == d_objectContentIsZero){
                        x=(extractNextShapeOutside(CPoint(x-1,y)));
                    }
                } else {
                    x = mask_row[x];
                }
            } 
        }

    } else {

        // Run around inside
        for(y=0;y<image->getH();y++){
            row=d_row[y];
            mask_row=d_detectedShapesMask->getRow(y);
            for(x=0;x<image->getW();x++)
                if( (row[x]==0) == d_objectContentIsZero){
                    if(!mask_row[x]){
                        x=(extractNextShapeInside(CPoint(x,y)));
                    } else {
                        x = mask_row[x];
                    }
                }
        }

    }

    delete d_detectedShapesMask;
    delete [] d_maskHashVector;
    return d_shapes;
}

/**********************************************************/
/*!
 *   Runs around an object's shape by successively
 *   finding contour points.
 *
 * \param firstrun internal flag used to determine if we have started running or not
 * \param firstdir first direction the automaton should look at
 * \param startpoint the point in the image just outside of an object
 *
 */
/**********************************************************/
void CExtractorShape::runAroundOutside(int firstdir, CPoint startpoint, bool firstrun)
{
    int testx,testy,newtestx,newtesty;
    const int xoffsets[4]={0,1,0,-1};
    const int yoffsets[4]={-1,0,1,0};

    if(startpoint==d_startPoint && !firstdir && !firstrun) // Halt criterion: If we are in the initial state (and not starting): return
        return;

    //UP
    newtestx=startpoint.getX()+xoffsets[(0+firstdir)%4];
    newtesty=startpoint.getY()+yoffsets[(0+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+0)+3)%4,startpoint,false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+1)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //RIGHT
    newtestx=testx+xoffsets[(1+firstdir)%4];
    newtesty=testy+yoffsets[(1+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+1)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+1)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //RIGHT
    newtestx=testx+xoffsets[(1+firstdir)%4];
    newtesty=testy+yoffsets[(1+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+1)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+2)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //DOWN
    newtestx=testx+xoffsets[(2+firstdir)%4];
    newtesty=testy+yoffsets[(2+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+2)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+2)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //DOWN
    newtestx=testx+xoffsets[(2+firstdir)%4];
    newtesty=testy+yoffsets[(2+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+2)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+3)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //LEFT
    newtestx=testx+xoffsets[(3+firstdir)%4];
    newtesty=testy+yoffsets[(3+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+3)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+3)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //LEFT
    newtestx=testx+xoffsets[(3+firstdir)%4];
    newtesty=testy+yoffsets[(3+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+3)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+0)%4) )
            return;
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
    //UP
    newtestx=testx+xoffsets[(0+firstdir)%4];
    newtesty=testy+yoffsets[(0+firstdir)%4];
    if(newtestx >=0 && newtestx < p_image->getW() && newtesty >= 0 && newtesty < p_image->getH() && d_row[newtesty][newtestx]!=(unsigned char) d_objectContentIsZero){
        runAroundOutside(((firstdir+0)+3)%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        updateHashVectorOutside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
}

/**********************************************************/
/*!
 *   Updates the incidence image when having finished (outside version)
 *   the extraction of one object. This way it can be determined
 *   if new candidates in the image belong to objects which have already
 *   been found
 *
 * \param startline internal parameter which describes the row object extraction has begun
 *
 */
/**********************************************************/
int CExtractorShape::updateShapesMaskOutside(int startline)
{
    int i;

    CPoint BBXUL(d_maskHashVector[startline].min,startline);
    CPoint BBXLR(d_maskHashVector[startline].min,startline);

    for(i=startline;i<p_image->getH();i++){
        if(d_maskHashVector[i].min==p_image->getW() && d_maskHashVector[i].max==0)
            break;
        // Bounding box stuff
        BBXLR.setY(i);
        if(d_maskHashVector[i].min<BBXUL.getX())
            BBXUL.setX(d_maskHashVector[i].min);
        if(d_maskHashVector[i].max>BBXLR.getX())
            BBXLR.setX(d_maskHashVector[i].max);

        // Shape Mask
        (d_detectedShapesMask->getRow(i))[d_maskHashVector[i].min]=d_maskHashVector[i].max;
    }

    d_currentShape->setBBX(CRect(BBXUL,BBXLR));
    return d_maskHashVector[startline].max-1;
}

/**********************************************************/
/*!
 *   Updates the incidence image when having finished (inside version)
 *   the extraction of one object. This way it can be determined
 *   if new candidates in the image belong to objects which have already
 *   been found
 *
 * \param startline internal parameter which describes the row object extraction has begun
 *
 */
/**********************************************************/
int CExtractorShape::updateShapesMaskInside(int startline)
{
    int i;
    CPoint BBXUL(d_maskHashVector[startline].min,startline);
    CPoint BBXLR(d_maskHashVector[startline].min,startline);

    for(i=startline;i<p_image->getH();i++){
        if(d_maskHashVector[i].min==p_image->getW() && d_maskHashVector[i].max==0)
            break;
        // Bounding box stuff
        BBXLR.setY(i);
        if(d_maskHashVector[i].min<BBXUL.getX())
            BBXUL.setX(d_maskHashVector[i].min);
        if(d_maskHashVector[i].max>BBXLR.getX())
            BBXLR.setX(d_maskHashVector[i].max);

        // Shape Mask
        (d_detectedShapesMask->getRow(i))[d_maskHashVector[i].min]=d_maskHashVector[i].max;
    }

    d_currentShape->setBBX(CRect(BBXUL,BBXLR));
    return d_maskHashVector[startline].max;
}

/**********************************************************/
/*!
 *   Starts the extraction of a shape (outside version).
 *   It starts the automaton, updates the incidence image and
 *   returns the next x-position which the global linear search
 *   is supposed to look at
 *
 * \param startPoint first pixel of the shape which has been found by the global search
 *
 */
/**********************************************************/
int CExtractorShape::extractNextShapeOutside(CPoint startPoint)
{
    int i;

    for(i=0;i<p_image->getH();i++){
        d_maskHashVector[i].min=p_image->getW();
        d_maskHashVector[i].max=0;
    }

    d_currentShape = new CPointShape();
    d_currentShape->addPoint(startPoint);
    updateHashVectorOutside(startPoint);
    d_startPoint=startPoint;
    runAroundOutside(0,startPoint,true);
    d_shapes->push_back(d_currentShape);
    return updateShapesMaskOutside(startPoint.getY());
}

/**********************************************************/
/*!
 *   Starts the extraction of a shape (inside version).
 *   It starts the automaton, updates the incidence image and
 *   returns the next x-position which the global linear search
 *   is supposed to look at
 *
 * \param startPoint first pixel of the shape which has been found by the global search
 *
 */
/**********************************************************/
int CExtractorShape::extractNextShapeInside(CPoint startPoint)
{
    int i;

    for(i=0;i<p_image->getH();i++){
        d_maskHashVector[i].min=p_image->getW();
        d_maskHashVector[i].max=0;
    }

    d_currentShape = new CPointShape();
    d_currentShape->addPoint(startPoint);
    updateHashVectorInside(startPoint);
    d_startPoint=startPoint;
    runAroundInside(0,startPoint,true);
    d_shapes->push_back(d_currentShape);
    return updateShapesMaskInside(startPoint.getY());
}

/**********************************************************/
/*!
 *   Runs around an object's shape inside of it
 *   finding contour points.
 *
 * \param firstrun internal flag used to determine if we have started running or not
 * \param firstdir first direction the automaton should look at
 * \param startpoint the point in the image just outside of an object
 *
 */
/**********************************************************/
void CExtractorShape::runAroundInside(int firstdir, CPoint startpoint, bool firstrun)
{
    int testx,testy,newtestx,newtesty;
    int xoffsets[4]={0,-1,0,1};
    int yoffsets[4]={-1,0,1,0};
    int xoffsetsdiag[4]={-1,-1,1,1};
    int yoffsetsdiag[4]={-1,1,1,-1};

    if(startpoint==d_startPoint && !firstdir && !firstrun) // Halt criterion: If we are in the initial state (and not starting): return
        return;

    //UP-LEFT (DIAGONAL)
    newtestx=startpoint.getX()+xoffsetsdiag[(0+firstdir)%4];
    newtesty=startpoint.getY()+yoffsetsdiag[(0+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        newtestx=startpoint.getX()+xoffsets[(0+firstdir)%4];
        newtesty=startpoint.getY()+yoffsets[(0+firstdir)%4];
        if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){ // UP
            runAroundInside(((firstdir+0)+3)%4,startpoint,false);
            return;
        } else {
            testx=newtestx;
            testy=newtesty;
            CPoint newPoint(testx,testy);
            d_currentShape->addPoint(newPoint);
            if( newPoint==d_startPoint && !((firstdir+0)%4) )
                return;
            updateHashVectorInside(newPoint);
            runAroundInside(((firstdir+0)%4),newPoint,false);
            return;
        }
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+1)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);
    }

    //LEFT
    newtestx=testx+xoffsets[(1+firstdir)%4];
    newtesty=testy+yoffsets[(1+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+1))%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+2)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);  
    }
    //DOWN
    newtestx=testx+xoffsets[(2+firstdir)%4];
    newtesty=testy+yoffsets[(2+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+2))%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+2)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);
    }
    //DOWN
    newtestx=testx+xoffsets[(2+firstdir)%4];
    newtesty=testy+yoffsets[(2+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+2))%4,CPoint(testx,testy),false);
        return;
    } else {
        if( CPoint(testx,testy)==d_startPoint && !((firstdir+2)%4) )
            return;
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+3)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);
    }
    //RIGHT
    newtestx=testx+xoffsets[(3+firstdir)%4];
    newtesty=testy+yoffsets[(3+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+3))%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+3)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);
    }
    //RIGHT
    newtestx=testx+xoffsets[(3+firstdir)%4];
    newtesty=testy+yoffsets[(3+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+3))%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        if( newPoint==d_startPoint && !((firstdir+0)%4) )
            return;
        d_currentShape->addPoint(newPoint);
        updateHashVectorInside(newPoint);
    }
    //UP
    newtestx=testx+xoffsets[(0+firstdir)%4];
    newtesty=testy+yoffsets[(0+firstdir)%4];
    if(newtestx <0 || newtestx >= p_image->getW() || newtesty < 0 || newtesty >= p_image->getH() || d_row[newtesty][newtestx]==(unsigned char) d_objectContentIsZero){
        runAroundInside(((firstdir+0))%4,CPoint(testx,testy),false);
        return;
    } else {
        testx=newtestx;
        testy=newtesty;
        CPoint newPoint(testx,testy);
        updateHashVectorInside(newPoint);
        d_currentShape->addPoint(newPoint);
    }
}

void CExtractorShape::updateHashVectorInside(CPoint &newPoint)
{

    if(d_maskHashVector[newPoint.getY()].min>newPoint.getX()){
        d_maskHashVector[newPoint.getY()].min=newPoint.getX();
    }

    if(d_maskHashVector[newPoint.getY()].max<newPoint.getX()){
        d_maskHashVector[newPoint.getY()].max=newPoint.getX();
    }

}

void CExtractorShape::updateHashVectorOutside(CPoint &newPoint)
{

    if(newPoint.getY()<0 || newPoint.getY()>=p_image->getH())
        return;

    if(d_maskHashVector[newPoint.getY()].min>newPoint.getX()){
        if(newPoint.getX()<0)
            d_maskHashVector[newPoint.getY()].min=0;
        else
            d_maskHashVector[newPoint.getY()].min=newPoint.getX();
    }

    if(d_maskHashVector[newPoint.getY()].max<newPoint.getX()){
        d_maskHashVector[newPoint.getY()].max=newPoint.getX();
    }

}
