/***************************************************************************
                           CHoughLineSearch.cpp  -  description
                              -------------------
     begin                : 30.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Diploma thesis Marc Hensel
 
***************************************************************************/


#include <tuv2d/CHoughLineSearch.h>

#include <cmath>

#if !defined(M_PI)
#define		M_PI	3.14159265358979323846
#endif

#if !defined(NULL)
#define		NULL	0
#endif


//using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

CHoughLineSearch::CHoughLineSearch()
{
    hough_space		= NULL;
    hwidth			= 0;
    hheight			= 0;
    s_max			= 0;
    s				= NULL;
    theta			= NULL;
    valid_results	= 0;
}


CHoughLineSearch::~CHoughLineSearch()
{
    if ( hough_space != NULL)		delete[] ( hough_space );
    if ( valid_results != 0)
    {
        delete[] ( s ) ;
        delete[] ( theta ) ;
    }
}


/**************************************************
 *
 * Return variables and results
 *
 **************************************************/

bool CHoughLineSearch::getLineDistance(unsigned char result_nr,double &distance)
{
    if (valid_results < result_nr)	return false;
    distance = s[result_nr - 1];
    return true;
}

bool CHoughLineSearch::getLineAngle(unsigned char result_nr,double &angle)
{
    if (valid_results < result_nr)	return false;
    angle = theta[result_nr - 1];
    return true;
}



/**********************************************************
 *
 * Create results (line parameters)
 *
 **********************************************************/

/*
 * Create array/matrix holding hough space data/image
 */

bool CHoughLineSearch::createHoughImage(CImageByteGray *image,unsigned char threshold)
{
    int		iwidth,iheight;			// Image dimensions
    int		x,y;					// Indices in image data matrix
    int		sc,tr;					// Indices for radius s (matrix columns) and angle theta (matrix rows)
    double	stepsize;				// Stepsize when running through theta (0 to pi)
    int		*hough_pnt;				// Address pointer in hough space matrix

    if (hough_space != NULL)	delete[](hough_space);

    /* Init image dimensions */
    iwidth	= image->getW();
    iheight	= image->getH();

    /* Translate origin to center of image matrix (reduces max. radius => reduces dimension of hough space) */
    center_x	= (int)(0.5 * iwidth);
    center_y	= (int)(0.5 * iheight);
    s_max		= (int)(0.5 * sqrt( (double)(iwidth * iwidth + iheight * iheight) ));

    /* Create hough space matrix */
    hwidth		= 180;
    hheight		= 2 * s_max + 1;
    hough_space	= new int[ hwidth * hheight ];
    if (hough_space == NULL)	return false;
    for (sc=0,hough_pnt=hough_space; sc<hheight; sc++)			// Clear matrix data
        for (tr=0; tr<hwidth; tr++,hough_pnt++)
            *hough_pnt = 0;

    /* Fill hough space image */
    stepsize = M_PI / 180.0;						// Note: hwidth = 180
    for (y=0; y<iheight; y++)
        for (x=0; x<iwidth; x++)
            if (image->getPixel(x,y,1) >= threshold)
            {
                for (tr=0; tr<180; tr+=1)			// Note: hwidth = 180
                {
                    // Swap x and y to compensate mirrowed coordinate system (euclidean: xy, image: x(-y))
                    sc = (int)( (double)(x - center_x) * cos((double)(stepsize * tr)) + (double)(y - center_y) * sin((double)(stepsize * tr)) );
                    if (sc>0)	sc++;
                    // Increase line count in hough space ("+rmax" to take s in [-rmax,rmax] into account)
                    hough_space[tr + (sc + s_max) * hwidth] += 1;
                }
            }
            return true;
}



/*
 * Find maximum value/line in hough space data
 */

bool CHoughLineSearch::findMaxLine()
{
    int		sc,tr;							// Indices for radius s (matrix columns) and angle theta (matrix rows)
    int		max_entries	= 0;				// Maximum entires (counts) in hough space element
    int		radius		= 0;				// Distance value s belonging to maximum (matrix index)
    int		angle		= 0;				// Angle theta belonging to maximum (matrix index)
    int		*hough_pnt;						// Address pointer in hough space matrix

    /* No valid hough space array in memory */
    if ( hough_space == 0 )	return false;

    /* Find maximum value */
    for (sc=0,hough_pnt=hough_space; sc<hheight; sc++)
    {
        for (tr=0; tr<hwidth; tr++,hough_pnt++)
        {
            if (*hough_pnt > max_entries)
            {
                radius		= sc;
                angle		= tr;
                max_entries	= *hough_pnt;
            }
        }
    }

    /* Set found line in global object data */
    if ( valid_results != 0)										// Delete old results
    {
        delete[] ( s ) ;
        delete[] ( theta ) ;
    }
    s				= new double(radius - s_max);					// Store new result
    theta			= new double(M_PI * (double)(angle) / 180.0 );	// Note: hwidth = 180
    valid_results	= 1;

    return true;
}


/*
 * Improve line parameters by subpixel maximum location in hough space.
 */

bool CHoughLineSearch::improveLineSubpixel	(unsigned char result_nr)
{
    if ( valid_results < result_nr)	return false;

    long int		s0,t0;			// Center position of window
    long int		sc,tr;			// Indices for radius s (matrix columns) and angle theta (matrix rows)
    long int		value;			// Value of entries in hough space matrix(tr,sc)
    long int		sum;			// Sum of pixel values in window
    long int		gs,gt;			// Gravity values for radius s (matrix columns) and angle theta (matrix rows)
    unsigned char	window;			// Half window edge length of pixels used to calculate center of gravity (1 = 3x3 window, 2 = 5x5 window, ...)

    /* Set center of window */
    t0	= (long int)( ((double)(hwidth)) * theta[result_nr - 1] / M_PI );
    s0	= (long int)( s[result_nr - 1] + s_max );

    /* Set window size */
    if		( (t0>=2) && (t0<=hwidth-3) && (s0>=2) && (s0<=hheight-3) )	window = 2;		// Is 5x5 window in matrix bounds?
    else if	( (t0>=1) && (t0<=hwidth-2) && (s0>=1) && (s0<=hheight-2) )	window = 1;		// Is 3x3 window in matrix bounds?
    else	return false;																// Center directly at matrix boarder

    /* Calculate center of gravity */
    sum = gs = gt = 0;
    for (sc=s0-window; sc<=s0+window; sc++)
        for (tr=t0-window; tr<=t0+window; tr++)
        {
            value	 = hough_space[tr + sc * hwidth];
            sum		+= value;
            gt		+= value * tr;
            gs		+= value * sc;
        }
        theta[result_nr - 1]	= M_PI * ((double)(gt)) / ((double)(sum * 180));			// Note: hwidth = 180
        s[result_nr - 1]		= (((double)(gs)) / (double)(sum)) - s_max;

        return true;
}



/**********************************************************
 *
 * Evaluate results
 *
 **********************************************************/

/*
 * Evaluate line function
 */

bool CHoughLineSearch::getFunctionOfX(unsigned char result_nr,double x0,double &y0)
{
    if (valid_results < result_nr)	return false;

    double	distance	= s[result_nr - 1];
    double	angle		= theta[result_nr - 1];

    if (sin(angle) != 0)
        y0	= ( (distance - (x0 - (double)(center_x)) * cos(angle)) / sin(angle) ) + (double)(center_y);
    else
        y0	= 0.0;
    return true;
}


bool CHoughLineSearch::getFunctionOfY(unsigned char result_nr,double y0,double &x0)
{
    if (valid_results < result_nr)	return false;

    double	distance	= s[result_nr - 1];
    double	angle		= theta[result_nr - 1];

    if (cos(angle) != 0)
        x0	= ( (distance - (y0 - (double)(center_y)) * sin(angle)) / cos(angle) ) + (double)(center_x);
    else
        x0	= 0.0;
    return true;
}



/*
 * Draw line in image
 */

bool CHoughLineSearch::drawLine(CImageByteGray *image,unsigned char result_nr,unsigned char gray_intensity)
{
    if ( valid_results < result_nr)	return false;

    double	distance,angle;				// Parameters (disrance to origin, angle of orthonarmal vector) of line to be drawn
    int		x,y;						// Run through image
    int		coord;						// Function value of line (coord=f(x) or coord=f(y) depending on theta)
    int		width,height;				// Image dimensions

    /* Init parameters */
    distance	= s[result_nr - 1];
    angle		= theta[result_nr - 1];
    width		= image->getW();
    height		= image->getH();

    /* Draw line */
    if ((angle >= M_PI/4.0) && (angle < M_PI*3.0/4.0))	// Avoid division by zero ( sin(theta), cos(theta) )
    {
        for (y=0; y<height; y++)
            for (x=0; x<width; x++)
            {
                coord = (int)( (distance - (double)(x - center_x) * cos(angle)) / sin(angle) ) + center_y;
                if ( (coord >= 0) && (coord < height) )
                    image->setPixel(x,coord,gray_intensity);
            }
    }
    else
    {
        for (x=0; x<width; x++)
            for (y=0; y<height; y++)
            {
                coord = (int)( (distance - (double)(y - center_y) * sin(angle)) / cos(angle) ) + center_x;
                if ( (coord >= 0) && (coord < width) )
                    image->setPixel(coord,y,gray_intensity);
            }
    }
    return true;
}
