/***************************************************************************
                          CFilterGauss.c  -  description
                             -------------------
    begin                : 19-4-2001
    copyright            : (C) 2001 by Islam Shdaifat
    email                : shdaifat@tuhh.de
    origin               : PhD Thesis Islam Shdaifat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include <cmath>
#include "tuv2d/CFilterGauss.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"

using namespace TUVision;

// Constructor
CFilterGauss::CFilterGauss(double sigma, int w)
{
    d_GaussArray = 0;
    d_intGaussArray = 0;
    if(w ==  -1) // if no value given fron user
    {
        w = 2*getGaussWidth(sigma)+1; // find width automatically
    }
    createGaussArray( sigma, w );
    d_MaskWidth = 2*w+1;
}

// Distructor
CFilterGauss::~CFilterGauss()
{
    delete [] d_GaussArray;
    delete [] d_intGaussArray;
}


double
CFilterGauss::getGauss(double x, double sigma)
{
    double xx;

    if (sigma == 0) return 0.0;
    xx = (double)exp((double) ((-x*x)/(2*sigma*sigma)));
    return xx;
}

int
CFilterGauss::getGaussWidth(double sigma)
{
    double GaussValue = 1.0;
    int x = 0;
    while(GaussValue > 0.1)
    {
        GaussValue = getGauss((double)x,sigma);
        x++;
    }
    return x;
}

void
CFilterGauss::createGaussArray(double sigma, int w )
{
    if( d_GaussArray != 0)
        delete d_GaussArray; // delete old mask

    if( d_intGaussArray != 0)
        delete d_intGaussArray; // delete old mask

    d_SumOfMask = 0; // reset the sum
    d_GaussArray = new double[2*w+1];
    d_intGaussArray = new int[2*w+1];
    for(int x = 0; x< 2*w+1; x++)
    {
        d_GaussArray[x] = getGauss((double) (x - w), sigma);
        d_intGaussArray[x] =(int) (10.0*d_GaussArray[x]);
        d_SumOfMask += d_intGaussArray[x];
    }

    //	d_intGaussArray+=w; // go to the center of the mask

}

void
CFilterGauss::filter(CImageByteGray &image)
{
    int x, y, i, sum;
    int w = image.getW();
    int h = image.getH();

    unsigned char *dst, *original_location;
    int *mask;
    int wm = (d_MaskWidth + 1)/2 - 1; // half of width of the mask 

    // filter in the x direction
    for(y = 0 ; y < h ; y++)
    {
        original_location = image.getRow(y) + wm;
        for(x = wm; x < w - wm; x++) 
        {
            sum = 0;
            dst = original_location - wm;
            mask = d_intGaussArray;
            for(i = -wm; i < wm; i++)
            {
                sum += (*dst++)*(*mask++); // convolution im(x-i)mask(i)
            }
            *original_location++ = sum/ d_SumOfMask; // ++ is the following pixel
        }
    }
    // filter in the y direction
    int rowdiff = image.getRow(1) - image.getRow(0);
    if(rowdiff < 0)
        rowdiff = -rowdiff ;

    for(x = 0 ; x < w ; x++)
    {
        original_location = image.getRow(0) + wm*rowdiff + x;
        for(y = wm; y < h - wm; y++) 
        {
            sum = 0;
            dst = original_location - wm*rowdiff;
            mask = d_intGaussArray;// + wm;
            for(i = -wm; i < wm; i++)
            {
                sum += (*dst)*(*mask++);
                dst += rowdiff;
            }
            *original_location = sum/ d_SumOfMask;
            original_location += rowdiff; // the following pixel
        }
    }
}

void
CFilterGauss::filter(CImageByteRGB &image)
{
    int x, y, i, sumr,sumg,sumb;
    int w = image.getW();
    int h = image.getH();

    unsigned char *dst, *original_location;
    int *mask;
    int wm = (d_MaskWidth + 1)/2 - 1; // half of width of the mask 

    // filter in the x direction
    for(y = 0 ; y < h ; y++)
    {
        original_location = image.getRow(y) + wm*3;
        for(x = wm; x < w - wm; x++) 
        {
            sumr = 0; sumg = 0; sumb = 0;
            dst = original_location - wm*3;
            mask = d_intGaussArray + wm;
            for(i = -wm; i < wm; i++)
            {
                sumr += (*dst++)*(* (mask + i)); // convolution im(x-i)mask(i)
                sumg += (*dst++)*(* (mask + i)); // convolution im(x-i)mask(i)
                sumb += (*dst++)*(* (mask + i)); // convolution im(x-i)mask(i)
            }
            *original_location++ = sumr/ d_SumOfMask; // ++ is the following pixel
            *original_location++ = sumg/ d_SumOfMask; // ++ is the following pixel
            *original_location++ = sumb/ d_SumOfMask; // ++ is the following pixel
        }
    }
    // filter in the y direction
    int rowdiff = (image.getRow(1) - image.getRow(0));
    if(rowdiff < 0)
        rowdiff = -rowdiff ;

    for(x = 0 ; x < w ; x++)
    {
        original_location = image.getRow(0) + wm*rowdiff + x*3;
        for(y = wm; y < h - wm - 1; y++) 
        {
            sumr = 0; sumg = 0; sumb = 0;
            dst = original_location - wm*rowdiff;
            mask = d_intGaussArray + wm;
            for(i = -wm; i < wm; i++)
            {
                sumr += (*dst++)*(* (mask + i));
                //dst  += rowdiff;
                sumg += (*dst++)*(* (mask + i));
                //dst  += rowdiff;
                sumb += (*dst++)*(* (mask + i));
                dst  += rowdiff-3;
            }
            *original_location++ = sumr/ d_SumOfMask;
            //original_location += rowdiff; // the following pixel
            *original_location++ = sumg/ d_SumOfMask;
            //original_location += rowdiff; // the following pixel
            *original_location++ = sumb/ d_SumOfMask;
            original_location += rowdiff-3; // the following pixel
        }
    }

    //image.clearImage();
}
