/***************************************************************************
                          CFilterSymmetricCovariance.cpp  -  description
                             -------------------
    begin                : 06.04.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv2d/CFilterSymmetricCovariance.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageIntGray.h"
#include "tuv/CImageDoubleGray.h"
#include <iostream>
#include <cmath>
#include <cstdlib>

using namespace TUVision;
using namespace std;

/**********************************************************/
/*!
 *  The default constructor initializes the filter window to 13x13 pixels
 *
 */
/**********************************************************/
CFilterSymmetricCovariance::CFilterSymmetricCovariance() : CFilterByteGray(), d_filterwidth(13){}

/**********************************************************/
/*!
 *  Sets the filter window to a new width
 *
 * \param filterwidth the new filterwidth which is used for the window
 *
 * \warning The filterwidth has to be odd
 */
/**********************************************************/
void CFilterSymmetricCovariance::init(int filterwidth){
    d_filterwidth=filterwidth;
}


/**********************************************************/
/*!
 *  Filters a gray image with the symmetric covariance filter
 *  It uses the filterwidth which has been set before and updates
 *  the current image (it does not return a new one).
 *
 * \param img the image to be filtered
 */
/**********************************************************/
void CFilterSymmetricCovariance::filter (CImageByteGray &img)
{
    if(!(d_filterwidth&1)) // filter length has to be odd
        return;

    CImageByteGray  *orig_img;
    orig_img = new CImageByteGray(img);

    img.clearImage();
    int x,y,i,j;
    double covariance;
    typedef unsigned char * charptr ;
    charptr *row = new charptr[d_filterwidth];
    unsigned char *row_target;
    int meanrow = d_filterwidth/2;

    // Loop through all rows
    for(y=d_filterwidth/2;y<orig_img->getH()-d_filterwidth/2;y++){

        // For speed reasons, save pointers to rows
        for(i=-d_filterwidth/2;i<=d_filterwidth/2;i++)
            row[meanrow+i]=orig_img->getRow(y+i);
        row_target=img.getRow(y);

        // Loop through columns
        for(x=d_filterwidth/2;x<orig_img->getW()-d_filterwidth/2;x++){

            // Two checks whether there is an odd symmetry: First one: Corner Points
            //      if(abs(row[0]              [x-d_filterwidth/2]-row[d_filterwidth-1][x+d_filterwidth/2])+
            //      	 abs(row[d_filterwidth-1][x-d_filterwidth/2]-row[0]              [x+d_filterwidth/2])<=50){


            // Second check: Calculate difference of the four quarter areas

            unsigned int sum_ulquarter=0, sum_urquarter=0, sum_llquarter=0, sum_lrquarter=0, sum_centerrows=0;
            double meanul=0, meanur=0, meanll=0, meanlr=0;
            double average  = 0;
            double variance = 0;


            for(j=d_filterwidth/2;j>0;j--)
                for(i=d_filterwidth/2;i>0;i--){
                    sum_ulquarter+=row[meanrow-j][x-i];
                    sum_lrquarter+=row[meanrow+j][x+i];
                    sum_urquarter+=row[meanrow-j][x+i];
                    sum_llquarter+=row[meanrow+j][x-i];
                }


                int mean_quarter_normalizefactor = (d_filterwidth/2) * (d_filterwidth/2);

                meanul = sum_ulquarter / (double)mean_quarter_normalizefactor;
                meanlr = sum_lrquarter / (double)mean_quarter_normalizefactor;
                meanur = sum_urquarter / (double)mean_quarter_normalizefactor;
                meanll = sum_llquarter / (double)mean_quarter_normalizefactor;


                for(j=d_filterwidth/2;j>0;j--)
                    sum_centerrows+=row[meanrow+j][x]+row[meanrow-j][x]+row[meanrow][x-j]+row[meanrow][x+j];
                sum_centerrows+=row[meanrow][x];

                average = (double) (sum_ulquarter+sum_urquarter+sum_lrquarter+sum_llquarter+sum_centerrows)/(d_filterwidth*d_filterwidth);

                for(j=-d_filterwidth/2;j<=d_filterwidth/2;j++)
                    for(i=-d_filterwidth/2;i<=d_filterwidth/2;i++)
                        variance+=(row[meanrow+j][x+i]-average)*(row[meanrow+j][x+i]-average);

                variance/=(255*d_filterwidth*d_filterwidth);


                //	if(!(fabs(meanul-meanlr)>3*variance || fabs(meanur-meanll)>3*variance)){

                // Calculate average as it has to be subtracted for covariance calculations

                covariance=0.0;

                // Apply filterwidth filter where x is the current center point
                // As the filter is symmetric it is sufficient to only process one half of the filter window
                // (j=-d_filterwidth/2 ... 0), double the value and add the middle line
                for(j=-d_filterwidth/2;j<0;j++)
                    for(i=-d_filterwidth/2;i<=d_filterwidth/2;i++)
                        covariance+=(row[meanrow+j][x+i]-average)*(row[meanrow-j][x-i]-average);

                covariance*=2;

                for(i=-d_filterwidth/2;i<=d_filterwidth/2;i++)
                    covariance+=(row[meanrow][x+i]-average)*(row[meanrow][x-i]-average);

                covariance = (covariance<0)?0:covariance; // Throw away negative parts
                covariance = covariance*4/(255*d_filterwidth*d_filterwidth);
                row_target[x]=(unsigned char)(covariance+0.5); // 0.5 as rounding to nearest = floor(value+0.5)
                //      }
        }
    }

    delete orig_img;
    delete [] row;
}
