/***************************************************************************
                           COperatorContrast.cpp  -  description
                              -------------------
     begin                : 30.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Diploma thesis Marc Hensel
 
***************************************************************************/


/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include <tuv2d/COperatorContrast.h>

#include <cmath>

// using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

//
//! Constructor.
//

COperatorContrast::COperatorContrast()
{
}


//
//! Destructor.
//

COperatorContrast::~COperatorContrast()
{
}


/**************************************************
 *
 * Public methods
 *
 **************************************************/

//
//! Linear transform so that the image intensities are stretched to the full dynamic range [0,255].
/*!
 *  The min. and max. intensities i_min and i_max in the image are determined
 *  and mapped linear i_min -> 0 and i_max -> 255 so that the full dynamic range
 *  is used.
 */

void COperatorContrast::maxContrast (CImageByteGray *img)
{
    int				N;					//!< Number of pixel in image
    unsigned char	value,max,min;		//!< Current, maximum and minimum value in image
    unsigned char	*row;				//!< Pointer: 1st pixel in row (image array)
    int				i;					//!< Current position in image array
    double			scale;				//!< Factor transforming (max-min) to 255

    row	= img->getRow(0);
    N	= img->getW() * img->getH();

    // Calculate dynamic range
    max = 0;
    min = 255;
    for (i=0; i<N; i++)
    {
        value = row[i];
        if (value<min) min = value;
        if (value>max) max = value;
    }

    // Transform dynamic range [min,max] -> [0,255]
    scale = 255.0 / ((double)max-(double)min);
    for (i=0; i<N; i++)
        row[i] = (unsigned char)((row[i] - min) * scale);
}


//
//! Logarithmic mapping of the dynamic range to itself [0,255]->[0,255].
/*!
 *  At first the image intensities are stretched linear to the full
 *  dynamic range according of the method description for maxContrst().
 *
 *  Afterwards for each pixel intensity i the function \f$ f(i) = 46 \cdot ln[i + 1] \f$
 *  is applied resulting in an logarithmic amplification towards higher intensities.
 *
 *  The parameters of f:
 *
 *  - Source data is of range [0,255], where log(0) is undefined. Therefore "1" is 
 *    added to each pixel intensity before applying the logarithm.
 *  - The factor 46 and taking the integer value leads to the desired
 *    dynamic range of [0,255] as f(0)=0 and (int)(f(255))=255.
 */

void COperatorContrast::maxLogContrast (CImageByteGray *img)
{
    int				N;					//!< Number of pixel in image
    unsigned char	value,max,min;		//!< Current, maximum and minimum value in image
    unsigned char	*row;				//!< Pointer: 1st pixel in row (image array)
    int				i;					//!< Current position in image array
    double			scale;				//!< Factor transforming (max-min) to 255

    row	= img->getRow(0);
    N	= img->getW() * img->getH();

    // Calculate dynamic range
    max = 0;
    min = 255;
    for (i=0; i<N; i++)
    {
        value	= (unsigned char)(46*log((double) row[i]+1));
        row[i]	= value;
        if (value<min) min = value;
        if (value>max) max = value;
    }

    // Transform dynamic range [min,max] -> [0,255]
    scale = 255.0 / ((double)max-(double)min);
    for (i=0; i<N; i++)
        row[i] = (unsigned char)((row[i] - min) * scale);
}

