/***************************************************************************
                          COperatorFFT.cpp  -  description
                             -------------------
    begin                : 27.11.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de
    origin               : Bildverarbeitung 1, WS 2001/2002

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv2d/COperatorFFT.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageDoubleGray.h"
#include "tuv/CImageComplexGray.h"
#include <complex>
#include <cmath>

using namespace std;
using namespace TUVision;

COperatorFFT::COperatorFFT()
{
}

COperatorFFT::~COperatorFFT()
{
}


/**********************************************************/
/*!  If there is just one "1" - bit in value, value
 *   is a power of 2
 *
 *   \param value The value which is to be checked
 *
 *   \return true if value is a power of 2, false else
 *
 **********************************************************/
inline bool COperatorFFT::is_ld(int value)
{
    int result=0;
    for(unsigned int i=0;i<sizeof(value)*8;i++)
        result+=((value>>i)&1);
    return (result==1);
}

/**********************************************************/
/*!  Gives back the index of the highest "1" - bit
 *
 *   \param value The value of which ld has to be computed
 *
 *   \return floor(ld(value))
 *
 **********************************************************/
inline int COperatorFFT::ld_floor(int value)
{
    int result;
    for(unsigned int i=0;i<sizeof(value)*8;i++)
        if(value & (1<<i))
            result=i;
    return result;
}

/**********************************************************/
/*!  Gives back the index of the highest "1" - bit plus
 *   one if value is not a power of 2
 *
 *   \param value The value of which ld_ceil has to be computed
 *
 *   \return ceil(ld(value))
 *
 **********************************************************/
inline int COperatorFFT::ld_ceil(int value)
{
    int result=ld_floor(value);
    return is_ld(value)?result:result+1;
}


/**********************************************************/
/*!  Butterfly structure can be found in many books on
 *   signal processing. See there for details. This operation
 *   is done "in place", i.e. the old values will be overwritten
 *   with the new values
 *
 *   \param firstwing The first input wing of the butterfly
 *   \param secondwing The second input wing of the butterfly
 *   \param twiddle The used complex coefficient
 *
 **********************************************************/
inline void
COperatorFFT::butterfly_in_place(complex<double> *firstwing,
                                 complex<double> *secondwing, const complex<double> &twiddle)
{
    complex<double> tmp=(*secondwing)*twiddle;
    *secondwing=*firstwing-tmp;
    *firstwing+=tmp;
}


/**********************************************************/
/*!  For each element in the bitreversal - table (0 ... width)
 *   the bitreversed index is computed. This is done by
 *   successively shifting the input index to the right,
 *   picking the LSB and ORing it to the bitreversed index.
 *   After each step, the bitreversed index gets shifted to
 *   the left.
 *
 *   \param width The width of the bitreversal table
 *   \param bitreversal_table The table which will be written to
 *
 **********************************************************/
void COperatorFFT::init_bitreversal(int width,
                                    vector<int> &bitreversal_table)
{
    int level=ld_floor(width);
    bitreversal_table.reserve(width);
    for(int i=0;i<width;i++){
        unsigned int bitreversed=0;
        unsigned int original=i;
        for(int j=0;j<level-1;j++){
            bitreversed|=(original&1);
            bitreversed<<=1;
            original>>=1;
        }
        bitreversed|=(original&1);
        bitreversal_table[i]=bitreversed;
    }
}


/**********************************************************/
/*!  The twiddle table gets initialized with the complex
 *   coefficients. The twiddles are the "W" - factors found
 *   in many books on digital signal processing. Only half
 *   of the factors for a N-vector actually have to be computed
 *   as the second half is the conjugated value of the first one.
 *   This is already included in the FFT algorithm.
 *
 *   \param width The width of the input vector
 *   \param twiddle_table The table which will be written to
 *
 **********************************************************/
void COperatorFFT::init_twiddletable(int width,
                                     vector< complex<double> > &twiddle_table)
{
    twiddle_table.reserve(width/2);
    for(int k=0;k<width/2;k++){
        twiddle_table[k]=polar(1.0,-2*M_PI*k/width);
    }
}

/**********************************************************/
/*!  If the dimensions of the input image are not even
 *   shifting can not take place. Otherwise, the image is
 *   divided into four quadrants which will be exchanged
 *   diagonally. The input image will be replaced by the
 *   result. This function resembles the fftshift of Matlab.
 *
 *   \param image The input image which will be replaced
 *   \return true if the operation has been done successfully
 *
 **********************************************************/
bool COperatorFFT::fftshift(CImageComplexGray &image)
{

    if( (image.getW() & 1) || (image.getH() & 1)){
        return false;
    }


    int blockwidth =image.getW()/2;
    int blockheight=image.getH()/2;

    CImageComplexGray tmpimage(image);

    tmpimage.setWindow(0,0,blockwidth,blockheight);
    image.insertSubImage(tmpimage,blockwidth,blockheight);
    tmpimage.resetWindow();

    tmpimage.setWindow(blockwidth,0,blockwidth,blockheight);
    image.insertSubImage(tmpimage,0,blockheight);
    tmpimage.resetWindow();

    tmpimage.setWindow(0,blockheight,blockwidth,blockheight);
    image.insertSubImage(tmpimage,blockwidth,0);
    tmpimage.resetWindow();

    tmpimage.setWindow(blockwidth,blockheight,blockwidth,blockheight);
    image.insertSubImage(tmpimage,0,0);
    tmpimage.resetWindow();

    return true;
}

/**********************************************************/
/*!  In place operations can be done in this class after
 *   the first step of the FFT (which cannot be done in place,
 *   as the input values have to be accessed bitreversed. This
 *   function completes a 1D - FFT of one complex input vector.
 *   The complex coefficients have to be initialized before
 *   using the correct width.
 *
 *   \param width The width of the current row which is processed
 *   \param row Pointer to the row
 *   \param twiddles The complex coefficients which are used.
 *
 **********************************************************/
inline void 
COperatorFFT::fft_in_place_core(int width,complex<double> *row,
                                vector< complex<double> > &twiddles)
{

    int iincrement=2;
    int coeffoffset=width/2;

    for(int level=1;level<ld_floor(width);level++){
        iincrement*=2;
        coeffoffset/=2;
        for(int i=0;i<width;i+=iincrement){ // basic butterfly loop
            complex<double> *butterflyfirst = row+i;
            complex<double> *butterflysecond = row+i+iincrement/2;
            complex<double> *currentcoefficient = &twiddles.front();
            for(int j=0;j<iincrement/2;j++){
                butterfly_in_place(butterflyfirst,
                    butterflysecond,*currentcoefficient);
                butterflyfirst++;
                butterflysecond++;
                currentcoefficient+=coeffoffset;
            }
        }
    }
}


/**********************************************************/
/*!  The main 1D - part of the FFT can be used by all
 *   fft - functions of this class. It takes a complex input
 *   image, computes a 1D FFT of each column(!) and stores the
 *   result rowwise(!) in a new image which will be returned.
 *   By calling this function twice, a 2D - FFT has been done.
 *
 *   \param width srcimage The input image
 *   \return A new 1D-FFT image with swapped dimensions
 *
 **********************************************************/
CImageComplexGray *COperatorFFT::fft_1d_main(CImageComplexGray &srcimage)
{

    // Initialize twiddle tables and bit reversal tables

    int width =srcimage.getW();
    int height=srcimage.getH();

    vector< complex<double> > twiddles;
    vector< int >             bitindex;

    init_twiddletable(height,twiddles);
    init_bitreversal (height,bitindex);

    CImageComplexGray *calcimage =  new CImageComplexGray(height,width);
    calcimage->clearImage();

    complex<double> *inputrow  = srcimage.getRow(0);
    complex<double> *outputrow = calcimage->getRow(0);

    // First pass: columwise 1D FFT of the input image

    for(int x=0;x<width;x++){

        // First "not-in-place" - stage
        {
            complex<double> *butterflyfirst = outputrow;
            complex<double> *butterflysecond = outputrow+1;
            for(int i=0;i<height;i+=2){ // basic butterfly loop
                complex<double> tmpinput = inputrow[width*bitindex[i]];
                complex<double> tmp      = inputrow[width*bitindex[i+1]];
                *butterflysecond=tmpinput-tmp;
                *butterflyfirst =tmpinput+tmp;
                butterflyfirst+=2;
                butterflysecond+=2;
            }
        }

        fft_in_place_core(height,outputrow, twiddles);

        outputrow+=height;
        inputrow++;
    }

    return calcimage;
}


/**********************************************************/
/*!  The IFFT can be done by calling fft_1d_main twice.
 *   The only specific thing is that real and imaginary part
 *   have to be swapped before and after the FFT operation,
 *   and normalization has to be done afterwards. This can be
 *   found in many books on digital signal processing.
 *   The output size of the image can be set by setting
 *   output_width and output_height to values other than 0
 *   (0 means: Use the appropriate dimensions corresponding to
 *   the input image). Output width and height in this case
 *   have to be power of 2 and must be greater than the size
 *   of the original image
 *
 *   \param spectrum The spectrum of which the IFFT has to be computed
 *   \param output_width The desired output width
 *   \param output_height The desired output height
 *   \return The complex result image. Usually only the real-part is needed later
 *
 **********************************************************/
CImageComplexGray COperatorFFT::ifft(CImageComplexGray &spectrum,
                                     int output_width,
                                     int output_height)
{

    int calculation_width= 1<<ld_ceil(spectrum.getW());
    int calculation_height=1<<ld_ceil(spectrum.getH());

    if(output_width>calculation_width && is_ld(output_width))
        calculation_width = output_width;

    if(output_height>calculation_height && is_ld(output_height))
        calculation_height = output_height;

    CImageComplexGray srcimage(calculation_width,calculation_height);
    srcimage.clearImage();

    {
        complex<double> *imagptr;
        complex<double> *dstptr;
        for(int y=0;y<spectrum.getH();y++){
            imagptr = spectrum.getRow(y);
            dstptr  = srcimage.getRow(y);
            for(int x=0;x<spectrum.getW(); x++){
                *dstptr = complex<double>(imagptr->imag(),imagptr->real());
                dstptr++,imagptr++;
            }
        }
    }

    CImageComplexGray *intermediate = fft_1d_main(srcimage);
    CImageComplexGray *almostresult = fft_1d_main(*intermediate);

    CImageComplexGray resultreal(calculation_width,calculation_height);
    resultreal.clearImage();
    {
        complex<double> *result = resultreal.getRow(0);
        complex<double> *srcptr = almostresult->getRow(0);
        for(int y=0;y<spectrum.getH();y++){
            for(int x=0;x<spectrum.getW(); x++){
                *result = complex<double> (srcptr->imag()/
                    (calculation_width*calculation_height),srcptr->real()/
                    (calculation_width*calculation_height));
                result++, srcptr++;
            }
        }
    }

    delete intermediate;
    delete almostresult;

    return CImageComplexGray(resultreal,true);
}

/**********************************************************/
/*!  This function basically sets up an image with "power-of-2"
 *   - dimensions which will be processed by fft_1d_main twice.
 *   The output size of the image can be set by setting
 *   output_width and output_height to values other than 0
 *   (0 means: Use the appropriate dimensions corresponding to
 *   the input image). Output width and height in this case
 *   have to be power of 2 and must be greater than the size
 *   of the original image
 *
 *   \param image The input image
 *   \param output_width The desired output width
 *   \param output_height The desired output height
 *   \return The (eventually resized) spectrum of the input image
 *
 **********************************************************/
CImageComplexGray COperatorFFT::fft(CImageComplexGray &image,
                                    int output_width,
                                    int output_height)
{

    int calculation_width= 1<<ld_ceil(image.getW());
    int calculation_height=1<<ld_ceil(image.getH());

    if(output_width>calculation_width && is_ld(output_width))
        calculation_width = output_width;

    if(output_height>calculation_height && is_ld(output_height))
        calculation_height = output_height;

    CImageComplexGray srcimage(calculation_width,calculation_height);
    srcimage.clearImage();

    {
        complex<double> *imagptr;
        complex<double> *dstptr;
        for(int y=0;y<image.getH();y++){
            imagptr = image.getRow(y);
            dstptr  = srcimage.getRow(y);
            for(int x=0;x<image.getW(); x++){
                *dstptr = *imagptr;
                dstptr++,imagptr++;
            }
        }
    }

    CImageComplexGray *intermediate = fft_1d_main(srcimage);
    CImageComplexGray *almostresult = fft_1d_main(*intermediate);

    CImageComplexGray resultreal(*almostresult,true);

    delete intermediate;
    delete almostresult;

    return CImageComplexGray(resultreal,true);
}

/**********************************************************/
/*!  This function is used by the two fft-versions taking
 *   a real image as input. The FFT is computed of the srcimage
 *   by taking into account that the complex input image has been
 *   setup by shrinking the original real image by a factor of 2.
 *   This means that after the first fft_1d_main() - step the output
 *   values have to preprocessed before continuing. Also, the
 *   mean frequency row of the result has to be calculated separately.
 *   Finally the result image can be combined of the calculated one
 *   and a conjugated mirrored version (as the input has been real
 *   in the first place).
 *
 *   \param srcimage The input image
 *   \return The complex spectrum of the input image
 *
 **********************************************************/
CImageComplexGray COperatorFFT::fft_main(CImageComplexGray &srcimage)
{
    // Now we have set up a complex image. We can call the fft_main
    // pass and get a column based dstimage

    CImageComplexGray *dstimage = fft_1d_main(srcimage);

    //----------------------------------------------------

    // Now we can set up the complex frequencyimage which will be
    // the target for the next calculations

    // This frequency image has twice the height of our calculation
    // height as we shrinked the original calculation height by
    // a factor of 2 as we processed a real input image

    int calculation_width = srcimage.getW();
    int calculation_height = srcimage.getH();

    int d_image_width  = calculation_width;
    int d_image_height = calculation_height*2;

    vector< complex<double> > twiddles;
    vector< int >             bitindex;

    init_twiddletable(calculation_width,twiddles);
    init_bitreversal (calculation_width,bitindex);


    CImageComplexGray d_complex_frequencyimage(d_image_width,d_image_height);

    // Second pass: rowwise 1D FFT of the dstimage

    complex<double> *horizontalpass_row = d_complex_frequencyimage.getRow(0);

    // First row has a special meaning as the complex coefficients of
    // the first pass have to be processed specifically
    // (see Kammeyer, Formel 7.1.17 and 7.1.18)

    {
        complex<double> *butterflyfirst   = horizontalpass_row;
        complex<double> *butterflysecond  = horizontalpass_row+1;
        for(int i=0;i<calculation_width;i+=2){
            complex<double> *scrambled = dstimage->getRow(bitindex[i]);
            complex<double> x_first = scrambled->real()+scrambled->imag();
            scrambled = dstimage->getRow(bitindex[i+1]);
            complex<double> x_second = scrambled->real()+scrambled->imag();
            *butterflysecond=x_first-x_second;
            *butterflyfirst=x_first+x_second;
            butterflyfirst+=2;
            butterflysecond+=2;
        }
    }

    fft_in_place_core(calculation_width, horizontalpass_row,twiddles);

    horizontalpass_row+=calculation_width;

    // now the remaining rows

    for(int y=1;y<calculation_height;y++){

        {
            complex<double> *butterflyfirst = horizontalpass_row;
            complex<double> *butterflysecond = horizontalpass_row+1;
            for(int i=0;i<calculation_width;i+=2){ // basic butterfly loop

                complex<double> x_n_first = *(dstimage->getRow(bitindex[i])+y);

                complex<double> x_N_n_first =
                    *(dstimage->getRow(bitindex[i])+calculation_height-y);

                complex<double> x_n_second =
                    *(dstimage->getRow(bitindex[i+1])+y);

                complex<double> x_N_n_second =
                    *(dstimage->getRow(bitindex[i+1])+calculation_height-y);

                complex<double> x_first = 0.5*(x_n_first+conj(x_N_n_first) +
                    polar(1.0,-M_PI*(0.5+(double)(y)/(double)calculation_height))
                    *(x_n_first-conj(x_N_n_first)));

                complex<double> tmp = 0.5*(x_n_second+conj(x_N_n_second) +
                    polar(1.0,-M_PI*(0.5+(double)(y)/(double)calculation_height))
                    *(x_n_second-conj(x_N_n_second)));
                *butterflysecond=x_first-tmp;
                *butterflyfirst=x_first+tmp;
                butterflyfirst+=2;
                butterflysecond+=2;
            }
        }

        fft_in_place_core(calculation_width,horizontalpass_row, twiddles);

        horizontalpass_row+=calculation_width;
    }

    // Up to now we can not fully complete the image as the
    // middle frequency is missing in the image

    complex<double> *verticalpass_row=dstimage->getRow(0);
    complex<double> *inputrow;
    // we use the dstimage for our calculations as
    // it is not needed anymore
    for(int m=0;m<srcimage.getW();m++){
        double myreal=0;
        inputrow=srcimage.getRow(0)+m;
        for(int y=0;y<srcimage.getH();y++){
            myreal+=inputrow->real()-inputrow->imag();
            inputrow+=srcimage.getW();
        }
        *verticalpass_row = complex<double>(myreal,0.0);
        verticalpass_row++;
    }

    verticalpass_row = dstimage->getRow(0);

    {
        complex<double> *butterflyfirst   = horizontalpass_row;
        complex<double> *butterflysecond  = horizontalpass_row+1;
        for(int i=0;i<calculation_width;i+=2){ // basic butterfly loop
            complex<double> x_first  = verticalpass_row[bitindex[i]];
            complex<double> x_second = verticalpass_row[bitindex[i+1]];
            *butterflysecond=x_first-x_second;
            *butterflyfirst =x_first+x_second;
            butterflyfirst +=2;
            butterflysecond+=2;
        }
    }

    fft_in_place_core(calculation_width,horizontalpass_row,
        twiddles);



    // Last step: Complete frequency image

    // First part of completion: Mirror part containing only positive
    // frequencies

    //  verticalpass_row = d_complex_frequencyimage+d_image_width*d_image_height-1;
    verticalpass_row =
        d_complex_frequencyimage.getRow(d_image_height-1)+d_image_width-1;

    horizontalpass_row=d_complex_frequencyimage.getRow(1)+1;
    int i;

    for(i=0; i<d_image_width*d_image_height/2; i++){
        *verticalpass_row=conj(*horizontalpass_row);
        horizontalpass_row++, verticalpass_row--;
    }

    // Second part of completion: Mirror part containing only
    // vertical frequencies
    //  verticalpass_row = d_complex_frequencyimage+d_image_width*(d_image_height-1);
    verticalpass_row = d_complex_frequencyimage.getRow(d_image_height-1);
    horizontalpass_row = d_complex_frequencyimage.getRow(1);
    for(i=0;i<d_image_width/2;i++){
        *verticalpass_row = conj(*horizontalpass_row);
        horizontalpass_row+=d_image_width, verticalpass_row-=d_image_width;
    }

    return CImageComplexGray(d_complex_frequencyimage,true);
}

/**********************************************************/
/*!  The input image (of type byte) is shrinked to a complex
 *   input image which will be processed by fft_main() later.
 *   The output size of the image can be set by setting
 *   output_width and output_height to values other than 0
 *   (0 means: Use the appropriate dimensions corresponding to
 *   the input image). Output width and height in this case
 *   have to be power of 2 and must be greater than the size
 *   of the original image
 *
 *   \param srcimage The real input image
 *   \param output_width The desired output width
 *   \param output_height The desired output height
 *   \return The complex spectrum of the input image
 *
 **********************************************************/
CImageComplexGray COperatorFFT::fft(CImageByteGray &image,
                                    int output_width,
                                    int output_height)
{

    // First step: Generate a radix-2 compatible complex image
    // which contains the original image in interleaved order

    int calculation_width =1<<ld_ceil(image.getW());
    int calculation_height=1<<(ld_ceil(image.getH())-1);

    if(output_width>calculation_width && is_ld(output_width))
        calculation_width = output_width;

    if(output_height>calculation_height*2 && is_ld(output_height))
        calculation_height = output_height/2;

    CImageComplexGray srcimage(calculation_width,calculation_height);
    srcimage.clearImage();

    {
        unsigned char *srcptr_real;
        unsigned char *srcptr_imag;
        complex<double> *dstptr;
        for(int y=0;y<image.getH()/2;y++){
            srcptr_real=image.getRow(2*y);
            srcptr_imag=image.getRow(2*y+1);
            dstptr=srcimage.getRow(y);
            for(int x=0;x<image.getW();x++){
                *dstptr=complex<double>((double) *srcptr_real/(double) 255 ,(double) *srcptr_imag/(double) 255);
                srcptr_real++, srcptr_imag++, dstptr++;
            }
        }

        // do we have an odd height? If so, take care of last row
        if(image.getH()&1){
            dstptr=srcimage.getRow(image.getH()/2);
            srcptr_real=image.getRow(image.getH()-1);
            for(int x=0;x<image.getW();x++){
                *dstptr=complex<double>((double) *srcptr_real/(double) 255,0.0);
                srcptr_real++, dstptr++;
            }
        }
    }

    return fft_main(srcimage);
}

/**********************************************************/
/*!  The input image (of type double) is shrinked to a complex
 *   input image which will be processed by fft_main() later.
 *   The output size of the image can be set by setting
 *   output_width and output_height to values other than 0
 *   (0 means: Use the appropriate dimensions corresponding to
 *   the input image). Output width and height in this case
 *   have to be power of 2 and must be greater than the size
 *   of the original image
 *
 *   \param srcimage The real input image
 *   \param output_width The desired output width
 *   \param output_height The desired output height
 *   \return The complex spectrum of the input image
 *
 **********************************************************/
CImageComplexGray COperatorFFT::fft(CImageDoubleGray &image,
                                    int output_width,
                                    int output_height)
{

    // First step: Generate a radix-2 compatible complex image
    // which contains the original image in interleaved order

    int calculation_width =1<<ld_ceil(image.getW());
    int calculation_height=1<<(ld_ceil(image.getH())-1);

    if(output_width>calculation_width && is_ld(output_width))
        calculation_width = output_width;

    if(output_height>calculation_height*2 && is_ld(output_height))
        calculation_height = output_height/2;

    CImageComplexGray srcimage(calculation_width,calculation_height);
    srcimage.clearImage();

    {
        double *srcptr_real;
        double *srcptr_imag;
        complex<double> *dstptr;
        for(int y=0;y<image.getH()/2;y++){
            srcptr_real=image.getRow(2*y);
            srcptr_imag=image.getRow(2*y+1);
            dstptr=srcimage.getRow(y);
            for(int x=0;x<image.getW();x++){
                *dstptr=complex<double>(*srcptr_real,*srcptr_imag);
                srcptr_real++, srcptr_imag++, dstptr++;
            }
        }

        // do we have an odd height? If so, take care of last row
        if(image.getH()&1){
            dstptr=srcimage.getRow(image.getH()/2);
            srcptr_real=image.getRow(image.getH()-1);
            for(int x=0;x<image.getW();x++){
                *dstptr=complex<double>(*srcptr_real,0.0);
                srcptr_real++, dstptr++;
            }
        }
    }

    return fft_main(srcimage);
}

