/***************************************************************************
                          COperatorMultiplex.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/COperatorMultiplex.h"
#include "tuv/CExceptionRange.h"
#include "tuv/CImage.h"
#include "tuv/CImageByte.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CImageByteYUV.h"
#include "tuv/CImageInt.h"
#include "tuv/CImageIntGray.h"
#include "tuv/CImageIntRGB.h"
#include "tuv/CImageIntYUV.h"
#include "tuv/CImageDouble.h"
#include "tuv/CImageDoubleGray.h"
#include "tuv/CImageDoubleRGB.h"
#include "tuv/CImageDoubleYUV.h"

#include <vector>


using namespace std;
using namespace TUVision;


////////////////////////////////////////////////////////////
// Generic getChannel()
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Extract one color channel from an image with more
 *   than one color channel (byte version).
 *
 *   \param src source image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 *   \return extracted color channel as image
 */
/**********************************************************/

CImageByteGray *COperatorMultiplex::getChannel 
(const CImageByte &src,
 int nchannels,
 int channel) const
{
    int                  x, y;
    const unsigned char  *row_s;
    unsigned char        *row_d;
    CImageByteGray       *dst;


    if (channel >= nchannels)
        throw CExceptionRange ("The requested channel of the image "
        "is not available.", channel, 0,
        nchannels);

    dst = new CImageByteGray (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst->getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x] = row_s[x*nchannels + channel];
    }

    return (dst);
}


/**********************************************************/
/*!
 *   Extract one color channel from an image with more
 *   than one color channel (integer version).
 *
 *   \param src source image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 *   \return extracted color channel as image
 */
/**********************************************************/

CImageIntGray *COperatorMultiplex::getChannel
(const CImageInt &src,
 int nchannels,
 int channel) const
{
    int                  x, y;
    const int            *row_s;
    int                  *row_d;
    CImageIntGray        *dst;


    if (channel >= nchannels)
        throw CExceptionRange ("The requested channel of the image "
        "is not available.", channel, 0,
        nchannels);

    dst = new CImageIntGray (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst->getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x] = row_s[x*nchannels + channel];
    }

    return (dst);
}


/**********************************************************/
/*!
 *   Extract one color channel from an image with more
 *   than one color channel (double version).
 *
 *   \param src source image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 *   \return extracted color channel as image
 */
/**********************************************************/

CImageDoubleGray *COperatorMultiplex::getChannel
(const CImageDouble &src,
 int nchannels,
 int channel) const
{
    int                  x, y;
    const double         *row_s;
    double               *row_d;
    CImageDoubleGray     *dst;


    if (channel >= nchannels)
        throw CExceptionRange ("The requested channel of the image "
        "is not available.", channel, 0,
        nchannels);

    dst = new CImageDoubleGray (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst->getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x] = row_s[x*nchannels + channel];
    }

    return (dst);
}


////////////////////////////////////////////////////////////
// Generic setChannel()
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Insert the source image in the destination image as
 *   one color channel (byte version).
 *
 *   \param src source image
 *   \param dst destination image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 */
/**********************************************************/

void COperatorMultiplex::setChannel (const CImageByteGray &src, 
                                     CImageByte &dst,
                                     int nchannels, int channel) const
{
    int                  x, y;
    const unsigned char  *row_s;
    unsigned char        *row_d;


    if (channel >= nchannels)
        throw CExceptionRange ("It is not possible to insert the "
        "image to this channel.", channel, 0,
        nchannels);

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst.getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x*nchannels + channel] = row_s[x];
    }
}


/**********************************************************/
/*!
 *   Insert the source image in the destination image as
 *   one color channel (integer version).
 *
 *   \param src source image
 *   \param dst destination image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 */
/**********************************************************/

void COperatorMultiplex::setChannel (const CImageIntGray &src, 
                                     CImageInt &dst,
                                     int nchannels, int channel) const
{
    int                  x, y;
    const int            *row_s;
    int                  *row_d;


    if (channel >= nchannels)
        throw CExceptionRange ("It is not possible to insert the "
        "image to this channel.", channel, 0,
        nchannels);

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst.getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x*nchannels + channel] = row_s[x];
    }
}


/**********************************************************/
/*!
 *   Insert the source image in the destination image as
 *   one color channel (double version).
 *
 *   \param src source image
 *   \param dst destination image
 *   \param nchannels number of color channels used by
 *                    the source image
 *   \param channel color channel to be extracted
 */
/**********************************************************/

void COperatorMultiplex::setChannel (const CImageDoubleGray &src, 
                                     CImageDouble &dst,
                                     int nchannels, int channel) const
{
    int                  x, y;
    const double         *row_s;
    double               *row_d;


    if (channel >= nchannels)
        throw CExceptionRange ("It is not possible to insert the "
        "image to this channel.", channel, 0,
        nchannels);

    for (y = 0; y < src.getH(); y++)
    {
        row_s = src.getRow (y);
        row_d = dst.getRow (y);

        for (x = 0; x < src.getW(); x++)
            row_d[x*nchannels + channel] = row_s[x];
    }
}


////////////////////////////////////////////////////////////
// Generic demultiplexers
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Demultiplex a byte image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 *   \param nchannels number of color channels used by
 *                    the source image
 */
/**********************************************************/

void COperatorMultiplex::demultiplexGeneric (const CImageByte &src,
                                             vector <CImageByteGray*> &dst,
                                             int nchannels) const
{
    int i;


    dst.clear();

    for (i = 0; i < nchannels; i++)
        dst.push_back (getChannel (src, nchannels, i));
}


/**********************************************************/
/*!
 *   Demultiplex an integer image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 *   \param nchannels number of color channels used by
 *                    the source image
 */
/**********************************************************/

void COperatorMultiplex::demultiplexGeneric (const CImageInt &src,
                                             vector <CImageIntGray*> &dst,
                                             int nchannels) const
{
    int i;


    dst.clear();

    for (i = 0; i < nchannels; i++)
        dst.push_back (getChannel (src, nchannels, i));
}


/**********************************************************/
/*!
 *   Demultiplex a double image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 *   \param nchannels number of color channels used by
 *                    the source image
 */
/**********************************************************/

void COperatorMultiplex::demultiplexGeneric (const CImageDouble &src,
                                             vector <CImageDoubleGray*> &dst,
                                             int nchannels) const
{
    int i;


    dst.clear();

    for (i = 0; i < nchannels; i++)
        dst.push_back (getChannel (src, nchannels, i));
}


////////////////////////////////////////////////////////////
// Generic multiplexers
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Multiplex a byte image.
 *
 *   \param src images to be multiplexed
 *   \param dst multiplexed image
 */
/**********************************************************/

void COperatorMultiplex::multiplexGeneric 
(const vector <CImageByteGray*> &src,
 CImageByte &dst) const
{
    unsigned int i;


    for (i = 0; i < src.size(); i++)
        setChannel (*(src[i]), dst, src.size(), i);
}


/**********************************************************/
/*!
 *   Multiplex an integer image.
 *
 *   \param src images to be multiplexed
 *   \param dst multiplexed image
 */
/**********************************************************/

void COperatorMultiplex::multiplexGeneric 
(const vector <CImageIntGray*> &src,
 CImageInt &dst) const
{
    unsigned int i;


    for (i = 0; i < src.size(); i++)
        setChannel (*(src[i]), dst, src.size(), i);
}


/**********************************************************/
/*!
 *   Multiplex a double image.
 *
 *   \param src images to be multiplexed
 *   \param dst multiplexed image
 */
/**********************************************************/

void COperatorMultiplex::multiplexGeneric 
(const vector <CImageDoubleGray*> &src,
 CImageDouble &dst) const
{
    unsigned int i;


    for (i = 0; i < src.size(); i++)
        setChannel (*(src[i]), dst, src.size(), i);
}


////////////////////////////////////////////////////////////
// Helper function
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Tests whether all images in the vector have got
 *   the same width and height and whether there are
 *   "nchannels" images in the vector. If this is not the 
 *   case, it throws an exception.
 *
 *   \param src vector of images to be compared
 *   \param nchannels number of color channels
 */
/**********************************************************/

void COperatorMultiplex::testDimensions (const vector <CImage*> &src,
                                         int nchannels) const
{
    unsigned int i;

    /*** Disabled by Fabian: I have no idea who wrote the next lines

    if (src.size() != (unsigned int) nchannels);  // fixme 

    for (i = 1; i < src.size(); i++)
    {
    if ((src[i])->getW() != (src[0])->getW()); // fixme
    if ((src[i])->getH() != (src[0])->getH()); // fixme
    }

    */
}


////////////////////////////////////////////////////////////
// RGB
////////////////////////////////////////////////////////////


/**********************************************************/
/*!
 *   Demultiplex a byte RGB image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageByteRGB &src,
                                      vector <CImageByteGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Demultiplex an integer RGB image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageIntRGB &src,
                                      vector <CImageIntGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Demultiplex a double RGB image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageDoubleRGB &src,
                                      vector <CImageDoubleGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Multiplex a byte RGB image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageByteGray*> &src,
                               CImageByteRGB **dst) const
{
    // testDimensions (src, 3);  // fixme

    *dst = new CImageByteRGB ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}


/**********************************************************/
/*!
 *   Multiplex an integer RGB image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageIntGray*> &src,
                               CImageIntRGB **dst) const
{
    // testDimensions (src, 3);   // fixme

    *dst = new CImageIntRGB ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}



/**********************************************************/
/*!
 *   Multiplex a double RGB image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageDoubleGray*> &src,
                               CImageDoubleRGB **dst) const
{
    // testDimensions (src, 3);  // fixme

    *dst = new CImageDoubleRGB ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}


////////////////////////////////////////////////////////////
// YUV
////////////////////////////////////////////////////////////


/**********************************************************/
/*!
 *   Demultiplex a byte YUV image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageByteYUV &src,
                                      vector <CImageByteGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Demultiplex an integer YUV image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageIntYUV &src,
                                      vector <CImageIntGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Demultiplex a double YUV image.
 *
 *   \param src image to be demultiplexed
 *   \param dst demultiplexed images
 */
/**********************************************************/

void COperatorMultiplex::demultiplex (const CImageDoubleYUV &src,
                                      vector <CImageDoubleGray*> &dst) const
{
    demultiplexGeneric (src, dst, 3);
}


/**********************************************************/
/*!
 *   Multiplex a byte YUV image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageByteGray*> &src,
                               CImageByteYUV **dst) const
{
    // testDimensions (src, 3);  // fixme

    *dst = new CImageByteYUV ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}


/**********************************************************/
/*!
 *   Multiplex an integer YUV image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageIntGray*> &src,
                               CImageIntYUV **dst) const
{
    // testDimensions (src, 3);   // fixme

    *dst = new CImageIntYUV ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}



/**********************************************************/
/*!
 *   Multiplex a double YUV image.
 *
 *   \param src images to be multiplexed
 *   \return multiplexed image
 */
/**********************************************************/

void
COperatorMultiplex::multiplex (const vector <CImageDoubleGray*> &src,
                               CImageDoubleYUV **dst) const
{
    // testDimensions (src, 3);  // fixme

    *dst = new CImageDoubleYUV ((src[0])->getW(), (src[0])->getH());

    try {
        multiplexGeneric (src, **dst);
    }
    catch (...)  {
        delete dst;
        throw;
    }
}
/**********************************************************/
/*! \example demultiplex.cpp 
 *     This gives an example on using the class 
 *     COperatorMultiplex.
 */
/**********************************************************/

