/***************************************************************************
                          COperatorRGB.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/COperatorRGB.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteYUV.h"

static unsigned char ClipArray[1024];
static unsigned char *Clip=0;

static void 
setupClip()
{
    if(Clip==0) {
        Clip=ClipArray;
        Clip += 384;
        int i;
        for (i=-384; i<640; i++)
            Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);	
    }
}


/* *******************************************************/
/*! 
 *   converts the image to gray.
 *
 *   The conversion is done with the formula
 *   <code>
 *     y=( 0.299*(1<<16)*R
 *        +0.587*(1<<16)*G
 *        +0.114*(1<<16)*B) >> 16
 *   </code>
 *
 *      \param rgb the image to be converted.
 *
 *      \return the gray valued image. 
 *
 *   \ingroup operator
 *
 */
/* *******************************************************/
TUVision::CImageByteGray 
TUVision::convertToGray (const TUVision::CImageByteRGB &rgb)
{   
    int w;
    int h;

    w=rgb.getW();
    h=rgb.getH();

    CImageByteGray gray(w,h);

    int curh=0;

    const unsigned char* src;
    unsigned char* dst;

    while(curh<h) {
        src=rgb.getRow(curh);
        dst=gray.getRow(curh);

        int size=w;
        int y;

        // next lines are formula (see above)
        // with already calculated intagers.
        // The red value has been corrected
        // due to rounding effects
        while(size--) {
            y =(int)( 19595 * *src++);
            y+=(int)( 38470 * *src++);
            y+=(int)(  7471 * *src++);

            *dst++=(unsigned char)(y>>16);
        }
        curh++;
    }
    return(CImageByteGray(gray,false));
}



/* *******************************************************/
/*! 
 *   converts the yuv image to rgb.
 *
 *   y cr cb conversion from 
 *   http://home.inforamp.net/~poynton color faq.
 *	 (Item 30)
 *
 *   The conversion is done with the formula
 *   <code>
 *     r=(unsigned char)clamp(1.000*y        +1.367*v,0.0,255.0);
 *     g=(unsigned char)clamp(1.000*y-0.336*u-0.698*v,0.0,255.0);
 *     b=(unsigned char)clamp(1.000*y+1.732*u        ,0.0,255.0);
 *   </code>
 *
 *   The function clamp() returns a value in the specified closed interval. 
 *
 *      \param yuv the image to be converted.
 *
 *      \return the rgb valued image. 
 *
 *   currently, 
 *
 *   \ingroup operator
 *
 */
/* *******************************************************/

TUVision::CImageByteRGB 
TUVision::convertToRGB (const CImageByteYUV &yuv)
{
    static const int scale=65536*256/219;
    static const int crv=(int)(65536.0*(+1.367)*256/219);
    static const int cgu=(int)(65536.0*(-0.336)*256/219);
    static const int cgv=(int)(65536.0*(-0.698)*256/219);
    static const int cbu=(int)(65536.0*(+1.732)*256/219);

    setupClip();

    CImageByteRGB rgb(yuv.getW(), yuv.getH());

    int i,j;
    int y,u,v;
    const unsigned char* src;
    unsigned char* dst;

    for(j=0;j<rgb.getH();++j) {
        src=yuv.getRow(j);
        dst=rgb.getRow(j);
        for(i=0;i<rgb.getW();++i) {

            y=scale * (*src++-16);
            u=*src++-128;
            v=*src++-128;
            // y cr cb conversion from 
            // http://home.inforamp.net/~poynton color faq.
            //
            *dst++=Clip[(y         + crv*v +32768)>>16];
            *dst++=Clip[(y + cgu*u + cgv*v +32768)>>16];
            *dst++=Clip[(y + cbu*u         +32768)>>16];
        }
    }
    return(CImageByteRGB(rgb,false));
}


TUVision::CImageByteRGB 
TUVision::convertToRGB (const CImageByteGray &gray)
{
    CImageByteRGB rgb(gray.getW(), gray.getH());

    int i,j;
    const unsigned char* src;
    unsigned char* dst;

    for(j=0;j<rgb.getH();++j) {
        src=gray.getRow(j);
        dst=rgb.getRow(j);
        for(i=0;i<rgb.getW();++i) {
            *dst++=*src;
            *dst++=*src;
            *dst++=*src++;
        }
    }

    return(CImageByteRGB(rgb,false));
}

