/***************************************************************************
                          COperatorThreshold.h  -  description
                             -------------------
    begin                : 22.08.2002
    copyright            : (C) 2001 by Islam Shdaifat, Marc Hensel (1)
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat, study thesis Marc Hensel (1)

 ***************************************************************************/

/*
 * (1) Following method added:
 *     void threshold(CImageByteGray *img,unsigned char level);
 */

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include <tuv2d/COperatorThreshold.h>

#include <tuv/CImageByteGray.h>
#include <tuv/CImageIntGray.h>
#include <tuv/CImageDoubleGray.h>
#include <tuv/CImageByteBilevel.h>


//using namespace std;
using namespace TUVision;


/**********************************************************/
/*!
 *   Convert the byte gray level image to a bilevel image.
 *   Pixels with a value greater than "level" will become
 *   white pixels in the new image. The other pixels will
 *   become black pixels.
 *
 *   \param src image to be thresholded
 *   \param tar returned bilevel image
 *   \param level threshold value  
 */
/**********************************************************/

void 
COperatorThreshold::threshold (const CImageByteGray &src, 
                               CImageByteBilevel **tar,
                               unsigned char level)
{
    int                 x, y;
    const unsigned char *row_src;
    unsigned char       *row_tar;


    *tar = new CImageByteBilevel (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_src = src.getRow (y);
        row_tar = (*tar)->getRow (y);

        for (x = 0; x < src.getW(); x++)  
        {
            if (row_src[x] > level)
                row_tar[x] = 1;
            else
                row_tar[x] = 0;
        }
    }
}; 


/**********************************************************/
/*!
 *   Threshold a byte gray level image.
 *   Pixels with a value greater equal "level" are assigned 
 *   the value specified in "high" (default: white = 255) in
 *   the new image. All other pixels are set to "low" (default: black = 0).
 *
 *   \param img		Image to be thresholded (gets overwritten with result).
 *   \param level	Threshold value.
 *   \param low		New value for pixels < level.
 *   \param high	New value for pixels >= level.
 */
/**********************************************************/

void COperatorThreshold::threshold (CImageByteGray *img,unsigned char level,unsigned char low,unsigned char high)
{
    int				elements	= img->getW() * img->getH();
    unsigned char	*position	= img->getRow(0);

    for (int i=0; i<elements; i++, position++)
        *(position) = (*(position)<level)? low : high;
}



/**********************************************************/
/*!
 *   Convert the integer gray level image to a bilevel 
 *   image. Pixels with a value greater than "level" will 
 *   become white pixels in the new image. The other pixels 
 *   will become black pixels.
 *
 *   \param src image to be thresholded
 *   \param tar returned bilevel image
 *   \param level threshold value  
 */
/**********************************************************/

void 
COperatorThreshold::threshold (const CImageIntGray &src, 
                               CImageByteBilevel **tar,
                               int level)
{
    int           x, y;
    const int     *row_src; 
    unsigned char *row_tar;


    *tar = new CImageByteBilevel (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_src = src.getRow (y);
        row_tar = (*tar)->getRow (y);

        for (x = 0; x < src.getW(); x++)  
        {
            if (row_src[x] > level)
                row_tar[x] = 1;
            else
                row_tar[x] = 0;
        }
    }
}; 



/**********************************************************/
/*!
 *   Convert the double gray level image to a bilevel 
 *   image. Pixels with a value greater than "level" will 
 *   become white pixels in the new image. The other pixels 
 *   will become black pixels.
 *
 *   \param src image to be thresholded
 *   \param tar returned bilevel image
 *   \param level threshold value  
 */
/**********************************************************/

void 
COperatorThreshold::threshold (const CImageDoubleGray &src, 
                               CImageByteBilevel **tar,
                               double level)
{
    int           x, y;
    const double  *row_src; 
    unsigned char *row_tar;


    *tar = new CImageByteBilevel (src.getW(), src.getH());

    for (y = 0; y < src.getH(); y++)
    {
        row_src = src.getRow (y);
        row_tar = (*tar)->getRow (y);

        for (x = 0; x < src.getW(); x++)  
        {
            if (row_src[x] > level)
                row_tar[x] = 1;
            else
                row_tar[x] = 0;
        }
    }
}; 


