/***************************************************************************
                           CCalibrationGapDetection.cpp  -  description
                              -------------------
     last modified        : 22.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel
 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include <tuv3d/CCalibrationGapDetection.h>

#if !defined(M_PI)
#define		M_PI	3.14159265358979323846
#endif

// using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

CCalibrationGapDetection::CCalibrationGapDetection()
{
    fWidthSymCov = 7;
    fWidthSumInv = 15;
    thresPoints  = 32;
    thresArea    = 245;

    gap_left = gap_right  = 0;		// Coordinates of bounding box
    gap_top  = gap_bottom = 0;
    gap_x    = gap_y      = 0;		// Starting position in gap

    dx = dy = 0.0;					// Direction (dx,dy)T in search for points

    Ox = Oy = Xx = Xy = 0.0;		// Image coordinates of world coordinate system
    Yx = Yy = Zx = Zy = 0.0;
}


CCalibrationGapDetection::~CCalibrationGapDetection()
{
}



/**************************************************
 *
 * Set parameters:
 *
 **************************************************/

void CCalibrationGapDetection::setWidthSCF(int width)
{
    fWidthSymCov = width;
}

void CCalibrationGapDetection::setWidthSIF(int width)
{
    fWidthSumInv = width;
}

void CCalibrationGapDetection::setThreshPoints(unsigned char threshold)
{
    thresPoints = threshold;
}

void CCalibrationGapDetection::setThreshArea(unsigned char threshold)
{
    thresArea = threshold;
}


/**************************************************
 *
 * Return parameters:
 *
 **************************************************/


int CCalibrationGapDetection::getWidthSCF()
{
    return fWidthSymCov;
}

int CCalibrationGapDetection::getWidthSIF()
{
    return fWidthSumInv;
}

unsigned char CCalibrationGapDetection::getThreshPoints()
{
    return thresPoints;
}

unsigned char CCalibrationGapDetection::getThreshArea()
{
    return thresArea;
}

void CCalibrationGapDetection::getGapFrame(int &left,int &right,int &top,int &bottom)
{
    left	= gap_left;
    right	= gap_right;
    top		= gap_top;
    bottom	= gap_bottom;
}

void CCalibrationGapDetection::getGapPosition(int &x,int &y)
{
    x = gap_x;
    y = gap_y;
}

void CCalibrationGapDetection::getDirection(double &x,double &y)
{
    x = dx;
    y = dy;
}

void CCalibrationGapDetection::getOrigin(double &x,double &y)
{
    x = Ox;
    y = Oy;
}

void CCalibrationGapDetection::getAxisX(double &x,double &y)
{
    x = Xx;
    y = Xy;
}

void CCalibrationGapDetection::getAxisY(double &x,double &y)
{
    x = Yx;
    y = Yy;
}

void CCalibrationGapDetection::getAxisZ(double &x,double &y)
{
    x = Zx;
    y = Zy;
}


/**************************************************
 *
 * Common method:
 * Rotate direction (dx,dy)T by angle (in rad).
 *
 **************************************************/

void CCalibrationGapDetection::rotateDirection(double angle)
{
    double vx = dx,vy = dy;
    dx = vx * cos(angle) - vy * sin(angle);
    dy = vx * sin(angle) + vy * cos(angle);
}


/**************************************************
 *
 * Detect and analyse gap object.
 * Approximates starting position at black/white edge in gap
 * and motion vector so that white area is left of motion.
 *
 **************************************************/

/*
 * Public method
 */

bool CCalibrationGapDetection::initGap(CImageByteGray &orig_img)
{
    CImageByteGray		*gap_img;		// Detected gap area pixels = 255, else 0
    gap_img = detectGap(orig_img);		// Detect gap area

    if (!frameGap(*gap_img))			// Set bounding box
    {
        delete(gap_img);
        return false;
    }

    delete(gap_img);
    initGapPosition (orig_img);			// Find starting point at black/white edge in gap
    initGapDirection(orig_img);			// Approx. vector so that white area is on left side

    return true;
}


/*
 * Private method: Detect gap area and set pixels to 255 (other pixels 0)
 */

CImageByteGray* CCalibrationGapDetection::detectGap(CImageByteGray &orig_img)
{
    CImageByteGray		*temp_img;				// Temporary image (processing steps)

    CFilterSymmetricCovariance	SymCovFilter;	// Filter and operator
    CFilterSumInversion			sumInversion;
    COperatorThreshold			threshold;

    int				x,y;					// Runs through image to find pixels with value = 255
    unsigned char	*row;					// Pointer: 1st pixel in row (image matrix)
    int				obj_size;				// Number of pixels in current object
    int				max_x,max_y;			// Object with largest number of pixels (canonian element)
    int				max_size = 0;			// Number of pixels in biggest object

    /* Create binary gap image */
    temp_img = new CImageByteGray(orig_img);
    SymCovFilter.init(fWidthSymCov);			// Symmetric covariance filter
    SymCovFilter.filter(*temp_img);
    threshold.threshold(temp_img,thresPoints);	// Threshold filter (Amplify points)
    sumInversion.setWidth(fWidthSumInv);		// Sum inversion filter
    sumInversion.filter(*temp_img);
    threshold.threshold(temp_img,thresArea);	// Threshold filter (Discard except for peaks)
    sumInversion.setWidth(fWidthSumInv);		// Sum inversion filter (Remove border areas)
    sumInversion.removeBorder(*temp_img);

    /* Remove everything but biggest connected area */
    for (y=0; y<temp_img->getH(); y++)
    {
        row = temp_img->getRow(y);
        for (x=0; x<temp_img->getW(); x++)
        {
            if (row[x]==255)								// Object pixel found
            {
                obj_size = transposeArea(*temp_img,x,y,1);	// Set area around (x,y) to 1
                if (obj_size>max_size)						// Test for maximum
                {
                    max_x		= x;
                    max_y		= y;
                    max_size	= obj_size;
                }
            }
        }
    }
    if (max_size>0) transposeArea(*temp_img,max_x,max_y,255);	// Set area around (max_x,max_y) to 255
    return temp_img;
}


/*
 * Private method: Change pixel values of connected area containing (cx,cy)
 */

int CCalibrationGapDetection::transposeArea(CImageByteGray &img,int cx,int cy,unsigned char new_value)
{
    unsigned char	old_value;					// Pixel value of area to be transposed
    unsigned char	*row;						// Pointer: 1st pixel in row (image matrix)
    int				x,y;						// Position to be checked/of neighboring pixel
    int				*check_x;					// List of pixels still to tested if they belong to current object
    int				*check_y;
    int				chk_pnt;					// Index pointer in test pixel array check_{x|y}[]
    int				obj_size;					// Number of pixels in area
    int				max_size;					// Max. pixels that can be in the "to check" list at a time

    max_size = img.getW()*img.getH();
    check_x = new int[max_size];
    check_y = new int[max_size];

    obj_size	= 0;							// Init: Number of pixels in area
    row			= img.getRow(cy);				// Init: Pixel value of area to be transposed
    old_value	= row[cx];
    chk_pnt		= 0;							// Init: List of pixels to be testet
    check_x[chk_pnt] = cx;
    check_y[chk_pnt] = cy;

    while (chk_pnt>=0)							// Check area (list of neighbouring pixels to be checked)
    {
        x	= check_x[chk_pnt];
        y	= check_y[chk_pnt];
        row	= img.getRow(y);
        if (chk_pnt>max_size-9)					// Avoid array overflow
        {
            delete[](check_x);
            delete[](check_y);
            return 0;
        }
        if (row[x]==old_value)					// Pixel belongs to object
        {
            row[x] = new_value;					// Mark pixel as "checked"
            obj_size++;
            chk_pnt--;
            if ((x>0)&&(y>0))					// Mark neighbours (coordinates in image!) as "to be checked"
            {
                chk_pnt++;
                check_x[chk_pnt] = x - 1;
                check_y[chk_pnt] = y - 1;
            }
            if (y>0)
            {
                chk_pnt++;
                check_x[chk_pnt] = x;
                check_y[chk_pnt] = y - 1;
            }
            if ((x<img.getW()-2)&&(y>0))
            {
                chk_pnt++;
                check_x[chk_pnt] = x + 1;
                check_y[chk_pnt] = y - 1;
            }
            if (x>0)
            {
                chk_pnt++;
                check_x[chk_pnt] = x - 1;
                check_y[chk_pnt] = y;
            }
            if (x<img.getW()-2)
            {
                chk_pnt++;
                check_x[chk_pnt] = x + 1;
                check_y[chk_pnt] = y;
            }
            if ((x>0)&&(y<img.getH()-2))
            {
                chk_pnt++;
                check_x[chk_pnt] = x - 1;
                check_y[chk_pnt] = y + 1;
            }
            if (y<img.getH()-2)
            {
                chk_pnt++;
                check_x[chk_pnt] = x;
                check_y[chk_pnt] = y + 1;
            }
            if ((x<img.getW()-2)&&(y<img.getH()-2))
            {
                chk_pnt++;
                check_x[chk_pnt] = x + 1;
                check_y[chk_pnt] = y + 1;
            }

        }
        else									// Remove pixel from check list
            chk_pnt--;
    }
    delete[](check_x);
    delete[](check_y);
    return 	obj_size;
}



/*
 * Private method: Find and set bounding box coordinates of detected gap area.
 */

bool CCalibrationGapDetection::frameGap(CImageByteGray &gap_img)
{
    int				x,y;
    unsigned char	*row;						// Pointer: 1st pixel in row (image matrix)

    gap_top		= gap_img.getH();				/* Find top and bottom */
    gap_bottom	= gap_img.getH();
    for (y=0; y<gap_img.getH(); y++)
    {
        row = gap_img.getRow(y);
        for (x=0; x<gap_img.getW(); x++)
        {
            if (row[x]==255)					// Object area pixel found in row
            {
                if (y<gap_top) gap_top = y;
                gap_bottom	= y;
                x			= gap_img.getW();	// Next row
            }
        }
    }
    if (gap_top>=gap_bottom) return false;

    gap_left	= gap_img.getW();				/* Find left and right */
    gap_right	= gap_img.getW();
    for (x=0; x<gap_img.getW(); x++)
    {
        for (y=0; y<gap_img.getH(); y++)
        {
            row = gap_img.getRow(y);
            if (row[x]==255)					// Object area pixel found in column
            {
                if (x<gap_left) gap_left = x;
                gap_right	= x;
                y			= gap_img.getH();	// Next column
            }
        }
    }
    if (gap_left>=gap_right) return false;

    return true;
}


/*
 * Private method: Find and set starting position at black/white edge in gap.
 */

void CCalibrationGapDetection::initGapPosition(CImageByteGray &orig_img)
{
    /*
     * Search center of box for middle of border between white and black.
     * (Moves max. 3 pixel from center of box)
     * Parameter img: Original image
     */

    int				x,y;
    unsigned char	val,min_val,max_val;	// Pixel values in bounding box
    unsigned char	threshold;				// Threshold: Middle between min. and max. value

    /* List of neighbouring points to be examined (allows special order of examination).
    List => priority of movement: Remain in center, horizontal/vertical, diagonal */
    int				x_list[9] = {0,1,0,-1, 0,-1,1,-1, 1};
    int				y_list[9] = {0,0,1, 0,-1,-1,1, 1,-1};
    int				d_index;				// Index to motion vector to improve point (xc,yc)

    int				xp,yp;					// Temporary position of examined point
    unsigned char	filterWidth,offset;
    unsigned char	win_count;				// Window: Number of pixels>threshold (=: white)
    int				win_diff,min_diff;		// Difference from equal number white and black pixels
    unsigned char	*row;					// Row pointer in matrix

    filterWidth	= 5;
    offset		= filterWidth/2;

    /* Calculate Threshold in bounding box */
    min_val = 255;
    max_val = 0;
    for (y=gap_top; y<=gap_bottom; y++)
    {
        row = orig_img.getRow(y);
        for (x=gap_left; x<=gap_right; x++)
        {
            val	= row[x];
            if (val<min_val) min_val = val;
            if (val>max_val) max_val = val;
        }
    }
    threshold = (min_val + max_val) / 2;

    /* Initial point */
    gap_x = (gap_left + gap_right) / 2;	// Center of box
    gap_y = (gap_top + gap_bottom) / 2;

    /* Improvement steps */
    for (int step=0; step<=3; step++)
    {
        /* Examine Neighbouring pixels */
        d_index		= 0;				// Init: (xc,yc) best fit
        min_diff	= filterWidth*filterWidth;
        for (int i=0; i<9; i++)
        {
            yp = gap_y + y_list[i];
            xp = gap_x + x_list[i];
            /* Count pixel>threshold in window around pixel */
            win_count = 0;
            for (y=-offset; y<=offset; y++)
            {
                row = orig_img.getRow(yp+y);
                for (x=-offset; x<=offset; x++)
                    if (row[xp+x]>threshold) win_count++;
            }
            /* Compare result to so far best fit */
            win_diff = (filterWidth*filterWidth)/2 - win_count;
            if (win_diff<0) win_diff = -win_diff;
            if (win_diff<min_diff)
            {
                min_diff = win_diff;
                d_index	 = i;
            }
        }
        gap_x += x_list[d_index];
        gap_y += y_list[d_index];
    }

}


/*
 * Private method: Find and set starting vector so that white is on left side of motion.
 */

void CCalibrationGapDetection::initGapDirection(CImageByteGray &orig_img)
{
    /*
     * Find approx. direction of edge between black and white using (xc,yc).
     * Orientation is chosen so that the white area is on the left side of the motion.
     * Vector (dx,dy)T is normalized to length 1.
     * Parameter img: Original image
     */

    int				x,y;
    unsigned char	val,min_val,max_val;	// Pixel values in bounding box
    unsigned char	threshold;				// Threshold: Middle between min. and max. value

    unsigned char	filterWidth,offset;
    unsigned char	*row;					// Row pointer in matrix

    double			wx,wy,bx,by;			// Center of gravity
    int				wSum,bSum;				// (white pixel:>threshold, black pixel:<=threshold)
    double			euclid_norm;			// Length of (dx,dy)T (to normalize to length 1)

    filterWidth	= 7;
    offset		= filterWidth/2;

    /* Calculate threshold in filter area around (xc,yc) */
    min_val = 255;
    max_val = 0;
    for (y=gap_y-offset; y<=gap_y+offset; y++)
    {
        row = orig_img.getRow(y);
        for (x=gap_x-offset; x<=gap_x+offset; x++)
        {
            val	= row[x];
            if (val<min_val) min_val = val;
            if (val>max_val) max_val = val;
        }
    }
    threshold = (min_val + max_val) / 2;

    /* Calculate centers of gravity in filter area around (xc,yc) */
    wx = wy = 0.0;
    bx = by = 0.0;
    wSum = bSum = 0;
    for (y=gap_y-offset; y<=gap_y+offset; y++)
    {
        row = orig_img.getRow(y);
        for (x=gap_x-offset; x<=gap_x+offset; x++)
        {
            if (row[x]>threshold)
            {
                wSum+= row[x];
                wx	+= row[x] * x;
                wy	+= row[x] * y;
            }
            else
            {
                bSum+= row[x];
                bx	+= row[x] * x;
                by	+= row[x] * y;
            }
        }
    }
    wx /= wSum;	wy /= wSum;				// Calculate centers of gravity
    bx /= bSum;	by /= bSum;

    /* Calculate direction vector */
    dx = wx - bx;						// Vector from black to white center of gravity
    dy = wy - by;
    rotateDirection(0.5*M_PI);			// Rotate direction 90

    /* Normalize vector */
    euclid_norm = sqrt(dx*dx+dy*dy);
    dx /= euclid_norm;				// Normalize to length=1
    dy /= euclid_norm;
}

/**************************************************
 *
 * Find feature points/coordinate systems.
 * Gap must have been analysed before calling methods.
 *
 **************************************************/

/*
 * Public method for 2D calibration object
 */

bool CCalibrationGapDetection::findAxes2D(CImageByteGray &orig_img)
{
    CImageByteGray		*pnt_img;				// Image with peaks at points (filtered by sym. covariance filter)
    CImageByteGray		*edge_img;				// Edge image (enable findNextPoint to move along edges)

    CFilterSobel				SobelFilter;
    CFilterSymmetricCovariance	SymCovFilter;
    COperatorThreshold			threshold;

    /* Create point and edge images */
    pnt_img = new CImageByteGray(orig_img);			// Point image
    SymCovFilter.init(fWidthSymCov);				// - Symmetric covariance filter
    SymCovFilter.filter(*pnt_img);
    threshold.threshold(pnt_img,thresPoints);		// - Threshold filter (Amplify points)
    edge_img = new CImageByteGray(orig_img);		// Edge image (Sobel filter)
    SobelFilter.filter(*edge_img);

    /* Follow direction to next point and update direction */
    if(!findNextPoint(*edge_img,*pnt_img,gap_x,gap_y,Ox,Oy))	// Find origin (0;0;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }
    if(!findNextPoint(*edge_img,*pnt_img,Ox,Oy,Yx,Yy))			// Find y-axis (0;1;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }
    rotateDirection(0.5*M_PI);									// Rotate direction 90
    if(!findNextPoint(*edge_img,*pnt_img,Ox,Oy,Xx,Xy))			// Find x-axis (1;0;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }

    delete(pnt_img);
    delete(edge_img);
    return true;
}


/*
 * Public method for 3D calibration object
 */

bool CCalibrationGapDetection::findAxes3D(CImageByteGray &orig_img)
{
    CImageByteGray		*pnt_img;				// Image with peaks at points (filtered by sym. covariance filter)
    CImageByteGray		*edge_img;				// Edge image (enable findNextPoint to move along edges)

    CFilterSobel				SobelFilter;
    CFilterSymmetricCovariance	SymCovFilter;
    COperatorThreshold			threshold;

    /* Create point and edge images */
    pnt_img = new CImageByteGray(orig_img);		// Point image
    SymCovFilter.init(fWidthSymCov);			// - Symmetric covariance filter
    SymCovFilter.filter(*pnt_img);
    threshold.threshold(pnt_img,thresPoints);	// - Threshold filter (Amplify points)
    edge_img = new CImageByteGray(orig_img);	// Edge image (Sobel filter)
    SobelFilter.filter(*edge_img);

    /* Follow direction to next point and update direction */
    if(!findNextPoint(*edge_img,*pnt_img,gap_x,gap_y,Ox,Oy))	// Find 2D-origin (0;0;0) (area deleted in pnt_img)
    {															// (Temporary)
        delete(pnt_img);
        return false;
    }
    rotateDirection(-0.5*M_PI);									// Rotate direction -90
    if(!findNextPoint(*edge_img,*pnt_img,Ox,Oy,Yx,Yy))			// Find y-axis (0;1;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }
    if(!findNextPoint(*edge_img,*pnt_img,Yx,Yy,Ox,Oy))			// Find origin (0;0;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }
    if(!findNextPoint(*edge_img,*pnt_img,Ox,Oy,Xx,Xy))			// Find x-axis (1;0;0) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }
    rotateDirection(0.5*M_PI);									// Rotate direction 90
    if(!findNextPoint(*edge_img,*pnt_img,Ox,Oy,Zx,Zy))			// Find z-axis (0;0;1) (area deleted in pnt_img)
    {
        delete(pnt_img);
        return false;
    }

    delete(pnt_img);
    delete(edge_img);
    return true;
}


/*
 * Private method: Find next point in direction (dx,dy)
 */

bool CCalibrationGapDetection::findNextPoint
(CImageByteGray &edge_img,CImageByteGray &pnt_img,double x0,double y0,double &Px,double &Py)
{
    int				x,y;
    double			xp,yp;					// Current position
    unsigned char	filterWidth,offset;
    bool			found;
    unsigned char	*row;					// Row pointer in matrix

    double			gx,gy;					// Point <- Center of gravity
    int				gSum;

    double			angle;					// Angle of direction (dx,dy)T in rad
    int				edge_x[2],edge_y[2];	// Relative position of neighboring pixels to move on edge after each step
    int				edge_pointer;			// Pointer to arrays to pixel better on edge than current position (xp,yp)T
    unsigned char	edge_value;				// Value of pixel in edge image

    double			euclid_norm;			// Length of (dx,dy)T (to normalize to length 1)

    filterWidth	= 7;
    offset		= filterWidth/2;
    found		= false;

    /* Point too close to image boarder */
    if ( ((int)x0-offset<0) || ((int)x0+offset>=edge_img.getW()) )	return false;
    if ( ((int)y0-offset<0) || ((int)y0+offset>=edge_img.getH()) )	return false;

    /* Set neighboring pixels to compare after each step (move on edge) */
    angle = acos(dx);					// Result in [0,PI]; cos(angle) = dx/sqrt(dx^2+dy^2) = dx/1.0 = dx
    angle /= M_PI;
    if ((1.0/8.0<=angle)&&(angle<3.0/8.0))
    {
        edge_x[0] = -1; edge_x[1] =  1;	// Relative neighboring points to compare
        edge_y[0] =  1; edge_y[1] = -1;
    }
    else if ((3.0/8.0<=angle)&&(angle<5.0/8.0))
    {
        edge_x[0] = -1; edge_x[1] =  1;
        edge_y[0] =  0; edge_y[1] =  0;
    }
    else if ((5.0/8.0<=angle)&&(angle<7.0/8.0))
    {
        edge_x[0] = -1; edge_x[1] =  1;
        edge_y[0] = -1; edge_y[1] =  1;
    }
    else
    {
        edge_x[0] =  0; edge_x[1] =  0;
        edge_y[0] = -1; edge_y[1] =  1;
    }

    /* Set window/point to 0 (otherwise no movement when starting at a point) */
    xp = x0;	// Starting point
    yp = y0;
    for (y=(int)yp-offset; y<=(int)yp+offset; y++)
    {
        row = pnt_img.getRow(y);
        for (x=(int)xp-offset; x<=(int)xp+offset; x++)
            row[x] = 0;
    }

    /* Follow direction until pixel=255 found */
    while (!found)
    {
        /* Search window for pixel=255 */
        for (y=(int)yp-offset; y<=(int)yp+offset; y++)
        {
            if ((y<0)||(y>=pnt_img.getH())) return false;
            row = pnt_img.getRow(y);
            for (x=(int)xp-offset; x<=(int)xp+offset; x++)
            {
                if ((x<0)||(x>=pnt_img.getW())) return false;
                if (row[x]==255)			// Point fount in window
                {
                    Px	= x;				// Set estimated origin coordinates
                    Py	= y;
                    found = true;			// Leave while-construct
                    x	= (int)xp+offset;	// Leave for-constructs
                    y	= (int)yp+offset;
                }
            }
        }
        /* Move forward in direction (dx,dy)T */
        xp += 2*dx;	// Step forward alomg direction vector (length((dx,dy)T)=1)
        yp += 2*dy;
        /*
         * Compare position with 2 direct neighbors approx. orthogonal to direction.
         * Move to pixel with max. value in edge image => move along edge.
         * (Important in case the quality of the direction is bad)
         */
        edge_pointer = -1;					// Current pixel is best on edge
        row = edge_img.getRow((int)yp);
        edge_value = row[(int)xp];
        row = edge_img.getRow((int)(yp+edge_y[0]));	// Compare 1st neighbor
        if (edge_value<row[(int)(xp+edge_x[0])])
        {
            edge_value = row[(int)(xp+edge_x[0])];
            edge_pointer = 0;
        }
        row = edge_img.getRow((int)(yp+edge_y[1]));	// Compare 2nd neighbor
        if (edge_value<row[(int)(xp+edge_x[1])])
        {
            edge_value = row[(int)(xp+edge_x[1])];
            edge_pointer = 1;
        }
        if (edge_pointer>=0)						// Move to neighboring pixel
        {
            xp = xp + edge_x[edge_pointer];
            yp = yp + edge_x[edge_pointer];
        }
    }

    /*
     * Improve position of detected point P:
     * Set to center of gravity in window around detected pixel
     */
    xp = Px;	// Starting position: Detected pixel of point
    yp = Py;
    gx = gy = 0.0;
    gSum = 0;
    for (y=(int)yp-offset; y<=(int)yp+offset; y++)
    {
        row = pnt_img.getRow(y);
        for (x=(int)xp-offset; x<=(int)xp+offset; x++)
        {
            gSum+= row[x];
            gx	+= row[x] * x;
            gy	+= row[x] * y;
        }
    }
    Px = gx / gSum;				// Calculate center of gravity
    Py = gy / gSum;

    /* Improve direction vector */
    dx = Px - x0;					// Vector from found point to starting point
    dy = Py - y0;
    euclid_norm = sqrt(dx*dx+dy*dy);
    dx /= euclid_norm;				// Normalize to length=1
    dy /= euclid_norm;

    return true;
}
