/***************************************************************************
                          CCorrespondence.cpp  -  description
                             -------------------
    begin                : 11.03.2003
    copyright            : (C) 2001 by Thomas Beilicke
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Thomas Beilicke

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CCorrespondence.h"

using namespace TUVision;

/**********************************************************/ 
/*! It is just a wrapper for vector access to write more 
 *  descriptive code. 
 *  \remark The bounds are not checked if the
 *			class is used directly. 
 *			Use the CCorrespondenceChecking template function 
 *			instead to enable checking.
 *
 *	\returns The according image point in homogeneous coordinates (3D).
 *
 *	\param index The camera number to which the image point belongs.
 */
CPoint3D<double> CCorrespondenceBase::getImagePoint(unsigned int index) const
{
    return (*this)[index];
}

/**********************************************************/ 
/*! It is just a wrapper for vector access to write more 
 *  descriptive code. The correspondence point is appended.
 *
 *	\param point An image point in homogeneous coordinates (3D).
 */
void CCorrespondenceBase::addImagePoint(const CPoint3D<double> &point)
{
    push_back(point);
}

/**********************************************************/ 
/*! It is just a wrapper for vector access to write more 
 *  descriptive code. Additionally the point is converted from
 *  homogeneous to euclidean coordinates.
 *
 *  \remark The bounds are not checked if the class is used directly. 
 *			Use the CCorrespondenceChecking template function 
 *			instead to enable checking.
 *
 *	\throws CExceptionIllegalPoint if the template 
 *			CCorrespondenceChecking is used and the point represents
 *			a direction (last component 0).
 *
 *	\returns The according image point in euclidean coordinates (2D).
 *
 *	\param index The camera number to which the image point belongs.
 */
CPoint2D<double> CCorrespondenceBase::getEuclideanImagePoint(unsigned int index) const
{
    CPoint3D<double> tmp = (*this)[index];
    return CPoint2D<double> (tmp.x/tmp.z, tmp.y/tmp.z);
}

/**********************************************************/ 
/*! It is just a wrapper for vector access to write more 
 *  descriptive code. The correspondence point is appended.
 *  Additionally the point is converted to homogeneous 
 *  coordinates by appending a third component 1.
 *
 *	\param point An image point in euclidean coordinates (2D).
 */
void CCorrespondenceBase::addEuclideanImagePoint(const CPoint2D<double> &point)
{
    push_back(CPoint3D<double> (point.x, point.y, 1.0));
}


/**********************************************************/ 
/*! It represents the point in world coordinates that projects
 *  to the image points of this correspondence for different 
 *  cameras.
 *
 *	\param point The corresponding world point in homogeneous coordinates (4D).
 */
void CCorrespondenceBase::setWorldPoint(const CPoint4D<double> &point)
{
    worldPoint=point;
}

/**********************************************************/ 
/*! If the world point has not been calculated it contains
 *  (0,0,0,0) which is an illegal homogeneous point!
 *
 *  \remark Use the CWorldCorrespondenceChecking template function 
 *			to throw an exception if the point has not been calculated yet.
 *
 *	\throws CExceptionIllegalPoint if the template 
 *			CWorldCorrespondenceChecking is used and the mentioned 
 *			condition occurs.
 *
 *	\returns The world point in homogeneous coordinates (4D).
 *
 */
CPoint4D<double> CCorrespondenceBase::getWorldPoint() const 
{
    return worldPoint;
}

/**********************************************************/ 
/*! The points is converted to euclidean coordinates.
 *  If the world point has not been calculated it contains
 *  (0,0,0,0) which is an illegal homogeneous point!
 *
 *  \remark Use the CWorldCorrespondenceChecking template 
 *			to throw an exception if the point has not been 
 *			calculated yet or it represents a direction 
 *			(last component 0).
 *
 *	\throws CExceptionIllegalPoint if the template 
 *			CWorldCorrespondenceChecking is used and 
 *			the mentioned condition occurs.
 *
 *	\returns The world point in euclidean coordinates (4D).
 *
 */
CPoint3D<double> CCorrespondenceBase::getEuclideanWorldPoint() const
{
    return CPoint3D<double> (worldPoint.x/worldPoint.t, worldPoint.y/worldPoint.t, worldPoint.z/worldPoint.t);
}

/**********************************************************/ 
/*! The point is converted to homogeneous coordinates 
 *  by appending a fourth component 1.
 *
 *	\param point The corresponding world point in euclidean coordinates (3D).
 */
void CCorrespondenceBase::setEuclideanWorldPoint(const CPoint3D<double> &point)
{
    worldPoint=CPoint4D<double> (point.x, point.y, point.z, 1.0);
}
