/***************************************************************************
                          CHomography.cpp  -  description
                             -------------------
    begin                : July 17 2002
    copyright            : (C) 2002 by Marco Grimm, Islam Shdaifat
    email                : grimm@tu-harburg.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CHomography.h"
#include "tuv3d/CPointListNormalizer.h"
#include "newmat/newmatap.h"

using namespace NEWMAT;
using namespace std;
using namespace TUVision;

bool CHomography::m_bUseNormalizer = true;

/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3
 *
 ***************************************************************************/
CHomography::CHomography() :
Matrix(3,3)
{
    //Set the homography to 0
    *this << 1.0 << 0.0 << 0.0
        <<   0.0 << 1.0 << 0.0
        <<   0.0 << 0.0 << 1.0;
}


/***************************************************************************/
/*!
 *  plane-to-image homography from correspondences
 *
 *  Initializes the matrix-part by 3x3 and calculates its values for
 *  the plane-to-image homography.
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 17 )
 *
 *        [xi]                [Xi]
 *  m = s [yi]        = [H] * [Yi]         = M
 *        [1 ]                [1 ]
 *
 *        list.first          list.second
 *        image point         worldc. point
 *
 ***************************************************************************/
CHomography::CHomography( const TDataVector &list,
                         CIntrinsicMatrix A ) :
Matrix(3,3)
{
    //Set the homography to 0
    *this << 1.0 << 0.0 << 0.0
        <<   0.0 << 1.0 << 0.0
        <<   0.0 << 0.0 << 1.0;

    intrinsicParametersFirst = A;

    if ( list.size() >= minimumdata )
    {
        if ( m_bUseNormalizer )
        {
            CPointPairListNormalizer normalizer( list );

            planeToImageAlgorithm( normalizer.getNormalizedlist() );

            Matrix H = normalizer.getfirstsidetransformation().i() * (*this) * normalizer.getsecondsidetransformation();

            *this << H;
        }
        else
        {
            planeToImageAlgorithm( list );
        }
    }
}


/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3 and calculates its values for
 *  the plane-to-image homography from R and t.
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 17 )
 *
 ***************************************************************************/
CHomography::CHomography( Matrix R,
                         ColumnVector t,
                         CIntrinsicMatrix A ) :
Matrix(3,3)
{
    intrinsicParametersFirst = A;

    Matrix H(3,3);

    H <<   R(1,1) << R(1,2) << t(1)
        << R(2,1) << R(2,2) << t(2)
        << R(3,1) << R(3,2) << t(3);
    H = intrinsicParametersFirst * H;

    // Normalize the homography to get always the solution that the plane
    // is in front of the camera.
    H /= -H(3,3);

    *this << H;
}


/***************************************************************************/
/*!
 *  Destructor does nothing as there is no dynamic data
 *
 ***************************************************************************/
CHomography::~CHomography()
{

}


/***************************************************************************/
/*!
 *  Computes the plane-to-image homography with the correspondences and
 *  sets the offset of the first point in the correspondences list because
 *  these values are image coordinates. The second value has to be the
 *  point in world coordinates.
 *
 ***************************************************************************/
void CHomography::planeToImageAlgorithm( const TDataVector &list )
{
    calculateHomographyFromCorrespondences( list );
}


/***************************************************************************/
/*!
 *  Helper function to calculate the homography from correspondences.
 *
 ***************************************************************************/
void CHomography::calculateHomographyFromCorrespondences(
    const TDataVector &list,
    unsigned int firstindex,
    unsigned int secondindex )
{
    TDataVector::const_iterator currentcorrespondence;

    try
    {
        // building the matrix A we want to solve here
        // Ax=0, x is the minimum eigen vector of A
        double Xi,Yi,xi,yi;
        int equation;
        Matrix A( 2 * list.size(), 9);

        for ( currentcorrespondence = list.begin(), equation = 1;
            currentcorrespondence < list.end();
        ++currentcorrespondence, equation++ )
        {
            xi = (*currentcorrespondence)[firstindex].x;
            yi = (*currentcorrespondence)[firstindex].y;

            Xi = (*currentcorrespondence)[secondindex].x;
            Yi = (*currentcorrespondence)[secondindex].y;

            //        [xi]         [Xi]
            //  m = s [yi] = [H] * [Yi] = M
            //        [1 ]         [1 ]
            //             [ Xi   Yi   1    0    0    0    -Xi*xi, -Yi*xi, -1*xi]
            //             [ 0    0    0    Xi   Yi   1    -Xi*yi, -Yi*yi, -1*yi]
            double eq1[] = { Xi,  Yi,  1.0, 0.0, 0.0, 0.0, -Xi*xi, -Yi*xi, -xi };
            double eq2[] = { 0.0, 0.0, 0.0, Xi,  Yi,  1.0, -Xi*yi, -Yi*yi, -yi };

            A.Row( equation ) << eq1;
            equation++;
            A.Row( equation ) << eq2;
        }

        // The eigen vector
        int m=A.Nrows();
        int n=A.Ncols();

        // if the matrix underdetermined
        if(m<n)
        {
            m=n;
            A=A.t()*A;
        }
        Matrix U(m,n);
        DiagonalMatrix D(9);
        Matrix V(n,n);

        // singular value decomposition
        SVD(A, D, U, V);

        // minimum singular value
        double minv = D(1);
        int minp = 1;
        double maxv = D(1);
        for ( int j=2; j<=9; j++ )
        {
            if ( D(j)<minv )
            {
                minv = D(j);
                minp = j;
            }
            if ( D(j)>maxv )
            {
                maxv = D(j);
            }
        }

        // condition of the equation system
        m_Condition = maxv / minv;

        // x is the minimum singular value
        ColumnVector x(9);
        x = V.Column( minp );

        // Normalize the homography to get always the solution that the plane
        // is in front of the camera.
        x /= -x(9);

        // assigning the homography matrix
        (*this) << x(1) << x(2) << x(3) << x(4) << x(5) << x(6) << x(7) << x(8) << x(9);

    }
    catch (...)
    {
    }
}


/***************************************************************************/
/*!
 *  Calculate the rotation matrix from the homography depending if the
 *  homography is a plane-to-image or a planar homography
 *
 ***************************************************************************/
Matrix CHomography::getR()
{
    return getRplaneToImageHomography();
}

/***************************************************************************/
/*!
 *  Calculate the translation vector from the homography depending if the
 *  homography is a plane-to-image or a planar homography
 *
 ***************************************************************************/
Matrix CHomography::getT()
{
    return getTplaneToImageHomography();
}


/***************************************************************************/
/*!
 *  Helper function to calculate the rotation matrix from a plane-to-image
 *  homography.
 *  (see paper Zhang, 1998, A Flexible New Technique for Camera
 *   Calibration, page 6)
 *
 ***************************************************************************/
Matrix CHomography::getRplaneToImageHomography()
{
    Matrix R(3,3);
    R = 0;

    try
    {
        Matrix Ha;

        //              -1
        // [Ha] =    [A]                  *   [H]
        Ha = intrinsicParametersFirst.i() * (*this);

        ColumnVector ha1 = Ha.Column(1);
        ColumnVector ha2 = Ha.Column(2);
        ColumnVector ha3 = Ha.Column(3);

        // the first scale factor lamda1 in Zhang paper
        double s1 = 1/sqrt( ha1.SumSquare() );

        // the second scale factor lamda1 in Zhang paper
        double s2 = 1/sqrt( ha2.SumSquare() );

        // mean of the two factors
        double s = (s1+s2)/2.0;


        // if the two scale factors are near this is a correct homography
        //grimm: to do -- if( abs((1-s2/s1)*100) < 30)
        {
            // column1 of the rotation matrix r1
            R.Column(1) << s * ha1;

            // column2 of the rotation matrix r2
            R.Column(2) << s * ha2;

            // column3 of the rotation matrix r3 = r1 X r3
            R.Column(3) << crossProduct( R.Column(1), R.Column(2) );

            calculateOrthogonalRotationMatrix( R );
        }

    }
    catch (...)
    {
    }

    return R;
}


/***************************************************************************/
/*!
 *  Helper function to calculate the translation vector from a
 *  plane-to-image homography.
 *  (see paper Zhang, 1998, A Flexible New Technique for Camera
 *   Calibration, page 6)
 *
 ***************************************************************************/
ColumnVector CHomography::getTplaneToImageHomography()
{
    ColumnVector t(3);
    t = 0;

    try
    {
        Matrix Ha;

        //              -1
        // [Ha] =    [A]                  *   [H]
        Ha = intrinsicParametersFirst.i() * (*this);

        ColumnVector ha1 = Ha.Column(1);
        ColumnVector ha2 = Ha.Column(2);
        ColumnVector ha3 = Ha.Column(3);

        // the first scale factor lamda1 in Zhang paper
        double s1 = 1/sqrt( ha1.SumSquare() );

        // the second scale factor lamda1 in Zhang paper
        double s2 = 1/sqrt( ha2.SumSquare() );

        // mean of the two factors
        double s = (s1+s2)/2.0;


        // if the two scale factors are near this is a correct homography
        //grimm: to do -- if( abs((1-s2/s1)*100) < 30)
        {
            // translation vector
            t << s * ha3;
        }

    }
    catch (...)
    {
    }

    return t;
}


/***************************************************************************/
/*!
 *  Helper function to calculate a orthogonal rotation matrix from an
 *  estimated rotation matrix with errors.
 *  (see paper Zhang, 1998, A Flexible New Technique for Camera
 *   Calibration, page 19)
 *
 ***************************************************************************/
void CHomography::calculateOrthogonalRotationMatrix( Matrix& R )
{
    Matrix ru(3,3);
    Matrix rv(3,3);
    DiagonalMatrix rs(3);
    Matrix Rortho(3,3);

    SVD( R, rs, ru, rv );

    // the corrected rotation matrix
    R = ru * ( rv.t() );
}


/***************************************************************************/
/*!
 *  Helper function to calculate cross product of two vectors.
 *
 ***************************************************************************/
ColumnVector CHomography::crossProduct(ColumnVector v1, ColumnVector v2)
{
    ColumnVector v3(3);

    v3(1) = v1(2) * v2(3) - v1(3) * v2(2);
    v3(2) = v1(3) * v2(1) - v1(1) * v2(3);
    v3(3) = v1(1) * v2(2) - v1(2) * v2(1);

    return v3;
}


/***************************************************************************/
/*!
 *  Set the homography to the input Matrix. The intrinsic parameters and
 *  the type of the homography are not changed.
 *
 ***************************************************************************/
void CHomography::setHomography( Matrix H, CIntrinsicMatrix A )
{
    *this << H;
    intrinsicParametersFirst = A;
}


/***************************************************************************/
/*!
 *  Set the intrinsic parameters of the first view
 *
 ***************************************************************************/
void CHomography::setIntrinsicParameters( CIntrinsicMatrix A )
{
    intrinsicParametersFirst = A;
}


/***************************************************************************/
/*!
 *  Get the intrinsic parameters of the first view
 *
 ***************************************************************************/
CIntrinsicMatrix CHomography::getIntrinsicParameters()
{
    return intrinsicParametersFirst;
}


/***************************************************************************/
/*!
 *  The return value is the projection with the homography of the
 *  input point.
 *
 ***************************************************************************/
CVertex2d CHomography::project( const CVertex2d& point )
{
    CVertex2d projectedPoint;

    try
    {
        ColumnVector m1(3), m2(3);
        m1 << point.x << point.y << 1;

        m2 = (*this) * m1;

        projectedPoint.x = m2(1) / m2(3);
        projectedPoint.y = m2(2) / m2(3);
    }
    catch (...)
    {
    }

    return projectedPoint;
}

/***************************************************************************/
/*!
 *  The return value is the projection with the inverse homography of the
 *  input point.
 *
 ***************************************************************************/
CVertex2d CHomography::projectInverse( const CVertex2d& point )
{
    CVertex2d projectedPoint;

    try
    {
        ColumnVector m1(3), m2(3);
        m2 << point.x << point.y << 1;

        m1 = (*this).i() * m2;

        projectedPoint.x = m1(1) / m1(3);
        projectedPoint.y = m1(2) / m1(3);
    }
    catch (...)
    {
    }

    return projectedPoint;
}


/***************************************************************************/
/*!
 *  Calculates the distance of a point pair to the homography
 *  based on geometric distances.
 *
 ***************************************************************************/
double CHomography::calculateDistance( const TDataVector::value_type &sample,
                                      unsigned int firstindex,
                                      unsigned int secondindex )
{
    double distance = 0.0;

    CVertex2d distancePoint;

    CVertex2d projectPoint = (*this).project( sample[secondindex] );

    distancePoint = projectPoint - sample[firstindex];
    distance = distancePoint.abs();

    distancePoint = (*this).projectInverse( sample[firstindex] ) - sample[secondindex];
    distance += distancePoint.abs();

    return distance;
}

double CHomography::getCondition()
{
    return m_Condition;
}


/***************************************************************************/
/*!
 *  Set the covariance of the homography as vector with the dimension 9x1.
 *
 ***************************************************************************/
void CHomography::setCovariance( const NEWMAT::Matrix& covariance )
{
    m_Covariance = covariance;
}

/***************************************************************************/
/*!
 *  Get the covariance of the homography as vector with the dimension 9x1.
 *
 ***************************************************************************/
NEWMAT::Matrix CHomography::getCovariance()
{
    return m_Covariance;
}
