/***************************************************************************
                          CHomographyPlanar.cpp  -  description
                             -------------------
    begin                : July 17 2002
    copyright            : (C) 2002 by Marco Grimm, Islam Shdaifat
    email                : grimm@tu-harburg.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CHomographyPlanar.h"
#include "tuv3d/CPointListNormalizer.h"

using namespace NEWMAT;
using namespace std;
using namespace TUVision;

/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3 
 *
 ***************************************************************************/
CHomographyPlanar::CHomographyPlanar()
: CHomography()
{
    //Set the homography to 0
    *this << 1.0 << 0.0 << 0.0
        <<   0.0 << 1.0 << 0.0
        <<   0.0 << 0.0 << 1.0;
}


/***************************************************************************/
/*!
 *  Planar homography from correspondences for the usage with CRansac
 * 
 *  Initializes the matrix-part by 3x3 and calculates its values for
 *  the planar homography between two images induced by a plane 
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 21 ) 
 *
 *         [xi]                [xi]
 *  m2 = s [yi]        = [H] * [yi]         = m1
 *         [1 ]                [1 ]
 *
 *         list.first          list.second
 *         image point 2       image point 1
 *
 ***************************************************************************/
CHomographyPlanar::CHomographyPlanar( const TDataVector &list )
: CHomography()
{
    //Set the homography to 0
    *this << 1.0 << 0.0 << 0.0
        <<   0.0 << 1.0 << 0.0
        <<   0.0 << 0.0 << 1.0;

    intrinsicParametersFirst << 1.0 << 0.0 << 0.0
        << 0.0 << 1.0 << 0.0
        << 0.0 << 0.0 << 1.0;
    intrinsicParametersSecond << 1.0 << 0.0 << 0.0
        << 0.0 << 1.0 << 0.0
        << 0.0 << 0.0 << 1.0;

    if ( list.size() >= minimumdata ) 
    {
        if ( m_bUseNormalizer )
        {
            CPointPairListNormalizer normalizer( list );

            planarHomography( normalizer.getNormalizedlist() );

            Matrix H = normalizer.getfirstsidetransformation().i() * (*this) * normalizer.getsecondsidetransformation();

            *this << H;
        }
        else
        {
            planarHomography( list );
        }

        //normalize the homography to H(3,3) = 1;
        (*this) /= (*this)(3,3);
    }
}


/***************************************************************************/
/*!
 *  Planar homography from correspondences
 * 
 *  Initializes the matrix-part by 3x3 and calculates its values for
 *  the planar homography between two images induced by a plane 
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 21 ) 
 *
 *         [xi]                [xi]
 *  m2 = s [yi]        = [H] * [yi]         = m1
 *         [1 ]                [1 ]
 *
 *         list.first          list.second
 *         image point 2       image point 1
 *
 ***************************************************************************/
CHomographyPlanar::CHomographyPlanar( const TDataVector &list, 
                                     CIntrinsicMatrix Afirst, 
                                     CIntrinsicMatrix Asecond )
                                     : CHomography()
{
    //Set the homography to 0
    *this << 1.0 << 0.0 << 0.0
        <<   0.0 << 1.0 << 0.0
        <<   0.0 << 0.0 << 1.0;

    intrinsicParametersFirst  = Afirst;
    intrinsicParametersSecond = Asecond;

    if ( list.size() >= minimumdata ) 
    {
        if ( m_bUseNormalizer )
        {
            CPointPairListNormalizer normalizer( list );

            planarHomography( normalizer.getNormalizedlist() );

            Matrix H = normalizer.getfirstsidetransformation().i() * (*this) * normalizer.getsecondsidetransformation();

            *this << H;
        }
        else
        {
            planarHomography( list );
        }

        //normalize the homography to H(3,3) = 1;
        (*this) /= (*this)(3,3);
    }
}


/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3 and calculates its values for 
 *  the planar homography induced by a plane. The plane is assumed at the 
 *  world coordinate position of Z=0;
 *  Input values are:
 *  R1 and t1 the transformation from the world coordinate frame to the 
 *    first camera.
 *  Rd and td the transformation from the fist to the second camera.
 *  A1 and A2 the intrinsic parameters of the two cameras.
 *
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 286 ) 
 *
 ***************************************************************************/
CHomographyPlanar::CHomographyPlanar( Matrix R1, 
                                     ColumnVector t1, 
                                     Matrix Rd, 
                                     ColumnVector td, 
                                     CIntrinsicMatrix Afirst, 
                                     CIntrinsicMatrix Asecond )
                                     : CHomography()
{
    intrinsicParametersFirst =  Afirst;
    intrinsicParametersSecond = Asecond;

    // The worldcoordinate frame is located for this homography at the 
    // origin of the first camera, so Rd and td are 
    // the rotation and translation from the first to the second camera.

    // The normal vector of the plane must be calculated due to the 
    // new world coordinate position
    // The plane is assumed to be at Z=0 so n = [0 0 1]'
    ColumnVector n(3);
    n << 0.0 << 0.0 << 1.0;
    n = R1.i() * n;

    // The distance between the first camera and the plane at Z=0
    double distance;
    ColumnVector OriginCamera1;
    OriginCamera1 = -R1 * t1;
    distance = OriginCamera1(3);

    planarAlgorithm( Rd, td, Afirst, Asecond, n, distance );
}


/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3 and calculates its values for 
 *  the planar homography from R,t and the distance to the plane and the 
 *  plane normal in worldcoordinates (n has to be in unit norm ||n||=1). 
 *  The worldcoordinate frame is located for this homography at the 
 *  origin of the first camera, so R and t are 
 *  the rotation and translation from the first to the second camera.
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 286 ) 
 *
 ***************************************************************************/
CHomographyPlanar::CHomographyPlanar( Matrix R, 
                                     ColumnVector t,
                                     CIntrinsicMatrix Afirst, 
                                     CIntrinsicMatrix Asecond, 
                                     ColumnVector n, 
                                     double distance )
                                     : CHomography()
{
    intrinsicParametersFirst =  Afirst;
    intrinsicParametersSecond = Asecond;

    planarAlgorithm( R, t, Afirst, Asecond, n, distance );
}


/***************************************************************************/
/*!
 *  Destructor does nothing as there is no dynamic data
 *
 ***************************************************************************/
CHomographyPlanar::~CHomographyPlanar()
{

}


/***************************************************************************/
/*!
 *  Calculation of the the planar homography from R,t and the distance 
 *  to the plane and the plane normal in worldcoordinates 
 *  (n has to be in unit norm ||n||=1). 
 *  The worldcoordinate frame is located for this homography at the 
 *  origin of the first camera, so R and t are 
 *  the rotation and translation from the first to the second camera.
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 286 ) 
 *
 ***************************************************************************/
void CHomographyPlanar::planarAlgorithm( Matrix R, 
                                        ColumnVector t, 
                                        CIntrinsicMatrix Afirst, 
                                        CIntrinsicMatrix Asecond, 
                                        ColumnVector n, 
                                        double distance )
{
    Matrix H(3,3);

    H = intrinsicParametersSecond * ( R - 1/distance * t * n.t() ) 
        * intrinsicParametersFirst.i();

    // Normalize the homography to get always the solution that the plane 
    // is in front of the camera.
    H /= H(3,3);

    *this << H;
}


/***************************************************************************/
/*!
 *  Computes the planar homography with the correspondences and 
 *  sets the offset of the first and the second point in the correspondences 
 *  list because these values are both image coordinates.
 *
 ***************************************************************************/
void CHomographyPlanar::planarHomography( const TDataVector &list )
{
    calculateHomographyFromCorrespondences( list );
}


/***************************************************************************/
/*!
 *  Calculate the rotation matrix from the homography depending if the 
 *  homography is a plane-to-image or a planar homography
 *
 ***************************************************************************/
Matrix CHomographyPlanar::getR()
{
    Matrix R(3,3);
    R = 0;

    //not implemented yet

    return R;
}

/***************************************************************************/
/*!
 *  Calculate the translation vector from the homography depending if the 
 *  homography is a plane-to-image or a planar homography
 *
 ***************************************************************************/
Matrix CHomographyPlanar::getT()
{
    ColumnVector t(3);
    t = 0;

    //not implemented yet

    return t;
}


/***************************************************************************/
/*!
 *  Set the homography to the input Matrix. The intrinsic parameters and 
 *  the type of the homography are not changed.
 *
 ***************************************************************************/
void CHomographyPlanar::setHomography( Matrix H, CIntrinsicMatrix A1, CIntrinsicMatrix A2 )
{
    *this << H;
    intrinsicParametersFirst = A1;
    intrinsicParametersSecond = A2;
}


/***************************************************************************/
/*!
 *  Set the intrinsic parameters of the first view
 *
 ***************************************************************************/
void CHomographyPlanar::setIntrinsicParameters( CIntrinsicMatrix A1, CIntrinsicMatrix A2 )
{
    intrinsicParametersFirst  = A1;
    intrinsicParametersSecond = A2;
}


/***************************************************************************/
/*!
 *  Get the intrinsic parameters of the first view
 *
 ***************************************************************************/
CIntrinsicMatrix CHomographyPlanar::getIntrinsicParameters( int index )
{
    CIntrinsicMatrix intMatrix = intrinsicParametersFirst;

    if ( 2 == index )
    {
        intMatrix = intrinsicParametersSecond;
    }

    return intMatrix;
}


/***************************************************************************/
/*!
 *  Calculates the distance of a point pair to the homography
 *  based on geometric distances. 
 *
 ***************************************************************************/
double CHomographyPlanar::calculateDistance( const TDataVector::value_type &sample,
                                            unsigned int firstindex,
                                            unsigned int secondindex )
{
    double distance = 0.0;

    CVertex2d distancePoint;

    CVertex2d projectPoint = (*this).project( sample[secondindex] );

    distancePoint = projectPoint - sample[firstindex];
    distance = distancePoint.abs();

    distancePoint = (*this).projectInverse( sample[firstindex] ) - sample[secondindex];
    distance += distancePoint.abs();

    return distance;
}
