/***************************************************************************
                           CImageByteGrayTsaiDistortion.cpp  -  description
                              -------------------
     begin                : 23.01.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel

***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifdef	TUV_USE_OPENGL

#include "CImageByteGrayTsaiDistortion.h"

#ifndef SQRT3
#define SQRT3   1.732050807568877293527446341505872366943	// Undistorted_to_distorted_sensor_coordinates
#endif

#define SQRT(x) sqrt(fabs(x))
#define C_IBGTD_LSB	1.0e-10		// End condition in logarithmic search (Error in triangles)
// Iteration stops when grid<=_TSAI_DIST_LSB
// using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

CImageByteGrayTsaiDistortion::CImageByteGrayTsaiDistortion()
{
    kappa1		= 0.0;		// Default values resulting in no undistortion at all
    Cx			= 0.0;
    Cy			= 0.0;
    sx			= 1.0;
    dpx			= 1.0;
    dpy			= 1.0;
    size		= 1;
    xsteps		= 0;
    ysteps		= 0;
    tex_width	= 0;
    tex_height	= 0;
}

CImageByteGrayTsaiDistortion::~CImageByteGrayTsaiDistortion()
{
}



/**************************************************
 *
 * Get parameters
 *
 **************************************************/

double CImageByteGrayTsaiDistortion::getKappa1()
{
    return (kappa1);
}

double CImageByteGrayTsaiDistortion::getSx()
{
    return (sx);
}

double CImageByteGrayTsaiDistortion::getDpx()
{
    return (dpx);
}

double CImageByteGrayTsaiDistortion::getDpy()
{
    return (dpy);
}

double CImageByteGrayTsaiDistortion::getCenterX()
{
    return (Cx);
}

double CImageByteGrayTsaiDistortion::getCenterY()
{
    return (Cy);
}

int CImageByteGrayTsaiDistortion::getBlockSize()
{
    return (size);
}

/*
 * Get geometric parameters used for texture
 */

int CImageByteGrayTsaiDistortion::getNumberOfCornersX()
{
    return (xsteps);
}

int CImageByteGrayTsaiDistortion::getNumberOfCornersY()
{
    return (ysteps);
}



/**************************************************
 *
 * Set parameters
 *
 **************************************************/

void CImageByteGrayTsaiDistortion::setKappa1(double value)
{
    kappa1=value;
}

void CImageByteGrayTsaiDistortion::setSx(double value)
{
    sx=value;
}

void CImageByteGrayTsaiDistortion::setDpx(double value)
{
    dpx=value;
}

void CImageByteGrayTsaiDistortion::setDpy(double value)
{
    dpy=value;
}

void CImageByteGrayTsaiDistortion::setCenterX(double value)
{
    Cx=value;
}

void CImageByteGrayTsaiDistortion::setCenterY(double value)
{
    Cy=value;
}

bool CImageByteGrayTsaiDistortion::setBlockSize(int value)
{
    if (value>0)
    {
        size = value;
        return true;
    }
    else
        return false;
}


/*
 * Set maximum texture block size that guarantees absolut error < max_error
 * and does not discard parts of the original image in undistortion
 */

bool CImageByteGrayTsaiDistortion::setMaxBlockSize(int width,int height,double max_error)
{
    /*
     * 1. Set texture block size to greatest common divisor of image width and height
     * 2. Reduce blocksize until error bound is fulfilled or blocksize=1
     *    - Uses maxAlignedError 1st, then triangle ErrorOrtho as the later error is normally smaller
     *    - Reduction: Devide by smallest integer divisor and check error bound again
     */
    int  divisor;				// Smallest integer divisor >1 of new_size

    if (max_error<0.0) max_error = -max_error;

    /* 1. Set texture block size to gcd(width,height) */
    size = height;
    for (int a=width,swap_tmp; a>0; a-=size)
    {
        if (a<size)
        {
            swap_tmp = a;
            a		 = size;
            size	 = swap_tmp;
        }
    }

    /* 2. Reduce block size until error bound is fulfilled */
    while ((maxAlignedError(width,height)>max_error) && (size!=1))		// Aligned triangle
    {
        divisor = 2;
        while (size%divisor!=0) divisor++;
        size = size/divisor;
    }
    while ((maxOrthogonalError(width,height)>max_error) && (size!=1))	// Orthogonal triangle
    {
        divisor = 2;
        while (size%divisor!=0) divisor++;
        size = size/divisor;
    }

    if (size==1)
        if ((maxAlignedError(width,height)>max_error)||(maxOrthogonalError(width,height)>max_error))
            return false;
        else
            return true;
    else
        return true;
}



/**************************************************
 *
 * Coordinate transforms:
 * The following methods were taken from TUVisionLib
 * (TsaiCalibrator) and modified slightly.
 *
 **************************************************/

/*
 * Convert from undistorted to distorted image coordinates
 */

void CImageByteGrayTsaiDistortion::undistorted_to_distorted_image_coord
(double Xfu, double Yfu, double *Xfd, double *Yfd)
{
    double    Xu, Yu, Xd, Yd;

    /* convert from image to sensor coordinates */
    Xu = dpx * (Xfu - Cx) / sx;
    Yu = dpy * (Yfu - Cy);

    /* convert from undistorted sensor to distorted sensor plane coordinates */
    undistorted_to_distorted_sensor_coord (Xu, Yu, &Xd, &Yd);

    /* convert from sensor to image coordinates */
     *Xfd = Xd * sx / dpx + Cx;
     *Yfd = Yd / dpy + Cy;
}


/*
 * Convert from distorted to undistorted image coordinates
 */

void  CImageByteGrayTsaiDistortion::distorted_to_undistorted_image_coord
(double Xfd, double Yfd, double *Xfu, double *Yfu)
{
    double Xd, Yd, Xu, Yu;

    /* convert from image to sensor coordinates */
    Xd = dpx * (Xfd - Cx) / sx;
    Yd = dpy * (Yfd - Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    distorted_to_undistorted_sensor_coord (Xd, Yd, &Xu, &Yu);

    /* convert from sensor to image coordinates */
     *Xfu = Xu * sx / dpx + Cx;
     *Yfu = Yu / dpy + Cy;
}


/*
 * Convert from distorted to undistorted sensor plane coordinates
 */

void  CImageByteGrayTsaiDistortion::distorted_to_undistorted_sensor_coord
(double Xd, double Yd, double *Xu, double *Yu)
{
    double    distortion_factor;
    distortion_factor = 1.0 + kappa1 * (Xd*Xd + Yd*Yd);
    *Xu = Xd * distortion_factor;
    *Yu = Yd * distortion_factor;
}


/*
 * Convert from undistorted to distorted sensor plane coordinates
 */

void  CImageByteGrayTsaiDistortion::undistorted_to_distorted_sensor_coord
(double Xu, double Yu, double *Xd, double *Yd)
{
    /*
     * Original comments in TUVisionLib:
     *
     * This routine converts from undistorted to distorted sensor coordinates.
     * The technique involves algebraically solving the cubic polynomial
     * Ru = Rd * (1 + kappa1 * Rd**2) using the Cardan method.
     *
     * Note: for kappa1 < 0 the distorted sensor plane extends out to a maximum
     * barrel distortion radius of  Rd = sqrt (-1/(3 * kappa1)).
     * To see the logic used in this routine try graphing the above polynomial
     * for positive and negative kappa1's
     */

    double    Ru, Rd, lambda, c, d, Q, R, D, S, T;

    if (((Xu == 0) && (Yu == 0)) || (kappa1 == 0))
    {
        *Xd = Xu;
        *Yd = Yu;
        return;
    }
    Ru = sqrt(Xu*Xu+Yu*Yu);
    c = 1.0 / kappa1;
    d = -c * Ru;
    Q = c / 3.0;
    R = -d / 2.0;
    D = pow(Q,3) + pow(R,2);
    if (D >= 0)			/* one real root */
    {
        D = sqrt (D);
        if		(R+D == 0)	S = 0;			// S = CubicRoot(R+D)
        else if	(R+D > 0)	S = pow (R+D, 1.0/3.0);
        else				S = -pow (-(R+D), 1.0/3.0);
        if		(R-D == 0)	T = 0;			// T = CubicRoot(R-D)
        else if	(R-D > 0)	T = pow (R-D, 1.0/3.0);
        else				T = -pow (-(R-D), 1.0/3.0);
        Rd = S + T;
        if (Rd < 0)
        {
            Rd = SQRT (-1.0 / (3.0 * kappa1));
            fprintf (stderr, "\nWarning: undistorted image point to distorted image point mapping limited by\n");
            fprintf (stderr, "         maximum barrel distortion radius of %lf\n", Rd);
            fprintf (stderr, "         (Xu = %lf, Yu = %lf) -> (Xd = %lf, Yd = %lf)\n\n",
                Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
        }
    }
    else				/* three real roots */
    {
        D = sqrt (-D);
        S = hypot (R, D);					// S = CubicRoot(hypot (R, D))
        if		(S == 0)	S = 0;
        else if	(S > 0)		S = pow (S, 1.0/3.0);
        else				S = -pow (-S, 1.0/3.0);
        T = atan2 (D, R) / 3.0;
        Rd = -S * cos(T) + SQRT3 * S * sin(T);	/* use the smaller positive root */
    }
    lambda = Rd / Ru;
    *Xd = Xu * lambda;
    *Yd = Yu * lambda;
}



/**************************************************
 *
 * Undistort image:
 * All methods not using OpenGL
 *
 **************************************************/

/*
 * Full undistortion: Calculate undistorted coordinates, undistort and store image
 */

bool CImageByteGrayTsaiDistortion::undistortImage(CImageByteGray *img)
{
    int		*new_x,*new_y;				// Undistorted coordinates of texture block corners

    setTextureParameters(img);			// Set xsteps, ysteps, tex_width and tex_height
    new_x = new int[xsteps*ysteps];
    new_y = new int[xsteps*ysteps];
    if(!undistortedCoordinates(img,new_x,new_y))	// Calculate coordinates
    {
        delete[] (new_x);
        delete[] (new_y);
        return false;
    }
    textureImage(img,new_x,new_y);		// Undistort image and write result to img
    delete[] (new_x);
    delete[] (new_y);
    return true;
}


/*
 * Set xsteps, ysteps (Size of array holding cordinates of texture block corners),
 * tex_width and tex_height
 */

void CImageByteGrayTsaiDistortion::setTextureParameters(CImageByteGray *img)
{
    /* Texture width (2^m x 2^n pixel) */
    tex_width  = 1;
    while (tex_width<img->getW())	tex_width*=2;
    tex_height = 1;
    while (tex_height<img->getH())	tex_height*=2;
    /* Number of texture block corners fitting in image x- and y-dimension */
    xsteps = (img->getW()/size) +1;
    ysteps = (img->getH()/size) + 1;
    /*
     * Blocks do not cover whole image, but texture dimension big enough
     * to embed a additional block and therefore cover whole image
     */
    if ( ((xsteps-1)*size<img->getW()) && (xsteps*size<=tex_width) )
        xsteps++;
    if ( ((ysteps-1)*size<img->getH()) && (ysteps*size<=tex_height) )
        ysteps++;
}


/*
 * Calculate undistorted coordinates
 */

bool CImageByteGrayTsaiDistortion::undistortedCoordinates
(CImageByteGray *img,int *new_x,int *new_y)
{
    /*
     * Example:
     *
     *  1 2 3 4 5=xsteps
     *  +-+-+-+-+ 1
     *  | | | | |
     *  +-+-+-+-+ 2
     *  | | | | |
     *  +-+-+-+-+ 3=ysteps
     * (| | | | |)<- part of image ommitted, if height no multiple of blocksize
     *                (same in x-direction, if width no multiple of blocksize)
     *
     * Calculation in double precision (Note: x(in pixel)=x*size):
     */
    int		x, y;			// Index: Run throught block corners (for-construct)
    double	xu,yu;			// Undistorted coordinates
    double	xd,yd;			// Distorted coordinates (in original image)

    if ((img->getW()<1)||(img->getH()<1)) return false;

    for (y=0,yd=0; y<ysteps; y++,yd+=size)
    {
        for (x=0,xd=0; x<xsteps; x++,xd+=size)
        {
            distorted_to_undistorted_image_coord(xd,yd,&xu,&yu);
            new_x[x+y*xsteps]=(int)(xu + 0.5);	// + 0.5 => Round to nearest integer
            new_y[x+y*xsteps]=(int)(yu + 0.5);
        }
    }
    return true;
}


/**************************************************
 *
 * Undistort image:
 * All methods using OpenGL
 *
 **************************************************/

/*
 * Init OpenGL and init image as texture
 */

void CImageByteGrayTsaiDistortion::initTexture(CImageByteGray *img,GLuint& texName)
{
    /* Init OpenGL */
    glClearColor (1.0, 1.0, 1.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);			// Bitmap format
    glPixelStorei(GL_PACK_ALIGNMENT, 1);			// Bitmap format

    /* Init Viewport and coordinate system */
    glViewport(0,0,img->getW(),img->getH());
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0,img->getW(),0,img->getH());
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* Init texture */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_2D, texName);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    if ((tex_width==img->getW())&&(tex_height==img->getH()))
    {
        glTexImage2D(GL_TEXTURE_2D, 0, 1, img->getW(), img->getH(), 
            0, GL_LUMINANCE, GL_UNSIGNED_BYTE, img->getRow(0));
    }
    else		// Embed image-file in white (2^n)*(2^m)-image
    {
        unsigned char *white_mat = new unsigned char[tex_width*tex_height];
        for (int i=0; i<tex_width*tex_height; i++)
            white_mat[i]=255;		// unsigned char = 1 byte => 255 = white
        glTexImage2D(GL_TEXTURE_2D, 0, 1,tex_width, tex_height, 
            0, GL_LUMINANCE, GL_UNSIGNED_BYTE, white_mat);
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, img->getW(), img->getH(), 
            GL_LUMINANCE, GL_UNSIGNED_BYTE, img->getRow(0));
        delete[] white_mat;
    }

    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#ifdef GL_VERSION_1_1
    glBindTexture(GL_TEXTURE_2D, texName);
#endif
    glFlush();
}


/*
 * Create undistorted image by texturing in frame buffer:
 * "new"-coordinates are used for polygons
 */

void CImageByteGrayTsaiDistortion::textureImage(CImageByteGray *img,int *new_x,int *new_y,bool isNext)
{
    int		x, y;				// Counter throught block corners (for-construct)
    double	xt,yt;				// Position in texture (original image in pixel)
    double	invW,invH;			// 1/Image dimensions (double for texture mapping to interval [0;1])
    int		yfed0,yfed1;		// = y*ysteps , = (y+1)*ysteps (array)
    invW = 1.0/tex_width;
    invH = 1.0/tex_height;

    /* Init texture */
    if (isNext)					// Parameters not changed => Simply overwrite texture image
    {
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, img->getW(), img->getH(), 
            GL_LUMINANCE, GL_UNSIGNED_BYTE, img->getRow(0));
    }
    else						// Parameters changed => Create new texture image
    {
#ifdef GL_VERSION_1_1
        static GLuint texName=0;
#endif
        initTexture(img,texName);
        glEnable(GL_TEXTURE_2D);
    }

    /* Create undistorted image in frame buffer */
    glBegin(GL_QUADS);
    for(y=0,yt=0; y<ysteps-1; y++,yt+=size)
    {
        yfed0	=  y*xsteps;			// Indices count added by rows (current position)
        yfed1	=  (y+1)*xsteps;		// (next coordinate in y-direction)
        for(x=0,xt=0; x<xsteps-1; x++,xt+=size)
        {
            glTexCoord2f((xt)     *invW,(yt)     *invH); glVertex2i(new_x[x  +yfed0],new_y[x  +yfed0]);
            glTexCoord2f((xt+size)*invW,(yt)     *invH); glVertex2i(new_x[x+1+yfed0],new_y[x+1+yfed0]);
            glTexCoord2f((xt+size)*invW,(yt+size)*invH); glVertex2i(new_x[x+1+yfed1],new_y[x+1+yfed1]);
            glTexCoord2f((xt)     *invW,(yt+size)*invH); glVertex2i(new_x[x  +yfed1],new_y[x  +yfed1]);
        }
    }
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glFlush();

    /* Store frame buffer image to img */
    /*
     * Note (compare [1], p. 313, chap. 8, Pixel transfer operations):
     * OpenGL calculates GL_LUMINANCE as 1.0*Red + 1.0*Green + 1.0*Blue instead of
     * e.g. 0.3*Red + 0.59*Green + 0.11*Blue in NTSC standard. Therefore any of
     * GL_RED, GL_GREEN or GL_BLUE corresponds to the true luminance value.
     */
    glReadPixels(0,0,img->getW(),img->getH(),GL_RED,GL_UNSIGNED_BYTE,img->getRow(0));
    glClear(GL_COLOR_BUFFER_BIT);
}




/**************************************************
 *
 * Error Calculations: Common method
 * (Calculate relevant point P furthest from center of distortion)
 *
 **************************************************/

void CImageByteGrayTsaiDistortion::maxCorner(int width,int height,double &xp,double &yp)
{
    double dx,dy;				// Maximal distance in x- and y-direction (no absolute value!)
    double rad,rad_tmp;			// Distances: Radius^2

    /*
     * (1) Find corner with maximum distance to center of distortion
     */
    rad		= Cx*Cx+Cy*Cy;										// Corner (0/0)
    dx		= 0.0-Cx;
    dy		= 0.0-Cy;
    rad_tmp	= (Cx*Cx+(height-Cy)*(height-Cy));					// Corner (0/height)
    if (rad_tmp > rad)
    {
        rad	= rad_tmp;
        dx	= 0.0-Cx;
        dy	= height-Cy;
    }
    rad_tmp	= ((width-Cx)*(width-Cx)+(height-Cy)*(height-Cy));	// Corner (width/height)
    if (rad_tmp > rad)
    {
        rad	= rad_tmp;
        dx	= width-Cx;
        dy	= height-Cy;
    }
    rad_tmp	= ((width-Cx)*(width-Cx)+Cy*Cy);					// Corner (width/0)
    if (rad_tmp > rad)
    {
        rad	= rad_tmp;
        dx	= width-Cx;
        dy	= 0.0-Cy;
    }

    /*
     * (2) kappa1<0: Contraction => full image displayed after undistortion
     *     (Nothing to do as dx and dy already describe corner with max. distance to center.)
     */

    /*
     * (3) kappa1>0: Only part of the undistorted image displayed
     *     (a) Calculate point (xd/yd) that will be mapped on furthest corner
     *		   when compensating distortion.
     *	   (b) Use furthest corner in the texture block that includes (xu/yu)
     */
    if (kappa1>0)
    {
        double	xu,yu;		// Coordinates of max. corner after distortion is compensated
        double	xd,yd;		// Coordinates of max. corner in camera image (distorted by lens)
        xu = dx + Cx;		// True coordinates (NOT relative to center of distortion)
        yu = dy + Cy;
        undistorted_to_distorted_image_coord(xu,yu,&xd,&yd);

        /* Furthest corner of texture block including (xd,yd) */
        if ((int)(xd+0.5)%size != 0)	// True x-coordinate in image not on block border
        {
            if (dx>0)				// Positive direction increases distance to center
                dx = (double)(((int)xd/size) * size + size - Cx);
            else					// Negative direction increases distance to center
                dx = (double)(((int)xd/size) * size - size - Cx);
        }
        else
            dx = (double)((int)(xd+0.5)) - Cx;	// 0.5: Round, Cx transform relative to center
        if ((int)(yd+0.5)%size != 0)	// True y-coordinate in image not on block border
        {
            if (dy>0)				// Positive direction increases distance to center
                dy = (double)(((int)yd/size) * size + size - Cy);
            else					// Negative direction increases distance to center
                dy = (double)(((int)yd/size) * size - size - Cy);
        }
        else
            dy = (double)((int)(yd+0.5)) - Cy;	// 0.5: Round, Cy transform relative to center
    }

    /* Set coordinates: Abstract point P (furthest from center of distortion) */
    if (dx<0)	dx = -dx;	// Absolute values of relevant corner with maximum distance
    if (dy<0)	dy = -dy;

    if (dx>dy)				// Abstract coordinates of P are defined as yp >= xp
    {
        xp = dy;
        yp = dx;
    }
    else
    {
        xp = dx;
        yp = dy;
    }
}


/**************************************************
 *
 * Upper bound for maximum absolute interpolation error:
 *
 **************************************************/

/*
 * Calculate maximum distance between textured and calculated coordinate
 * of all other error methods
 */

double CImageByteGrayTsaiDistortion::maxAbsoluteError(int width,int height)
{
    if (kappa1==0) return(0.0);		// No undistortion taking place

    double aligned,orthogonal;
    aligned		= maxAlignedError	(width,height);
    orthogonal	= maxOrthogonalError(width,height);
    return ((aligned>orthogonal)?aligned:orthogonal);
}



/*
 * Aligned triangle that is furthest from center of distortion
 * (Calculation in euclidian coordinates)
 */

double CImageByteGrayTsaiDistortion::maxAlignedError
(int width,int height,double *lambda_ret,double *mu_ret)
{
    /*
     *   y ^
     *     | Q       v
     *  yp +-+---+P  -
     * =yq | |  /.   |
     *     | |/  .   | size
     *     | +...+   -
     *     | T   |   ^
     *     |     |
     *     |     |
     *     | |<->|
     *     | size|
     *    -+-----+----> x
     *           xp
     *
     * Note: yp defined as: yp>= xp
     *
     * Error function (in frame/image coordinates):
     *
     * e(lambda,mu) = su - qu + lambda*(qu-pu) + mu*(qu-tu)
     *
     * || e(lambda,mu) || = sqrt[ ex(lambda,mu)^2 + ey(lambda,mu)^2 ]
     *
     * where 0 <= lambda,mu <= 1 and lambda + mu <= 1
     *
     * Maximum is calculated by logarithmic search (repeated evaluation in decreasing grid)
     */

    int i,j;							// "for"-counters

    double pd[2],qd[2],td[2],sd[2];		// Distorted frame/image coordinates (index 0=x, 1=y)
    double pu[2],qu[2],tu[2],su[2];		// Undistorted frame/image coordinates
    double qu_minus_pu[2];				// Term often used => efficient in extra variable
    double qu_minus_tu[2];

    double abs_err_2,lambda_tmp,mu_tmp;	// Error^2 at point S and parameters of S
    double e[2];						// Error in x- and in y-direction at S
    double delta;						// Grid size in logarithmic search
    double lambda,mu;					// Parameters of point S with maximum error
    double max_err_2;					// Maximum error at S with parameters lambda and mu

    /* Calculate frame buffer (image) coordinates of edge points */
    maxCorner(width,height,pd[0],pd[1]);// Calculate coordinates of furthest relevant corner point P
    qd[0] = pd[0] - size;
    qd[1] = pd[1];
    td[0] = pd[0] - size;
    td[1] = pd[1] - size;

    /* Calculate coordinates of undistorted edge points */
    distorted_to_undistorted_image_coord(pd[0],pd[1],&pu[0],&pu[1]);
    distorted_to_undistorted_image_coord(qd[0],qd[1],&qu[0],&qu[1]);
    distorted_to_undistorted_image_coord(td[0],td[1],&tu[0],&tu[1]);

    qu_minus_pu[0] = qu[0] - pu[0];
    qu_minus_pu[1] = qu[1] - pu[1];
    qu_minus_tu[0] = qu[0] - tu[0];
    qu_minus_tu[1] = qu[1] - tu[1];

    /* Find start values for parameters lambda (indices 2*i) and mu (indices 2*i+1):
     * Evaluate 0.25-grid except for edge points (error = 0.0 on P,Q and T) */
    double par[2*12] = {0.0, 0.25,		0.0, 0.5,		0.0,0.75,
        0.25,0.0,		0.25,0.25,		0.25,0.5,		0.25,0.75,
        0.5,0.0,		0.5,0.25,		0.5,0.5,
        0.75,0.0,		0.75,0.25};
    max_err_2 = 0.0;				// Initialize maximum error
    for (i=0; i<12; i++)
    {
        /* Parameters of currently examined point S */
        lambda_tmp	= par[2*i];
        mu_tmp		= par[2*i+1];
        sd[0]		= qd[0] + size * lambda_tmp;
        sd[1]		= qd[1] - size * mu_tmp;
        distorted_to_undistorted_image_coord(sd[0],sd[1],&su[0],&su[1]);
        e[0]		= su[0] - qu[0]+lambda_tmp*(qu_minus_pu[0])+mu_tmp*(qu_minus_tu[0]);
        e[1]		= su[1] - qu[1]+lambda_tmp*(qu_minus_pu[1])+mu_tmp*(qu_minus_tu[1]);
        abs_err_2	= e[0] * e[0] + e[1] * e[1];
        /* Error at S max. so far */
        if (abs_err_2>max_err_2)
        {
            max_err_2	= abs_err_2;
            lambda		= lambda_tmp;
            mu			= mu_tmp;
        }
    }

    /* Logarithmic search for maximum error */
    for (delta=0.125; delta>=C_IBGTD_LSB; delta/=2)
    {
        for (i=-1; i<=1; i++)		// Check nine neighbouring grid points
        {
            for (j=-1; j<=1; j++)
            {
                /* Parameters of currently examined point S */
                lambda_tmp	= i * delta + lambda;
                mu_tmp		= j * delta + mu;
                /* Evaluate onley if S in triangle */
                if ((lambda_tmp>=0)&&(mu_tmp>=0)&&(lambda_tmp+mu_tmp<=1))
                {
                    sd[0]		= qd[0] + size * lambda_tmp;
                    sd[1]		= qd[1] - size * mu_tmp;
                    distorted_to_undistorted_image_coord(sd[0],sd[1],&su[0],&su[1]);
                    e[0]		= su[0] - qu[0]+lambda_tmp*(qu_minus_pu[0])+mu_tmp*(qu_minus_tu[0]);
                    e[1]		= su[1] - qu[1]+lambda_tmp*(qu_minus_pu[1])+mu_tmp*(qu_minus_tu[1]);
                    abs_err_2	= e[0] * e[0] + e[1] * e[1];
                    /* Error at S max. so far */
                    if (abs_err_2>max_err_2)
                    {
                        max_err_2	= abs_err_2;
                        lambda		= lambda_tmp;
                        mu			= mu_tmp;
                    }
                }
            }
        }  // Check nine neighbours
    }  // Logarithmic search

    /* Return results */
    if (lambda_ret != NULL)		*lambda_ret = lambda;
    if (mu_ret != NULL)			*mu_ret = mu;
    return (sqrt(max_err_2) + 0.70710678118655);	// + sqrt(0.5) for worst case rounding of textured coordinate
}													// (1/2 diagonal of 1-pixel block directed directly from true undistorted value)


/*
 * Orthogonal triangle that is furthest from center of distortion
 * (Calculation in euclidian coordinates)
 */

double CImageByteGrayTsaiDistortion::maxOrthogonalError
(int width,int height,double *lambda_ret,double *mu_ret)
{
    /*
     *   y ^
     *     | Q       v
     *  yp +-+---+P  -
     * =yq | .\  |   |
     *     | .  \|   | size
     *     | ....+T   -
     *     |     |   ^
     *     |     |
     *     |     |
     *     | |<->|
     *     | size|
     *    -+-----+----> x
     *           xp
     *
     * Note: yp defined as: yp>= xp
     *
     * Error function:
     *
     * e(lambda,mu) = su - pu + lambda*(pu-qu) + mu*(pu-tu)
     *
     * || e(lambda,mu) || = sqrt[ ex(lambda,mu)^2 + ey(lambda,mu)^2 ]
     *
     * where 0 <= lambda,mu <= 1 and lambda + mu <= 1
     *
     * Maximum is calculated by logarithmic search (repeated evaluation in decreasing grid)
     */

    int i,j;							// "for"-counters

    double pd[2],qd[2],td[2],sd[2];		// Distorted frame/image coordinates (index 0=x, 1=y)
    double pu[2],qu[2],tu[2],su[2];		// Undistorted frame/image coordinates
    double pu_minus_qu[2];				// Term often used => efficient in extra variable
    double pu_minus_tu[2];

    double abs_err_2,lambda_tmp,mu_tmp;	// Error^2 at point S and parameters of S
    double e[2];						// Error in x- and in y-direction at S
    double delta;						// Grid size in logarithmic search
    double lambda,mu;					// Parameters of point S with maximum error
    double max_err_2;					// Maximum error at S with parameters lambda and mu

    /* Calculate frame buffer (image) coordinates of edge points */
    maxCorner(width,height,pd[0],pd[1]);// Calculate coordinates of furthest relevant corner point P
    qd[0] = pd[0] - size;
    qd[1] = pd[1];
    td[0] = pd[0];
    td[1] = pd[1] - size;

    /* Calculate coordinates of undistorted edge points */
    distorted_to_undistorted_image_coord(pd[0],pd[1],&pu[0],&pu[1]);
    distorted_to_undistorted_image_coord(qd[0],qd[1],&qu[0],&qu[1]);
    distorted_to_undistorted_image_coord(td[0],td[1],&tu[0],&tu[1]);

    pu_minus_qu[0] = pu[0] - qu[0];
    pu_minus_qu[1] = pu[1] - qu[1];
    pu_minus_tu[0] = pu[0] - tu[0];
    pu_minus_tu[1] = pu[1] - tu[1];

    /* Find start values for parameters lambda (indices 2*i) and mu (indices 2*i+1):
     * Evaluate 0.25-grid except for edge points (error = 0.0 on P,Q and T) */
    double par[2*12] = {0.0, 0.25,		0.0, 0.5,		0.0,0.75,
        0.25,0.0,		0.25,0.25,		0.25,0.5,		0.25,0.75,
        0.5,0.0,		0.5,0.25,		0.5,0.5,
        0.75,0.0,		0.75,0.25};
    max_err_2 = 0.0;				// Initialize maximum error
    for (i=0; i<12; i++)
    {
        /* Parameters of currently examined point S */
        lambda_tmp	= par[2*i];
        mu_tmp		= par[2*i+1];
        sd[0]		= pd[0] - size * lambda_tmp;
        sd[1]		= pd[1] - size * mu_tmp;
        distorted_to_undistorted_image_coord(sd[0],sd[1],&su[0],&su[1]);
        e[0]		= su[0] - pu[0]+lambda_tmp*(pu_minus_qu[0])+mu_tmp*(pu_minus_tu[0]);
        e[1]		= su[1] - pu[1]+lambda_tmp*(pu_minus_qu[1])+mu_tmp*(pu_minus_tu[1]);
        abs_err_2	= e[0] * e[0] + e[1] * e[1];
        /* Error at S max. so far */
        if (abs_err_2>max_err_2)
        {
            max_err_2	= abs_err_2;
            lambda		= lambda_tmp;
            mu			= mu_tmp;
        }
    }

    /* Logarithmic search for maximum error */
    for (delta=0.125; delta>=C_IBGTD_LSB; delta/=2)
    {
        for (i=-1; i<=1; i++)		// Check nine neighbouring grid points
        {
            for (j=-1; j<=1; j++)
            {
                /* Parameters of currently examined point S */
                lambda_tmp	= i * delta + lambda;
                mu_tmp		= j * delta + mu;
                /* Evaluate onley if S in triangle */
                if ((lambda_tmp>=0)&&(mu_tmp>=0)&&(lambda_tmp+mu_tmp<=1))
                {
                    sd[0]		= pd[0] - size * lambda_tmp;
                    sd[1]		= pd[1] - size * mu_tmp;
                    distorted_to_undistorted_image_coord(sd[0],sd[1],&su[0],&su[1]);
                    e[0]		= su[0] - pu[0]+lambda_tmp*(pu_minus_qu[0])+mu_tmp*(pu_minus_tu[0]);
                    e[1]		= su[1] - pu[1]+lambda_tmp*(pu_minus_qu[1])+mu_tmp*(pu_minus_tu[1]);
                    abs_err_2	= e[0] * e[0] + e[1] * e[1];
                    /* Error at S max. so far */
                    if (abs_err_2>max_err_2)
                    {
                        max_err_2	= abs_err_2;
                        lambda		= lambda_tmp;
                        mu			= mu_tmp;
                    }
                }
            }
        }  // Check nine neighbours
    }  // Logarithmic search

    /* Return results */
    if (lambda_ret != NULL)		*lambda_ret = lambda;
    if (mu_ret != NULL)			*mu_ret = mu;
    return (sqrt(max_err_2) + 0.70710678118655);	// + sqrt(0.5) for worst case rounding of textured coordinate
}													// (1/2 diagonal of 1-pixel block directed directly from true undistorted value)

#endif // TUV_USE_OPENGL
