/***************************************************************************
                           CTsaiDistortion.cpp  -  description
                              -------------------
     begin                : 05.09.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Marc Hensel

***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CTsaiDistortion.h"

#ifndef SQRT3
#define SQRT3   1.732050807568877293527446341505872366943	// Undistorted_to_distorted_sensor_coordinates
#endif

#define SQRT(x) sqrt(fabs(x))

using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

CTsaiDistortion::CTsaiDistortion()
{
    kappa1		= 0.0;		// Default values resulting in no undistortion at all
    Cx			= 0.0;
    Cy			= 0.0;
    sx			= 1.0;
    dpx			= 1.0;
    dpy			= 1.0;
    streamOpen	= false;	// Default mapping values
    width		= 0;
    height		= 0;
}

CTsaiDistortion::~CTsaiDistortion()
{
    closeStream();
}



/**************************************************
 *
 * Get parameters
 *
 **************************************************/

double CTsaiDistortion::getKappa1()
{ return (kappa1); }

double CTsaiDistortion::getSx()
{ return (sx); }

double CTsaiDistortion::getDpx()
{ return (dpx); }

double CTsaiDistortion::getDpy()
{ return (dpy); }

double CTsaiDistortion::getCenterX()
{ return (Cx); }

double CTsaiDistortion::getCenterY()
{ return (Cy); }



/**************************************************
 *
 * Set parameters
 *
 **************************************************/

void CTsaiDistortion::setKappa1(double value)
{ kappa1=value; }

void CTsaiDistortion::setSx(double value)
{ sx=value; }

void CTsaiDistortion::setDpx(double value)
{ dpx=value; }

void CTsaiDistortion::setDpy(double value)
{ dpy=value; }

void CTsaiDistortion::setCenterX(double value)
{ Cx=value; }

void CTsaiDistortion::setCenterY(double value)
{ Cy=value; }




/**************************************************
 *
 * Coordinate transforms:
 * The following methods were taken from TUVisionLib
 * (TsaiCalibrator) and modified slightly.
 *
 **************************************************/

/*
 * Convert from undistorted to distorted image coordinates
 */

void CTsaiDistortion::undistorted_to_distorted_image_coord
(double Xfu, double Yfu, double *Xfd, double *Yfd)
{
    double    Xu, Yu, Xd, Yd;
    /* convert from image to sensor coordinates */
    Xu = dpx * (Xfu - Cx) / sx;
    Yu = dpy * (Yfu - Cy);
    /* convert from undistorted sensor to distorted sensor plane coordinates */
    undistorted_to_distorted_sensor_coord (Xu, Yu, &Xd, &Yd);
    /* convert from sensor to image coordinates */
     *Xfd = Xd * sx / dpx + Cx;
     *Yfd = Yd / dpy + Cy;
}


/*
 * Convert from distorted to undistorted image coordinates
 */

void  CTsaiDistortion::distorted_to_undistorted_image_coord
(double Xfd, double Yfd, double *Xfu, double *Yfu)
{
    double Xd, Yd, Xu, Yu;
    /* convert from image to sensor coordinates */
    Xd = dpx * (Xfd - Cx) / sx;
    Yd = dpy * (Yfd - Cy);
    /* convert from distorted sensor to undistorted sensor plane coordinates */
    distorted_to_undistorted_sensor_coord (Xd, Yd, &Xu, &Yu);
    /* convert from sensor to image coordinates */
     *Xfu = Xu * sx / dpx + Cx;
     *Yfu = Yu / dpy + Cy;
}


/*
 * Convert from distorted to undistorted sensor plane coordinates
 */

void  CTsaiDistortion::distorted_to_undistorted_sensor_coord
(double Xd, double Yd, double *Xu, double *Yu)
{
    double    distortion_factor;
    distortion_factor = 1.0 + kappa1 * (Xd*Xd + Yd*Yd);
    *Xu = Xd * distortion_factor;
    *Yu = Yd * distortion_factor;
}


/*
 * Convert from undistorted to distorted sensor plane coordinates
 */

void  CTsaiDistortion::undistorted_to_distorted_sensor_coord
(double Xu, double Yu, double *Xd, double *Yd)
{
    /*
     * Original comments in TUVisionLib:
     *
     * This routine converts from undistorted to distorted sensor coordinates.
     * The technique involves algebraically solving the cubic polynomial
     * Ru = Rd * (1 + kappa1 * Rd**2) using the Cardan method.
     *
     * Note: for kappa1 < 0 the distorted sensor plane extends out to a maximum
     * barrel distortion radius of  Rd = sqrt (-1/(3 * kappa1)).
     * To see the logic used in this routine try graphing the above polynomial
     * for positive and negative kappa1's
     */

    double    Ru, Rd, lambda, c, d, Q, R, D, S, T;

    if (((Xu == 0) && (Yu == 0)) || (kappa1 == 0))
    {
        *Xd = Xu;
        *Yd = Yu;
        return;
    }
    Ru = sqrt(Xu*Xu+Yu*Yu);
    c = 1.0 / kappa1;
    d = -c * Ru;
    Q = c / 3.0;
    R = -d / 2.0;
    D = pow(Q,3) + pow(R,2);
    if (D >= 0)			/* one real root */
    {
        D = sqrt (D);
        if		(R+D == 0)	S = 0;			// S = CubicRoot(R+D)
        else if	(R+D > 0)	S = pow (R+D, 1.0/3.0);
        else				S = -pow (-(R+D), 1.0/3.0);
        if		(R-D == 0)	T = 0;			// T = CubicRoot(R-D)
        else if	(R-D > 0)	T = pow (R-D, 1.0/3.0);
        else				T = -pow (-(R-D), 1.0/3.0);
        Rd = S + T;
        if (Rd < 0)
        {
            Rd = SQRT (-1.0 / (3.0 * kappa1));
            fprintf (stderr, "\nWarning: undistorted image point to distorted image point mapping limited by\n");
            fprintf (stderr, "         maximum barrel distortion radius of %f\n", Rd);
            fprintf (stderr, "         (Xu = %f, Yu = %f) -> (Xd = %f, Yd = %f)\n\n",
                Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
        }
    }
    else				/* three real roots */
    {
        D = sqrt (-D);
        S = hypot (R, D);					// S = CubicRoot(hypot (R, D))
        if		(S == 0)	S = 0;
        else if	(S > 0)		S = pow (S, 1.0/3.0);
        else				S = -pow (-S, 1.0/3.0);
        T = atan2 (D, R) / 3.0;
        Rd = -S * cos(T) + SQRT3 * S * sin(T);	/* use the smaller positive root */
    }
    lambda = Rd / Ru;
    *Xd = Xu * lambda;
    *Yd = Yu * lambda;
}



/**************************************************
 *
 * Undistort image:
 *
 **************************************************/

/*
 * Full undistortion: Undistort and store image
 */

double CTsaiDistortion::undistortImage(CImageByteGray &img)
{
    double	max_error;					// Maximum interpolation error [pixel]
    max_error = openStream(img.getW(),img.getH());
    undistort(img);
    closeStream();
    return max_error;
}

double CTsaiDistortion::undistortImage(CImageByteRGB &img)
{
    double	max_error;					// Maximum interpolation error [pixel]
    max_error = openStream(img.getW(),img.getH());
    undistort(img);
    closeStream();
    return max_error;
}


/*
 * Open stream: Calculate coordinate mapping
 */

double CTsaiDistortion::openStream(int iwidth,int iheight)
{
    int		xu,yu;				// Undistorted image coordinates (exact pixel position in new image)
    double	xd,yd;				// - Corresponding distorted image coordinate
    double	xnd[4],ynd[4];		// 4 Corners of neighbours (2x2 pixel) containing distortion of (xu,yu)
    double	xnu[4],ynu[4];		// - 4 corners in undistorted image
    double	max_error;			// (Maximum interpolation error in image)^2 [pixel^2]
    double	error[4];			// (Interpolation error to 4 corner points)^2 [pixel^2]
    int		i;					// Pointer to corner point arrays
    int		nearest_pointer;	// Nearest neighbor: Pointer to corner point with minimum error (nearest neighbor)
    int		xd_final,yd_final;	// - Result: Distorted image coordinates of nearest neighbor in UNDISTORTED image
    int		ptr;				// Pointer in mapping arrays (x+y*width)

    /* Set image dimensions */
    width	= iwidth;
    height	= iheight;

    /* Close last stream if neccessary (free mapping structures) */
    if (streamOpen==1)
    {
        delete[] (map_x);
        delete[] (map_y);
    }

    /* Open stream: Create mapping undistorted -> distorted coordinates */
    max_error	= 0.0;
    map_x		= new int[width * height];
    map_y		= new int[width * height];
    ptr			= 0;
    for (yu=0; yu<height; yu++)
    {
        for (xu=0; xu<width; xu++,ptr++)
        {
            nearest_pointer = 0;
            /* Exact distortion of undistorted coordinates */
            undistorted_to_distorted_image_coord(xu,yu,&xd,&yd);
            /* 4 nearest neighbors of exact distortion of undistorted coordinates */
            xnd[0] = (xd>=0)?(double)((int)(xd)):(double)((int)(xd-1.0));	// Round down (upper left corner)
            ynd[0] = (yd>=0)?(double)((int)(yd)):(double)((int)(yd-1.0));	// ("-1.0" neccessary for kappa1<0!!!)
            xnd[1] = xnd[0] + 1.0;	ynd[1] = ynd[0];		// Upper right
            xnd[2] = xnd[0];		ynd[2] = ynd[0] + 1.0;	// Lower left
            xnd[3] = xnd[0] + 1.0;	ynd[3] = ynd[0] + 1.0;	// lower right
            /* Distance (error) of undistorted coordinate to undistorted corners */
            for (i=0; i<4; i++)
            {
                distorted_to_undistorted_image_coord(xnd[i],ynd[i],&xnu[i],&ynu[i]);
                error[i] = (xu-xnu[i])*(xu-xnu[i])+(yu-ynu[i])*(yu-ynu[i]);
                if (error[i]<error[nearest_pointer]) nearest_pointer = i;
            }
            /* Coordinates of nearest neighbor */
            xd_final = (int)(xnd[nearest_pointer]);
            yd_final = (int)(ynd[nearest_pointer]);
            /* Coordinate of nearest neighbour are valid, i.e. in the image */
            if ( (xd_final>=0)&&(xd_final<width) && (yd_final>=0)&&(yd_final<height) )
            {
                map_x[ptr] = xd_final;
                map_y[ptr] = yd_final;
                if (error[nearest_pointer]>max_error) max_error = error[nearest_pointer];
            }
            /* Coordinate of nearest neighbour are NOT valid, i.e. in the image (Note: Kappa1<0 = contraction) */
            else
                map_x[ptr] = -1;
        }
    }
    streamOpen = true;
    return (sqrt(max_error));
}


/*
 * Stream: Undistort image
 */

bool CTsaiDistortion::undistort(CImageByteGray &img)
{
    if (streamOpen==0) openStream(img.getW(),img.getH());

    CImageByteGray	*orig_img;			// Original image
    int		xu,yu;						// Undistorted image coordinates (new image)
    int		ptr;						// Pointer in mapping arrays (x+y*width)

    /* Create destination (undistorted) image data structure */
    orig_img = new CImageByteGray (img);	// Distorted image
    img.clearImage();						// Undistorted image

    /* Undistort image */
    if (kappa1>=0)
    {
        ptr = 0;
        for (yu=0; yu<height; yu++)
            for (xu=0; xu<width; xu++,ptr++)
                img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],0),0);
    }
    else
    {
        ptr = 0;
        for (yu=0; yu<height; yu++)
            for (xu=0; xu<width; xu++,ptr++)
                if (map_x[ptr]>=0)				// Note: "map_x[ptr]=-1" if data not valid
                    img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],0),0);
                else
                    img.setPixel(xu,yu,0,0);	// Contraction: No corresponding original image coordinate
    }

    delete orig_img;
    return true;
}

bool CTsaiDistortion::undistort(CImageByteRGB &img)
{
    if (streamOpen==0) openStream(img.getW(),img.getH());

    CImageByteRGB	*orig_img;			// Original image
    int		xu,yu;						// Undistorted image coordinates (new image)
    int		ptr;						// Pointer in mapping arrays (x+y*width)

    /* Create destination (undistorted) image data structure */
    orig_img = new CImageByteRGB (img);	// Distorted image
    img.clearImage();					// Undistorted image

    /* Undistort image */
    if (kappa1>=0)
    {
        ptr = 0;
        for (yu=0; yu<height; yu++)
        {
            for (xu=0; xu<width; xu++,ptr++)
            {
                img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],0),0);
                img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],1),1);
                img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],2),2);
            }
        }
    }
    else
    {
        ptr = 0;
        for (yu=0; yu<height; yu++)
        {
            for (xu=0; xu<width; xu++,ptr++)
            {
                if (map_x[ptr]>=0)		// Note: "map_x[ptr]=-1" if data not valid
                {
                    img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],0),0);
                    img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],1),1);
                    img.setPixel(xu,yu,orig_img->getPixel(map_x[ptr],map_y[ptr],2),2);
                }
                else
                {
                    img.setPixel(xu,yu,0,0);	// Contraction: No corresponding original image coordinate
                    img.setPixel(xu,yu,0,1);
                    img.setPixel(xu,yu,0,2);
                }
            }
        }
    }

    delete orig_img;
    return true;
}



/*
 * Close stream: Free mapping arrays
 */

bool CTsaiDistortion::closeStream()
{
    if (streamOpen==1)
    {
        delete[] (map_x);
        delete[] (map_y);
    }
    streamOpen=0;
    return true;
}



//-------------------------------------------------------------------------------------------

/**********************************************************
 *
 * Semantic documentation ("Related page"):
 *
 **********************************************************/

/*! \page ctsaidistortionsemantic CTsaiDistortion
 *     -  \ref ctsaidistortionsyntactic "Syntactic documentation"
 *			
 *
 * 		\section ctsaidistortionexpic Example picture
 *			
 *			\image html CTsaiDistortion.jpg "(a) Distorted. (b) Undistorted (compare left and right pillars)."
 *			\image latex CTsaiDistortion.jpg "(a) Distorted. (b) Undistorted (compare left and right pillars)."
 *
 *		\section ctsaidistortionbe Brief explanation
 *			Class to compensate radial distortion using the Tsai camera model and nearest neighbor interpolation.
 *
 *		\section ctsaidistortionde Detailled explanation
 *
 *		  \subsection ctsaidistortiontsaimodel Tsai camera model
 *	 	    The mapping of a real 3d-scenery to a 2d-image is described by camera models.
 *	 	    The pinhole camera corresponds to the ideal mapping by central projection. Physical
 *	 	    cameras, however, are subject to a large variety of effects, e.g. radial
 *	 	    nonlinear distortion due to curved lenses. Tsai's camera model extends the
 *	 	    pinhole camera by modelling radial distortion and hardware factors as e.g.
 *	 	    synchronization inaccuracies.
 *
 *		  \subsection ctsaidistortiontsaiparameters Parameters
 *          - \f$ \kappa_1 \f$: Distortion coefficient [1/mm^2]
 *          - \f$ C_x \f$: Optical camera centrum (i.e. centrum of distortion) in x-image-coordinate [pixel]
 *          - \f$ C_y \f$: Optical camera centrum (i.e. centrum of distortion) in y-image-coordinate [pixel]
 *          - \f$ d_x \f$: Horizontal distance between adjacent CCD cells [mm]
 *          - \f$ d_y \f$: Vertical distance between adjacent CCD cells [mm]
 *          - \f$ s_x \f$: Uncertainty factor for scaling the x-coordinate
 *          - \f$ N_{cx} \f$: Number of sensor elements in x-direction
 *          - \f$ N_{fx} \f$: Number of pixel in x-direction
 *
 *          The camera parameters and calibration constants must be calculated using appropriate
 *          calibration tools/algorithms (e.g. TUCal in the TUHH Vision Systems department).
 *
 *		  \subsection ctsaidistortionradialdistortion Modelling radial distortion
 *	 	    Radial distortion can be modelled exactly by
 *
 *	 	    \f[
 *	 	        {\bf x}_u = \left ( 1 + \sum \limits_{i=1}^{\infty}{\kappa_i \cdot r^{2i}} \right ) {\bf x}_d \; ,
 *	 	    \f]
 *
 *	 	    with undistorted centered image coordinates \f$ {\bf x}_u \f$, distorted centered image coordinates
 *          \f$ {\bf x}_d \f$, distortion coefficients \f$ \kappa_i \f$, and the euclidean distance
 *          to the center of distortion (i.e. the coordinate system origin)
 *
 *          \f[
 *            r = r({\bf x}_d) = \left \| {\bf x}_d \right \|_2 = \sqrt{x_d^2 + y_d^2} \; .
 *          \f]
 * 
 *          The term "centered image coordinates" stresses the fact that the coordinates refer to a
 *          coordinate system whose origin matches with the center of distortion.
 *
 *          Tsai's work has shown that \f$ \kappa_1 \f$ takes into account all substantially effects
 *          and the usage of further coefficients might lead to numerical instability. Therefore, radial
 *          distortion is modelled:
 *
 *	 	    \f[
 *	 	        {\bf x}_u = \left ( 1 + \kappa_1 \cdot r^2 \right ) {\bf x}_d
 *	 	    \f]
 *
 *	 	  \subsection ctsaidistortionundistortion Compensation of radial distortion
 *          The compensation of radial distortion takes place in two steps:
 *          -# Calculate the exact undistorted coordinates \f$ {\bf x}_u \f$
 *          -# Interpolate the undistorted image by choosing the nearest undistorted pixel \f$ {\bf x}_u \f$
 *
 *	 	  \subsection ctsaidistortionreferences References
 *	 	    -# Marc Hensel. "Komponenten fr die Kalibrierung nichtlinearer Kameramodelle".
 *	 	       Studienarbeit, TUHH, Vision Systems, 2002.
 *	 	    -# Roger Y. Tsai. "An Efficient and Accurate Camera Calibration Technique for 3D Machine Vision".
 *	 	       Proceedings of IEEE Conference on Computer Vision and Pattern Recognition, 1986, pages 364-374.
 *
 *		\section ctsaidistortionconcl Conclusion
 *          In the TUVisionlib there are two classes to compensate nonlinear radial distortion modelled
 *          by Tsai's camera model: CTsaiDistortion, CImageByteGrayTsaiDistortion.
 *
 *			Advantages of CTsaiDistortion
 *            - accepts CImagebyteGray and CImageByteRGB
 *            - faster than CImageByteGrayTsaiDistortion due to nearest neighbor interpolation
 *            - in contrast to CImageByteGrayTsaiDistortion the open graphic library (OpenGL) is not needed
 *
 *			Advantages of CImageByteGrayTsaiDistortion
 *            - better results due to bilinear interpolation (using OpenGL)
 */
