/*******************************************************************************\
 *                                                                               *
 * This file contains routines for calibrating Tsai's 11 parameter camera model. *
 * The camera model is based on the pin hole model of 3D-2D perspective          *
 * projection with 1st order radial lens distortion.  The model consists of      *
 * 5 internal (also called intrinsic or interior) camera parameters:             *
 *                                                                               *
 *       f       - effective focal length of the pin hole camera                 *
 *       kappa1  - 1st order radial lens distortion coefficient                  *
 *       Cx, Cy  - coordinates of center of radial lens distortion               *
 *		  (also used as the piercing point of the camera coordinate
 *		   frame's Z axis with the camera's sensor plane)               *
 *       sx      - uncertainty factor for scale of horizontal scanline           *
 *                                                                               *
 * and 6 external (also called extrinsic or exterior) camera parameters:         *
 *                                                                               *
 *       Rx, Ry, Rz, Tx, Ty, Tz  - rotational and translational components of    *
 *                                 the transform between the world's coordinate  *
 *                                 frame and the camera's coordinate frame.      *
 *                                                                               *
 * Data for model calibration consists of the (x,y,z) world coordinates of a     *
 * feature point (in mm) and the corresponding coordinates (Xf,Yf) (in pixels)   *
 * of the feature point in the image.  Two types of calibration are available:   *
 *                                                                               *
 *       coplanar     - all of the calibration points lie in a single plane      *
 *       non-coplanar - the calibration points do not lie in a single plane      *
 *                                                                               *
 * This file contains routines for two levels of calibration.  The first level   *
 * of calibration is a direct implementation of Tsai's algorithm in which only   *
 * the f, Tz and kappa1 parameters are optimized for.  The routines are:         *
 *                                                                               *
 *       coplanar_calibration ()                                                 *
 *       noncoplanar_calibration ()                                              *
 *                                                                               *
 * The second level of calibration optimizes for everything.  This level is      *
 * very slow but provides the most accurate calibration.  The routines are:      *
 *                                                                               *
 *       coplanar_calibration_with_full_optimization ()                          *
 *       noncoplanar_calibration_with_full_optimization ()                       *
 *                                                                               *
 * Routines are also provided for initializing camera parameter variables        *
 * for five of our camera/frame grabber systems.  These routines are:            *
 *                                                                               *
 *       initialize_photometrics_parms ()                                        *
 *       initialize_general_imaging_mos5300_matrox_parms ()                      *
 *       initialize_panasonic_gp_mf702_matrox_parms ()                           *
 *       initialize_sony_xc75_matrox_parms ()                                    *
 *       initialize_sony_xc77_matrox_parms ()                                    *
 *       initialize_sony_xc57_androx_parms ()                                    *
 *                                                                               *
 *                                                                               *
 * External routines                                                             *
 * -----------------                                                             *
 *                                                                               *
 * Nonlinear optimization for these camera calibration routines is performed     *
 * by the MINPACK lmdif subroutine.  lmdif uses a modified Levenberg-Marquardt   *
 * with a jacobian calculated by a forward-difference approximation.             *
 * The MINPACK FORTRAN routines were translated into C generated using f2c.      *
 *                                                                               *
 * Matrix operations (inversions, multiplications, etc.) are also provided by    *
 * external routines.                                                            *
 *                                                                               *
 *                                                                               *
 * Extra notes                                                                   *
 * -----------                                                                   *
 *                                                                               *
 * An explanation of the basic algorithms and description of the variables       *
 * can be found in several publications, including:                              *
 *                                                                               *
 * "An Efficient and Accurate Camera Calibration Technique for 3D Machine        *
 *  Vision", Roger Y. Tsai, Proceedings of IEEE Conference on Computer Vision    *
 *  and Pattern Recognition, Miami Beach, FL, 1986, pages 364-374.               *
 *                                                                               *
 *  and                                                                          *
 *                                                                               *
 * "A versatile Camera Calibration Technique for High-Accuracy 3D Machine        *
 *  Vision Metrology Using Off-the-Shelf TV Cameras and Lenses", Roger Y. Tsai,  *
 *  IEEE Journal of Robotics and Automation, Vol. RA-3, No. 4, August 1987,      *
 *  pages 323-344.                                                               *
 *                                                                               *
 *                                                                               *
 * Notation                                                                      *
 * --------                                                                      *
 *                                                                               *
 * The camera's X axis runs along increasing column coordinates in the           *
 * image/frame.  The Y axis runs along increasing row coordinates.               *
 *                                                                               *
 * pix == image/frame grabber picture element                                    *
 * sel == camera sensor element                                                  *
 *                                                                               *
 * Internal routines starting with "cc_" are for coplanar calibration.           *
 * Internal routines starting with "ncc_" are for noncoplanar calibration.       *
 *                                                                               *
 *                                                                               *
 * History                                                                       *
 * -------                                                                       *
 *                                                                               *
 * 20-May-95  Reg Willson (rgwillson@mmm.com) at 3M St. Paul, MN                 *
 *       Return the error to lmdif rather than the squared error.                *
 *         lmdif calculates the squared error internally during optimization.    *
 *         Before this change calibration was essentially optimizing error^4.    *
 *       Put transform and evaluation routines into separate files.              *
 *                                                                               *
 * 02-Apr-95  Reg Willson (rgwillson@mmm.com) at 3M St. Paul, MN                 *
 *       Rewrite memory allocation to avoid memory alignment problems            *
 *       on some machines.                                                       *
 *       Strip out IMSL code.  MINPACK seems to work fine.                       *
 *       Filename changes for DOS port.                                          *
 *                                                                               *
 * 04-Jun-94  Reg Willson (rgwillson@mmm.com) at 3M St. Paul, MN                 *
 *       Replaced ncc_compute_Xdp_and_Ydp with ncc_compute_Xd_Yd_and_r_squared.  *
 *         (effectively propagates the 22-Mar-94 to the non-coplanar routines)   *
 *       Added alternate macro definitions for less common math functions.       *
 *                                                                               *
 * 25-Mar-94  Torfi Thorhallsson (torfit@verk.hi.is) at the University of Iceland*
 *       Added a new version of the routines:                                    *
 *            cc_compute_exact_f_and_Tz ()                                       *
 *            cc_five_parm_optimization_with_late_distortion_removal ()          *
 *            cc_five_parm_optimization_with_early_distortion_removal ()         *
 *            cc_nic_optimization ()                                             *
 *            cc_full_optimization ()                                            *
 *            ncc_compute_exact_f_and_Tz ()                                      *
 *            ncc_nic_optimization ()                                            *
 *            ncc_full_optimization ()                                           *
 *                                                                               *
 *       The new routines use the *public domain* MINPACK library for            *
 *       optimization instead of the commercial IMSL library.                    *
 *       To select the new routines, compile this file with the flag -DMINPACK   *
 *                                                                               *
 * 22-Mar-94  Torfi Thorhallsson (torfit@verk.hi.is) at the University of Iceland*
 *       Fixed a bug in cc_nic_optimization_error and cc_full_optimization_error.*
 *       A division by cp.sx was missing.                                        *
 *                                                                               *
 * 15-Feb-94  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *       Included Frederic Devernay's (<Frederic.Devernay@sophia.inria.fr>)
 *	significantly improved routine for converting from undistorted to
 *	distorted sensor coordinates.  Rather than iteratively solving a
 *	system of two non-linear equations to perform the conversion, the 
 *	new routine algebraically solves a cubic polynomial in Rd (using
 *	the Cardan method).
 *                                                                               *
 * 14-Feb-94  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *	Fixed a coding bug in ncc_compute_R and ncc_compute_better_R.           *
 *       The r4, r5, and r6 terms should not be divided by cp.sx.                *
 *       Bug reported by: Volker Rodehorst <vr@cs.tu-berlin.de>                  *
 *                                                                               *
 * 04-Jul-93  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *	Added new routines to evaluate the accuracy of camera calibration.      *
 *                                                                               *
 *       Added check for coordinate handedness problem in calibration data.      *
 *                                                                               *
 * 01-May-93  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *	For efficiency the non-linear optimizations are now all based on
 *	the minimization of squared error in undistorted image coordinates      *
 *	instead of the squared error in distorted image coordinates.
 *
 *	New routine for inverse perspective projection.
 *
 * 14-Feb-93  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *       Bug fixes and speed ups.                                                *
 *                                                                               *
 * 07-Feb-93  Reg Willson (rgw@cs.cmu.edu) at Carnegie-Mellon University         *
 *       Original implementation.                                                *
 *                                                                               *
\*******************************************************************************/

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cerrno>

#include "tuv3d/TsaiCalibrationBackend.h"

/* The following structs and bindings refer to the c-function defined in Tsai's
   matrix and f2c - subdirectories */

typedef struct {
    int       lb1,
        ub1,
        lb2,
        ub2;
    char     *mat_sto;
    double  **el;
} dmat;

extern "C" {
    dmat newdmat(int rs, int re, int cs, int ce, int *error);
    int solve_system (dmat M, dmat a, dmat b);
    int lmdif_(int (*fcn)(integer *, integer *, doublereal *, doublereal *, integer *), integer *m, integer *n, doublereal *x,
        doublereal *fvec, doublereal *ftol, doublereal *xtol, doublereal *gtol, integer *maxfev, 
        doublereal *epsfcn, doublereal *diag, integer *mode, doublereal *factor, integer *nprint, integer *info,
        integer *nfev, doublereal *fjac, integer *ldfjac, integer *ipvt, 
        doublereal *qtf, doublereal *wa1, doublereal *wa2, doublereal *wa3, doublereal *wa4);
#define freemat(m) free((m).mat_sto) ; free((m).el)
}

/* Local working storage */
double    Xd[MAX_POINTS],
Yd[MAX_POINTS],
r_squared[MAX_POINTS],
U[7];

#define SQRT(x) sqrt(fabs(x))

using namespace TUVision;

// Static members have to be initialized
SCameraParameters TsaiCalibrator::cp;
calibration_data TsaiCalibrator::cd;
SCalibrationConstants TsaiCalibrator::cc;

/************************************************************************/
/*!
 * This routine solves for the roll, pitch and yaw angles (in radians)
 * for a given orthonormal rotation matrix (from Richard P. Paul,
 * Robot Manipulators: Mathematics, Programming and Control, p70).
 * Note 1, should the rotation matrix not be orthonormal these will not
 * be the "best fit" roll, pitch and yaw angles.
 * Note 2, there are actually two possible solutions for the matrix.
 * The second solution can be found by adding 180 degrees to Rz before
 * Ry and Rx are calculated.
 ***********************************************************************/
void      TsaiCalibrator::Matrix2Euler (){
    double    sg, cg;

    cc.Rz = atan2 (cc.r4, cc.r1);

    SINCOS (cc.Rz, sg, cg);

    cc.Ry = atan2 (-cc.r7, cc.r1 * cg + cc.r4 * sg);

    cc.Rx = atan2 (cc.r3 * sg - cc.r6 * cg, cc.r5 * cg - cc.r2 * sg);
}


/***********************************************************************/
/*! This routine simply takes the roll, pitch and yaw angles and fills in
 * the rotation matrix elements r1-r9.
\***********************************************************************/
void TsaiCalibrator::Euler2Matrix (){
    double    sa, ca, sb, cb, sg, cg;

    SINCOS (cc.Rx, sa, ca);
    SINCOS (cc.Ry, sb, cb);
    SINCOS (cc.Rz, sg, cg);

    cc.r1 = cb * cg;
    cc.r2 = cg * sa * sb - ca * sg;
    cc.r3 = sa * sg + ca * cg * sb;
    cc.r4 = cb * sg;
    cc.r5 = sa * sb * sg + ca * cg;
    cc.r6 = ca * sb * sg - cg * sa;
    cc.r7 = -sb;
    cc.r8 = cb * sa;
    cc.r9 = ca * cb;
}


void TsaiCalibratorCoplanar::compute_Xd_Yd_and_r_squared (){
    int       i;
    double    Xd_, Yd_;

    for (i = 0; i < cd.point_count; i++) {
        Xd[i] = Xd_ = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;	/* [mm] */
        Yd[i] = Yd_ = cp.dpy * (cd.Yf[i] - cp.Cy);	        /* [mm] */
        r_squared[i] = SQR (Xd_) + SQR (Yd_);                   /* [mm^2] */
    }
}


bool TsaiCalibratorCoplanar::compute_U (){
    int       i;
    dmat      M, a, b;

    M = newdmat (0, (cd.point_count - 1), 0, 4, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "cc compute U: unable to allocate matrix M\n");
        exit (-1); */
    }

    a = newdmat (0, 4, 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "cc compute U: unable to allocate vector a\n");
        exit (-1);*/
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        /*        fprintf (stderr, "cc compute U: unable to allocate vector b\n");
        exit (-1);*/
    }

    for (i = 0; i < cd.point_count; i++) {
        M.el[i][0] = Yd[i] * cd.xw[i];
        M.el[i][1] = Yd[i] * cd.yw[i];
        M.el[i][2] = Yd[i];
        M.el[i][3] = -Xd[i] * cd.xw[i];
        M.el[i][4] = -Xd[i] * cd.yw[i];
        b.el[i][0] = Xd[i];
    }

    if (solve_system (M, a, b)) {
        return false;
        /*	fprintf (stderr, "cc compute U: unable to solve system  Ma=b\n");
        exit (-1);*/
    }

    U[0] = a.el[0][0];
    U[1] = a.el[1][0];
    U[2] = a.el[2][0];
    U[3] = a.el[3][0];
    U[4] = a.el[4][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


void TsaiCalibratorCoplanar::compute_Tx_and_Ty ()
{
    int       i, far_point;
    double    Tx, Ty, Ty_squared, x, y, Sr, r1p, r2p, r4p, r5p, r1, r2, r4, r5, distance, far_distance;

    r1p = U[0];
    r2p = U[1];
    r4p = U[3];
    r5p = U[4];

    /* first find the square of the magnitude of Ty */
    if ((fabs (r1p) < EPSILON) && (fabs (r2p) < EPSILON))
        Ty_squared = 1 / (SQR (r4p) + SQR (r5p));
    else if ((fabs (r4p) < EPSILON) && (fabs (r5p) < EPSILON))
        Ty_squared = 1 / (SQR (r1p) + SQR (r2p));
    else if ((fabs (r1p) < EPSILON) && (fabs (r4p) < EPSILON))
        Ty_squared = 1 / (SQR (r2p) + SQR (r5p));
    else if ((fabs (r2p) < EPSILON) && (fabs (r5p) < EPSILON))
        Ty_squared = 1 / (SQR (r1p) + SQR (r4p));
    else {
        Sr = SQR (r1p) + SQR (r2p) + SQR (r4p) + SQR (r5p);
        Ty_squared = (Sr - sqrt (SQR (Sr) - 4 * SQR (r1p * r5p - r4p * r2p))) /
            (2 * SQR (r1p * r5p - r4p * r2p));
    }

    /* find a point that is far from the image center */
    far_distance = 0;
    far_point = 0;
    for (i = 0; i < cd.point_count; i++)
        if ((distance = r_squared[i]) > far_distance) {
            far_point = i;
            far_distance = distance;
        }

        /* now find the sign for Ty */
        /* start by assuming Ty > 0 */
        Ty = sqrt (Ty_squared);
        r1 = U[0] * Ty;
        r2 = U[1] * Ty;
        Tx = U[2] * Ty;
        r4 = U[3] * Ty;
        r5 = U[4] * Ty;
        x = r1 * cd.xw[far_point] + r2 * cd.yw[far_point] + Tx;
        y = r4 * cd.xw[far_point] + r5 * cd.yw[far_point] + Ty;

        /* flip Ty if we guessed wrong */
        if ((SIGNBIT (x) != SIGNBIT (Xd[far_point])) ||
            (SIGNBIT (y) != SIGNBIT (Yd[far_point])))
            Ty = -Ty;

        /* update the calibration constants */
        cc.Tx = U[2] * Ty;
        cc.Ty = Ty;
}


void TsaiCalibratorCoplanar::compute_R (){
    double    r1, r2, r3, r4, r5, r6, r7, r8, r9;

    r1 = U[0] * cc.Ty;
    r2 = U[1] * cc.Ty;
    r3 = sqrt (1 - SQR (r1) - SQR (r2));

    r4 = U[3] * cc.Ty;
    r5 = U[4] * cc.Ty;
    r6 = sqrt (1 - SQR (r4) - SQR (r5));
    if (!SIGNBIT (r1 * r4 + r2 * r5))
        r6 = -r6;

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;
    r8 = r3 * r4 - r1 * r6;
    r9 = r1 * r5 - r2 * r4;

    /* update the calibration constants */
    cc.r1 = r1;
    cc.r2 = r2;
    cc.r3 = r3;
    cc.r4 = r4;
    cc.r5 = r5;
    cc.r6 = r6;
    cc.r7 = r7;
    cc.r8 = r8;
    cc.r9 = r9;

    /* fill in cc.Rx, cc.Ry and cc.Rz */
    Matrix2Euler ();
}


bool TsaiCalibratorCoplanar::compute_approximate_f_and_Tz (){
    int       i;
    dmat      M, a, b;

    M = newdmat (0, (cd.point_count - 1), 0, 1, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "cc compute apx: unable to allocate matrix M\n");
        exit (-1);*/
    }

    a = newdmat (0, 1, 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "cc compute apx: unable to allocate vector a\n");
        exit (-1);*/
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "cc compute apx: unable to allocate vector b\n");
        exit (-1);*/
    }

    for (i = 0; i < cd.point_count; i++) {
        M.el[i][0] = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.Ty;
        M.el[i][1] = -Yd[i];
        b.el[i][0] = (cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i]) * Yd[i];
    }

    if (solve_system (M, a, b)) {
        return false;
        /*	fprintf (stderr, "cc compute apx: unable to solve system  Ma=b\n");
        exit (-1);*/
    }

    /* update the calibration constants */
    cc.f = a.el[0][0];
    cc.Tz = a.el[1][0];
    cc.kappa1 = 0.0;		/* this is the assumption that our calculation was made under */

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


/*
\param m_ptr  pointer to number of points to fit
\param n_ptr  pointer to number of parameters
\param params vector of parameters
\param err    vector of error from data */
/************************************************************************/
int  TsaiCalibratorCoplanar::compute_exact_f_and_Tz_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    f, Tz, kappa1, xc, yc, zc, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor;

    f = params[0];
    Tz = params[1];
    kappa1 = params[2];

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        /* Note: zw is implicitly assumed to be zero for these (coplanar) calculations */
        xc = cc.r1 * cd.xw[i] + cc.r2 * cd.yw[i] + cc.Tx;
        yc = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.Ty;
        zc = cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd[i]) + SQR (Yd[i]));
        Xu_2 = Xd[i] * distortion_factor;
        Yu_2 = Yd[i] * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorCoplanar::compute_exact_f_and_Tz (){
#define NPARAMS 3

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*     fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration constants as an initial guess */
    x[0] = cc.f;
    x[1] = cc.Tz;
    x[2] = cc.kappa1;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */ 
    lmdif_ (compute_exact_f_and_Tz_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration constants */
    cc.f = x[0];
    cc.Tz = x[1];
    cc.kappa1 = x[2];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*  fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev); */
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
bool  TsaiCalibratorCoplanar::three_parm_optimization (){
    int       i;

    for (i = 0; i < cd.point_count; i++)
        if (cd.zw[i]) {
            return false;
            /*	    fprintf (stderr, "error - coplanar calibration tried with data outside of Z plane\n\n");
            exit (-1);*/
        }

        compute_Xd_Yd_and_r_squared ();

        compute_U ();

        compute_Tx_and_Ty ();

        compute_R ();

        if(!compute_approximate_f_and_Tz ())
            return false;

        if (cc.f < 0) {
            /* try the other solution for the orthonormal matrix */
            cc.r3 = -cc.r3;
            cc.r6 = -cc.r6;
            cc.r7 = -cc.r7;
            cc.r8 = -cc.r8;
            Matrix2Euler();

            if(!compute_approximate_f_and_Tz ())
                return false;

            if (cc.f < 0) {
                return false;
                /*	    fprintf (stderr, "error - possible handedness problem with data\n");
                exit (-1);*/
            }
        }

        if(!compute_exact_f_and_Tz ())
            return false;

        return true;
}


/************************************************************************/
void TsaiCalibratorCoplanar::remove_sensor_plane_distortion_from_Xd_and_Yd (){
    int       i;
    double    Xu, Yu;

    for (i = 0; i < cd.point_count; i++) {
        distorted_to_undistorted_sensor_coord (Xd[i], Yd[i], &Xu, &Yu);
        Xd[i] = Xu;
        Yd[i] = Yu;
        r_squared[i] = SQR (Xu) + SQR (Yu);
    }
}


/************************************************************************/
int TsaiCalibratorCoplanar::five_parm_optimization_with_late_distortion_removal_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    f, Tz, kappa1, xc, yc, zc, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor;

    /* in this routine radial lens distortion is only taken into account */
    /* after the rotation and translation constants have been determined */

    f = params[0];
    Tz = params[1];
    kappa1 = params[2];

    cp.Cx = params[3];
    cp.Cy = params[4];

    compute_Xd_Yd_and_r_squared ();

    compute_U ();

    compute_Tx_and_Ty ();

    compute_R ();

    compute_approximate_f_and_Tz ();

    if (cc.f < 0) {
        /* try the other solution for the orthonormal matrix */
        cc.r3 = -cc.r3;
        cc.r6 = -cc.r6;
        cc.r7 = -cc.r7;
        cc.r8 = -cc.r8;
        Matrix2Euler ();

        compute_approximate_f_and_Tz ();

        if (cc.f < 0) {
            return -1;
            /*            fprintf (stderr, "error - possible handedness problem with data\n");
            exit (-1);*/
        }
    }

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        /* Note: zw is implicitly assumed to be zero for these (coplanar) calculations */
        xc = cc.r1 * cd.xw[i] + cc.r2 * cd.yw[i] + cc.Tx;
        yc = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.Ty;
        zc = cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + kappa1 * r_squared[i];
        Xu_2 = Xd[i] * distortion_factor;
        Yu_2 = Yd[i] * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorCoplanar::five_parm_optimization_with_late_distortion_removal (){
#define NPARAMS 5

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*     fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*     fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration constants as an initial guess */
    x[0] = cc.f;
    x[1] = cc.Tz;
    x[2] = cc.kappa1;
    x[3] = cp.Cx;
    x[4] = cp.Cy;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (five_parm_optimization_with_late_distortion_removal_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.f = x[0];
    cc.Tz = x[1];
    cc.kappa1 = x[2];
    cp.Cx = x[3];
    cp.Cy = x[4];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
int TsaiCalibratorCoplanar::five_parm_optimization_with_early_distortion_removal_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    f, Tz, xc, yc, zc, Xu_1, Yu_1, Xu_2, Yu_2;

    /* in this routine radial lens distortion is taken into account */
    /* before the rotation and translation constants are determined */
    /* (this assumes we have the distortion reasonably modelled)    */

    f = params[0];
    Tz = params[1];

    cc.kappa1 = params[2];
    cp.Cx = params[3];
    cp.Cy = params[4];

    compute_Xd_Yd_and_r_squared ();

    /* remove the sensor distortion before computing the translation and rotation stuff */
    remove_sensor_plane_distortion_from_Xd_and_Yd ();

    compute_U ();

    compute_Tx_and_Ty ();

    compute_R ();

    /* we need to do this just to see if we have to flip the rotation matrix */
    compute_approximate_f_and_Tz ();

    if (cc.f < 0) {
        /* try the other solution for the orthonormal matrix */
        cc.r3 = -cc.r3;
        cc.r6 = -cc.r6;
        cc.r7 = -cc.r7;
        cc.r8 = -cc.r8;
        Matrix2Euler ();

        compute_approximate_f_and_Tz ();

        if (cc.f < 0) {
            return -1;
            /*            fprintf (stderr, "error - possible handedness problem with data\n");
            exit (-1);*/
        }
    }

    /* now calculate the squared error assuming zero distortion */
    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        /* Note: zw is implicitly assumed to be zero for these (coplanar) calculations */
        xc = cc.r1 * cd.xw[i] + cc.r2 * cd.yw[i] + cc.Tx;
        yc = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.Ty;
        zc = cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from distorted sensor coordinates to undistorted sensor coordinates  */
        /* (already done, actually)							 */
        Xu_2 = Xd[i];
        Yu_2 = Yd[i];

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorCoplanar::five_parm_optimization_with_early_distortion_removal (){
#define NPARAMS 5

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.f;
    x[1] = cc.Tz;
    x[2] = cc.kappa1;
    x[3] = cp.Cx;
    x[4] = cp.Cy;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (five_parm_optimization_with_early_distortion_removal_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.f = x[0];
    cc.Tz = x[1];
    cc.kappa1 = x[2];
    cp.Cx = x[3];
    cp.Cy = x[4];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
int TsaiCalibratorCoplanar::nic_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    xc, yc, zc, Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor, Rx, Ry, Rz, Tx, Ty, Tz, kappa1, f,
        r1, r2, r4, r5, r7, r8, sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r7 = -sb;
    r8 = cb * sa;

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        /* Note: zw is implicitly assumed to be zero for these (coplanar) calculations */
        xc = r1 * cd.xw[i] + r2 * cd.yw[i] + Tx;
        yc = r4 * cd.xw[i] + r5 * cd.yw[i] + Ty;
        zc = r7 * cd.xw[i] + r8 * cd.yw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd_ = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx; 
        Yd_ = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorCoplanar::nic_optimization (){
#define NPARAMS 8

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.Rx;
    x[1] = cc.Ry;
    x[2] = cc.Rz;
    x[3] = cc.Tx;
    x[4] = cc.Ty;
    x[5] = cc.Tz;
    x[6] = cc.kappa1;
    x[7] = cc.f;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (nic_optimization_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.Rx = x[0];
    cc.Ry = x[1];
    cc.Rz = x[2];
    Euler2Matrix ();

    cc.Tx = x[3];
    cc.Ty = x[4];
    cc.Tz = x[5];
    cc.kappa1 = x[6];
    cc.f = x[7];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
int TsaiCalibratorCoplanar::full_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    xc, yc, zc, Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor, Rx, Ry, Rz, Tx,
        Ty, Tz, kappa1, f, Cx, Cy, r1, r2, r4, r5, r7, r8, sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];
    Cx = params[8];
    Cy = params[9];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r7 = -sb;
    r8 = cb * sa;

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        /* Note: zw is implicitly assumed to be zero for these (coplanar) calculations */
        xc = r1 * cd.xw[i] + r2 * cd.yw[i] + Tx;
        yc = r4 * cd.xw[i] + r5 * cd.yw[i] + Ty;
        zc = r7 * cd.xw[i] + r8 * cd.yw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd_ = cp.dpx * (cd.Xf[i] - Cx) / cp.sx; 
        Yd_ = cp.dpy * (cd.Yf[i] - Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorCoplanar::full_optimization (){
#define NPARAMS 10

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.Rx;
    x[1] = cc.Ry;
    x[2] = cc.Rz;
    x[3] = cc.Tx;
    x[4] = cc.Ty;
    x[5] = cc.Tz;
    x[6] = cc.kappa1;
    x[7] = cc.f;
    x[8] = cp.Cx;
    x[9] = cp.Cy;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (full_optimization_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.Rx = x[0];
    cc.Ry = x[1];
    cc.Rz = x[2];
    Euler2Matrix ();

    cc.Tx = x[3];
    cc.Ty = x[4];
    cc.Tz = x[5];
    cc.kappa1 = x[6];
    cc.f = x[7];
    cp.Cx = x[8];
    cp.Cy = x[9];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/***********************************************************************\
 * Routines for noncoplanar camera calibration	 
\***********************************************************************/
void   TsaiCalibratorNoncoplanar::compute_Xd_Yd_and_r_squared (){
    int       i;
    double    Xd_, Yd_;

    for (i = 0; i < cd.point_count; i++) {
        Xd[i] = Xd_ = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;      /* [mm] */
        Yd[i] = Yd_ = cp.dpy * (cd.Yf[i] - cp.Cy);              /* [mm] */
        r_squared[i] = SQR (Xd_) + SQR (Yd_);                   /* [mm^2] */
    }
}


bool      TsaiCalibratorNoncoplanar::compute_U (){
    int       i;
    dmat      M, a, b;

    M = newdmat (0, (cd.point_count - 1), 0, 6, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute U: unable to allocate matrix M\n");
        exit (-1);*/
    }

    a = newdmat (0, 6, 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute U: unable to allocate vector a\n");
        exit (-1);*/
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute U: unable to allocate vector b\n");
        exit (-1);*/
    }

    for (i = 0; i < cd.point_count; i++) {
        M.el[i][0] = Yd[i] * cd.xw[i];
        M.el[i][1] = Yd[i] * cd.yw[i];
        M.el[i][2] = Yd[i] * cd.zw[i];
        M.el[i][3] = Yd[i];
        M.el[i][4] = -Xd[i] * cd.xw[i];
        M.el[i][5] = -Xd[i] * cd.yw[i];
        M.el[i][6] = -Xd[i] * cd.zw[i];
        b.el[i][0] = Xd[i];
    }

    if (solve_system (M, a, b)) {
        return false;
        /*	fprintf (stderr, "ncc compute U: error - non-coplanar calibration tried with data which may possibly be coplanar\n\n");
        exit (-1);*/
    }

    U[0] = a.el[0][0];
    U[1] = a.el[1][0];
    U[2] = a.el[2][0];
    U[3] = a.el[3][0];
    U[4] = a.el[4][0];
    U[5] = a.el[5][0];
    U[6] = a.el[6][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


void      TsaiCalibratorNoncoplanar::compute_Tx_and_Ty (){
    int       i, far_point;
    double    Tx, Ty, Ty_squared, x, y, r1, r2, r3, r4, r5, r6, distance, far_distance;

    /* first find the square of the magnitude of Ty */
    Ty_squared = 1 / (SQR (U[4]) + SQR (U[5]) + SQR (U[6]));

    /* find a point that is far from the image center */
    far_distance = 0;
    far_point = 0;
    for (i = 0; i < cd.point_count; i++)
        if ((distance = r_squared[i]) > far_distance) {
            far_point = i;
            far_distance = distance;
        }

        /* now find the sign for Ty */
        /* start by assuming Ty > 0 */
        Ty = sqrt (Ty_squared);
        r1 = U[0] * Ty;
        r2 = U[1] * Ty;
        r3 = U[2] * Ty;
        Tx = U[3] * Ty;
        r4 = U[4] * Ty;
        r5 = U[5] * Ty;
        r6 = U[6] * Ty;
        x = r1 * cd.xw[far_point] + r2 * cd.yw[far_point] + r3 * cd.zw[far_point] + Tx;
        y = r4 * cd.xw[far_point] + r5 * cd.yw[far_point] + r6 * cd.zw[far_point] + Ty;

        /* flip Ty if we guessed wrong */
        if ((SIGNBIT (x) != SIGNBIT (Xd[far_point])) ||
            (SIGNBIT (y) != SIGNBIT (Yd[far_point])))
            Ty = -Ty;

        /* update the calibration constants */
        cc.Tx = U[3] * Ty;
        cc.Ty = Ty;
}


void      TsaiCalibratorNoncoplanar::compute_sx (){
    cp.sx = sqrt (SQR (U[0]) + SQR (U[1]) + SQR (U[2])) * fabs (cc.Ty);
}


void      TsaiCalibratorNoncoplanar::compute_R (){
    double    r1, r2, r3, r4, r5, r6, r7, r8, r9;

    r1 = U[0] * cc.Ty / cp.sx;
    r2 = U[1] * cc.Ty / cp.sx;
    r3 = U[2] * cc.Ty / cp.sx;

    r4 = U[4] * cc.Ty;
    r5 = U[5] * cc.Ty;
    r6 = U[6] * cc.Ty;

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;
    r8 = r3 * r4 - r1 * r6;
    r9 = r1 * r5 - r2 * r4;

    /* update the calibration constants */
    cc.r1 = r1;
    cc.r2 = r2;
    cc.r3 = r3;
    cc.r4 = r4;
    cc.r5 = r5;
    cc.r6 = r6;
    cc.r7 = r7;
    cc.r8 = r8;
    cc.r9 = r9;

    /* fill in cc.Rx, cc.Ry and cc.Rz */
    Matrix2Euler ();
}


void      TsaiCalibratorNoncoplanar::compute_better_R (){
    double    r1, r2, r3, r4, r5, r6, r7, sa, ca, sb, cb, sg, cg;

    r1 = U[0] * cc.Ty / cp.sx;
    r2 = U[1] * cc.Ty / cp.sx;
    r3 = U[2] * cc.Ty / cp.sx;

    r4 = U[4] * cc.Ty;
    r5 = U[5] * cc.Ty;
    r6 = U[6] * cc.Ty;

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;

    /* now find the RPY angles corresponding to the estimated rotation matrix */
    cc.Rz = atan2 (r4, r1);

    SINCOS (cc.Rz, sg, cg);

    cc.Ry = atan2 (-r7, r1 * cg + r4 * sg);

    cc.Rx = atan2 (r3 * sg - r6 * cg, r5 * cg - r2 * sg);

    SINCOS (cc.Rx, sa, ca);

    SINCOS (cc.Ry, sb, cb);

    /* now generate a more orthonormal rotation matrix from the RPY angles */
    cc.r1 = cb * cg;
    cc.r2 = cg * sa * sb - ca * sg;
    cc.r3 = sa * sg + ca * cg * sb;
    cc.r4 = cb * sg;
    cc.r5 = sa * sb * sg + ca * cg;
    cc.r6 = ca * sb * sg - cg * sa;
    cc.r7 = -sb;
    cc.r8 = cb * sa;
    cc.r9 = ca * cb;
}


bool      TsaiCalibratorNoncoplanar::compute_approximate_f_and_Tz (){
    int       i;
    dmat      M, a, b;

    M = newdmat (0, (cd.point_count - 1), 0, 1, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute apx: unable to allocate matrix M\n");
        exit (-1);*/
    }

    a = newdmat (0, 1, 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute apx: unable to allocate vector a\n");
        exit (-1);*/
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        /*	fprintf (stderr, "ncc compute apx: unable to allocate vector b\n");
        exit (-1);*/
    }

    for (i = 0; i < cd.point_count; i++) {
        M.el[i][0] = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.r6 * cd.zw[i] + cc.Ty;
        M.el[i][1] = -Yd[i];
        b.el[i][0] = (cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + cc.r9 * cd.zw[i]) * Yd[i];
    }

    if (solve_system (M, a, b)) {
        return false;
        /*	fprintf (stderr, "ncc compute apx: unable to solve system  Ma=b\n");
        exit (-1);*/
    }

    /* update the calibration constants */
    cc.f = a.el[0][0];
    cc.Tz = a.el[1][0];
    cc.kappa1 = 0.0;		/* this is the assumption that our calculation was made under */

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


/************************************************************************/
int TsaiCalibratorNoncoplanar::compute_exact_f_and_Tz_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    xc, yc, zc, Xu_1, Yu_1, Xu_2,
        Yu_2, distortion_factor, f, Tz, kappa1;

    f = params[0];
    Tz = params[1];
    kappa1 = params[2];

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        xc = cc.r1 * cd.xw[i] + cc.r2 * cd.yw[i] + cc.r3 * cd.zw[i] + cc.Tx;
        yc = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.r6 * cd.zw[i] + cc.Ty;
        zc = cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + cc.r9 * cd.zw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + kappa1 * r_squared[i];
        Xu_2 = Xd[i] * distortion_factor;
        Yu_2 = Yd[i] * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool      TsaiCalibratorNoncoplanar::compute_exact_f_and_Tz (){
#define NPARAMS 3

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration constants as an initial guess */
    x[0] = cc.f;
    x[1] = cc.Tz;
    x[2] = cc.kappa1;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (compute_exact_f_and_Tz_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration constants */
    cc.f = x[0];
    cc.Tz = x[1];
    cc.kappa1 = x[2];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
bool      TsaiCalibratorNoncoplanar::three_parm_optimization ()
{
    compute_Xd_Yd_and_r_squared ();

    compute_U ();

    compute_Tx_and_Ty ();

    compute_sx ();

    compute_Xd_Yd_and_r_squared ();

    compute_better_R ();

    compute_approximate_f_and_Tz ();

    if (cc.f < 0) {
        /* try the other solution for the orthonormal matrix */
        cc.r3 = -cc.r3;
        cc.r6 = -cc.r6;
        cc.r7 = -cc.r7;
        cc.r8 = -cc.r8;
        Matrix2Euler ();

        compute_approximate_f_and_Tz ();

        if (cc.f < 0) {
            return false;
            /*            fprintf (stderr, "error - possible handedness problem with data\n");
            exit (-1);*/
        }
    }

    compute_exact_f_and_Tz ();
    return true;
}


/************************************************************************/
int TsaiCalibratorNoncoplanar::nic_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    xc, yc, zc, Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor, Rx, Ry, Rz, Tx, Ty, Tz, kappa1, sx,
        f, r1, r2, r3, r4, r5, r6, r7, r8, r9, sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];
    sx = params[8];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        xc = r1 * cd.xw[i] + r2 * cd.yw[i] + r3 * cd.zw[i] + Tx;
        yc = r4 * cd.xw[i] + r5 * cd.yw[i] + r6 * cd.zw[i] + Ty;
        zc = r7 * cd.xw[i] + r8 * cd.yw[i] + r9 * cd.zw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd_ = cp.dpx * (cd.Xf[i] - cp.Cx) / sx;
        Yd_ = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibratorNoncoplanar::nic_optimization ()
{
#define NPARAMS 9

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.Rx;
    x[1] = cc.Ry;
    x[2] = cc.Rz;
    x[3] = cc.Tx;
    x[4] = cc.Ty;
    x[5] = cc.Tz;
    x[6] = cc.kappa1;
    x[7] = cc.f;
    x[8] = cp.sx;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (nic_optimization_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.Rx = x[0];
    cc.Ry = x[1];
    cc.Rz = x[2];
    Euler2Matrix ();

    cc.Tx = x[3];
    cc.Ty = x[4];
    cc.Tz = x[5];
    cc.kappa1 = x[6];
    cc.f = x[7];
    cp.sx = x[8];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
int TsaiCalibratorNoncoplanar::full_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *notused){
    int       i;
    double    xc, yc, zc, Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor, Rx, Ry, Rz, Tx, Ty, Tz,
        kappa1, sx, f, Cx, Cy, r1, r2, r3, r4, r5, r6, r7, r8, r9, sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];
    sx = params[8];
    Cx = params[9];
    Cy = params[10];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        xc = r1 * cd.xw[i] + r2 * cd.yw[i] + r3 * cd.zw[i] + Tx;
        yc = r4 * cd.xw[i] + r5 * cd.yw[i] + r6 * cd.zw[i] + Ty;
        zc = r7 * cd.xw[i] + r8 * cd.yw[i] + r9 * cd.zw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd_ = cp.dpx * (cd.Xf[i] - Cx) / sx;
        Yd_ = cp.dpy * (cd.Yf[i] - Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool      TsaiCalibratorNoncoplanar::full_optimization ()
{
#define NPARAMS 11

    int     i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        exit(-1);*/
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        exit(-1);*/
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        /*       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        exit(-1);*/
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.Rx;
    x[1] = cc.Ry;
    x[2] = cc.Rz;
    x[3] = cc.Tx;
    x[4] = cc.Ty;
    x[5] = cc.Tz;
    x[6] = cc.kappa1;
    x[7] = cc.f;
    x[8] = cp.sx;
    x[9] = cp.Cx;
    x[10] = cp.Cy;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (full_optimization_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.Rx = x[0];
    cc.Ry = x[1];
    cc.Rz = x[2];
    Euler2Matrix ();

    cc.Tx = x[3];
    cc.Ty = x[4];
    cc.Tz = x[5];
    cc.kappa1 = x[6];
    cc.f = x[7];
    cp.sx = x[8];
    cp.Cx = x[9];
    cp.Cy = x[10];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    /*    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);*/
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/

bool TsaiCalibratorCoplanar::calibration ()
{
    /* just do the basic 3 parameter (Tz, f, kappa1) optimization */
    return three_parm_optimization ();
}


bool TsaiCalibratorCoplanar::calibration_with_full_optimization ()
{
    /* start with a 3 parameter (Tz, f, kappa1) optimization */
    if(!three_parm_optimization ())
        return false;

    /* do a 5 parameter (Tz, f, kappa1, Cx, Cy) optimization */
    if(!five_parm_optimization_with_late_distortion_removal ())
        return false;

    /* do a better 5 parameter (Tz, f, kappa1, Cx, Cy) optimization */
    if(!five_parm_optimization_with_early_distortion_removal ())
        return false;

    /* do a full optimization minus the image center */
    if(!nic_optimization ())
        return false;

    /* do a full optimization including the image center */
    return full_optimization ();

}


bool TsaiCalibratorNoncoplanar::calibration ()
{
    /* just do the basic 3 parameter (Tz, f, kappa1) optimization */
    return three_parm_optimization ();
}


bool  TsaiCalibratorNoncoplanar::calibration_with_full_optimization ()
{
    /* start with a 3 parameter (Tz, f, kappa1) optimization */
    if(!three_parm_optimization ())
        return false;

    /* do a full optimization minus the image center */
    if(!nic_optimization ())
        return false;

    /* do a full optimization including the image center */
    return full_optimization ();
}


/************************************************************************/
/*
This routine converts from undistorted to distorted sensor coordinates.
The technique involves algebraically solving the cubic polynomial

Ru = Rd * (1 + kappa1 * Rd**2)

using the Cardan method.

Note: for kappa1 < 0 the distorted sensor plane extends out to a maximum
barrel distortion radius of  Rd = sqrt (-1/(3 * kappa1)).

To see the logic used in this routine try graphing the above polynomial
for positive and negative kappa1's
*/

void  TsaiCalibrator::undistorted_to_distorted_sensor_coord (double Xu, double Yu, double *Xd, double *Yd){
#define SQRT3   1.732050807568877293527446341505872366943

    double    Ru, Rd, lambda, c, d, Q, R, D, S, T, sinT, cosT;

    if (((Xu == 0) && (Yu == 0)) || (cc.kappa1 == 0)) {
        *Xd = Xu;
        *Yd = Yu;
        return;
    }

    Ru = hypot (Xu, Yu);	/* SQRT(Xu*Xu+Yu*Yu) */

    c = 1 / cc.kappa1;
    d = -c * Ru;

    Q = c / 3;
    R = -d / 2;
    D = CUB (Q) + SQR (R);

    if (D >= 0) {		/* one real root */
        D = SQRT (D);
        S = CBRT (R + D);
        T = CBRT (R - D);
        Rd = S + T;

        if (Rd < 0) {
            Rd = SQRT (-1 / (3 * cc.kappa1));
            fprintf (stderr, "\nWarning: undistorted image point to distorted image point mapping limited by\n");
            fprintf (stderr, "         maximum barrel distortion radius of %f\n", Rd);
            fprintf (stderr, "         (Xu = %f, Yu = %f) -> (Xd = %f, Yd = %f)\n\n",
                Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
        }
    } else {			/* three real roots */
        D = SQRT (-D);
        S = CBRT (hypot (R, D));
        T = atan2 (D, R) / 3;
        SINCOS (T, sinT, cosT);

        /* the larger positive root is    2*S*cos(T)                   */
        /* the smaller positive root is   -S*cos(T) + SQRT(3)*S*sin(T) */
        /* the negative root is           -S*cos(T) - SQRT(3)*S*sin(T) */

        Rd = -S * cosT + SQRT3 * S * sinT;	/* use the smaller positive root */
    }

    lambda = Rd / Ru;

    *Xd = Xu * lambda;
    *Yd = Yu * lambda;
}


/************************************************************************/
void  TsaiCalibrator::distorted_to_undistorted_sensor_coord (double Xd, double Yd, double *Xu, double *Yu){
    double    distortion_factor;

    /* convert from distorted to undistorted sensor plane coordinates */
    distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
    *Xu = Xd * distortion_factor;
    *Yu = Yd * distortion_factor;
}


/************************************************************************/
void TsaiCalibrator::undistorted_to_distorted_image_coord (double Xfu, double Yfu, double *Xfd, double *Yfd){
    double    Xu, Yu, Xd, Yd;

    /* convert from image to sensor coordinates */
    Xu = cp.dpx * (Xfu - cp.Cx) / cp.sx;
    Yu = cp.dpy * (Yfu - cp.Cy);

    /* convert from undistorted sensor to distorted sensor plane coordinates */
    undistorted_to_distorted_sensor_coord (Xu, Yu, &Xd, &Yd);

    /* convert from sensor to image coordinates */
     *Xfd = Xd * cp.sx / cp.dpx + cp.Cx;
     *Yfd = Yd / cp.dpy + cp.Cy;
}


/************************************************************************/
void  TsaiCalibrator::distorted_to_undistorted_image_coord (double Xfd, double Yfd, double *Xfu, double *Yfu){
    double Xd, Yd, Xu, Yu;

    /* convert from image to sensor coordinates */
    Xd = cp.dpx * (Xfd - cp.Cx) / cp.sx;
    Yd = cp.dpy * (Yfd - cp.Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    distorted_to_undistorted_sensor_coord (Xd, Yd, &Xu, &Yu);

    /* convert from sensor to image coordinates */
     *Xfu = Xu * cp.sx / cp.dpx + cp.Cx;
     *Yfu = Yu / cp.dpy + cp.Cy;
}


/***********************************************************************\
 * This routine takes the position of a point in world coordinates [mm]
 * and determines the position of its image in image coordinates [pix].
\***********************************************************************/
void TsaiCalibrator::world_coord_to_image_coord (double xw, double yw, double zw, double *Xf, double *Yf){
    double xc, yc, zc, Xu, Yu, Xd, Yd;

    /* convert from world coordinates to camera coordinates */
    xc = cc.r1 * xw + cc.r2 * yw + cc.r3 * zw + cc.Tx;
    yc = cc.r4 * xw + cc.r5 * yw + cc.r6 * zw + cc.Ty;
    zc = cc.r7 * xw + cc.r8 * yw + cc.r9 * zw + cc.Tz;

    /* convert from camera coordinates to undistorted sensor plane coordinates */
    Xu = cc.f * xc / zc;
    Yu = cc.f * yc / zc;

    /* convert from undistorted to distorted sensor plane coordinates */
    undistorted_to_distorted_sensor_coord (Xu, Yu, &Xd, &Yd);

    /* convert from distorted sensor plane coordinates to image coordinates */
     *Xf = Xd * cp.sx / cp.dpx + cp.Cx;
     *Yf = Yd / cp.dpy + cp.Cy;
}


/***********************************************************************\
 * This routine performs an inverse perspective projection to determine
 * the position of a point in world coordinates that corresponds to a 
 * given position in image coordinates.  To constrain the inverse
 * projection to a single point the routine requires a Z world	 
 * coordinate for the point in addition to the X and Y image coordinates.* 
\***********************************************************************/
void TsaiCalibrator::image_coord_to_world_coord (double Xfd, double Yfd, double zw, double *xw, double *yw){
    double Xd, Yd, Xu, Yu, common_denominator;

    /* convert from image to distorted sensor coordinates */
    Xd = cp.dpx * (Xfd - cp.Cx) / cp.sx;
    Yd = cp.dpy * (Yfd - cp.Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    distorted_to_undistorted_sensor_coord (Xd, Yd, &Xu, &Yu);

    /* calculate the corresponding xw and yw world coordinates	 */
    /* (these equations were derived by simply inverting	 */
    /* the perspective projection equations using Macsyma)	 */
    common_denominator = ((cc.r1 * cc.r8 - cc.r2 * cc.r7) * Yu +
        (cc.r5 * cc.r7 - cc.r4 * cc.r8) * Xu -
        cc.f * cc.r1 * cc.r5 + cc.f * cc.r2 * cc.r4);

    *xw = (((cc.r2 * cc.r9 - cc.r3 * cc.r8) * Yu +
        (cc.r6 * cc.r8 - cc.r5 * cc.r9) * Xu -
        cc.f * cc.r2 * cc.r6 + cc.f * cc.r3 * cc.r5) * zw +
        (cc.r2 * cc.Tz - cc.r8 * cc.Tx) * Yu +
        (cc.r8 * cc.Ty - cc.r5 * cc.Tz) * Xu -
        cc.f * cc.r2 * cc.Ty + cc.f * cc.r5 * cc.Tx) / common_denominator;

    *yw = -(((cc.r1 * cc.r9 - cc.r3 * cc.r7) * Yu +
        (cc.r6 * cc.r7 - cc.r4 * cc.r9) * Xu -
        cc.f * cc.r1 * cc.r6 + cc.f * cc.r3 * cc.r4) * zw +
        (cc.r1 * cc.Tz - cc.r7 * cc.Tx) * Yu +
        (cc.r7 * cc.Ty - cc.r4 * cc.Tz) * Xu -
        cc.f * cc.r1 * cc.Ty + cc.f * cc.r4 * cc.Tx) / common_denominator;
}


/***********************************************************************\
 * This routine takes the position of a point in world coordinates [mm]
 * and determines its position in camera coordinates [mm].
\***********************************************************************/
void  TsaiCalibrator::world_coord_to_camera_coord (double xw, double yw, double zw, double *xc, double *yc, double *zc){
    *xc = cc.r1 * xw + cc.r2 * yw + cc.r3 * zw + cc.Tx;
    *yc = cc.r4 * xw + cc.r5 * yw + cc.r6 * zw + cc.Ty;
    *zc = cc.r7 * xw + cc.r8 * yw + cc.r9 * zw + cc.Tz;
}


/***********************************************************************\
 * This routine takes the position of a point in camera coordinates [mm]
 * and determines its position in world coordinates [mm].
\***********************************************************************/
void TsaiCalibrator::camera_coord_to_world_coord (double xc, double yc, double zc, double *xw, double *yw, double *zw){
    double    common_denominator;

    /* these equations were found by simply inverting the previous routine using Macsyma */

    common_denominator = ((cc.r1 * cc.r5 - cc.r2 * cc.r4) * cc.r9 +
        (cc.r3 * cc.r4 - cc.r1 * cc.r6) * cc.r8 +
        (cc.r2 * cc.r6 - cc.r3 * cc.r5) * cc.r7);

    *xw = ((cc.r2 * cc.r6 - cc.r3 * cc.r5) * zc +
        (cc.r3 * cc.r8 - cc.r2 * cc.r9) * yc +
        (cc.r5 * cc.r9 - cc.r6 * cc.r8) * xc +
        (cc.r3 * cc.r5 - cc.r2 * cc.r6) * cc.Tz +
        (cc.r2 * cc.r9 - cc.r3 * cc.r8) * cc.Ty +
        (cc.r6 * cc.r8 - cc.r5 * cc.r9) * cc.Tx) / common_denominator;

    *yw = -((cc.r1 * cc.r6 - cc.r3 * cc.r4) * zc +
        (cc.r3 * cc.r7 - cc.r1 * cc.r9) * yc +
        (cc.r4 * cc.r9 - cc.r6 * cc.r7) * xc +
        (cc.r3 * cc.r4 - cc.r1 * cc.r6) * cc.Tz +
        (cc.r1 * cc.r9 - cc.r3 * cc.r7) * cc.Ty +
        (cc.r6 * cc.r7 - cc.r4 * cc.r9) * cc.Tx) / common_denominator;

    *zw = ((cc.r1 * cc.r5 - cc.r2 * cc.r4) * zc +
        (cc.r2 * cc.r7 - cc.r1 * cc.r8) * yc +
        (cc.r4 * cc.r8 - cc.r5 * cc.r7) * xc +
        (cc.r2 * cc.r4 - cc.r1 * cc.r5) * cc.Tz +
        (cc.r1 * cc.r8 - cc.r2 * cc.r7) * cc.Ty +
        (cc.r5 * cc.r7 - cc.r4 * cc.r8) * cc.Tx) / common_denominator;
}



/************************************************************************/
/* This cube root routine handles negative arguments (unlike cbrt).     */

double TsaiCalibrator::CBRT (double x){
    if (x == 0)
        return (0);
    else if (x > 0)
        return (pow (x, (double) 1.0 / 3.0));
    else
        return (-pow (-x, (double) 1.0 / 3.0));
} 


/***********************************************************************\
 * Routines for coplanar extrinsic parameter estimation
\***********************************************************************/
bool  TsaiCalibratorCoplanar::epe_compute_U (double U[])
{
    dmat      M, a, b;

    double    Xd,
        Yd,
        Xu,
        Yu,
        distortion_factor;

    int       i;

    M = newdmat (0, (cd.point_count - 1), 0, 4, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute U: unable to allocate matrix M\n");
        //	exit (-1);
    }

    a = newdmat (0, 4, 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute U: unable to allocate vector a\n");
        //	exit (-1);
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute U: unable to allocate vector b\n");
        //	exit (-1);
    }

    for (i = 0; i < cd.point_count; i++) {
        /* convert from image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu = Xd * distortion_factor;
        Yu = Yd * distortion_factor;

        M.el[i][0] = Yu * cd.xw[i];
        M.el[i][1] = Yu * cd.yw[i];
        M.el[i][2] = Yu;
        M.el[i][3] = -Xu * cd.xw[i];
        M.el[i][4] = -Xu * cd.yw[i];
        b.el[i][0] = Xu;
    }

    if (solve_system (M, a, b)) {
        return false;
        //	fprintf (stderr, "cepe compute U: unable to solve system  Ma=b\n");
        //	exit (-1);
    }

    U[0] = a.el[0][0];
    U[1] = a.el[1][0];
    U[2] = a.el[2][0];
    U[3] = a.el[3][0];
    U[4] = a.el[4][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


void TsaiCalibratorCoplanar::epe_compute_Tx_and_Ty (double U[])
{
    double    Tx,
        Ty,
        Ty_squared,
        x,
        y,
        Sr,
        r1p,
        r2p,
        r4p,
        r5p,
        r1,
        r2,
        r4,
        r5,
        distance,
        far_distance;

    int       i,
        far_point;

    r1p = U[0];
    r2p = U[1];
    r4p = U[3];
    r5p = U[4];

    /* first find the square of the magnitude of Ty */
    if ((fabs (r1p) < EPSILON) && (fabs (r2p) < EPSILON))
        Ty_squared = 1 / (SQR (r4p) + SQR (r5p));
    else if ((fabs (r4p) < EPSILON) && (fabs (r5p) < EPSILON))
        Ty_squared = 1 / (SQR (r1p) + SQR (r2p));
    else if ((fabs (r1p) < EPSILON) && (fabs (r4p) < EPSILON))
        Ty_squared = 1 / (SQR (r2p) + SQR (r5p));
    else if ((fabs (r2p) < EPSILON) && (fabs (r5p) < EPSILON))
        Ty_squared = 1 / (SQR (r1p) + SQR (r4p));
    else {
        Sr = SQR (r1p) + SQR (r2p) + SQR (r4p) + SQR (r5p);
        Ty_squared = (Sr - sqrt (SQR (Sr) - 4 * SQR (r1p * r5p - r4p * r2p))) /
            (2 * SQR (r1p * r5p - r4p * r2p));
    }

    /* find a point that is far from the image center */
    far_distance = 0;
    far_point = 0;
    for (i = 0; i < cd.point_count; i++)
        if ((distance = SQR (cd.Xf[i] - cp.Cx) + SQR (cd.Yf[i] - cp.Cy)) > far_distance) {
            far_point = i;
            far_distance = distance;
        }

        /* now find the sign for Ty */
        /* start by assuming Ty > 0 */
        Ty = sqrt (Ty_squared);
        r1 = U[0] * Ty;
        r2 = U[1] * Ty;
        Tx = U[2] * Ty;
        r4 = U[3] * Ty;
        r5 = U[4] * Ty;
        x = r1 * cd.xw[far_point] + r2 * cd.yw[far_point] + Tx;
        y = r4 * cd.xw[far_point] + r5 * cd.yw[far_point] + Ty;

        /* flip Ty if we guessed wrong */
        if ((SIGNBIT (x) != SIGNBIT (cd.Xf[far_point] - cp.Cx)) ||
            (SIGNBIT (y) != SIGNBIT (cd.Yf[far_point] - cp.Cy)))
            Ty = -Ty;

        /* update the calibration constants */
        cc.Tx = U[2] * Ty;
        cc.Ty = Ty;
}


void  TsaiCalibratorCoplanar::epe_compute_R (double U[]){
    double    r1,
        r2,
        r3,
        r4,
        r5,
        r6,
        r7,
        r8,
        r9;

    r1 = U[0] * cc.Ty;
    r2 = U[1] * cc.Ty;
    r3 = sqrt (1 - SQR (r1) - SQR (r2));

    r4 = U[3] * cc.Ty;
    r5 = U[4] * cc.Ty;
    r6 = sqrt (1 - SQR (r4) - SQR (r5));
    if (!SIGNBIT (r1 * r4 + r2 * r5))
        r6 = -r6;

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;
    r8 = r3 * r4 - r1 * r6;
    r9 = r1 * r5 - r2 * r4;

    /* update the calibration constants */
    cc.r1 = r1;
    cc.r2 = r2;
    cc.r3 = r3;
    cc.r4 = r4;
    cc.r5 = r5;
    cc.r6 = r6;
    cc.r7 = r7;
    cc.r8 = r8;
    cc.r9 = r9;

    /* fill in cc.Rx, cc.Ry and cc.Rz */
    Matrix2Euler ();
}


bool TsaiCalibratorCoplanar::epe_compute_approximate_f (double *f)
{
    dmat      M,
        a,
        b;

    double    Yd;

    int       i;

    M = newdmat (0, (cd.point_count - 1), 0, 1, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute apx: unable to allocate matrix M\n");
        //	exit (-1);
    }

    a = newdmat (0, 1, 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute apx: unable to allocate vector a\n");
        //	exit (-1);
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "cepe compute apx: unable to allocate vector b\n");
        //	exit (-1);
    }

    for (i = 0; i < cd.point_count; i++) {
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        M.el[i][0] = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.Ty;
        M.el[i][1] = -Yd;
        b.el[i][0] = (cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i]) * Yd;
    }

    if (solve_system (M, a, b)) {
        return false;
        //	fprintf (stderr, "cepe compute apx: unable to solve system  Ma=b\n");
        //	exit (-1);
    }

    /* return the approximate effective focal length */
     *f = a.el[0][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


/***********************************************************************\
 * Routines for noncoplanar extrinsic parameter estimation
\***********************************************************************/
bool TsaiCalibratorNoncoplanar::epe_compute_U (double U[])
{
    dmat      M,
        a,
        b;

    double    Xu,
        Yu,
        Xd,
        Yd,
        distortion_factor;

    int       i;

    M = newdmat (0, (cd.point_count - 1), 0, 6, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "ncepe compute U: unable to allocate matrix M\n");
        //	exit (-1);
    }

    a = newdmat (0, 6, 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "ncepe compute U: unable to allocate vector a\n");
        //	exit (-1);
    }

    b = newdmat (0, (cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "ncepe compute U: unable to allocate vector b\n");
        //	exit (-1);
    }

    for (i = 0; i < cd.point_count; i++) {
        /* convert from image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu = Xd * distortion_factor;
        Yu = Yd * distortion_factor;

        M.el[i][0] = Yu * cd.xw[i];
        M.el[i][1] = Yu * cd.yw[i];
        M.el[i][2] = Yu * cd.zw[i];
        M.el[i][3] = Yu;
        M.el[i][4] = -Xu * cd.xw[i];
        M.el[i][5] = -Xu * cd.yw[i];
        M.el[i][6] = -Xu * cd.zw[i];
        b.el[i][0] = Xu;
    }

    if (solve_system (M, a, b)) {
        return false;
        //	fprintf (stderr, "ncepe compute U: unable to solve system  Ma=b\n");
        //	exit (-1);
    }

    U[0] = a.el[0][0];
    U[1] = a.el[1][0];
    U[2] = a.el[2][0];
    U[3] = a.el[3][0];
    U[4] = a.el[4][0];
    U[5] = a.el[5][0];
    U[6] = a.el[6][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


void TsaiCalibratorNoncoplanar::epe_compute_Tx_and_Ty (double U[])
{
    double    Tx,
        Ty,
        Ty_squared,
        x,
        y,
        r1,
        r2,
        r3,
        r4,
        r5,
        r6,
        distance,
        far_distance;

    int       i,
        far_point;

    /* first find the square of the magnitude of Ty */
    Ty_squared = 1 / (SQR (U[4]) + SQR (U[5]) + SQR (U[6]));

    /* find a point that is far from the image center */
    far_distance = 0;
    far_point = 0;
    for (i = 0; i < cd.point_count; i++)
        if ((distance = SQR (cd.Xf[i] - cp.Cx) + SQR (cd.Yf[i] - cp.Cy)) > far_distance) {
            far_point = i;
            far_distance = distance;
        }

        /* now find the sign for Ty */
        /* start by assuming Ty > 0 */
        Ty = sqrt (Ty_squared);
        r1 = U[0] * Ty;
        r2 = U[1] * Ty;
        r3 = U[2] * Ty;
        Tx = U[3] * Ty;
        r4 = U[4] * Ty;
        r5 = U[5] * Ty;
        r6 = U[6] * Ty;
        x = r1 * cd.xw[far_point] + r2 * cd.yw[far_point] + r3 * cd.zw[far_point] + Tx;
        y = r4 * cd.xw[far_point] + r5 * cd.yw[far_point] + r6 * cd.zw[far_point] + Ty;

        /* flip Ty if we guessed wrong */
        if ((SIGNBIT (x) != SIGNBIT (cd.Xf[far_point] - cp.Cx)) ||
            (SIGNBIT (y) != SIGNBIT (cd.Yf[far_point] - cp.Cy)))
            Ty = -Ty;

        /* update the calibration constants */
        cc.Tx = U[3] * Ty;
        cc.Ty = Ty;
}


void  TsaiCalibratorNoncoplanar::epe_compute_R (double U[])
{
    double    r1,
        r2,
        r3,
        r4,
        r5,
        r6,
        r7,
        r8,
        r9;

    r1 = U[0] * cc.Ty;
    r2 = U[1] * cc.Ty;
    r3 = U[2] * cc.Ty;

    r4 = U[4] * cc.Ty;
    r5 = U[5] * cc.Ty;
    r6 = U[6] * cc.Ty;

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;
    r8 = r3 * r4 - r1 * r6;
    r9 = r1 * r5 - r2 * r4;

    /* update the calibration constants */
    cc.r1 = r1;
    cc.r2 = r2;
    cc.r3 = r3;
    cc.r4 = r4;
    cc.r5 = r5;
    cc.r6 = r6;
    cc.r7 = r7;
    cc.r8 = r8;
    cc.r9 = r9;

    /* fill in cc.Rx, cc.Ry and cc.Rz */
    Matrix2Euler ();
}


/************************************************************************/
bool TsaiCalibrator::epe_compute_Tx_Ty_Tz ()
{
    dmat      M,
        a,
        b;

    double    xk,
        yk,
        zk,
        Xu,
        Yu,
        Xd,
        Yd,
        distortion_factor;

    int       i,
        j;

    M = newdmat (0, (2 * cd.point_count - 1), 0, 2, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "epe compute Tx Ty Tz: unable to allocate matrix M\n");
        //	exit (-1);
    }

    a = newdmat (0, 2, 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "epe compute Tx Ty Tz: unable to allocate vector a\n");
        //	exit (-1);
    }

    b = newdmat (0, (2 * cd.point_count - 1), 0, 0, &errno);
    if (errno) {
        return false;
        //	fprintf (stderr, "epe compute Tx Ty Tz: unable to allocate vector b\n");
        //	exit (-1);
    }

    for (i = 0, j = cd.point_count; i < cd.point_count; i++, j++) {
        /* convert from world coordinates to untranslated camera coordinates */
        xk = cc.r1 * cd.xw[i] + cc.r2 * cd.yw[i] + cc.r3 * cd.zw[i];
        yk = cc.r4 * cd.xw[i] + cc.r5 * cd.yw[i] + cc.r6 * cd.zw[i];
        zk = cc.r7 * cd.xw[i] + cc.r8 * cd.yw[i] + cc.r9 * cd.zw[i];

        /* convert from image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu = Xd * distortion_factor;
        Yu = Yd * distortion_factor;

        M.el[i][0] = cc.f;
        M.el[i][1] = 0;
        M.el[i][2] = -Xu;
        b.el[i][0] = Xu * zk - cc.f * xk;

        M.el[j][0] = 0;
        M.el[j][1] = cc.f;
        M.el[j][2] = -Yu;
        b.el[j][0] = Yu * zk - cc.f * yk;
    }

    if (solve_system (M, a, b)) {
        return false;
        //	fprintf (stderr, "epe compute Tx Ty Tz: unable to solve system  Ma=b\n");
        //	exit (-1);
    }

    cc.Tx = a.el[0][0];
    cc.Ty = a.el[1][0];
    cc.Tz = a.el[2][0];

    freemat (M);
    freemat (a);
    freemat (b);
    return true;
}


/************************************************************************/
int TsaiCalibrator::epe_optimize_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err,integer *notused)
{
    int       i;

    double    xc,
        yc,
        zc,
        Xd,
        Yd,
        Xu_1,
        Yu_1,
        Xu_2,
        Yu_2,
        distortion_factor,
        Rx,
        Ry,
        Rz,
        Tx,
        Ty,
        Tz,
        r1,
        r2,
        r3,
        r4,
        r5,
        r6,
        r7,
        r8,
        r9,
        sa,
        sb,
        sg,
        ca,
        cb,
        cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);

    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;

    for (i = 0; i < cd.point_count; i++) {
        /* convert from world coordinates to camera coordinates */
        xc = r1 * cd.xw[i] + r2 * cd.yw[i] + r3 * cd.zw[i] + Tx;
        yc = r4 * cd.xw[i] + r5 * cd.yw[i] + r6 * cd.zw[i] + Ty;
        zc = r7 * cd.xw[i] + r8 * cd.yw[i] + r9 * cd.zw[i] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = cc.f * xc / zc;
        Yu_1 = cc.f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu_2 = Xd * distortion_factor;
        Yu_2 = Yd * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
    return 0;
}


bool TsaiCalibrator::epe_optimize ()
{
#define NPARAMS 6

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cd.point_count;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        //       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
        //       exit(-1);
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        //       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
        //       exit(-1);
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
        return false;
        //       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
        //       exit(-1);
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = cc.Rx;
    x[1] = cc.Ry;
    x[2] = cc.Rz;
    x[3] = cc.Tx;
    x[4] = cc.Ty;
    x[5] = cc.Tz;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */
    lmdif_ (epe_optimize_error,
        &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
        diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
        ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    cc.Rx = x[0];
    cc.Ry = x[1];
    cc.Rz = x[2];
    Euler2Matrix ();

    cc.Tx = x[3];
    cc.Ty = x[4];
    cc.Tz = x[5];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    //    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);
#endif

#undef NPARAMS
    return true;
}


/************************************************************************/
bool TsaiCalibratorCoplanar::extrinsic_parameter_estimation ()
{
    double    trial_f,
        U[5];

    if(!epe_compute_U (U))
        return false;

    epe_compute_Tx_and_Ty (U);

    epe_compute_R (U);

    if(!epe_compute_approximate_f (&trial_f))
        return false;

    if (trial_f < 0) {
        /* try the other rotation matrix solution */
        cc.r3 = -cc.r3;
        cc.r6 = -cc.r6;
        cc.r7 = -cc.r7;
        cc.r8 = -cc.r8;
        Matrix2Euler ();

        if(!epe_compute_approximate_f (&trial_f))
            return false;

        if (trial_f < 0) {
            return false;
            //	    fprintf (stderr, "error - possible handedness problem with data\n");
            //	    exit (-1);
        }
    }

    if(!epe_compute_Tx_Ty_Tz ())
        return false;

    return epe_optimize ();
}


/************************************************************************/
bool TsaiCalibratorNoncoplanar::extrinsic_parameter_estimation ()
{
    double    U[7];

    if(!epe_compute_U (U))
        return false;;

    epe_compute_Tx_and_Ty (U);
    epe_compute_R (U);

    if(!epe_compute_Tx_Ty_Tz ())
        return false;

    return epe_optimize ();
}



void  TsaiCalibrator::distorted_image_plane_error_stats (double *mean, double *stddev, double *max, double *sse)
{
    int       i;

    double    Xf,
        Yf,
        error,
        squared_error,
        max_error = 0,
        sum_error = 0,
        sum_squared_error = 0;

    if (cd.point_count < 1) {
        *mean = *stddev = *max = *sse = 0;
        return;
    }

    for (i = 0; i < cd.point_count; i++) {
        /* calculate the ideal location of the image of the data point */
        world_coord_to_image_coord (cd.xw[i], cd.yw[i], cd.zw[i], &Xf, &Yf);

        /* determine the error between the ideal and actual location of the data point	 */
        /* (in distorted image coordinates)						 */
        squared_error = SQR (Xf - cd.Xf[i]) + SQR (Yf - cd.Yf[i]);
        error = sqrt (squared_error);
        sum_error += error;
        sum_squared_error += squared_error;
        max_error = MAX (max_error, error);
    }

    *mean = sum_error / cd.point_count;
    *max = max_error;
    *sse = sum_squared_error;

    if (cd.point_count == 1)
        *stddev = 0;
    else
        *stddev = sqrt ((sum_squared_error - SQR (sum_error) / cd.point_count) / (cd.point_count - 1));
}


/*******************************************************************************\
 * This routine calculates the mean, standard deviation, max, and sum-of-squared *
 * error of the magnitude of the error, in undistorted image coordinates, between*
 * the measured location of a feature point in the image plane and the image of  *
 * the 3D feature point as projected through the calibrated model.               *
 * The calculation is for all of the points in the calibration data set.         *
\*******************************************************************************/
void TsaiCalibrator::undistorted_image_plane_error_stats (double *mean, double *stddev, double *max, double *sse)
{
    int       i;

    double    xc,
        yc,
        zc,
        Xu_1,
        Yu_1,
        Xu_2,
        Yu_2,
        Xd,
        Yd,
        distortion_factor,
        x_pixel_error,
        y_pixel_error,
        error,
        squared_error,
        max_error = 0,
        sum_error = 0,
        sum_squared_error = 0;

    if (cd.point_count < 1) {
        *mean = *stddev = *max = *sse = 0;
        return;
    }

    for (i = 0; i < cd.point_count; i++) {
        /* calculate the ideal location of the image of the data point */
        world_coord_to_camera_coord (cd.xw[i], cd.yw[i], cd.zw[i], &xc, &yc, &zc);

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = cc.f * xc / zc;
        Yu_1 = cc.f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu_2 = Xd * distortion_factor;
        Yu_2 = Yd * distortion_factor;

        /* determine the error between the ideal and actual location of the data point	 */
        /* (in undistorted image coordinates)						 */
        x_pixel_error = cp.sx * (Xu_1 - Xu_2) / cp.dpx;
        y_pixel_error = (Yu_1 - Yu_2) / cp.dpy;
        squared_error = SQR (x_pixel_error) + SQR (y_pixel_error);
        error = sqrt (squared_error);
        sum_error += error;
        sum_squared_error += squared_error;
        max_error = MAX (max_error, error);
    }

    *mean = sum_error / cd.point_count;
    *max = max_error;
    *sse = sum_squared_error;

    if (cd.point_count == 1)
        *stddev = 0;
    else
        *stddev = sqrt ((sum_squared_error - SQR (sum_error) / cd.point_count) / (cd.point_count - 1));
}


/*******************************************************************************\
 * This routine calculates the mean, standard deviation, max, and sum-of-squared *
 * error of the distance of closest approach (i.e. 3D error) between the point   *
 * in object space and the line of sight formed by back projecting the measured  *
 * 2D coordinates out through the camera model.                                  *
 * The calculation is for all of the points in the calibration data set.         *
\*******************************************************************************/
void  TsaiCalibrator::object_space_error_stats (double *mean, double *stddev, double *max, double *sse)
{
    int       i;

    double    xc,
        yc,
        zc,
        Xu,
        Yu,
        Xd,
        Yd,
        t,
        distortion_factor,
        error,
        squared_error,
        max_error = 0,
        sum_error = 0,
        sum_squared_error = 0;

    if (cd.point_count < 1) {
        *mean = *stddev = *max = *sse = 0;
        return;
    }

    for (i = 0; i < cd.point_count; i++) {
        /* determine the position of the 3D object space point in camera coordinates */
        world_coord_to_camera_coord (cd.xw[i], cd.yw[i], cd.zw[i], &xc, &yc, &zc);

        /* convert the measured 2D image coordinates into distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates into undistorted sensor plane coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu = Xd * distortion_factor;
        Yu = Yd * distortion_factor;

        /* find the magnitude of the distance (error) of closest approach */
        /* between the undistorted line of sight and the point in 3 space */
        t = (xc * Xu + yc * Yu + zc * cc.f) / (SQR (Xu) + SQR (Yu) + SQR (cc.f));
        squared_error = SQR (xc - Xu * t) + SQR (yc - Yu * t) + SQR (zc - cc.f * t);
        error = sqrt (squared_error);
        sum_error += error;
        sum_squared_error += squared_error;
        max_error = MAX (max_error, error);
    }

    *mean = sum_error / cd.point_count;
    *max = max_error;
    *sse = sum_squared_error;

    if (cd.point_count == 1)
        *stddev = 0;
    else
        *stddev = sqrt ((sum_squared_error - SQR (sum_error) / cd.point_count) / (cd.point_count - 1));
}


/*******************************************************************************\
 * This routine performs an error measure proposed by Weng in IEEE PAMI,         *
 * October 1992.                                                                 *
\*******************************************************************************/
void TsaiCalibrator::normalized_calibration_error (double *mean, double *stddev)
{
    int       i;

    double    xc,
        yc,
        zc,
        xc_est,
        yc_est,
        zc_est,
        fu,
        fv,
        Xd,
        Yd,
        Xu,
        Yu,
        distortion_factor,
        squared_error,
        sum_error = 0,
        sum_squared_error = 0;

    if (cd.point_count < 1) {
        *mean = *stddev = 0;
        return;
    }

    /* This error metric is taken from						 */
    /* "Camera Calibration with Distortion Models and Accuracy Evaluation"	 */
    /* J. Weng, P. Cohen, and M. Herniou					 */
    /* IEEE Transactions on PAMI, Vol. 14, No. 10, October 1992, pp965-980	 */

    for (i = 0; i < cd.point_count; i++) {
        /* estimate the 3D coordinates of the calibration data point by back 	 */
        /* projecting its measured image location through the model to the	 */
        /* plane formed by the original z world component.			 */

        /* calculate the location of the data point in camera coordinates */
        world_coord_to_camera_coord (cd.xw[i], cd.yw[i], cd.zw[i], &xc, &yc, &zc);

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd = cp.dpx * (cd.Xf[i] - cp.Cx) / cp.sx;
        Yd = cp.dpy * (cd.Yf[i] - cp.Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + cc.kappa1 * (SQR (Xd) + SQR (Yd));
        Xu = Xd * distortion_factor;
        Yu = Yd * distortion_factor;

        /* estimate the location of the data point by back projecting the image position */
        zc_est = zc;
        xc_est = zc_est * Xu / cc.f;
        yc_est = zc_est * Yu / cc.f;

        fu = cp.sx * cc.f / cp.dpx;
        fv = cc.f / cp.dpy;

        squared_error = (SQR (xc_est - xc) + SQR (yc_est - yc)) /
            (SQR (zc_est) * (1 / SQR (fu) + 1 / SQR (fv)) / 12);
        sum_error += sqrt (squared_error);
        sum_squared_error += squared_error;
    }

    *mean = sum_error / cd.point_count;

    if (cd.point_count == 1)
        *stddev = 0;
    else
        *stddev = sqrt ((sum_squared_error - SQR (sum_error) / cd.point_count) / (cd.point_count - 1));
}
