/***************************************************************************
                          TsaiTransform.cpp  -  description
                             -------------------
    begin                : 23.04.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv3d/TsaiTransform.h"
#include <cmath>
#include <iostream>
#include <cstring>

using namespace TUVision;

#define SQRT3   1.732050807568877293527446341505872366943

#define SQR(a)          ((a) * (a))
#define CUB(a)          ((a)*(a)*(a))
#define SQRT(x) sqrt(fabs(x))
/* Some math libraries may not have the sincos() routine */
#ifdef _SINCOS_
void sincos();
#define SINCOS(x,s,c)   sincos(x,&s,&c)
#else
double sin(), cos();
#define SINCOS(x,s,c)   s=sin(x);c=cos(x)
#endif


/************************************************************************/
/* This cube root routine handles negative arguments (unlike cbrt).     */

double    CBRT (double x)
{
    if (x == 0)
        return (0);
    else if (x > 0)
        return (pow (x, (double) 1.0 / 3.0));
    else
        return (-pow (-x, (double) 1.0 / 3.0));
} 

/************************************************************************/
/*!
 * Default constructor sets cameraparameters and calibration constants
 * (both define Tsai's camera model) to zero, i.e. they have to be
 * initialized first or passed to static routines
 *
 ************************************************************************/
CTsaiCoordTransform::CTsaiCoordTransform(){
    d_cameraParameters     = 0;
    d_calibrationConstants = 0;
}

/************************************************************************/
/*!
 * Constructor sets cameraparameters to the passed structs.
 * They have to be initialized first with correct values
 * Note: This constructor is only needed for non-static calls.
 *
 * \param cameraParameters camera parameters which should be used for following transformations
 * \param calibrationConstants calibration constants which should be used for following transformations
 *
 ************************************************************************/
CTsaiCoordTransform::CTsaiCoordTransform(SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    Init(cameraParameters,calibrationConstants);
}

/************************************************************************/
/*!
 * Destructor deletes camera parameters and calibration constants if
 * if they have been used
 *
 ************************************************************************/
CTsaiCoordTransform::~CTsaiCoordTransform(){
    if(d_cameraParameters)
        delete d_cameraParameters;
    if(d_calibrationConstants)
        delete d_calibrationConstants;
}

/************************************************************************/
/*!
 * Sets cameraparameters to the passed struct.
 *
 * \param cameraParameters camera parameters which should be used for following transformations
 *
 ************************************************************************/
void CTsaiCoordTransform::setCameraParameters(SCameraParameters &cameraParameters){
    if(!d_cameraParameters)
        d_cameraParameters = new SCameraParameters;
    memcpy(d_cameraParameters,&cameraParameters,sizeof(SCameraParameters));
}

/************************************************************************/
/*!
 * Sets calibration constants to the passed struct.
 *
 * \param calibrationConstants calibration constants which should be used for following transformations
 *
 ************************************************************************/
void CTsaiCoordTransform::setCalibrationConstants(SCalibrationConstants &calibrationConstants){
    if(!d_calibrationConstants)
        d_calibrationConstants = new SCalibrationConstants;
    memcpy(d_calibrationConstants,&calibrationConstants,sizeof(SCalibrationConstants));
}

/************************************************************************/
/*!
 * Sets cameraparameters and calibration constants to the passed structs. Both
 * arguments have to be initialized before with correct values
 *
 * \param cameraParameters camera parameters which should be used for following transformations
 * \param calibrationConstants calibration constants which should be used for following transformations
 *
 ************************************************************************/
void CTsaiCoordTransform::Init(SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    setCameraParameters(cameraParameters);
    setCalibrationConstants(calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates undistorted sensor coordinates given distorted coordinates.
 * Processing nonlinear lens distortion takes care at this level.
 * Usually a user would not call this function directly but
 * undistorted2distortedImage instead.
 *
 * \param undistortedCoords undistorted coordinates which are given
 * \param distortedCoords pointer to distorted coordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::undistorted2distortedSensor(TUVision::CVertex2d &undistortedCoords, TUVision::CVertex2d *distortedCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return undistorted2distortedSensor(undistortedCoords,distortedCoords,*d_cameraParameters,*d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates distorted sensor coordinates given undistorted coordinates.
 * Processing nonlinear lens distortion takes care at this level (static version)
 * Usually a user would not call this function directly but
 * distorted2undistortedImage instead.
 *
 * \param distortedCoords undistorted coordinates which are given
 * \param undistortedCoords pointer to distorted coordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::distorted2undistortedSensor(TUVision::CVertex2d &distortedCoords, TUVision::CVertex2d *undistortedCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return distorted2undistortedSensor(distortedCoords,undistortedCoords,*d_cameraParameters,*d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates distorted image coordinates given undistorted image coordinates.
 *
 * \param undistortedCoords undistorted coordinates which are given
 * \param distortedCoords pointer to distorted coordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::undistorted2distortedImage(TUVision::CVertex2d &undistortedCoords, TUVision::CVertex2d *distortedCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return undistorted2distortedImage(undistortedCoords,distortedCoords,*d_cameraParameters,*d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates undistorted image coordinates given distorted image coordinates.
 *
 * \param distortedCoords undistorted coordinates which are given
 * \param undistortedCoords pointer to distorted coordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::distorted2undistortedImage(TUVision::CVertex2d &distortedCoords, TUVision::CVertex2d *undistortedCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return distorted2undistortedImage(distortedCoords,undistortedCoords,*d_cameraParameters,*d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates (distorted) image coordinates given world coordinates.
 *
 * \param worldCoords world oordinates which are given
 * \param imageCoords pointer to distorted coordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::world2image(TUVision::CVertex3d &worldCoords, TUVision::CVertex2d *imageCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return world2image(worldCoords,imageCoords,*d_cameraParameters,*d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates world coordinates given (distorted) image coordinates.
 * This of course is not possible per se as points on an image can
 * originate from a ray. Thus, the distance of the camera to the 
 * point in world coordinates has to be passed to this function too such
 * that the world coordinate can be identified.
 *
 * \param imageCoords image coordinated which are given
 * \param worldCoords pointer to world oordinates which should be calculated
 * \param cameraDistance distance of the camera to the desired world coordinate
 *
 ************************************************************************/
bool CTsaiCoordTransform::image2world(TUVision::CVertex2d &imageCoords, double cameraDistanceZ, TUVision::CVertex3d *worldCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return image2world(imageCoords, cameraDistanceZ, worldCoords, *d_cameraParameters, *d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates camera coordinates given world ones
 * By having defined two coordinate systems (one inside the camera and one
 * world coordinate system which defines positions in space), a position
 * in world space can be transformed to camera space by rotation and translation.
 *
 * \param worldCoords world coordinated which are given
 * \param cameraCoords pointer to camera oordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::world2camera(TUVision::CVertex3d &worldCoords, TUVision::CVertex3d *cameraCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return world2camera(worldCoords, cameraCoords, *d_cameraParameters, *d_calibrationConstants);
}

/************************************************************************/
/*!
 * Calculates world coordinates given camera ones
 * By having defined two coordinate systems (one inside the camera and one
 * world coordinate system which defines positions in space), a position
 * in camera space can be transformed to world space by rotation and translation.
 *
 * \param cameraCoords camera coordinated which are given
 * \param worldCoords pointer to world oordinates which should be calculated
 *
 ************************************************************************/
bool CTsaiCoordTransform::camera2world(TUVision::CVertex3d &cameraCoords, TUVision::CVertex3d *worldCoords){
    if(!d_calibrationConstants || !d_cameraParameters)
        return false;
    else
        return camera2world(cameraCoords, worldCoords, *d_cameraParameters, *d_calibrationConstants);
}

bool CTsaiCoordTransform::undistorted2distortedSensor(TUVision::CVertex2d &undistortedCoords, TUVision::CVertex2d *distortedCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    double Ru, Rd, lambda, c, d, Q, R, D, S, T, sinT, cosT;

    if (((undistortedCoords.x == 0) && (undistortedCoords.y == 0)) || (calibrationConstants.kappa1 == 0)) {
        (*distortedCoords) = undistortedCoords;
        return true;
    }

    Ru = hypot (undistortedCoords.x, undistortedCoords.y);	/* SQRT(undistortedCoords[0]*undistortedCoords[0]+undistortedCoords[1]*undistortedCoords[1]) */

    c = 1 / calibrationConstants.kappa1;
    d = -c * Ru;

    Q = c / 3;
    R = -d / 2;
    D = CUB (Q) + SQR (R);

    if (D >= 0) {		/* one real root */
        D = SQRT (D);
        S = CBRT (R + D);
        T = CBRT (R - D);
        Rd = S + T;

        if (Rd < 0) {
            Rd = SQRT (-1 / (3 * calibrationConstants.kappa1));
        }
    } else {			/* three real roots */
        D = SQRT (-D);
        S = CBRT (hypot (R, D));
        T = atan2 (D, R) / 3;
        SINCOS (T, sinT, cosT);

        /* the larger positive root is    2*S*cos(T)                   */
        /* the smaller positive root is   -S*cos(T) + SQRT(3)*S*sin(T) */
        /* the negative root is           -S*cos(T) - SQRT(3)*S*sin(T) */

        Rd = -S * cosT + SQRT3 * S * sinT;	/* use the smaller positive root */
    }

    lambda = Rd / Ru;

    (*distortedCoords) = undistortedCoords * lambda;
    return true;
}

/************************************************************************/
/*!
 * Calculates undistorted sensor coordinates given distorted coordinates.
 * Processing nonlinear lens distortion takes care at this level.
 * Usually a user would not call this function directly but
 * undistorted2distortedImage instead.
 * (static version which can be used without having to initialize an object)
 *
 * \param undistortedCoords undistorted coordinates which are given
 * \param distortedCoords pointer to distorted coordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::distorted2undistortedSensor(TUVision::CVertex2d &distortedCoords, TUVision::CVertex2d *undistortedCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    double    distortion_factor;

    /* convert from distorted to undistorted sensor plane coordinates */
    distortion_factor = 1 + calibrationConstants.kappa1 * (SQR (distortedCoords.x) + SQR (distortedCoords.y));
    (*undistortedCoords) = distortedCoords * distortion_factor;
    return true;
}

/************************************************************************/
/*!
 * Calculates distorted sensor coordinates given undistorted coordinates.
 * Processing nonlinear lens distortion takes care at this level (static version)
 * Usually a user would not call this function directly but
 * distorted2undistortedImage instead.
 * (static version which can be used without having to initialize an object)
 *
 * \param distortedCoords undistorted coordinates which are given
 * \param undistortedCoords pointer to distorted coordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/

bool CTsaiCoordTransform::undistorted2distortedImage(TUVision::CVertex2d &undistortedCoords, TUVision::CVertex2d *distortedCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    TUVision::CVertex2d undistortedSensor;
    TUVision::CVertex2d distortedSensor;

    /* convert from image to sensor coordinates */
    undistortedSensor.x = cameraParameters.dpx * (undistortedCoords.x - cameraParameters.Cx) / cameraParameters.sx;
    undistortedSensor.y = cameraParameters.dpy * (undistortedCoords.y - cameraParameters.Cy);

    /* convert from undistorted sensor to distorted sensor plane coordinates */
    if(!undistorted2distortedSensor (undistortedSensor, &distortedSensor, cameraParameters, calibrationConstants))
        return false;

    /* convert from sensor to image coordinates */
    distortedCoords->x = distortedSensor.x * cameraParameters.sx / cameraParameters.dpx + cameraParameters.Cx;
    distortedCoords->y = distortedSensor.y / cameraParameters.dpy + cameraParameters.Cy;
    return true;
}

/************************************************************************/
/*!
 * Calculates distorted image coordinates given undistorted image coordinates.
 * (static version which can be used without having to initialize an object)
 *
 * \param undistortedCoords undistorted coordinates which are given
 * \param distortedCoords pointer to distorted coordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::distorted2undistortedImage(TUVision::CVertex2d &distortedCoords, TUVision::CVertex2d *undistortedCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    TUVision::CVertex2d distortedSensor;
    TUVision::CVertex2d undistortedSensor;

    /* convert from image to sensor coordinates */
    distortedSensor.x = cameraParameters.dpx * (distortedCoords.x - cameraParameters.Cx) / cameraParameters.sx;
    distortedSensor.y = cameraParameters.dpy * (distortedCoords.y - cameraParameters.Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    if(!distorted2undistortedSensor (distortedSensor, &undistortedSensor, cameraParameters, calibrationConstants))
        return false;

    /* convert from sensor to image coordinates */
    undistortedCoords->x = undistortedSensor.x * cameraParameters.sx / cameraParameters.dpx + cameraParameters.Cx;
    undistortedCoords->y = undistortedSensor.y / cameraParameters.dpy + cameraParameters.Cy;
    return true;
}

/************************************************************************/
/*!
 * Calculates (distorted) image coordinates given world coordinates.
 * (static version which can be used without having to initialize an object)
 *
 * \param worldCoords world oordinates which are given
 * \param imageCoords pointer to distorted coordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::world2image(TUVision::CVertex3d &worldCoords, TUVision::CVertex2d *imageCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    TUVision::CVertex3d cameraCoords;
    TUVision::CVertex2d undistortedSensor;
    TUVision::CVertex2d distortedSensor;

    euler2Matrix(&calibrationConstants);

    /* convert from world coordinates to camera coordinates */
    cameraCoords.x = calibrationConstants.r1 * worldCoords.x + calibrationConstants.r2 * worldCoords.y + calibrationConstants.r3 * worldCoords.z + calibrationConstants.Tx;
    cameraCoords.y = calibrationConstants.r4 * worldCoords.x + calibrationConstants.r5 * worldCoords.y + calibrationConstants.r6 * worldCoords.z + calibrationConstants.Ty;
    cameraCoords.z = calibrationConstants.r7 * worldCoords.x + calibrationConstants.r8 * worldCoords.y + calibrationConstants.r9 * worldCoords.z + calibrationConstants.Tz;

    /* convert from camera coordinates to undistorted sensor plane coordinates */
    undistortedSensor.x = calibrationConstants.f * cameraCoords.x / cameraCoords.z;
    undistortedSensor.y = calibrationConstants.f * cameraCoords.y / cameraCoords.z;

    /* convert from undistorted to distorted sensor plane coordinates */
    if(!undistorted2distortedSensor (undistortedSensor, &distortedSensor, cameraParameters, calibrationConstants))
        return false;

    /* convert from distorted sensor plane coordinates to image coordinates */
    imageCoords->x = distortedSensor.x * cameraParameters.sx / cameraParameters.dpx + cameraParameters.Cx;
    imageCoords->y = distortedSensor.y/ cameraParameters.dpy + cameraParameters.Cy;
    return true;
}

/************************************************************************/
/*!
 * Calculates world coordinates given (distorted) image coordinates.
 * This of course is not possible per se as points on an image can
 * originate from a ray. Thus, the distance of the camera to the 
 * point in world coordinates has to be passed to this function too such
 * that the world coordinate can be identified.
 * (static version which can be used without having to initialize an object)
 *
 * \param imageCoords image coordinated which are given
 * \param worldCoords pointer to world oordinates which should be calculated
 * \param cameraDistance distance of the camera to the desired world coordinate
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::image2world(TUVision::CVertex2d &imageCoords, double cameraDistanceZ, TUVision::CVertex3d *worldCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){

    TUVision::CVertex2d distortedSensor;
    TUVision::CVertex2d undistortedSensor;

    double bTx,bTy,bTz,scaleX,scaleY;;

    euler2Matrix(&calibrationConstants);

    /* convert from image to distorted sensor coordinates */
    distortedSensor.x = cameraParameters.dpx * (imageCoords.x - cameraParameters.Cx) / cameraParameters.sx;
    distortedSensor.y = cameraParameters.dpy * (imageCoords.y - cameraParameters.Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    if(!distorted2undistortedSensor (distortedSensor, &undistortedSensor, cameraParameters, calibrationConstants))
        return false;


    bTx = -(calibrationConstants.Tx*calibrationConstants.r1+calibrationConstants.Ty*calibrationConstants.r4+calibrationConstants.Tz*calibrationConstants.r7);
    bTy = -(calibrationConstants.Tx*calibrationConstants.r2+calibrationConstants.Ty*calibrationConstants.r5+calibrationConstants.Tz*calibrationConstants.r8);
    bTz = -(calibrationConstants.Tx*calibrationConstants.r3+calibrationConstants.Ty*calibrationConstants.r6+calibrationConstants.Tz*calibrationConstants.r9);

    scaleX=undistortedSensor.x*cameraDistanceZ/calibrationConstants.f;
    scaleY=undistortedSensor.y*cameraDistanceZ/calibrationConstants.f;

    worldCoords->x = scaleX*calibrationConstants.r1+scaleY*calibrationConstants.r4+cameraDistanceZ*calibrationConstants.r7+bTx;
    worldCoords->y = scaleX*calibrationConstants.r2+scaleY*calibrationConstants.r5+cameraDistanceZ*calibrationConstants.r8+bTy;
    worldCoords->z = scaleX*calibrationConstants.r3+scaleY*calibrationConstants.r6+cameraDistanceZ*calibrationConstants.r9+bTz;

    return true;
}

/************************************************************************/
/*!
 * Calculates camera coordinates given world ones
 * By having defined two coordinate systems (one inside the camera and one
 * world coordinate system which defines positions in space), a position
 * in world space can be transformed to camera space by rotation and translation.
 * (static version which can be used without having to initialize an object)
 *
 * \param worldCoords world coordinated which are given
 * \param cameraCoords pointer to camera oordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::world2camera(TUVision::CVertex3d &worldCoords, TUVision::CVertex3d *cameraCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){

    euler2Matrix(&calibrationConstants);

    cameraCoords->x = calibrationConstants.r1 * worldCoords.x + calibrationConstants.r2 * worldCoords.y + calibrationConstants.r3 * worldCoords.z + calibrationConstants.Tx;
    cameraCoords->y = calibrationConstants.r4 * worldCoords.x + calibrationConstants.r5 * worldCoords.y + calibrationConstants.r6 * worldCoords.z + calibrationConstants.Ty;
    cameraCoords->z = calibrationConstants.r7 * worldCoords.x + calibrationConstants.r8 * worldCoords.y + calibrationConstants.r9 * worldCoords.z + calibrationConstants.Tz;

    return true;
}

/************************************************************************/
/*!
 * Calculates world coordinates given camera ones
 * By having defined two coordinate systems (one inside the camera and one
 * world coordinate system which defines positions in space), a position
 * in camera space can be transformed to world space by rotation and translation.
 * (static version which can be used without having to initialize an object)
 *
 * \param cameraCoords camera coordinated which are given
 * \param worldCoords pointer to world oordinates which should be calculated
 * \param cameraParameters camera parameters which should be used for this transformation
 * \param calibrationConstants calibration constants which should be used for this transformation
 *
 ************************************************************************/
bool CTsaiCoordTransform::camera2world(TUVision::CVertex3d &cameraCoords, TUVision::CVertex3d *worldCoords, SCameraParameters &cameraParameters, SCalibrationConstants &calibrationConstants){
    double txc, tyc, tzc;

    euler2Matrix(&calibrationConstants);

    txc=cameraCoords.x-calibrationConstants.Tx;
    tyc=cameraCoords.y-calibrationConstants.Ty;
    tzc=cameraCoords.z-calibrationConstants.Tz;

    worldCoords->x = calibrationConstants.r1 * txc + calibrationConstants.r4 * tyc + calibrationConstants.r7 * tzc;
    worldCoords->y = calibrationConstants.r2 * txc + calibrationConstants.r5 * tyc + calibrationConstants.r8 * tzc;
    worldCoords->z = calibrationConstants.r3 * txc + calibrationConstants.r6 * tyc + calibrationConstants.r9 * tzc;

    return true;
}

void CTsaiCoordTransform::euler2Matrix(SCalibrationConstants *calibrationConstants){
    double sa, ca, sb, cb, sg, cg;

    SINCOS (calibrationConstants->Rx, sa, ca);
    SINCOS (calibrationConstants->Ry, sb, cb);
    SINCOS (calibrationConstants->Rz, sg, cg);

    calibrationConstants->r1 = cb * cg;
    calibrationConstants->r2 = cg * sa * sb - ca * sg;
    calibrationConstants->r3 = sa * sg + ca * cg * sb;
    calibrationConstants->r4 = cb * sg;
    calibrationConstants->r5 = sa * sb * sg + ca * cg;
    calibrationConstants->r6 = ca * sb * sg - cg * sa;
    calibrationConstants->r7 = -sb;
    calibrationConstants->r8 = cb * sa;
    calibrationConstants->r9 = ca * cb;
}
